// This file is @generated by crates/varcon/tests/codegen.rs
#![allow(clippy::unreadable_literal)]

use crate::{Category, Cluster, Entry, Pos, Tag, Type, Variant};

pub static VARCON: &[Cluster] = &[
Cluster {
  header: "abettor ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "abettor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "abetter",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "abettors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "abetters",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "abettor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "abetter's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "abnormalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "abnormalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "abnormalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "abnormalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "abnormalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "abnormalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "abnormalizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "abnormalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "abnormalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "abolitionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "abolitionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "abolitionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "abolitionized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "abolitionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "abolitionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "abolitionizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "abolitionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "abolitionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "abridgment ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "abridgment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "abridgement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "abridgments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "abridgements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "abridgment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "abridgement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "absinthe ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "absinthe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "absinth",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "absinthe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "absinthes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "absinths",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "absinthes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "absinthe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "absinth's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "absinthe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "academize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "academize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "academise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "academized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "academized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "academised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "academizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "academizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "academising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acalephe ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acalephe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acalephae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "accessorize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "accessorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accessorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "accessorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accessorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "accessorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accessorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "accessorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accessorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acclimatizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acclimatizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acclimatisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acclimatization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acclimatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acclimatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "acclimatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acclimatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acclimatize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acclimatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acclimatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "acclimatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acclimatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "acclimatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acclimatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "acclimatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acclimatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acclimatizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acclimatizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acclimatiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "acclimatizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acclimatisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "accorage ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "accorage",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accourage",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "accoraged",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accouraged",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "accoraging",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accouraging",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "accorages",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accourages",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "accouter ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "accouter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accoutre",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "accoutered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accoutred",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "accoutering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accoutring",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "accouters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accoutres",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "accouterment ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "accouterment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accoutrement",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "accouterments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accoutrements",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "accouterment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accoutrement's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "accursed ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "accursed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "accurst",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: Some("ODE: archaic, M-W: \"or\" but can find little evidence of use"),
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Acemetae ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Acemetae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Acoemetae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Acemetae's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Acoemetae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Acemetic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Acemetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Acoemetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Acemetic's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Acoemetic's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Acer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Acer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Acre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Acer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Acre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acetonemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acetonemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "acetonaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acetonemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acetonemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "acetonaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "achenia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "achenia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "achaenia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "achenium ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "achenium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "achaenium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "acheniums",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "achaeniums",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "achenocarp ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "achenocarp",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "achaenocarp",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "achroiocythemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "achroiocythemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "achroiocythaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "achromatization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "achromatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "achromatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "achromatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "achromatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "achromatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "achromatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "achromatize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "achromatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "achromatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "achromatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "achromatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "achromatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "achromatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "achromatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "achromatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acidemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acidemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "acidaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acknowledgment ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acknowledgment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "acknowledgement",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "acknowledgments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "acknowledgements",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "acknowledgment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "acknowledgement's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acmesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acmesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acmaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acroanesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acroanesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acroanaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acroesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acroesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acroaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "actualization ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "actualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "actualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "actualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "actualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "actualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "actualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "actualize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "actualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "actualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "actualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "actualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "actualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "actualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "actualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "actualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "acuesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "acuesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "acuaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "adapter ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "adapter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "adaptor",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adapters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "adaptors",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adapter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "adaptor's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "addable ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "addable",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "addible",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "addenda ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "addenda",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "addendums",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "adjuster ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "adjuster",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adjustor",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adjusters",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adjustors",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adjuster's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adjustor's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "adonize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "adonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "adulterize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "adulterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adulterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adulterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adulterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adulterizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adulterising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adulterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adulterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "advene ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "advene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "advenae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "adverbialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "adverbialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adverbialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adverbialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adverbialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adverbializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adverbialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adverbializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adverbialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "advertise ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "advertise",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "advertize",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "advertised",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "advertized",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "advertising",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "advertizing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "advertises",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "advertizes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "advertisement ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "advertisement",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "advertizement",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "advertisements",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "advertizements",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "advertisement's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "advertizement's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "advertizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "advertizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "advertisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "adviser ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "adviser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "advisor",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "advisers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "advisors",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adviser's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "advisor's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "The oxford dictionary has this to say:",
    "The spellings adviser and advisor are both correct. Adviser is more",
    "common, but advisor is also widely used , especially in North",
    "America. Adviser may be seen as less formal, while advisor often",
    "suggests an official position.",
  ],
  },
Cluster {
  header: "adze ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "adze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adz",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adze's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adz's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Aedon ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Aedon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aaedon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Aedon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aaedon's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Aeetes ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Aeetes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aeaetes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Aeetes's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aeaetes's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aegis ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aegis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "egis",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aegises",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "egises",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aegis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "egis's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerialness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerialness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aaerialness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aerialness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aaerialness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aeric ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aeric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerie ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerie",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aery",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "eyry",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
      Variant {
        word: "eyrie",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aeries",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eyries",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aerie's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aery's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "eyry's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
      Variant {
        word: "eyrie's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aery",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerogram ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerogram",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aerogramme",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aerograms",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aerogrammes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerographer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerographer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areographer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerographic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerographic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areographic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerographical ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerographical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areographical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerography ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerography",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areography",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aerographies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areographies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aerography's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areography's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerologic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerological ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerologist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aerologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areologist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerology ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aerologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aerology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areology's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerometer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areometer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aerometers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areometers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aerometric ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aerometric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areometric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "areometric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "araeometric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Aerope ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Aerope",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aaerope",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Aerope's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aaerope's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aesthete ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aesthete",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "esthete",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aesthetes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "esthetes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aesthete's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "esthete's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aestheticize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aestheticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aestheticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aestheticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aestheticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aestheticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aestheticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aestheticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aestheticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "affeer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "affeer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "affere",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Africanization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Africanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Africanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Africanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Africanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Africanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Africanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Africanize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Africanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Africanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Africanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Africanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Africanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Africanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Africanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Africanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Afrikanerization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Afrikanerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Afrikanerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Afrikanerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Afrikanerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Afrikanerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Afrikanerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Afrikanerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Afrikanerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Afrikanerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Afrikanerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Afrikanerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "afterwards ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "afterwards",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "afterward",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adverb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "agathodemon ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "agathodemon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agathodaemon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "agatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "agatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ageism ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ageism",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agism",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "agenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "agenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ager ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ager",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aggrandizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aggrandizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aggrandizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aggrandizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aggrandization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aggrandization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aggrandize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aggrandize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aggrandized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aggrandizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aggrandizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aggrandizement ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aggrandizement",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandisement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aggrandizements",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandisements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aggrandizement's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandisement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aggrandizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aggrandizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aggrandizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aggrandizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aggrandiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aging ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aging",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ageing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ageings",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aging's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ageing's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "agnize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "agnize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agnise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agnized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agnised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agnizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agnising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agnizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agnises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "agonize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "agonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "agonizedly ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "agonizedly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agonisedly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agonizedlies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agonisedlies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "agonizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "agonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agoniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agonizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agonisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "agonizingly ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "agonizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agonisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agonizinglier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agonisinglier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agonizingliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agonisingliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "agrarianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "agrarianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agrarianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agrarianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agrarianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "agrize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "agrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agrizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agrising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "agrizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "agrises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aguize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aguize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aguise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aguized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aguised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aguizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aguising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aguizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aguises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aiger ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aiger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aigre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "airfoil ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "airfoil",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aerofoil",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "airfoils",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aerofoils",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "airfoil's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aerofoil's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "airgel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "airgel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aerogel",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "air_gun ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "air gun",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "airgun",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "air_letter ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "air letter",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "airletter",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "airplane ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "airplane",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "aeroplane",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "airplanes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "aeroplanes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "airplane's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "aeroplane's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "airstrike ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "airstrike",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "air strike",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "airstrikes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "air strikes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "airstrike's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "air strike's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aker ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aker",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "akre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Aktistete ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Aktistete",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aktistetae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Aktistete's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aktistetae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "albitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "albitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "albitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "albitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "albitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "albumenization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "albumenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminization",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "albumenizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminizations",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "albumenization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminization's",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "albumenize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "albumenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminize",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "albumenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminized",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "albumenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminizing",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "albumenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albuminizes",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "albumenizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "albumenizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "albumeniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alchemize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alchemize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alchemise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alchemized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alchemised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alchemizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alchemising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alchemizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alchemises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Alcide ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Alcide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Alcidae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Alcide's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Alcidae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Alcithoe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Alcithoe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Alcithoae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Alcithoe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Alcithoae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alcoholizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alcoholizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alcoholisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alcoholizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alcoholisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alcoholizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alcoholisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alcoholization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alcoholization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alcoholisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alcoholizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alcoholisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alcoholization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alcoholisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alcoholize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alcoholize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alcoholise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alcoholized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alcoholised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alcoholizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alcoholising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alcoholizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alcoholises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aleucemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aleucemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aleucaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aleukemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aleukemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aleukaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Aleus ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Aleus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aloeus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Aleus's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aloeus's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "algebraization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "algebraization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "algebraisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "algebraizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "algebraisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "algebraization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "algebraisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "algebraize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "algebraize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "algebraise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "algebraizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "algebraises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "algesthesis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "algesthesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "algaesthesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alienize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alienize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alienise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alienizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alienises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "align ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "align",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aline",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aligned",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alined",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aligning",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alining",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aligns",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alines",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alignment ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alignment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alinement",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alignments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alinements",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alignment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alinement's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alkalinization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alkalinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alkalinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alkalinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alkalinize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alkalinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alkalinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alkalinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alkalinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alkalizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alkalizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alkalization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alkalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alkalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alkalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alkalize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alkalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alkalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alkalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alkalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alkalizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alkalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alkalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alkaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "allegorization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "allegorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "allegorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "allegorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "allegorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "allegorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "allegorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "allegorize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "allegorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "allegorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "allegorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "allegorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "allegorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "allegorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "allegorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "allegorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "allegorizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "allegorizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "allegoriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "allegorizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "allegorisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "allegorizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "allegoriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alleluia ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alleluia",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "halleluiah",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alleluias",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "halleluiahs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alleluia's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "halleluiah's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alloxuremia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alloxuremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alloxuraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "almanac ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "almanac",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "almanack",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "almanacs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "almanacks",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alphabetization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alphabetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alphabetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alphabetizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alphabetisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alphabetization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alphabetisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alphabetize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alphabetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alphabetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alphabetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alphabetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alphabetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alphabetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alphabetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alphabetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alphabetizer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alphabetizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alphabetiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alphabetizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alphabetisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alphabetizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alphabetiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alternize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alternize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alternise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "alternizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alternises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "althea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "althea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "althaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "altheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "althaeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "althea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "althaea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "althein ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "althein",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "althaein",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "although ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "although",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "altho",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aluminization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aluminization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aluminisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aluminize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aluminize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aluminise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aluminized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aluminised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aluminizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aluminising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aluminizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aluminises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aluminum ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aluminum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aluminium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aluminums",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aluminiums",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aluminum's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aluminium's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alumni ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alumni",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alumnuses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "alveole ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "alveole",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "alveolae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amalgamatize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amalgamatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amalgamatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amalgamatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amalgamatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amalgamization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amalgamization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amalgamisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amalgamizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amalgamisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amalgamization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amalgamisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amalgamize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amalgamize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amalgamise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amalgamizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amalgamises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Amalthea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Amalthea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Amalthaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Amalthea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Amalthaea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ambiance ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ambiance",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ambience",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ambiances",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ambiences",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ambiance's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ambience's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ambicolorate ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ambicolorate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ambicolourate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ambicoloration ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ambicoloration",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ambicolouration",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ambilevous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ambilevous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ambilaevous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ameban ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ameban",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoeban",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amebean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amebean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amebian ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amebian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amebiasis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amebiasis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebiasis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amebiases",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebiases",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amebicidal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amebicidal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebicidal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amebicide ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amebicide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebicide",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amebid ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amebid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amebiform ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amebiform",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebiform",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amebiosis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amebiosis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebiosis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amebocyte ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amebocyte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebocyte",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amebocytes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebocytes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amebocyte's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebocyte's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ameboidism ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ameboidism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoeboidism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ameboidism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoeboidism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amebous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amebous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amebula ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amebula",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amoebula",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amenorrhea ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amenorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amenorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amenorrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amenorrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amenorrhea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amenorrhoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amenorrheal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amenorrheal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amenorrhoeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amenorrheic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amenorrheic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amenorrhoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Americanization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Americanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Americanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Americanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Americanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Americanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Americanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "americanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "americanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Americanize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Americanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Americanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Americanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Americanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Americanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Americanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Americanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Americanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "americanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "americanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Americanizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Americanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Americaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Americanizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Americanisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Americanizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Americaniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amid ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amidst",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amidst",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("literary variant"),
  comment: None,
  },
  ],
  notes: &[
    "amidst literary variant according to oxford dictionaries",
  ],
  },
Cluster {
  header: "ammocete ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ammocete",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ammocoete",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ammocetes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ammocoetes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ammocete's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ammocoete's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ammoniemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ammoniemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ammoniaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amoeba ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amoeba",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ameba",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amoebas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amoebae",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amebas",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amebae",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amoeba's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ameba's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "If ameba not variant than: A2: amebas / Av2: amebae",
  ],
  },
Cluster {
  header: "amoebic ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amoebic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amebic",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amoeboid ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amoeboid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ameboid",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amok ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amok",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amuck",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amoks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amucks",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amok's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amuck's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "among ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "among",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "amongst",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amoralize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amoralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amoralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amoralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amoralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amoret ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amoret",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amouret",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amorism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amorism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amourism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amorisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amourisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amorism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amourism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amorist ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amorist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amourist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amorists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amourists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amorist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amourist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amoristic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amoristic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amouristic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amoristics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amouristics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amorphization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amorphization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amorphisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amorphize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amorphize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amorphise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amortizable ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amortizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amortizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amortizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amortizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amortizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amortization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amortization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amortizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amortization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amortize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amortize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amortized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amortizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amortizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amortizement ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amortizement",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortisement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amortizements",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortisements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amortizement's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amortisement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amphigean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amphigean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amphigaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amphitheater ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amphitheater",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amphitheatre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amphitheaters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amphitheatres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "amphitheater's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "amphitheatre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ampule ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ampule",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ampoule",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ampul",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ampules",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ampoules",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ampuls",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ampule's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ampoule's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ampul's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amygdale ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amygdale",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amygdalae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "amyxorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "amyxorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amyxorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anabaptize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anabaptize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anabaptise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anabaptized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anabaptised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anabaptizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anabaptising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anabaptizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anabaptises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anagrammatize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anagrammatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anagrammatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anagrammatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anagrammatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anagrammatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anagrammatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anagrammatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anagrammatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "analog ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "analogue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analog",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("analogous"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analogue's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analog's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("analogous"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analogues",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analogs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("analogous"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analog",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "analogue",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("vs. digital"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analog's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "analogue's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("vs. digital"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analogs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "analogues",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("vs. digital"),
  comment: None,
  },
  ],
  notes: &[
    "Oxford thinks the US spelling of analogue is always analog, but M-W",
    "gives analogue as the correct spelling when relating to analogous,",
    "and it doesn't even offer analog as a variant.",
    "Macquarie prefers analog to analogue in the sense \"vs digital\".",
  ],
  },
Cluster {
  header: "analogize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "analogize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analogise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analogized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analogised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analogizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analogising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analogizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analogises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "analytic ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "analytic",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("linguistics"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analytical",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analytic",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "analytically ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "analytically",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analyticalally",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "analyticiti ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "analyticiti",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analyticaliti",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analyticitis",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analyticalitis",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "analyticity ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "analyticity",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analyticality",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "analyzability ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "analyzability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analysability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analyzability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analysability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "analyzable ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "analyzable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analysable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analyzabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analysabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analyzablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analysablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "analyzation ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "analyzation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analysation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analyzations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analysations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analyzation's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analysation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "analyze ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "analyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analyses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: analysis"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "analyzer ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "analyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analyzers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analysers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "analyzer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "analyser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anapestically ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anapestically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anapaestically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anapneic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anapneic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anapnoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anarchize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anarchize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anarchise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anarchized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anarchised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anarchizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anarchising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anarchizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anarchises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anathematization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anathematization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anathematisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anathematizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anathematisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anathematization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anathematisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anathematize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anathematize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anathematise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anathematized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anathematised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anathematizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anathematising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anathematizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anathematises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anathematizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anathematizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anathematiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anathematizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anathematiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anatomizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anatomizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anatomisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anatomization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anatomization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anatomisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anatomization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anatomisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anatomize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anatomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anatomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anatomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anatomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anatomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anatomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anatomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anatomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anatomizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anatomizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anatomiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anatomizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anatomiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Anchinoe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Anchinoe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Anchinoae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Anchinoe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Anchinoae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ancille ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ancille",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ancillae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anematosis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anematosis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaematosis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anemia ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anemia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anemic ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anemicer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaemicer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anemicest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaemicest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anemics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaemics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anemically ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anemically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaemically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aneretic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aneretic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaeretic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anoesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesthesia ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthesia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesthesiant ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesthesiant",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthesiant",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesthesiologist ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesthesiologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthesiologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthesiologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthesiologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthesiologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthesiologist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesthesiology ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesthesiology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthesiology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthesiologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthesiologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthesiology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthesiology's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesthesis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesthesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesthetic ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anestheticer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaestheticer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anestheticest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaestheticest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthetics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthetics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthetic's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthetic's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesthetically ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesthetically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthetically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesthetist ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesthetist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthetist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthetists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthetists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthetist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthetist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesthetization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesthetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetization",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anesthetisation",
        types: &[
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthetizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetizations",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anesthetisations",
        types: &[
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthetization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetization's",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anesthetisation's",
        types: &[
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesthetize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesthetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetize",
        types: &[
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anesthetise",
        types: &[
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetized",
        types: &[
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anesthetised",
        types: &[
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetizing",
        types: &[
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anesthetising",
        types: &[
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaesthetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetizes",
        types: &[
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anesthetises",
        types: &[
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesthetizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesthetizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetizer",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anesthetiser",
        types: &[
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anesthetizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthetizers",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anesthetisers",
        types: &[
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anesthyl ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anesthyl",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anaesthyl",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anestri ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anestri",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anoestri",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anestrous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anestrous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anoestrous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anestrum ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anestrum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anoestrum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anestra",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anoestra",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anestrus ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anestrus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anoestrus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anestruses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anoestruses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anestrus's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anoestrus's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anoetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anetiological ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anetiological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anaetiological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aneurysm ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aneurysm",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aneurism",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aneurysms",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aneurisms",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aneurysm's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aneurism's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "angelicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "angelicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "angelicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "angelicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "angelicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "angelize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "angelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "angelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "angelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "angelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Anglicanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Anglicanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Anglicanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Anglicanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Anglicanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Anglicanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Anglicanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anglicization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anglicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anglicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anglicizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anglicisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anglicization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anglicisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anglicize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anglicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anglicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anglicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anglicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anglicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anglicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anglicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anglicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "angrecum ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "angrecum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "angraecum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "angularization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "angularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "angularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "angularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "angularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "angularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "angularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "angularize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "angularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "angularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "angularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "angularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anhematopoiesis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anhematopoiesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anhaematopoiesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anhematopoiesis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anhaematopoiesis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anhematosis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anhematosis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anhaematosis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anhemolytic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anhemolytic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anhaemolytic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anhydremia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anhydremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anhydraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anhydremic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anhydremic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anhydraemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anhydridization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anhydridization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anhydridisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anhydridizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anhydridisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anhydridization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anhydridisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anhydridize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anhydridize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anhydridise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anhydridizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anhydridises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "animalization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "animalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "animalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "animalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "animalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "animalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "animalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "animalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "animalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "animalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "animalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "animalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "animalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "animalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "animalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "animalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "animized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "animized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "animised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "annalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "annalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "annalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "annalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "annalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "annalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "annalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "annalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "annalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "annualization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "annualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "annualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "annualize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "annualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "annualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "annualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "annualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "annualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "annualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "annualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "annualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anodization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anodization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anodisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anodize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anodize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anodise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anodized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anodised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anodizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anodising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anodizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anodises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anonymize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anonymize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anonymise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anonymized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anonymised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anonymizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anonymising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anonymizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anonymises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Anora ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Anora",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Anoura",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Anora's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Anoura's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anoxemia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anoxemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anoxaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anoxemia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anoxaemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anoxemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anoxemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "anoxaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anserine ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anserine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anserinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antagonizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antagonizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antagonisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antagonization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antagonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antagonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antagonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antagonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antagonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antagonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antagonize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antagonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antagonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antagonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antagonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antagonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antagonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antagonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antagonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antagonizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antagonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antagoniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antagonizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antagonisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antenna ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antennae",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antennas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: antenna"),
  description: Some("insect "),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antennas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antennae",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: antenna"),
  description: Some("TV "),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anthecological ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anthecological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthoecological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anthecologist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anthecologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthoecologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anthecology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anthecology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthoecology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anthologization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anthologization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthologisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anthologize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anthologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anthologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anthologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anthologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anthologizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anthologizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthologiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anthracemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anthracemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthracaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anthracitization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anthracitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthracitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anthracitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthracitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anthracitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthracitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anthropomorphization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anthropomorphization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthropomorphisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anthropomorphizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthropomorphisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anthropomorphization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthropomorphisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anthropomorphize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anthropomorphize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthropomorphise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anthropomorphized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthropomorphised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anthropomorphizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthropomorphising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anthropomorphizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anthropomorphises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anticatalyzer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anticatalyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anticatalyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anticatalyzers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anticatalysers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anticatalyzer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anticatalyser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anticentralization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anticentralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anticentralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anticentralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anticentralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anticentralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anticentralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antiepicenter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antiepicenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antiepicentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antiepicenters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antiepicentres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antiepicenter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antiepicentre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antifertilizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antifertilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antifertiliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antifertilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antifertilisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antifertilizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antifertiliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antilabor ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antilabor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "antilabour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antilabors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "antilabours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antilabor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "antilabour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antilemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antilemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antiloemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antimedieval ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antimedieval",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antimediaeval",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antimedievalism ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antimedievalism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antimediaevalism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antimedievalism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antimediaevalism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antimedievalist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antimedievalist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antimediaevalist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antimedievalist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antimediaevalist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antimedievally ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antimedievally",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antimediaevally",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antioxidizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antioxidizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antioxidiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antioxidizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antioxidisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antioxidizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antioxidiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antioxidizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antioxidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antioxidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antioxidizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antioxidisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antioxidizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antioxidising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antipathize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antipathize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antipathise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antipathizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antipathises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antiquarianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antiquarianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antiquarianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antiquarianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antiquarianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antirumor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antirumor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "antirumour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antirumors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "antirumours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antirumor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "antirumour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antisensitize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antisensitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antisensitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antisensitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antisensitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antisensitizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antisensitizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antisensitiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antisensitizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antisensitisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antisensitizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antisensitiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antisepticize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antisepticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antisepticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antisepticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antisepticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antisepticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antisepticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antisepticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antisepticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antiseptize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antiseptize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antiseptise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antiseptizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antiseptises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antisiphon ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antisiphon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antisyphon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antisiphons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antisyphons",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antisiphon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antisyphon's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "antithesize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "antithesize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antithesise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "antithesizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "antithesises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "anvil ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "anviled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anvilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "anviling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "anvilling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apesthetic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apheresis ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apheresis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aphaeresis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aphereses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aphaereses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apheresis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aphaeresis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apheretic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apheretic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aphaeretic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aphetize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aphetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aphetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aphetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aphetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aphetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aphetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aphetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aphetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aphorize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aphorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aphorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aphorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aphorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aphorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aphorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aphorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aphorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aphorizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aphorizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aphoriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aphorizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aphorisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aphorizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aphoriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apnea ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apnea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "apnoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apneas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "apnoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apnea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "apnoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apneal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apneal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "apnoeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apneic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apneic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "apnoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apocenter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apocenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apocentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apogeic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apogeic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apogaeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apologize ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apologizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apologizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apologiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apologizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apologisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apologizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apologiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apophthegmatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apophthegmatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apophthegmatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apophthegmatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apophthegmatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apophthegmatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apophthegmatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apophthegmatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apophthegmatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aporrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aporrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aporrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apostatization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apostatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apostatize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apostatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apostatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apostatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apostatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apostolize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apostolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apostolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apostolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apostolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apostrophize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apostrophize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostrophise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apostrophized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostrophised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apostrophizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostrophising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apostrophizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apostrophises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apothegmatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apothegmatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apothegmatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apothegmatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apothegmatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apothegmatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apothegmatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apothegmatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apothegmatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apotheosize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apotheosize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apotheosise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apotheosized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apotheosised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apotheosizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apotheosising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apotheosizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apotheosises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "appall ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "appall",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "appal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "appalls",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "appals",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apparel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "appareled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apparelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "appareling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apparelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "appendix ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "appendices",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appendixes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: appendix"),
  description: Some("book "),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "appendixes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appendices",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: appendix"),
  description: Some("organ "),
  comment: None,
  },
  ],
  notes: &[
    "A few sources say the usage depends on context. M-w says appendixes",
    "is the preferred form and does not distinguish based on context, Ox",
    "(all versions) say appendices is the preferred form.",
  ],
  },
Cluster {
  header: "appetize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "appetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "appetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "appetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "appetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "appetizement ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "appetizement",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appetisement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "appetizements",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appetisements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "appetizement's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appetisement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "appetizer ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "appetizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appetiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "appetizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appetisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "appetizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appetiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "appetizingly ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "appetizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appetisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apprizal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apprizal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apprisal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apprize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apprize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apprise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("value highly"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apprize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apprise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("value highly"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apprizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apprising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("value highly"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apprized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apprised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("value highly"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apprizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apprises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("value highly"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apprise",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("inform"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apprise's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("inform"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apprising",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("inform"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apprised",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("inform"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apprises",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("inform"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apprizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apprizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "appriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "apprizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apprisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "apprizings ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "apprizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "apprisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aquaculture ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aquaculture",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aquiculture",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aquaculture's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aquiculture's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Arabianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Arabianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arabianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arabianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arabianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arabianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arabianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Arabicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Arabicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arabicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arabicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arabicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arabization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arabization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arabisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arabizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arabisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arabization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arabisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Arabize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Arabize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arabise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arabized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arabised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arabizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arabising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arabizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arabises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arabize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arabise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arabized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arabised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arabizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arabising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arabizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arabises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Aramean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Aramean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Aramaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arameans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Aramaeans",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arbalester ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arbalester",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arbalestre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arbor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arbor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "arbour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arbors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "arbours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arbores",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "arboures",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arbor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "arbour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arbored ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arbored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arboured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arborization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arborization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arborisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arborizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arborisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arborization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arborisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arborize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arborize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arborise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arborized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arborised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arborizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arborising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arborizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arborises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archaeologically ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archaeologically",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archeologically",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archaeologist ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archaeologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archeologist",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archaeologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archeologists",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archaeologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archeologist's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archaeology ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archaeology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archeology",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archaeologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archeologies",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archaeology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archeology's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archaize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archaize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archaise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archaized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archaised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archaizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archaising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archaizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archaises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archaizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archaizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archaiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archaizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archaisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archaizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archaiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arche ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arche",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archeal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archeal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archeoastronomy ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archeoastronomy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaeoastronomy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archeoastronomies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaeoastronomies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archeocyte ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archeocyte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaeocyte",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archeocyte's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaeocyte's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archeol ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archeol",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaeol",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archeolithic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archeolithic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaeolithic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archeologian ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archeologian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaeologian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archeologic ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archaeologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archeologic",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archeological ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archaeological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archeological",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archeometry ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archeometry",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaeometry",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archeometries",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaeometries",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Archeozic ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Archeozic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Archaeozic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Archeozoic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Archeozoic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Archaeozoic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archeozoic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaeozoic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archetypal ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archetypal",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archetypical",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archetypally ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archetypally",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "archetypically",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "archeus ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "archeus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaeus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "archei",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "archaei",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arcticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arcticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arcticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arcticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arcticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Arctogea ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Arctogea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Arctogaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arctogea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Arctogaea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Arctogean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Arctogean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Arctogaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arctogean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Arctogaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Arctogeic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Arctogeic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Arctogaeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arctogeic's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Arctogaeic's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ardor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ardor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "ardour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ardors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "ardours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ardor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "ardour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arenicolor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arenicolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "arenicolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "areole ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "areole",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "areolae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "areometrical ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "areometrical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "araeometrical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "areostyle ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "areostyle",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "araeostyle",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "areostyles",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "araeostyles",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "areosystyle ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "areosystyle",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "araeosystyle",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Arianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Arianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arianized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arianised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arianizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arianising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Arimathea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Arimathea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Arimathaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arimathea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Arimathaea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Arimathean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Arimathean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Arimathaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arimathean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Arimathaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ariste ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ariste",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aristae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arithmetization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arithmetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arithmetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arithmetizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arithmetisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arithmetization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arithmetisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arithmetize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arithmetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arithmetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arithmetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arithmetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arithmetized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arithmetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arithmetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "armor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "armor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "armored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "armoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "armors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "armor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "armorbearer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "armorbearer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armourbearer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "armorbearer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armourbearer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "armorer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "armorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "armorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "armorer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "armoried ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "armoried",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armouried",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "armorless ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "armorless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "armory ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "armory",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armoury",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "armories",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armouries",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: Some("Ox claims \"armories\" is the British plural but that has got to be a mistake "),
  },
  Entry {
    variants: &[
      Variant {
        word: "armory's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "armoury's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aromatite ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aromatite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aromatitae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aromatization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aromatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aromatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aromatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aromatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aromatize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aromatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aromatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aromatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aromatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aromatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aromatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aromatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aromatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aromatizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aromatizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aromatiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aromatizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aromatiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arsenicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arsenicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arsenicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arsenicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arsenicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Arsinoe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Arsinoe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arsinoae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Arsinoe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Arsinoae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arterialization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arterialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arterialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arterializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arterialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arterialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arterialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arterialize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arterialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arterialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arterialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arterialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arterializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arterialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arterializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arterialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arthrempyesis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arthrempyesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arthroempyesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "artifact ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "artifact",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "artefact",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "artifact's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "artefact's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "artifacts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "artefacts",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "artificialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "artificialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "artificialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "artificialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "artificialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "artificializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "artificialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "artificializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "artificialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Aryanization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Aryanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aryanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Aryanize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Aryanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aryanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Aryanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aryanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Aryanized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aryanised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Aryanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aryanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Aryanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aryanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Aryanizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Aryanising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "arytenoid ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "arytenoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arytaenoid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "arytenoids",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arytaenoids",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "asafetida ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "asafetida",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asafoetida",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "asafetidas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asafoetidas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "asafetida's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asafoetida's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ascendance ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ascendance",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ascendence",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ascendance's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ascendence's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ascendancy ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ascendancy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ascendency",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ascendancy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ascendency's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ascendant ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ascendant",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ascendent",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ascendants",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ascendents",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ascendant's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ascendent's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Asclepiade ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Asclepiade",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Asclepiadae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Asclepiade's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Asclepiadae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Asel ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Asel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Asael",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Asel's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Asael's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "asepticize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "asepticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asepticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "asepticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asepticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "asepticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asepticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "asepticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asepticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "asexualization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "asexualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asexualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "asexualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asexualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "asexualize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "asexualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asexualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "asexualized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "asexualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asexualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "asexualizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "asexualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asexualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Asiaticization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Asiaticization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Asiaticisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Asiaticizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Asiaticisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Asiaticization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Asiaticisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Asiaticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Asiaticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Asiaticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Asiaticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Asiaticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Asiaticize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Asiaticise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Asmonean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Asmonean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Asmonaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Asmonean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Asmonaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aspheterize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aspheterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aspheterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aspheterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aspheterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aspheterizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aspheterising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aspheterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aspheterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ass ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ass",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("donkey"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "asses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("doneky"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ass's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("donkey"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ass",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("buttocks"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "asses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("buttocks"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ass's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arse's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("buttocks"),
  comment: None,
  },
  ],
  notes: &[
    "ass and arse have additional regional specific meanings and are",
    "considered seperate words when not used to refer to a donkey or",
    "someone buttock",
  ],
  },
Cluster {
  header: "assafetida ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "assafetida",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "assafoetida",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "assafetidas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "assafoetidas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "asshole ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "asshole",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arsehole",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "assholes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arseholes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "asshole's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "arsehole's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Assidean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Assidean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Assidaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Assidean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Assidaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Assyrianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Assyrianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Assyrianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Assyrianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Assyrianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Assyrianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Assyrianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Astera ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Astera",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Astrea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Astera's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Astrea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Astrea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Astraea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Astrea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Astraea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "astigmatizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "astigmatizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "astigmatiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "astigmatizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "astigmatisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "astigmatizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "astigmatiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "astrean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "astrean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "astraean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "astronomize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "astronomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "astronomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "astronomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "astronomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "astronomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "astronomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "astronomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "astronomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "asynchronize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "asynchronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asynchronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "asynchronized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asynchronised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "asynchronizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asynchronising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "asynchronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "asynchronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "atheize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "atheize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atheise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atheized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atheised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atheizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atheising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atheizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atheises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "athetize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "athetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "athetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "athetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "athetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "athetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "athetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "athetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "athetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "atmolyze ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "atmolyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atmolyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atmolyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atmolysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atmolyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atmolysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atmolyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atmolyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "atomizability ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "atomizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atomisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "atomizable ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "atomizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atomisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "atomization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "atomization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atomisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atomizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atomisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atomization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atomisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "atomize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "atomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "atomizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "atomizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atomiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atomizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atomisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atomizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atomiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "atroceruleus ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "atroceruleus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atrocoeruleus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "attemper ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "attemper",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "attempre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Atticize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Atticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Atticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Atticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Atticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Atticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Atticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Atticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Atticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "atticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "atticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "attitudinization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "attitudinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "attitudinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "attitudinize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "attitudinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "attitudinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "attitudinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "attitudinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "attitudinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "attitudinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "attitudinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "attitudinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "attitudinizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "attitudinisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "attitudinizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "attitudinizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "attitudiniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "attitudinizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "attitudinisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "attitudinizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "attitudiniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aunter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aunter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "auntre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "auntie ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "auntie",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aunty",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aunties",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "auntie's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aunty's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aureole ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aureole",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aureola",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aureole's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aureola's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Australianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Australianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Australianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Australianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Australianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Australianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Australianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Australopithecine ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Australopithecine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Australopithecinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Australopithecine's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Australopithecinae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Austrianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Austrianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Austrianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Austrianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Austrianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Austrianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Austrianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "autecious ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "autecious",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autoecious",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "auteciously ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "auteciously",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autoeciously",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "auteciousness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "auteciousness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autoeciousness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "auteciousness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autoeciousness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "autecism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "autecism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autoecism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autecisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autoecisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autecism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autoecism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "autecy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "autecy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autoecy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "authorizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "authorizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "authorisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "authorization ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "authorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "authorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "authorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "authorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "authorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "authorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "authorize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "authorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "authorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "authorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "authorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "authorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "authorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "authorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "authorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "authorizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "authorizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "authoriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "authorizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "authorisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "authorizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "authoriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "autocatalyze ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "autocatalyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autocatalyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autocatalyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autocatalysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autocatalyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autocatalysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autocatalyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autocatalyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "autodialer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "autodialer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autodialler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "autoimmunization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "autoimmunization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autoimmunisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "autoionization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "autoionization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autoionisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autoionization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autoionisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "autolyzate ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "autolyzate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autolysate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autolyzates",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autolysates",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "autolyze ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "autolyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autolyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autolyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autolysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autolyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autolysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autolyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autolyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "automatization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "automatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "automatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "automatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "automatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "automatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "automatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "automatize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "automatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "automatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "automatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "automatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "automatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "automatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "automatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "automatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Autonoe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Autonoe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Autonoae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Autonoe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Autonoae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "autotomize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "autotomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autotomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autotomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autotomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autotomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autotomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autotomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autotomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "autotoxemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "autotoxemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autotoxaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autotoxemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autotoxaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "autotoxemia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "autotoxaemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "avianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "avianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "avianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "avianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "avianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "avianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "avianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "avianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "avianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "avizandum ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "avizandum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "avisandum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "avizandums",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "avisandums",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "avize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "avize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "avise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "avized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "avised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "avizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "avising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "avizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "avises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "awestruck ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "awestruck",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "awestricken",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "AWOL ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "AWOL",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "awol",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ax ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ax",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "axe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ax's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "axe's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "axiomatization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "axiomatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "axiomatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "axiomatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "axiomatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "axiomatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "axiomatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "axiomatize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "axiomatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "axiomatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "axiomatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "axiomatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "axiomatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "axiomatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "axiomatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "axiomatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "aye ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aye",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ay",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aye's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ay's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "azotemia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "azotemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "azotaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "azotization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "azotization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "azotisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "azotize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "azotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "azotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "azotized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "azotised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "azotizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "azotising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "azotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "azotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "azotorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "azotorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "azotorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Babelization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Babelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Babelisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Babelization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Babelisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Babelize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Babelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Babelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Babelize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Babelise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Babelized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Babelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Babelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Babelized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Babelised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Babelizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Babelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Babelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Babelizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Babelising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Babylonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Babylonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Babylonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Babylonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Babylonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Babylonize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Babylonise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bachelorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bachelorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bachelorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bachelorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bachelorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bacillemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bacillemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bacillaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bacillemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bacillaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bacilli ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bacilli",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bacilluses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "backpedal ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "backpedaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "backpedalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "backpedaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "backpedalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "backward ",
  verified: false,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "backward",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "backward",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "backwards",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adverb), archaic: false, note: None,
  description: Some(""),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "baconize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "baconize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baconise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baconizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baconises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bacteremia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bacteremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bacteraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bacteremias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bacteraemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bacteremia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bacteraemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bacteria ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bacteria",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bacterias",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: bacteria"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bacteriemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bacteriemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bacteriaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bacterize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bacterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bacterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bacterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bacterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bacterizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bacterising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bacterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bacterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "baddie ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "baddie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baddy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baddie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baddy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Baer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Baer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Bare",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Baer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Bare's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "balk ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "balk",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "baulk",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balk's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "baulk's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balked",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "baulked",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balking",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "baulking",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "baulks",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Balkanization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Balkanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Balkanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Balkanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Balkanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Balkanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Balkanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balkanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balkanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balkanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balkanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balkanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balkanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Balkanize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Balkanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Balkanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Balkanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Balkanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Balkanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Balkanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Balkanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Balkanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balkanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balkanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balkanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balkanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balkanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balkanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balkanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balkanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "balladize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "balladize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balladise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balladized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balladised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balladizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balladising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balladizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balladises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "baloney ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "baloney",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boloney",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baloney's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boloney's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "balsamize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "balsamize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balsamise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "balsamizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "balsamises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "banalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "banalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "banalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "banalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "banalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "banalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "banalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "banalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "banalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "banalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "banalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "banalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "banalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "banalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bandanna ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bandanna",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bandana",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bandannas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bandanas",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bandanna's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bandana's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bandoleer ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bandoleer",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bandolier",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bandoleers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bandoliers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bandoleer's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bandolier's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "banister ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "banister",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bannister",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "banisters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bannisters",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "banister's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bannister's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "banshee ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "banshee",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "banshie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "banshees",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "banshies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "banshee's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "banshie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bantamize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bantamize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bantamise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bantamizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bantamises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "baptistery ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "baptistery",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptistry",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baptisteries",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptistries",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baptistery's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptistry's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "baptizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "baptizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baptizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baptizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "baptize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "baptize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baptized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baptizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baptizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "baptizement ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "baptizement",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptisement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baptizements",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptisements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baptizement's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptisement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "baptizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "baptizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baptizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baptizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baptiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "barbarianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "barbarianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbarianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barbarianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbarianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "barbarization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "barbarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barbarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barbarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "barbarize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "barbarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barbarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barbarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barbarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "barbecue ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "barbecue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbeque",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barbecued",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbequed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barbecuing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbequing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barbecues",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbeques",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barbecue's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barbeque's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "barcarole ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "barcarole",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barcarolle",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barcaroles",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barcarolles",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barcarole's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barcarolle's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bark ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bark",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bark's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bark",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barque",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: true, note: None,
  description: Some("ship"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barques",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: true, note: None,
  description: Some("ship"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bark's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barque's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: true, note: None,
  description: Some("ship"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "baronetize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "baronetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baronetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "baronetized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "baronetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baronetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "baronetizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "baronetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baronetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "baronize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "baronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "baronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "barrel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "barreled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barrelled",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "barreling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barrelling",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bartizan ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bartizan",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bartisan",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bartizans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bartisans",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bartizaned ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bartizaned",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bartisaned",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "barycenter ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "barycenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "barycentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bastardization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bastardization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bastardisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bastardizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bastardisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bastardization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bastardisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bastardize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bastardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bastardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bastardized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bastardised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bastardizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bastardising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bastardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bastardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bastinade ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bastinaded",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bastinadoed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bastinades",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bastinadoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bathyscaphe ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bathyscaphe",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bathyscaph",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bathyscaphes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bathyscaphs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bathyscaphe's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bathyscaph's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "battleaxe ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "battleaxe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "battleax",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "battleaxe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "battleax's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "battleaxes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "battologize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "battologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "battologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "battologized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "battologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "battologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "battologizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "battologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "battologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "beaverize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "beaverize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "beaverise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "beaverizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "beaverises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "beclamor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "beclamor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "beclamour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "beclamored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "beclamoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "beclamoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "beclamouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "beclamors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "beclamours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "becolor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "becolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "becolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "becudgel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "becudgeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "becudgelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "becudgeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "becudgelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bedevil ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bedeviled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bedevilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bedeviling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bedevilling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bedlamize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bedlamize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bedlamise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bedlamized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bedlamised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bedlamizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bedlamising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bedlamizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bedlamises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bedrivel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bedriveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bedrivelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bedriveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bedrivelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Beera ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Beera",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Berea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Beera's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Berea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "beeves ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "beeves",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "beefs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: beef"),
  description: Some("animal "),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "beefs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: beef"),
  description: Some("complaint "),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "befavor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "befavor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "befavour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "beglamor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "beglamor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "beglamour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "beglamored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "beglamoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "beglamoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "beglamouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "beglamors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "beglamours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "behavior ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "behavior",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behaviors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behavior's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "behavioral ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "behavioral",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behavioural",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behavioraler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviouraler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behavioralest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviouralest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "behaviorally ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "behaviorally",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviourally",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "behaviored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "behaviored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behavioured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "behaviorism ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "behaviorism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviourism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behaviorisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviourisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behaviorism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviourism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "behaviorist ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "behaviorist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviourist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behaviorists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviourists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behaviorist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviourist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "behavioristic ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "behavioristic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviouristic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behavioristicer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviouristicer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behavioristicest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviouristicest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behavioristics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "behaviouristics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Behmenism ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Behmenism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boehmenism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Behmenism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boehmenism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behmenism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boehmenism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Behmenist ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Behmenist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boehmenist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Behmenist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boehmenist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Behmenite ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Behmenite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boehmenite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Behmenite's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boehmenite's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "behoove ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "behoove",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "behove",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behooved",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "behoved",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behooving",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "behoving",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behooves",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "behoves",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behoove's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "behove's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behoovings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "behovings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "behooving's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "behoving's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "behoovingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "behoovingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "behovingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bejewel ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bejeweled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bejewelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bejeweling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bejewelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bel ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bael",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bels",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "baels",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bel's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bael's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "belabor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "belabor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "belabour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "belabored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "belaboured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "belaboring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "belabouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "belabors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "belabours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "belabor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "belabour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Belter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Belter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boelter",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Belter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boelter's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Belus ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Belus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boelus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Belus's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boelus's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bemedaled ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bemedaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bemedalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Beria ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Beria",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Baeria",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Beria's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Baeria's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "berk ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "berk",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burk",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "berks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Berl ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Berl",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Baerl",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Berl's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Baerl's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Berlinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Berlinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Berlinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Berlinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Berlinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Berlinize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Berlinise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "berloque ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "berloque",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "breloque",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "berme ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "berme",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "breme",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bern ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bern",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bren",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Bessemerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Bessemerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Bessemerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Bessemerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Bessemerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Bessemerize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Bessemerise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bestialize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bestialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bestialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bestialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bestialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bestializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bestialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bestializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bestialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bevel ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "beveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bevelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "beveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "beveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bevelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bevellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "beveler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "beveller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "beveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bevelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bevelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bevellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Beyer ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Beyer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Baeyer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Beyer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Baeyer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bias ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "biased",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "biassed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "biasing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "biassing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bichromatize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bichromatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bichromatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bichromatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bichromatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bicolor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bicolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bicolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bicolorrer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bicolourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bicolorrest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bicolourest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bicolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bicolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bicolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bicolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bicolored ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bicolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bicoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bicoloredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bicolouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bicoloreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bicoloureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bicolorous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bicolorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bicolourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bimetalist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bimetalist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bimetallist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bimetalistic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bimetalistic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bimetallistic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "binge ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bingeing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "binging",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "biocenology ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "biocenology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "biocoenology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "biocenose ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "biocenose",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "biocoenose",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "biocenosis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "biocoenosis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "biocenoses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "biocoenoses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "biocenotic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "biocenotic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "biocoenotic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "biographize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "biographize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "biographise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "biographizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "biographises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "biologize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "biologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "biologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "biologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "biologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "biosynthesized ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "biosynthesized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "biosynthesised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bipolarization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bipolarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bipolarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bipolarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bipolarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bipolarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bipolarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bipolarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Birminghamize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Birminghamize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Birminghamise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Birminghamizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Birminghamises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Birminghamize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Birminghamise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bister ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bister",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bistre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bisters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bistres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bister's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bistre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bistered ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bistered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bistred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bisteredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bistredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bistereddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bistreddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bittor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bittor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bittour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bittors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bittours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bituminization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bituminization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bituminisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bituminizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bituminisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bituminization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bituminisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bituminize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bituminize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bituminise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bituminized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bituminised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bituminizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bituminising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bituminizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bituminises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bizet ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bizet",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "biset",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "blaize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "blaize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "blaise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "blamable ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "blamable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "blameable",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "blastocele ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "blastocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "blastocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "blastule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "blastule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "blastulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "blennemesis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "blennemesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "blennoemesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "blennorrhea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "blennorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "blennorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "blond ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "blond",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "blonde",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("form generally used"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blonds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "blondes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("form generally used"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blond's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "blonde's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("form generally used"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blonder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("form generally used"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blondest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("form generally used"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blond",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("often masculine forms, especially in Brit. English"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blonder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("often masculine forms, especially in Brit. English"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blondest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("often masculine forms, especially in Brit. English"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blonds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("often masculine forms, especially in Brit. English"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blond's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("often masculine forms, especially in Brit. English"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blonde",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("often feminine forms, especially in Brit. English"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blonder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("often feminine forms, especially in Brit. English"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blondest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("often feminine forms, especially in Brit. English"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blondes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("often feminine forms, especially in Brit. English"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blonde's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("often feminine forms, especially in Brit. English"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "blowzy ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "blowzy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "blowsy",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blowzier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "blowsier",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "blowziest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "blowsiest",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bluing ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bluing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "blueing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bluing's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "blueing's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "boatswain ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "boatswain",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bosun",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bo'sun",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bos'n",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bo's'n",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boatswains",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bosuns",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bo's'ns",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bo'suns",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bos'ns",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boatswain's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bosun's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bo'sun's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bo's'n's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bos'n's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bobbysoxer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bobbysoxer",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bobby-soxer",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bobbysoxers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bobby-soxers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bobbysoxer's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bobby-soxer's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "boccie ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "boccie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bocce",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bocci",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boccie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bocce's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bocci's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Boedromius ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Boedromius",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boaedromius",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Boedromius's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boaedromius's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bogie ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bogie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bogey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bogies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bogeys",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bogie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bogey's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bogey",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bogy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bogie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bogeys",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bogies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bogeyed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bogied",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bogeying",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bogieing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bogey's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bogy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "bogie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "boled ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "boled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boloed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Bolshevize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Bolshevize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Bolshevise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Bolshevized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Bolshevised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Bolshevizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Bolshevising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bolshevize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bolshevise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bolshevized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bolshevised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bolshevizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bolshevising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bolshevizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bolshevises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bolshie ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bolshie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bolshy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bonderize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bonderize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bonderise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bonny ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bonny",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bonnie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bony ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bony",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boney",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "boo-boo ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "boo-boo",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "booboo",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boo-boos",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "booboos",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boo-boo's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "booboo's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "boogeyman ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "boogeyman",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boogerman",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boogeymen",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boogermans",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boogeyman's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boogerman's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "boombox ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "boombox",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boom box",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boomboxes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boom boxes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boombox's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boom box's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bootee ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bootee",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bootie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("boot"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bootees",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "booties",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("boot"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bootee's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bootie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("boot"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "booty",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bootie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("buttocks"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "booties",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("buttocks"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "booty's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bootie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("buttocks"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "booty",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("plunder taken"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "booties",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("plunder taken"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "booty's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("plunder taken"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "borage ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "borage",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bourage",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "borasque ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "borasque",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bourasque",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bord ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bord",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bourd",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bords",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bourds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "borg ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "borg",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bourg",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "borize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "borize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "borise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "borizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "borises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "borscht ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "borscht",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "borsch",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "borscht's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "borsch's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Boswellize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Boswellize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boswellise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Boswellized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boswellised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Boswellizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boswellising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Boswellizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Boswellises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "botanize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "botanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "botanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "botanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "botanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "botanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "botanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "botanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "botanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "botanizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "botanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "botaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "botanizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "botaniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Bougainvillea ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Bougainvillea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Bougainvillaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":upper"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Bougainvillea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Bougainvillaea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":upper"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bougainvillea",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bougainvillaea",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":lower"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bougainvilleas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bougainvillaeas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":lower"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "boulder ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "boulder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowlder",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boulders",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowlders",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boulder's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowlder's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "boulevardize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "boulevardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boulevardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boulevardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "boulevardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bourbonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bourbonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bourbonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bourbonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bourbonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bowdlerization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bowdlerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowdlerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bowdlerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowdlerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bowdlerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowdlerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bowdlerize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bowdlerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowdlerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bowdlerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowdlerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bowdlerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowdlerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bowdlerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowdlerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bowdlerizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bowdlerizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowdleriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bowdlerizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowdlerisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bowdlerizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowdleriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bowel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "boweled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "boweling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bowelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bradypnea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bradypnea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bradypnoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "breathalyze ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "breathalyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "breathalyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "breathalyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "breathalysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "breathalyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "breathalysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "breathalyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "breathalyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "breathalyzer ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "breathalyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "breathalyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "breathalyzer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "breathalyser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Breathalyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("trademark"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "breathalyzers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "breathalysers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "brier ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "brier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "briar",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "briers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "briars",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "brier's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "briar's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "brimful ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "brimful",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brimfull",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "brinkmanship ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "brinkmanship",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brinksmanship",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "brinkmanship's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brinksmanship's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "briquette ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "briquette",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "briquet",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "briquettes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "briquets",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "briquette's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "briquet's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "briza ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "briza",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brisa",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "briza's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brisa's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bromethylene ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bromethylene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bromoethylene",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "brominize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "brominize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brominise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "brominizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brominises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bromization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bromization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bromisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bromization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bromisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bromize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bromize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bromise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bromized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bromised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bromizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bromising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bronco ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bronco",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "broncho",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bronco's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "broncho's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "broncos",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bronchos",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "brusque ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "brusque",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brusk",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "brusquer",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brusker",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "brusquest",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bruskest",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "brusquely ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "brusquely",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bruskly",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "brusqueness ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "brusqueness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bruskness",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "brusquenesses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brusknesses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "brusqueness's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bruskness's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "brutalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "brutalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brutalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "brutalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brutalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "brutalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brutalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "brutalize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "brutalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brutalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "brutalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brutalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "brutalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brutalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "brutalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "brutalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Budenny ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Budenny",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Budaenny",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Budenny's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Budaenny's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bunco ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bunco",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bunko",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "buncos",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bunkos",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bunco's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bunko's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "buncoed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bunkoed",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "buncoing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bunkoing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bunkum ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bunkum",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "buncombe",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bunkums",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "buncombes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bunkum's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "buncombe's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bureaucratization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bureaucratization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bureaucratisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bureaucratizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bureaucratisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bureaucratization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bureaucratisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bureaucratize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bureaucratize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bureaucratise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bureaucratized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bureaucratised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bureaucratizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bureaucratising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bureaucratizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bureaucratises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "burglarize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "burglarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burglarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "burglarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burglarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "burglarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burglarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "burglarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burglarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "burka ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "burka",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burqa",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "burkha",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "burkas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burqas",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "burkhas",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "burka's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burqa's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "burkha's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "burnettize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "burnettize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burnettise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "burnettized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burnettised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "burnettizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burnettising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "burnettizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burnettises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "burnoose ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "burnoose",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burnous",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "burnooses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burnouses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "burnoose's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "burnous's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bus ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(""),
  comment: Some("to avoid confusion with buss"),
  },
  Entry {
    variants: &[
      Variant {
        word: "buses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "busses",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "busing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bussing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "busings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bussings",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "busing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bussing's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: Some(""),
  comment: Some("to avoid confusion with buss"),
  },
  Entry {
    variants: &[
      Variant {
        word: "buses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "busses",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "busing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bussing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bused",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "bussed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "buss",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "busses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bussing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bussed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bushel ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "busheled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bushelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "busheling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bushelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bushelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bushellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "busheler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "busheler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "busheller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bushelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "bushellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "busheler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "busheller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "bylaw ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "bylaw",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "byelaw",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bylaws",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "byelaws",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "bylaw's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "byelaw's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Byronize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Byronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Byronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Byronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Byronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Byronize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Byronise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Byzantinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Byzantinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Byzantinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Byzantinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Byzantinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Byzantinize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Byzantinise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cabby ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cabby",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cabbie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cabby's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cabbie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cabbies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cacesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cacesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cacaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cachemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cachemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cachaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cachemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cachemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cachaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cacodemon ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cacodemon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cacodaemon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cacodemons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cacodaemons",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cacodemoniac ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cacodemoniac",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cacodaemoniac",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cacodemonial ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cacodemonial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cacodaemonial",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cacodemonic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cacodemonic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cacodaemonic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cacoethes ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cacoethes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cacoaethes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cacoethes's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cacoaethes's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cadaster ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cadaster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cadastre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cadasters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cadastres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cadaster's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cadastre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cadaverize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cadaverize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cadaverise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cadaverizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cadaverises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cadmiumize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cadmiumize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cadmiumise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cadmiumizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cadmiumises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Caesarize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Caesarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Caesarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Caesarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Caesarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Caesarize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Caesarise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caffer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caffer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caffre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caftan ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caftan",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kaftan",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caftans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kaftans",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caftan's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kaftan's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cagey ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cagey",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cagy",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caginess ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caginess",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cageyness",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caginesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cageynesses",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caginess's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cageyness's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caiman ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caiman",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cayman",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caimans",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caymans",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caiman's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cayman's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "calander ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "calander",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calandre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "calcemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "calcemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "calcaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "calendarization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "calendarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calendarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calendarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calendarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "calendarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "calendarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calendarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calendarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calendarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calendarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calendarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calendarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calendarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Calender ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Calender",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Calendre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Calender's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Calendre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caliber ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caliber",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calibre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calibers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calibres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caliber's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calibre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caliper ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caliper",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calliper",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calipered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "callipered",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calipering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "callipering",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calipers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "callipers",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caliper's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calliper's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calipers's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "callipers's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caliph ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caliph",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calif",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "khalif",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caliph's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calif's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "khalif's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caliphs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "califs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "khalifs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "calisthenics ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "calisthenic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "callisthenic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calisthenics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "callisthenics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "calk ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "calk",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caulk",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calked",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caulked",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calking",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caulking",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caulks",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "calk's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caulk's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caulk",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calk",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caulked",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calked",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caulking",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calking",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caulks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calks",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caulk",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calk",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caulking",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "calking",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caulks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calks",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caulkings",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "calkings",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caulk's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calk's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caulking's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "calking's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Callirrhoe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Callirrhoe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Callirrhoae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Callirrhoe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Callirrhoae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "calodemon ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "calodemon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "calodaemon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Calvinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Calvinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Calvinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Calvinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Calvinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Calvinize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Calvinise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "camelhair ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "camelhair",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: Some("camel hair (camel's hair)"),
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "camisades ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "camisades",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "camisadoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "camize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "camize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "camise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Canadianization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Canadianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Canadianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Canadianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Canadianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Canadianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Canadianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Canadianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Canadianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Canadianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Canadianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Canadianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Canadianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Canadianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canaler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canaler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canaller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canalers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canallers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canalize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Cananean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Cananean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Cananaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Cananean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Cananaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cancel ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canceled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cancelled",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canceling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cancelling",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cancelable ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cancelable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cancellable",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cancelate ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cancelate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cancellate",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cancelated ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cancelated",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cancellated",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canceler ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canceler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canceller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cancelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cancellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canceler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canceller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cancellation ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cancellation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cancelation",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cancelous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cancelous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cancellous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "candor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "candor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "candour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "candors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "candours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "candor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "candour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canephore ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canephore",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canephorae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canephore",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canephoroe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canisterization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canisterization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canisterisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canisterizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canisterisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canisterize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canisterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canisterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canisterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canisterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canisterizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canisterising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canisterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canisterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cannibalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cannibalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cannibalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cannibalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cannibalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cannibalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cannibalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cannibalize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cannibalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cannibalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cannibalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cannibalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cannibalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cannibalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cannibalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cannibalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canonicalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canonicalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonicalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canonicalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canonicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canonicalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonicalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canonicalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonicalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canonicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canonization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canonize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "canonizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "canonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canoniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canonizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canonisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "canonizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "canoniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cantaloupe ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cantaloupe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cantaloup",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cantaloupes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cantaloups",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cantaloupe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cantaloup's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cantonization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cantonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cantonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cantonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cantonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cantonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cantonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cantonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cantonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cantonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cantonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cantonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cantonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cantonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "capitalizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "capitalizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capitalizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capitalizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capitalizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capitalizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "capitalization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "capitalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capitalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capitalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "capitalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "capitalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capitalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capitalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "capitalizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "capitalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capitalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capitalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capitaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Caponization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Caponization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Caponisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caponization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caponisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caponization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caponisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caponize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caponize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caponise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caponized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caponised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caponizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caponising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caponizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caponises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caponizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caponizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caponiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caponizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caponiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "capsulization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "capsulization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capsulisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "capsulize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "capsulize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capsulise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capsulized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capsulised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capsulizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capsulising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "capsulizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "capsulises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caracol ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caracoled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caracolled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caracoling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caracolling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caramelization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caramelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caramelisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caramelizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caramelisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caramelization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caramelisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caramelize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caramelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caramelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caramelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caramelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caramelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caramelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caramelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caramelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caravansary ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caravansary",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caravanserai",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caravansaries",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caravansarais",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caravansarai",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caravansary's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caravanserai's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carbolization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carbolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carbolize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carbolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carbonades ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carbonades",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonadoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carbonatization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carbonatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carbonizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carbonizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carbonization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carbonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carbonize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carbonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carbonizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carbonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carboniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carbonisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carbonizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carboniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carburetor ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carburetor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburettor",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburetter",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carburetors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburettors",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburetters",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carburetor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburettor's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburetter's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carburization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carburization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carburizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carburization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carburize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carburize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carburized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carburizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carburizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carburizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carburizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carburizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carburiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cardie ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cardie",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cardiae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cardueline ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cardueline",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carduelinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carnalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carnalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carnalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carnalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carnalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carnalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carnalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carnalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carnalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carny ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carny",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carney",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "carnie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carnies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carneys",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carny's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carney's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "carnie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carol ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caroled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carolled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caroling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carolling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caroler ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caroler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caroller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carolers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carollers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caroler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caroller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carotinemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carotinemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "carotinaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "carousel ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "carousel",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carrousel",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carousels",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carrousels",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "carousel's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "carrousel's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cartelization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cartelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cartelisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cartelizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cartelisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cartelization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cartelisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cartelize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cartelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cartelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cartelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cartelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cartelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cartelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cartelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cartelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "casino ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "casino",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "casinos",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "casino's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "casino",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cassino",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("card game"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "casinos",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cassinos",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("card game"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "casino's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cassino's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("card game"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caster ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "castor",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "casters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "castors",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caster's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "castor's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "castor",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "casters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "castors",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caster's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "castor's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "casters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caster's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "castor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":4"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "castors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":4"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "castor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":4"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "castorized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "castorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "castorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "castorizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "castoriseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "castorized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "castorised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "casualization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "casualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "casualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "casualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "casualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "casualize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "casualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "casualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "casualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "casualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "casualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "casualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "casualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "casualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "catabolize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "catabolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catabolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "catalog ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "catalog",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catalogue",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cataloged",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catalogued",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cataloging",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cataloguing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catalogs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catalogues",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catalog's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catalogue's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cataloger ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cataloger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cataloguer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catalogers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cataloguers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cataloger's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cataloguer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cataloguize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cataloguize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cataloguise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cataloguized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cataloguised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cataloguizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cataloguising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cataloguizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cataloguises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "catalyze ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "catalyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catalyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catalyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catalysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catalyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catalysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catalyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catalyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catalyses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: catalysis"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "catalyzer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "catalyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catalyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catalyzers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catalysers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catalyzer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catalyser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "catechizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "catechizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catechizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catechizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catechizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catechizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "catechization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "catechization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catechizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catechization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "catechize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "catechize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catechized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catechizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catechizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catechizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "catechizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "catechizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catechizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catechizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catechiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "categorization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "categorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "categorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "categorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "categorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "categorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "categorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "categorize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "categorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "categorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "categorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "categorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "categorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "categorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "categorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "categorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "categorizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "categorizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "categoriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "categorizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "categorisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "catharize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "catharize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catharise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catharized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catharised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catharizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catharising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catharizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catharises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "catheterization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "catheterization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catheterisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catheterizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catheterisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catheterization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catheterisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "catheterize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "catheterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catheterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catheterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catheterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catheterizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catheterising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catheterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catheterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Catholicization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Catholicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Catholicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catholicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catholicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catholicizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catholicisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catholicization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catholicisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Catholicized ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Catholicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Catholicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Catholicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Catholicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catholicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catholicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catholicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catholicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catholicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catholicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catholicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catholicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "catholicizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "catholicizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catholiciser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catholicizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catholicisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "catholicizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catholiciser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cauldron ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cauldron",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caldron",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cauldrons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caldrons",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cauldron's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caldron's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "causticization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "causticization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "causticisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "causticizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "causticisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "causticization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "causticisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "causticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "causticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "causticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "causticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "causticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "causticizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "causticizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "causticiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "causticizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "causticisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cauterization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cauterization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cauterisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cauterizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cauterisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cauterization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cauterisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cauterize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cauterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cauterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cauterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cauterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cauterizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cauterising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cauterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cauterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caviar ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caviar",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caviare",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caviars",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caviares",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caviar's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caviare's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cavil ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caviled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cavilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caviling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cavilling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cavilings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cavillings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cavilation ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cavilation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cavillation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "caviler ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "caviler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caviller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cavilers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cavillers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "caviler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caviller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cecally ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cecally",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caecally",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cecitis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cecitis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caecitis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cecitises",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caecitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cecity ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cecity",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caecity",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cecostomy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cecostomy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caecostomy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cecotomy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cecotomy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caecotomy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Celastraceae ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Celastraceae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Coelastraceae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Celastraceae's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Coelastraceae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "celastraceous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "celastraceous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coelastraceous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "celestialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "celestialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "celestialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "celestializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "celestialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "celestine ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "celestine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coelestine",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "celiac ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "celiac",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coeliac",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "celiacs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coeliacs",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "celialgia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "celialgia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coelialgia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "celiomyalgia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "celiomyalgia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coeliomyalgia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "celiorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "celiorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coeliorrhea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "celioscopy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "celioscopy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coelioscopy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "celiotomy ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "celiotomy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coeliotomy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "celom ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "celom",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coelom",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "celoms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coeloms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coelomata",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "celom's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coelom's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "celoma ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "celoma",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coeloma",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "celomata",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coelomata",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "celoscope ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "celoscope",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coeloscope",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "celoscope's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coeloscope's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Celticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Celticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Celticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Celticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Celticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Celticize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Celticise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenacle ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenacle",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenacle",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cenacle's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenacle's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenaculum ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenaculum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenaculum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenesthesia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cenesthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cenesthesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenesthesia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenesthesis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenesthesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenesthesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Cenis ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Cenis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Caenis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Cenis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Caenis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenobe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenobe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenobe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenobite ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenobite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenobite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cenobites",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenobites",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cenobite's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenobite's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenobitic ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenobitic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenobitic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenobitical ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenobitical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenobitical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenobitism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenobitism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenobitism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cenobitism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenobitism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenobium ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenobium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenobium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenoby ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenoby",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenoby",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenogenetic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenogenetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caenogenetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cenogenetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenogenetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenogenetically ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenogenetically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caenogenetically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenosite ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenosite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenosite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenospecies ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenospecies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenospecies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenospecific ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenospecific",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenospecific",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenospecifically ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenospecifically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coenospecifically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cenozoic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cenozoic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caenozoic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "center ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "center",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "centre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "centred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "centring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "center's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "centre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "centrings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centering's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "centring's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "centres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centerable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centerable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centreable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centreables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centreable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centerboard ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centerboard",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centreboard",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerboards",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centreboards",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerboard's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centreboard's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centerer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centerer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centerfold ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centerfold",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrefold",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerfolds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrefolds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerfold's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrefold's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centerless ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centerless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centreless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrelesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrelessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centerline ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centerline",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centreline",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerlines",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrelines",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerline's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centreline's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centermost ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centermost",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centremost",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centerpiece ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centerpiece",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrepiece",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerpieces",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrepieces",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centerpiece's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrepiece's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Centerville ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Centerville",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Centreville",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Centerville's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Centreville's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centigram ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centigram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centigramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centigrams",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centigrammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centigram's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centigramme's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centiliter ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centiliter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centilitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centiliters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centilitres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centiliter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centilitre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centimeter ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centimeter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "centimetre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centimeters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "centimetres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centimeter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "centimetre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centralization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centralize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centralizer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centralizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centraliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centralizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centralisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centralizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centraliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centrifugalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centrifugalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrifugalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centrifugalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrifugalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centrifugalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrifugalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "centrifugalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "centrifugalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrifugalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centrifugalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrifugalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centrifugalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrifugalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "centrifugalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "centrifugalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cephalization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cephalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cephalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cephalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cephalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cephalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cephalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cere",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cree",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ceres",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crees",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cere's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cree's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cera ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cera",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cerat ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cerat",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "creat",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ceratin ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ceratin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "creatin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ceratins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "creatins",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cerebralization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cerebralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cerebralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cerebralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cerebralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cerebralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cerebralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cerebralize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cerebralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cerebralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cerebralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cerebralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ceremonialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ceremonialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ceremonialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ceremonializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ceremonialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cerolite ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cerolite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "creolite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cerule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cerule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caerule",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ceruleolactite ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ceruleolactite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coeruleolactite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cesar ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cesar",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caesar",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cesarean ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cesarean",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caesarean",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caesarian",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "cesarian",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cesareans",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caesareans",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caesarians",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "cesarians",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cesarean's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caesarean's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caesarian's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "cesarian's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Cesaria ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Cesaria",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Caesaria",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Cesaria's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Caesaria's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cesious ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cesious",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "caesious",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cesium ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cesium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caesium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cesiums",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caesiums",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cesium's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caesium's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cespitose ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cespitose",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caespitose",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cespitosely ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cespitosely",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caespitosely",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cestus ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cestus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caestus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cestuses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caestuses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cesural ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cesural",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "caesural",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cetus ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cetus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "coetus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chaces ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chaces",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chacoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chalaze ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chalaze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chalazae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Chaldeans ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Chaldeans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Chaldaeans",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chameleonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chameleonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chameleonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chameleonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chameleonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Chamizal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Chamizal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Chamisal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Chamizal's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Chamisal's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chamois ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chamois",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shammy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "chammy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chamois",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chamoix",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "shammies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "chammies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl"),
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chamois's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shammy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "chammy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chamomile ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chamomile",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "camomile",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chamomiles",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "camomiles",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chamomile's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "camomile's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "championize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "championize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "championise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "championizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "championises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "channel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "channeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "channelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "channeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "channelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "channeler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "channeler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "channeller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "channelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "channellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "channeler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "channeller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "channelization ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "channelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "channelisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "channelizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "channelisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "channelization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "channelisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "channelize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "channelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "channelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "channelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "channelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "channelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "channelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "channelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "channelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chaperon ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chaperon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chaperone",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chaperons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chaperones",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chaperon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chaperone's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chaptalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chaptalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chaptalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chaptalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chaptalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chaptalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chaptalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chaptalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chaptalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chaptalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chaptalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chaptalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chaptalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chaptalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "characterizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "characterizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "characterizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "characterizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "characterizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "characterizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "characterization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "characterization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "characterizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "characterization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "characterize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "characterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "characterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "characterizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "characterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "characterizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "characterizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characteriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "characterizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characterisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "characterizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "characteriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "charer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "charer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "charre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chattelization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chattelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chattelisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chattelizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chattelisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chattelization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chattelisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chattelize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chattelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chattelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chattelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chattelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "check ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "check",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "cheque",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("bank"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checked",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "chequed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("bank"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checking",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "chequing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("bank"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "cheques",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("bank"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "check's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "cheque's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("bank"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checker",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chequer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("pattern"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checkers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chequers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("pattern"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checker's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chequer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("pattern"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checkers's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chequers's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("pattern"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checkered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chequered",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("pattern"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checkering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chequering",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("pattern"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "check",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("verify"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checked",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("verify"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checker",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("verify"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "check's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("verify"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checkers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("verify"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checker's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("verify"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checkers's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("verify"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checking",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("verify"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("verify"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "checkbook ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "checkbook",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chequebook",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checkbooks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chequebooks",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checkbook's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chequebook's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "checkerboard ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "checkerboard",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chequerboard",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checkerboards",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chequerboards",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "checkerboard's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chequerboard's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cheerfulize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cheerfulize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cheerfulise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cheerfulizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cheerfulises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cheerly ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cheerly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cherely",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chelicere ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chelicere",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chelicerae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chemicalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chemicalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chemicalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chemicalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chemicalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chemicalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chemicalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chemicalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chemicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chemicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chemicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chemicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chenix ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chenix",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "choenix",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chenixes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "choenixes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chetopod ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chetopod",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chaetopod",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chiffer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chiffer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chiffre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "childer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "childer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "childre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chili ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chili",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chile",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chilli",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chilies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chiles",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chilis",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "chillies",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chili's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chile's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chilli's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chimera ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chimera",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chimaera",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chimera's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chimaera's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chimeras",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chimaeras",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chimerid ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chimerid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chimaerid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chimerids",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chimaerids",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chimerism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chimerism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chimaerism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chimerisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chimaerisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chisel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chiseled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chiselled",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chiseling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chiselling",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chiseler ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chiseler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chiseller",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chiselers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chisellers",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chiseler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chiseller's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chitterlings ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chitterlings",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chitlins",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chitlings",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chitterlings's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chitlins's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chitlings's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chivy ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chivy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chivvy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chivied",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chivvied",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chivying",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chivvying",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chivies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chivvies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chloranemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chloranemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chloranaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chloremia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chloremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chloraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chloridize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chloridize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chloridise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chloridized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chloridised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chloridizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chloridising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chloridizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chloridises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chlorinize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chlorinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chlorinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chlorinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chlorinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chlorinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chlorinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chlorinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chlorinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chloritization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chloritization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chloritisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chloritizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chloritisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chloroanemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chloroanemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chloroanaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chloroformization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chloroformization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chloroformisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chloroformizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chloroformisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chloroformization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chloroformisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chloroformize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chloroformize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chloroformise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chloroformizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chloroformises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chlorophyll ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chlorophyll",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chlorophyl",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chlorophylls",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chlorophyls",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chlorophyll's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chlorophyl's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chocoholic ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chocoholic",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chocaholic",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chocoholics",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chocaholics",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chocoholic's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chocaholic's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chocolaty ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chocolaty",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "chocolatey",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Choephori ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Choephori",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Choaephori",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Choephori's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Choaephori's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cholemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cholemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cholaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cholemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cholaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cholophein ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cholophein",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cholophaein",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "choosy ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "choosy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "choosey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chorization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Christianization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Christianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Christianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Christianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Christianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Christianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Christianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "christianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "christianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "christianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "christianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Christianize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Christianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Christianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Christianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Christianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Christianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Christianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Christianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Christianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "christianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "christianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "christianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "christianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "christianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "christianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "christianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "christianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Christianizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Christianizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Christianiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Christianizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Christianisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Christianizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Christianiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "christianizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "christianiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "christianizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "christianisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chromatize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chromatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chromatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chromatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chromatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chromesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chromesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chromaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chromicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chromicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chromicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chromicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chromicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chromize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chromize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chromise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chromized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chromised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chromizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chromising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chromizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chromises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chronologize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chronologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chronologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chronologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chronologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chronologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chronologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chronologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "chronologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "chutzpah ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "chutzpah",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hutzpa",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "chutzpa",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "hutzpah",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chutzpah's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hutzpa's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "chutzpa's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "hutzpah's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "chutzpahes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hutzpas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "chutzpas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "hutzpahes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cicatricule ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cicatricule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatriculae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cicatrix ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cicatrix",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatrice",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cicatrices",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatrixes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cicatrix's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatrice's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cicatrizant ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cicatrizant",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatrisant",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cicatrizate ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cicatrizate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatrisate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cicatrization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cicatrization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatrisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cicatrizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatrisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cicatrization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatrisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cicatrize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cicatrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cicatrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cicatrizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatrising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cicatrizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatrises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cicatrizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cicatrizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cicatrizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cicatriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ciceronianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ciceronianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ciceronianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ciceronianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ciceronianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ciceronianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ciceronianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cigarette ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cigarette",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cigaret",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cigarettes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cigarets",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cigarette's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cigaret's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cimicide ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cimicide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cimicidae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cinchonization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cinchonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cinchonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cinchonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cinchonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cinchonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cinchonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cinchonize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cinchonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cinchonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cinchonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cinchonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cinchonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cinchonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cinchonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cinchonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cinematize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cinematize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cinematise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cinter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cinter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cintre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cipher ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cipher",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cypher",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ciphered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyphered",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ciphering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyphering",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ciphers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyphers",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cipher's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cypher's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Circean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Circean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Circaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "circularization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "circularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "circularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "circularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "circularize ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "circularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "circularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "circularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "circularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "circularizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "circularizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circulariser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "circularizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circularisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "circularizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circulariser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "circumcenter ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "circumcenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circumcentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "circumcenters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circumcentres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "circumesophagal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "circumesophagal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "circumoesophagal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cithern ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cithern",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cithren",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "citherns",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cithrens",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "citizenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "citizenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "citizenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "citizenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "citizenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "citizenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "citizenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "citizenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "citizenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "civilianization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "civilianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "civilianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "civilianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "civilizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "civilizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "civilization ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "civilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "civilizational ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "civilizational",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisational",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizationaler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisationaler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizationalest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisationalest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizationals",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisationals",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizational's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisational's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "civilizatory ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "civilizatory",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisatory",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "civilize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "civilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "civilizedness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "civilizedness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisedness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizednesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisednesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizedness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisedness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "civilizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "civilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civiliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civilisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civilizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civiliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "civvy ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "civvy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civvies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "civvy's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "civie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "clamor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "clamor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clamour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clamored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clamoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clamoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clamouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clamors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clamours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clamor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clamour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clamoring's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clamouring's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "clamorer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "clamorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clamourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clamorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clamourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clamorer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clamourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "clamorist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "clamorist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "clamourist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clamorist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "clamourist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "clamorsome ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "clamorsome",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clamoursome",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "clangor ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "clangor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clangour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clangored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clangoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clangoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clangouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clangors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clangours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clangor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "clangour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "clares ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "clares",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "claroes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "clarinetist ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "clarinetist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "clarinettist",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clarinetists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "clarinettists",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clarinetist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "clarinettist's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "classicalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "classicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "classicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "classicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "classicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "classicization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "classicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "classicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "classicize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "classicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "classicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "classicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "classicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "classicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "classicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "classicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "classicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "clausule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "clausule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "clausulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "clericalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "clericalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "clericalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "clericalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "clericalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Clete ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Clete",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Cloete",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Clete's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Cloete's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "climatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "climatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "climatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "climatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "climatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "climatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "climatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "climatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "climatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "clower ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "clower",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "clowre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Clytie ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Clytie",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Clytiae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Clytie's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Clytiae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coalizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cocaine ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cocaine",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocain",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cocaine's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocain's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cocainization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cocainization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocainisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cocainizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocainisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cocainization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocainisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cocainize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cocainize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocainise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cocainized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocainised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cocainizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocainising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cocainizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocainises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coconut ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coconut",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocoanut",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coconuts",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocoanuts",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coconut's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocoanut's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cocuiza ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cocuiza",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cocuisa",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Coe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Coe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Cooe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Coe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Cooe's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ce",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coeducationalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coeducationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coeducationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coeducationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coeducationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coenamor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coenamor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "coenamour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coenamored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "coenamoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coenamoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "coenamouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coenamors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "coenamours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coenamorment ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coenamorment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "coenamourment",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coequalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coequalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coequalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coequalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coequalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cognizability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cognizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cognizable ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cognizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cognizableness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cognizableness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisableness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizableness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisableness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cognizably ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cognizably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizablier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisablier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizabliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisabliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cognizance ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cognizance",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisance",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizances",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisances",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizance's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisance's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cognizant ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cognizant",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisant",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizanter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisanter",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizantest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisantest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cognize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cognize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cognizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cognizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cogniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cognisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cognizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cogniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colander ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colander",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cullender",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colanders",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cullenders",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colander's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cullender's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "collateralize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "collateralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "collateralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "collectibility ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "collectibility",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "collectability",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "collectibilities",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "collectabilities",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "collectibility's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "collectability's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "collectible ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "collectible",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "collectable",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "collectibles",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "collectables",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "collectible's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "collectable's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "collectivization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "collectivization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "collectivisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "collectivizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "collectivisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "collectivization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "collectivisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "collectivize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "collectivize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "collectivise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "collectivized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "collectivised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "collectivizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "collectivising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "collectivizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "collectivises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colloquialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colloquialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colloquialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colloquializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colloquialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colloquize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colloquize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colloquise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colloquized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colloquised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colloquizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colloquising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colloquizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colloquises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colonialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colonialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colonizability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colonizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colonizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colonizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colonization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colonizationist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colonizationist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisationist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizationists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisationists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizationist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisationist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colonize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colonizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coloniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colonisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colonizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coloniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "color ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "color",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "coloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colored's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "coloured's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "color's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coloreds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "coloureds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coloring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coloring's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colouring's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorrest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colors's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colours's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorability ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorable ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorableness ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorableness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourableness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorablenesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourablenesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorableness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourableness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorably ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorablier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourablier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorablies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourablies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorabliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourabliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coloradoite ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coloradoite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colouradoite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coloradoite's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colouradoite's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorama ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorama",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourama",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorant ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorant",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourant",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorants",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourants",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorant's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourant's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorate ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coloration ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coloration",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colouration",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourations",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coloration's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colouration's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "OED has coloration as the preferred spelling and discolouration as a",
    "variant for British English for some reason",
  ],
  },
Cluster {
  header: "colorational ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorational",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourational",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorationaler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourationaler",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorationalest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourationalest",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorationally ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorationally",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourationally",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorationallier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourationallier",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorationalliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourationalliest",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorative ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorative",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourative",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorbearer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorbearer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourbearer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorbearer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourbearer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorblind ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorblind",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourblind",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorblindness ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorblindness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourblindness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorblindnesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourblindnesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorblindness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourblindness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorbreed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorbreed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourbreed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorbred",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourbred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorbreeding",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourbreeding",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorbreeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourbreeds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorcast ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorcast",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourcast",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorcasting",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourcasting",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorcasts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourcasts",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorcast's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourcast's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorcasted ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorcasted",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourcasted",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorcaster ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorcaster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourcaster",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorcasters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourcasters",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorectal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorectal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourectal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorectitis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorectitis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourectitis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorectostomy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorectostomy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourectostomy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorfast ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorfast",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourfast",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorfastness ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorfastness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourfastness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorfastnesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourfastnesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorfastness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourfastness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorful ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorful",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourful",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorfuler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourfuler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorfulest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourfulest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorfully ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorfully",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourfully",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorfullier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourfullier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorfulliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourfulliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorfulness ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorfulness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourfulness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorfulnesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourfulnesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorfulness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourfulness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorific ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorific",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourific",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorifics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourifics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorific's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourific's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorimeter ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorimeter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourimeter",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorimeters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourimeters",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorimeter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourimeter's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorimetric ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorimetric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourimetric",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorimetrics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourimetrics",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorimetrical ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorimetrical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourimetrical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorimetrically ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorimetrically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourimetrically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorimetrist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorimetrist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourimetrist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorimetrist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourimetrist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorimetry ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorimetry",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourimetry",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorimetries",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourimetries",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorist ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "colourist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "colourists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "colourist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coloristic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coloristic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colouristic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coloristicer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colouristicer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coloristicest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colouristicest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coloristics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colouristics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coloristically ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coloristically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colouristically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorization ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourization",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourizations",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourization's",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorize ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourize",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourized",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourizing",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourizes",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorless ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourlesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourlessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorlessly ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorlessly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourlessly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorlesslier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourlesslier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorlessliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourlessliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorlessness ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorlessness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourlessness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorlessnesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourlessnesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorlessness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourlessness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colormaker ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colormaker",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourmaker",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colormaking ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colormaking",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourmaking",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorman ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorman",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourman",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colormen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourmen",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorman's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourman's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colormap ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colormap",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourmap",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colormaps",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourmaps",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colormap's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourmap's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coloroto ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coloroto",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colouroto",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorpoint ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorpoint",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourpoint",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorpoints",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourpoints",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorrhaphy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorrhaphy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "colourrhaphy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colortype ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colortype",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourtype",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colorway ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colorway",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourway",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "colorways",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "colourways",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "colory ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "colory",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coloury",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "columnization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "columnization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "columnisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "columnizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "columnisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "columnization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "columnisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "columnize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "columnize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "columnise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "columnized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "columnised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "columnizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "columnising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "columnizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "columnises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "comedia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "comedia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "comoedia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "commercialization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "commercialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "commercialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "commercializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "commercialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "commercialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "commercialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "commercialize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "commercialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "commercialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "commercialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "commercialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "commercializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "commercialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "commercializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "commercialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "commonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "commonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "commonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "commonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "commonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "communalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "communalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "communalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "communalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "communalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "communalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "communalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "communalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "communalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "communalizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "communalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "communalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "communalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "communization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "communization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "communizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "communization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "communize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "communize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "communized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "communizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "communizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "communises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "companionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "companionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "companionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "companionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "companionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "compartmentalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "compartmentalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "compartmentalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "compartmentalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "compartmentalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "compartmentalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "compartmentalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "compartmentalize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "compartmentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "compartmentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "compartmentalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "compartmentalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "compartmentalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "compartmentalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "compartmentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "compartmentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "compartmentize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "compartmentize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "compartmentise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "compartmentizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "compartmentises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "comper ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "comper",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "compoer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "complementizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "complementizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "complementiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "complementizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "complementiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Composite ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Composite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Compositae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "comprehensivization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "comprehensivization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "comprehensivisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "comprehensivizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "comprehensivisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "comprehensivize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "comprehensivize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "comprehensivise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "comprehensivized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "comprehensivised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "comprehensivizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "comprehensivising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "comprehensivizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "comprehensivises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "comprizable ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "comprizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "comprisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "comprizal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "comprizal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "comprisal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "comprizal's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "comprisal's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "computerizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "computerizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "computerisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "computerization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "computerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "computerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "computerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "computerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "computerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "computerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "computerize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "computerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "computerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "computerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "computerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "computerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "computerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "computerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "computerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "concenter ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "concenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concentered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concentred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concentering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concentring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concenters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concentres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conceptualization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conceptualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conceptualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conceptualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conceptualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conceptualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conceptualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conceptualize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conceptualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conceptualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conceptualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conceptualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conceptualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conceptualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conceptualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conceptualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conceptualizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conceptualizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conceptualiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "concertist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "concertist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concretist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "concertize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "concertize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concertise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concertized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concertised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concertizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concertising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concertizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concertises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "concertizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "concertizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concertiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concertizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concertisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concertizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concertiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conche ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conche",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conchae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "concolor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "concolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "concolorous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "concolorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concolourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "concretization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "concretization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concretisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concretizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concretisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concretization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concretisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "concretize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "concretize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concretise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concretized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concretised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concretizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concretising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "concretizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "concretises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conditionalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conditionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conditionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conditionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conditionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coneys ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coneys",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cony",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coney",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cony's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coney's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "confederatize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "confederatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "confederatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "confederatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "confederatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "congenialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "congenialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "congenialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "congenializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "congenialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conges ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conges",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "congoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "congregationalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "congregationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "congregationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "congregationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "congregationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conjurer ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conjurer",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conjuror",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conjurers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conjurors",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conjurer's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conjuror's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "connectable ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "connectable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "connectible",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "connection ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "connection",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "connexion",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "connection's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "connexion's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "connections",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "connexions",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "connector ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "connector",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "connecter",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "connectors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "connecters",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "connector's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "connecter's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conservatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conservatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conservatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conservatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conservatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conservatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conservatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conservatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conservatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "consonantize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "consonantize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "consonantise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "consonantized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "consonantised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "consonantizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "consonantising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "consonantizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "consonantises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "constitutionalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "constitutionalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "constitutionalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "constitutionalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "constitutionalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "constitutionalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "constitutionalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "constitutionalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "constitutionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "constitutionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "constitutionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "constitutionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "constitutionalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "constitutionalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "constitutionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "constitutionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "containerization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "containerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "containerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "containerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "containerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "containerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "containerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "containerize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "containerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "containerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "containerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "containerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "containerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "containerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "containerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "containerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "contemporization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "contemporization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "contemporisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "contemporize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "contemporize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "contemporise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "contemporized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "contemporised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "contemporizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "contemporising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "contemporizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "contemporises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "contextualization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "contextualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "contextualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "contextualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "contextualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "contextualize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "contextualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "contextualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "contextualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "contextualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "contextualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "contextualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "contextualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "contextualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Continentalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Continentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Continentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Continentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Continentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "controversialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "controversialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "controversialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "controversializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "controversialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conundrumize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conundrumize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conundrumise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conundrumizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conundrumises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "convener ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "convener",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "convenor",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conveners",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "convenors",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "convener's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "convenor's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conventionalization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conventionalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conventionalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conventionalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conventionalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conventionalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conventionalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conventionalize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conventionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conventionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conventionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conventionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conventionalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conventionalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conventionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conventionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conventionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conventionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conventionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conventionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conventionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conversationize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conversationize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conversationise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conversationizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conversationises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "converter ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "converter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "convertor",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "converters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "convertors",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "converter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "convertor's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "converter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "convertor",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "converters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "convertors",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "converter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "convertor's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conveyor ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conveyor",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conveyer",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conveyors",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conveyers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conveyor's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conveyer's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "conveyorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "conveyorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conveyorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conveyorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conveyorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conveyorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conveyorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "conveyorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "conveyorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "convivialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "convivialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "convivialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "convivializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "convivialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cookie ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cookie",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cooky",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cookie's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cooky's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "copolymerization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "copolymerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "copolymerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "copolymerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "copolymerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "copolymerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "copolymerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "copolymerize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "copolymerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "copolymerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "copolymerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "copolymerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "copolymerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "copolymerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "copolymerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "copolymerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "copperization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "copperization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "copperisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "copperizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "copperisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "copperization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "copperisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "copperize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "copperize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "copperise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "copperizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "copperises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "copremia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "copremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "copraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "copremic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "copremic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "copraemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coraled ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coraled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coralled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coranto ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coranto",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "couranto",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corantoes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "courantoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corantos",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "courantos",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "corbe ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "corbe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "courbe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "corbed ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "corbed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "courbed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "corbel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "corbeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corbelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corbeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corbelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corbeling's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corbelling's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corbelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corbellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cordialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cordialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cordialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cordialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cordialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cordializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cordialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cordializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cordialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "corge ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "corge",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "courge",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Corinthianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Corinthianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Corinthianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Corinthianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Corinthianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Corinthianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Corinthianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corinthianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corinthianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corinthianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corinthianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corinthianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corinthianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corinthianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corinthianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "corporealization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "corporealization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corporealisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corporealizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corporealisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corporealization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corporealisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "corporealize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "corporealize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corporealise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corporealized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corporealised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corporealizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corporealising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "corporealizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corporealises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "corsy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "corsy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coursy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cortage ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cortage",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "courtage",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cortin ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cortin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "courtin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cortine ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cortine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cortinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "corve ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "corve",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "corvae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coryphe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coryphe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coryphae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coryphes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "coryphaes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cosmeticize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cosmeticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosmeticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cosmeticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosmeticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cosmeticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosmeticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cosmeticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosmeticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cosmopolitanization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cosmopolitanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosmopolitanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cosmopolitanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosmopolitanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cosmopolitanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosmopolitanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cosmopolitanize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cosmopolitanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosmopolitanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cosmopolitanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosmopolitanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cosmopolitanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosmopolitanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cosmopolitanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosmopolitanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cosplendor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cosplendor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "cosplendour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coste ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coste",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "costae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "costumier ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "costumier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "costumire",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cottar ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cottar",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cotter",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cottars",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cotters",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cottar's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cotter's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cotter",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cottered",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cottering",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cotters",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cotter's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cotte ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cotte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cottae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cottonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cottonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cottonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cottonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cottonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cottonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cottonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cottonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cottonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cottonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cottonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cottonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cotyle ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cotyle",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cotylae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "councilor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "councilor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "councillor",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "councilors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "councillors",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "councilor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "councillor's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "councilorship ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "councilorship",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "councillorship",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "councilorships",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "councillorships",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "councilorship's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "councillorship's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "counsel ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "counseled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "counselled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "counseling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "counselling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "counselee ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "counselee",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "counsellee",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "counselor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "counselor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "counsellor",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "counselors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "counsellors",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "counselor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "counsellor's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "counselorship ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "counselorship",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "counsellorship",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "counselorships",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "counsellorships",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "counselorship's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "counsellorship's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "countercolored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "countercolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "countercoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "countor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "countor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "countour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cozily ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cozily",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosily",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "coziness ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "coziness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosiness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coziness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosiness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cozy ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cozy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cozied",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosied",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cozier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "coziest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosiest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cozying",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosying",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cozies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cozy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cosy's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "craizey ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "craizey",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "craisey",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crambes ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crambes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cramboes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crawlerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crawlerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crawlerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crawlerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crawlerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "creaturize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "creaturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "creaturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "creaturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "creaturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crenel ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "creneled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crenelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "creneling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crenelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crenelate ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crenelate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crenellate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crenelated",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crenellated",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crenelater",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crenellater",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crenelates",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crenellates",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crenelatest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crenellatest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crenelating",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crenellating",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crenelation ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crenelation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crenellation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crenelations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crenellations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crenelation's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crenellation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Creolization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Creolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Creolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "creolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "creolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "creolizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "creolisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Creolize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Creolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Creolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Creolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Creolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Creolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Creolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Creolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Creolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "creolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "creolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "creolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "creolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "creolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "creolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "creolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "creolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cretinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cretinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cretinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cretinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cretinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cretinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cretinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cretinize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cretinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cretinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cretinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cretinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cretinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cretinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cretinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cretinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "criminalization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "criminalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criminalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criminalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criminalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criminalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criminalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "criminalize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "criminalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criminalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criminalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criminalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criminalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criminalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criminalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criminalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "criticizable ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "criticizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criticizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criticizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criticizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criticizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "criticize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "criticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "criticizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "criticizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criticizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criticizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "criticizingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "criticizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criticizinglier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticisinglier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criticizinglies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticisinglies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "criticizingliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "criticisingliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crofterization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crofterization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crofterisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crofterizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crofterisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crofterization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crofterisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crofterize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crofterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crofterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crofterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crofterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crosier ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crosier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crozier",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crosiers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "croziers",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crosier's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crozier's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cruel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crueler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "crueller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cruelest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cruellest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cruelize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cruelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cruelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cruelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cruelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cryesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cryesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cryaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cryoanesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cryoanesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cryoanaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cryptanalyzes ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cryptanalyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cryptanalyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cryptesthesia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cryptesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cryptaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cryptesthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cryptaesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cryptesthetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cryptesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cryptaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crystallizability ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crystallizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystallisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crystallizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystallisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crystallizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystallisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crystallizable ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crystallizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystallisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crystallizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystallisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crystallizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystallisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crystallizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystallisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crystallizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystallisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crystallization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crystallization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystallisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crystallizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystallisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crystallization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystallisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crystallize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crystallize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystalize",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "crystallise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crystallized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystalized",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "crystallised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crystallizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystalizing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "crystallising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crystallizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystalizes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "crystallises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "crystallizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "crystallizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystalliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "crystallizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "crystallisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Cubanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Cubanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Cubanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Cubanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Cubanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Cubanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Cubanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cuckoldize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cuckoldize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cuckoldise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cuckoldized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cuckoldised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cuckoldizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cuckoldising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cuckoldizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cuckoldises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cudgel ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cudgeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cudgelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cudgeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cudgelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cudgelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cudgellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cudgeler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cudgeler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cudgeller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cudgelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cudgellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cudgeler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cudgeller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "culicide ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "culicide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "culicidae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "culturization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "culturization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "culturisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "culturizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "culturisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "culturization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "culturisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "culturize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "culturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "culturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "culturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "culturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cupel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cupeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cupelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cupeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cupelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cupelation ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cupelation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cupellation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cupeler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cupeler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cupeller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cupelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cupellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cupeler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cupeller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cupule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cupule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cupulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "curarize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "curarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "curatize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "curatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "curb ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "curb",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("restrain"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curbed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("restrain"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curbing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("restrain"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curbs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("restrain"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curb's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("restrain"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curb",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kerb",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("street edge"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curbed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kerbed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("street edge"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curbing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kerbing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("street edge"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curbs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kerbs",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("street edge"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curb's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kerb's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("street edge"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "curlicue ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "curlicue",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curlycue",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curlicues",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curlycues",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curlicue's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curlycue's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curlicue",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curlicues",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "curricularization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "curricularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curricularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curricularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curricularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curricularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curricularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "curricularize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "curricularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curricularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curricularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curricularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "curtsy ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "curtsy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curtsey",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curtsied",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curtseyed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curtsying",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curtseying",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curtsies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curtseys",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "curtsy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curtsey's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "curvaceous ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "curvaceous",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "curvacious",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "customizable ",
  verified: true,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "customizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "customisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "customization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "customization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "customisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "customizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "customisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "customization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "customisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "customize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "customize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "customise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "customized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "customised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "customizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "customising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "customizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "customises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "customizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "customizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "customiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "customizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "customisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cutesy ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cutesy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutesie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cuticolor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cuticolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cuticolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cutinization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cutinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cutinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cutinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cutinize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cutinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cutinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cutinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cutinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cutization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cutization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cutizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cutization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cutlass ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cutlass",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutlas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cutlasses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutlases",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cutlass's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cutlas's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cuve ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cuve",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cuvae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Cyanee ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Cyanee",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Cyaneae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Cyanee's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Cyaneae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cyanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cyanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cyanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cyanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cyanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cyanomethemoglobin ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cyanomethemoglobin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "cyanomethaemoglobin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cyclization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cyclization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cyclizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cyclization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cyclize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cyclize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cyclized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cyclizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cyclizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cyclopedia ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cyclopedia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclopaedia",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cyclopedias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclopaedias",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cyclopedia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclopaedia's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cyclopedic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cyclopedic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclopaedic",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cyclopedically ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cyclopedically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclopaedically",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cyclopedist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cyclopedist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclopaedist",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "cyclopedist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cyclopaedist's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "cyme ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "cyme",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cymae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "czar ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "czar",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tzar",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tsar",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "czars",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tzars",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tsars",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "czar's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tzar's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tsar's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "czardom ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "czardom",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tzardom",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tsardom",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "czardoms",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tzardoms",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tsardoms",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "czardom's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tzardom's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tsardom's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "czarina ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "czarina",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tzarina",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tsarina",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "czarinas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tsarinas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tzarinas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "czarina's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tzarina's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tsarina's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "czarism ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "czarism",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tsarism",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tzarism",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "czarisms",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tsarisms",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tzarisms",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "czarism's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tzarism's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tsarism's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "czarist ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "czarist",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tzarist",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tsarist",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Czechization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Czechization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Czechisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Czechizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Czechisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Czechization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Czechisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "danaide ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "danaide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "danaidae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dandizette ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dandizette",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dandisette",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dandyize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dandyize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dandyise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dandyizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dandyises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dane ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dane",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "danae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Danization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Danization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Danisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Danizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Danisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Danization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Danisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Danize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Danize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Danise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Danizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Danises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Danize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Danise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Darwinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Darwinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Darwinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Darwinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Darwinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Darwinize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Darwinise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dastardize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dastardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dastardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dastardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dastardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deaconize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deaconize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deaconise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deaconizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deaconises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deadrize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deadrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deadrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dealcoholize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dealcoholize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dealcoholise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dealcoholized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dealcoholised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dealcoholizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dealcoholising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dealcoholizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dealcoholises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deaminize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deaminize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deaminise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deary ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deary",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dearie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deary's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dearie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decapitalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decapitalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decapitalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decapitalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decapitalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decapitalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decapitalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decapitalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decapitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decapitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decapitalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decapitalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decapitalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decapitalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decarbonization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decarbonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarbonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decarbonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarbonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decarbonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarbonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decarbonize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decarbonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarbonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decarbonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarbonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decarbonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarbonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decarbonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarbonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decarbonizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decarbonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarboniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decarbonizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarbonisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decarbonizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarboniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decarburization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decarburization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarburisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decarburizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarburisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decarburization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarburisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decarburize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decarburize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarburise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decarburized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarburised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decarburizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarburising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decarburizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decarburises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decasualization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decasualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decasualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decasualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decasualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decasualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decasualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decasualize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decasualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decasualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decasualized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decasualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decasualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decasualizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decasualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decasualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decenters ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decenters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decentres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decentralization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decentralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decentralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decentralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decentralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decentralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decentralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decentralizationist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decentralizationist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decentralisationist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decentralize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decentralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decentralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decentralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decentralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decentralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decentralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decentralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decentralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decerebrize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decerebrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decerebrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decerebrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decerebrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decerebrizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decerebrising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decerebrizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decerebrises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dechristianization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dechristianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dechristianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dechristianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dechristianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dechristianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dechristianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dechristianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dechristianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dechristianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dechristianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dechristianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dechristianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dechristianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decigram ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decigram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decigramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decigrams",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decigrammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decigram's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decigramme's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deciliter ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deciliter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decilitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deciliters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decilitres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deciliter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decilitre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decimalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decimalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decimalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decimalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decimalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decimalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decimalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decimalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decimalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decimalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decimalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decimalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decimalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decimalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decimalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decimalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decimeter ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decimeter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "decimetre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decimeters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "decimetres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decimeter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "decimetre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decivilize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decivilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decivilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decivilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decivilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decivilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decivilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decivilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decivilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decolonization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decolonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decolonize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decolonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decolor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "decolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "decoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decoloring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "decolouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "decolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decolorant ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decolorant",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "decolourant",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolorants",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "decolourants",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decolorate ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decolorate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolorated",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourated",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolorates",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourates",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decoloration ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decoloration",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolouration",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolorations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decoloration's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolouration's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decolorimeter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decolorimeter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourimeter",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decolorization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decolorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourization",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourizations",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourization's",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decolorize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decolorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourize",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourized",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourizing",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourizes",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decolorizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decolorizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolouriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourizer",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolorizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourizers",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decolorizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolouriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decolourizer's",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decriminalization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decriminalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decriminalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decriminalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decriminalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decriminalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decriminalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "decriminalize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "decriminalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decriminalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decriminalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decriminalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decriminalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decriminalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "decriminalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decriminalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dedal ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dedal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "daedal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dedalian ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dedalian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "daedalian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dedramatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dedramatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dedramatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dedramatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dedramatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dedramatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dedramatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dedramatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dedramatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deemphasization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deemphasization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deemphasisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deemphasization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deemphasisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deemphasize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deemphasize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deemphasise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deemphasized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deemphasised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deemphasizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deemphasising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deemphasizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deemphasises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deemphasizer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deemphasizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deemphasiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deemphasizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deemphasisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deenergize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deenergize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deenergise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deenergized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deenergised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deenergizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deenergising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deenergizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deenergises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deenergizer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deenergizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deenergiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "defedation ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "defedation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "defoedation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "defeminization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "defeminization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defeminisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defeminization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defeminisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "defeminize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "defeminize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defeminise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defeminized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defeminised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defeminizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defeminising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defeminizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defeminises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "defense ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "defense",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "defence",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defensed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "defenced",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defensing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "defencing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defenses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "defences",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defense's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "defence's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "defenseless ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "defenseless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defenceless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defenselesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defencelesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defenselessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defencelessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "defenselessly ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "defenselessly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defencelessly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defenselesslier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defencelesslier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defenselessliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defencelessliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "defenselessness ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "defenselessness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defencelessness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defenselessnesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defencelessnesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defenselessness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defencelessness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "defenseman ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "defenseman",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defenceman",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defensemen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defencemen",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defenseman's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defenceman's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "defeudalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "defeudalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defeudalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defeudalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defeudalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defeudalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defeudalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defeudalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defeudalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "defibrinize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "defibrinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defibrinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defibrinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defibrinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defibrinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defibrinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defibrinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defibrinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "definitization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "definitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "definitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "definitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "definitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "definitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "definitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "definitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "definitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "definitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "definitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "definitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "definitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "definitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "definitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "definitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deflectionization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deflectionization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deflectionisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deflectionizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deflectionisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deflectionization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deflectionisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deflectionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deflectionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deflectionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deflectionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deflectionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deformalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deformalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deformalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "defunctionalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "defunctionalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defunctionalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defunctionalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defunctionalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defunctionalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defunctionalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "defunctionalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "defunctionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defunctionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "defunctionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "defunctionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deglycerolized ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deglycerolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deglycerolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dehumanization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dehumanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehumanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehumanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehumanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehumanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehumanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dehumanize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dehumanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehumanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehumanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehumanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehumanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehumanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehumanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehumanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dehydrogenization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dehydrogenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehydrogenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehydrogenization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehydrogenisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dehydrogenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dehydrogenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehydrogenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehydrogenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehydrogenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehydrogenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehydrogenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dehydrogenizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dehydrogenizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehydrogeniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehydrogenizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehydrogeniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dehypnotization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dehypnotization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehypnotisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehypnotizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehypnotisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehypnotization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehypnotisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dehypnotize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dehypnotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehypnotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehypnotized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehypnotised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehypnotizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehypnotising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dehypnotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dehypnotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deindustrialization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deindustrialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deindustrialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deindustrializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deindustrialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deindustrialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deindustrialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deindustrialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deindustrialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deindustrialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deindustrialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deindustrialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deindustrializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deindustrialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deindustrializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deindustrialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deionization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deionization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deionisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deionization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deionisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deionize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dekagram ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dekagram's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decagram",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decagramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dekagram's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decagrams",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decagrammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dekagram's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decagram's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decagramme's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dekaliter ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dekaliter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decaliter",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "decalitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dekaliters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decaliter",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "decalitres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dekaliter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decaliter",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "decalitre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dekameter ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dekameter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decameter",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "decametre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dekameters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decameter",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "decametres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dekameter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "decameter",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "decametre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "delimitize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "delimitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "delimitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "delimitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "delimitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "delocalization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "delocalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "delocalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "delocalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "delocalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "delocalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "delocalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "delocalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "delocalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "delocalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "delocalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "delocalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "delocalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "delocalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "delocalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "delocalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deluster ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deluster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "delustre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demagnetizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demagnetizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagnetizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagnetizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagnetizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagnetizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demagnetization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demagnetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagnetizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagnetization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demagnetize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demagnetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagnetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagnetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagnetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demagnetizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demagnetizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagnetizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagnetizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagnetiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demagogue ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demagogue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagog",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagogued",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagoged",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagoguing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagoging",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagogues",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagogs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagogue's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagog's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demagogueri ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demagogueri",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagogri",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demagogueris",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagogris",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demagoguery ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demagoguery",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demagogry",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demasculinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demasculinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demasculinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demasculinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demasculinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demasculinize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demasculinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demasculinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demasculinized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demasculinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demasculinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demasculinizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demasculinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demasculinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dematerialization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dematerialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dematerialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dematerializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dematerialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dematerialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dematerialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dematerialize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dematerialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dematerialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dematerialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dematerialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dematerializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dematerialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dematerializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dematerialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demeanor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demeanor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "demeanour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demeanors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "demeanours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demeanor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "demeanour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demilitarization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demilitarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demilitarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demilitarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demilitarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demilitarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demilitarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demilitarize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demilitarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demilitarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demilitarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demilitarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demilitarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demilitarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demilitarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demilitarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demiluster ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demiluster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demilustre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demineralization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demineralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demineralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demineralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demineralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demineralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demineralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demineralize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demineralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demineralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demineralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demineralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demineralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demineralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demineralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demineralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demineralizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demineralizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demineraliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demobilization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demobilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demobilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demobilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demobilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demobilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demobilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demobilize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demobilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demobilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demobilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demobilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demobilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demobilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demobilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demobilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "democratization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "democratization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "democratisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "democratizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "democratisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "democratization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "democratisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "democratize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "democratize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "democratise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "democratized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "democratised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "democratizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "democratising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "democratizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "democratises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "democratizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "democratizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "democratiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demonetization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demonetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demonetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demonetizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demonetisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demonetization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demonetisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demonetize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demonetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demonetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demonetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demonetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demonetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demonetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demonetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demonetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demonian ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demonian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "daemonian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demonization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demonize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demoralization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demoralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demoralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demoralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demoralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demoralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demoralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demoralize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demoralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demoralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demoralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demoralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demoralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demoralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demoralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demoralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demoralizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demoralizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demoraliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demoralizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demoralisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demoralizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demoraliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demoralizingly ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demoralizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demoralisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demutualization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demutualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demutualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demutualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demutualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demutualize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demutualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demutualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demutualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demutualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demutualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demutualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demutualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demutualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demythologization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demythologization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demythologisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demythologizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demythologisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demythologization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demythologisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demythologize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demythologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demythologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demythologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demythologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demythologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demythologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "demythologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demythologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "demythologizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "demythologizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "demythologiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denationalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denationalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denationalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denationalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denationalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denationalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denationalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denationalize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denationalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denationalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denationalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denationalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denaturalization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denaturalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denaturalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denaturalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denaturalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denaturalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denaturalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denaturalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denaturalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denaturization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denaturization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denaturizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denaturization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denaturize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denaturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denaturized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denaturizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denaturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denaturizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denaturizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denaturizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denaturizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denaturiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denicotinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denicotinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denicotinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denizenize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denizenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denizenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denizenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denizenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denominationalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denominationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denominationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denominationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denominationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denormalize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denormalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denormalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dentalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dentalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dentalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dentalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dentalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dentalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dentalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dentalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dentalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dentalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dentalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dentalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dentin ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dentin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dentine",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dentins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dentines",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dentin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dentine's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denuclearization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denuclearization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denuclearisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denuclearizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denuclearisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denuclearization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denuclearisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "denuclearize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "denuclearize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denuclearise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denuclearized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denuclearised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denuclearizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denuclearising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "denuclearizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "denuclearises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deodorization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deodorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deodorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deodorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deodorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deodorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deodorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deodorize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deodorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deodorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deodorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deodorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deodorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deodorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deodorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deodorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deodorizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deodorizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deodoriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deodorizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deodorisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deodorizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deodoriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deoxidization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deoxidization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxidisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deoxidizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxidisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deoxidization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxidisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deoxidize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deoxidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deoxidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deoxidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deoxidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deoxidizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deoxidizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxidiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deoxidizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxidisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deoxidizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxidiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deoxygenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deoxygenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxygenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deoxygenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxygenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deoxygenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxygenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deoxygenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deoxygenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "departmentalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "departmentalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "departmentalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "departmentalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "departmentalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "departmentalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "departmentalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "departmentalize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "departmentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "departmentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "departmentalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "departmentalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "departmentalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "departmentalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "departmentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "departmentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "departmentization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "departmentization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "departmentisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "departmentizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "departmentisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "departmentization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "departmentisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "departmentize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "departmentize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "departmentise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "departmentizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "departmentises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depauperization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depauperization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depauperisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depauperize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depauperize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depauperise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depauperized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depauperised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depauperizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depauperising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depauperizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depauperises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dependence ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dependence",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dependance",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dependences",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dependances",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dependence's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dependance's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dependent ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dependent",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dependent",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dependant",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dependents",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dependants",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dependent's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dependant's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depersonalization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depersonalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depersonalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depersonalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depersonalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depersonalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depersonalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depersonalize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depersonalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depersonalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depersonalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depersonalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depersonalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depersonalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depersonalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depersonalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depolarization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depolarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depolarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depolarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depolarize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depolarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depolarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depolarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depolarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depolarizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depolarizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolariser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depolarizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolarisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depolarizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolariser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depoliticization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depoliticization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depoliticisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depoliticize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depoliticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depoliticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depoliticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depoliticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depoliticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depoliticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depoliticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depoliticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depolymerization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depolymerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolymerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depolymerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolymerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depolymerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolymerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depolymerize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depolymerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolymerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depolymerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolymerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depolymerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolymerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depolymerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depolymerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depressurization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depressurization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depressurisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depressurizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depressurisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depressurization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depressurisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "depressurize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "depressurize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depressurise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depressurized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depressurised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depressurizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depressurising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "depressurizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "depressurises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deputationize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deputationize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deputationise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deputationizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deputationises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deputization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deputization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deputisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deputize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deputize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deputise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deputized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deputised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deputizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deputising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deputizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deputises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deracialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deracialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deracialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deracialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deracialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deracializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deracialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deracializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deracialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "derationalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "derationalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "derationalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "derationalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "derationalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "derationalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "derationalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "derationalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "derationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "derationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "derationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "derationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deratization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deratization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deratisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deratizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deratisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deratization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deratisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "derealization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "derealization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "derealisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "derecognize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "derecognize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "derecognise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "derecognized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "derecognised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "derecognizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "derecognising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "derecognizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "derecognises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deregulationize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deregulationize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deregulationise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deregulationizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deregulationises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dereligionize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dereligionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dereligionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dereligionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dereligionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dereligionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dereligionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dereligionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dereligionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "derk ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "derk",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "drek",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dermatorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dermatorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dermatorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Derr ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Derr",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Drer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Derr's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Drer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desacralization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desacralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desacralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desacralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desacralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desacralize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desacralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desacralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desacralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desacralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desacralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desacralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desacralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desacralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desalinization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desalinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desalinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desalinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desalinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desalinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desalinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desalinize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desalinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desalinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desalinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desalinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desalinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desalinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desalinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desalinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "descendant ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "descendant",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "descendent",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "descendants",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "descendents",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "descendant's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "descendent's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "descendant",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "descendent",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desensitization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desensitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desensitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desensitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desensitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desensitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desensitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desensitize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desensitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desensitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desensitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desensitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desensitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desensitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desensitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desensitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desensitizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desensitizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desensitiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desensitizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desensitisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desensitizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desensitiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desertization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desertization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desertisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desertizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desertisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desexualization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desexualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desexualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desexualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desexualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desexualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desexualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desexualize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desexualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desexualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desexualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desexualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desexualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desexualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desexualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desexualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desilverization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desilverization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desilverisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desilverizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desilverisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desilverize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desilverize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desilverise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desilverized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desilverised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desilverizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desilverising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desilverizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desilverises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "despiritualization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "despiritualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "despiritualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "despiritualize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "despiritualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "despiritualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "despotize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "despotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "despotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "despotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "despotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "destabilization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "destabilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "destabilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "destabilize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "destabilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "destabilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "destabilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "destabilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "destabilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "destabilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "destabilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "destabilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "destabilizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "destabilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "destabiliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "destabilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "destabilisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "destalinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "destalinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "destalinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "destalinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "destalinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "destalinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "destalinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "destalinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "destalinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desterilize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desterilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desterilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "destigmatization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "destigmatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "destigmatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desulfurization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desulfurization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurization",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desulfurizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurizations",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desulfurization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurization's",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desulfurize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desulfurize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurize",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desulfurized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurized",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desulfurizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurizing",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desulfurizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurizes",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desulfurizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desulfurizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulfuriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desulfurizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulfuriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desulphurizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desulphurizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphuriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desulphurizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desulphurisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desynchronization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desynchronization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desynchronisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desynchronizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desynchronisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desynchronization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desynchronisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "desynchronize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "desynchronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desynchronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desynchronized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desynchronised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desynchronizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desynchronising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "desynchronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "desynchronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "detribalization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "detribalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "detribalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "detribalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "detribalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "detribalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "detribalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "detribalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "detribalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "detribalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "detribalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "detribalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "detribalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "detribalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "detribalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "detribalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "deva ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deva",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "daeva",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "devalorization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "devalorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devalorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devalorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devalorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "devalorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "devalorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devalorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devalorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devalorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devalorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devalorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devalorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devalorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "devel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "develed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "develled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "develing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "develling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "devil ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "deviled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "deviling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devilling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "devilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "devilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "devitalization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "devitalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devitalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devitalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devitalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devitalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devitalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "devitalize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "devitalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devitalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devitalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devitalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devitalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devitalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "devocalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "devocalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devocalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devocalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devocalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "devocalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "devocalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devocalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devocalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devocalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devocalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devocalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devocalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devocalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "devolatilization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "devolatilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devolatilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devolatilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devolatilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devolatilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devolatilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "devolatilize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "devolatilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devolatilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devolatilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devolatilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devolatilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devolatilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "devolatilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "devolatilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dexterous ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dexterous",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dextrous",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dexterously ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dexterously",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dextrously",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dexterousness ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dexterousness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dextrousness",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diabolization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diabolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diabolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diabolizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diabolisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diabolization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diabolisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diabolize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diabolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diabolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diabolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diabolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diabolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diabolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diabolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diabolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diagonalizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diagonalizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diagonalisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diagonalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diagonalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diagonalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diagonalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diagonalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diagonalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diagonalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diagonalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diagonalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diagonalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diagonalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diagonalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dial ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dialed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diallings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialling's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dialecticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dialecticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialecticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialecticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialecticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dialer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dialer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diallers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialler's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dialist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dialist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diallist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diallists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diallist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dialogize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dialogize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialogise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialogized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialogised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialogizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialogising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialogizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialogises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dialogue ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dialogue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialog",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialogued",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialogged",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialoguing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialogging",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialogues",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialogs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialogue's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialog's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialog",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialogue",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("dialog box"),
  comment: None,
  },
  ],
  notes: &[
    "Macquarie prefers dialogue (box) even for computing.",
  ],
  },
Cluster {
  header: "dialyzability ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dialyzability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dialyzable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dialyzable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dialyzate ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dialyzate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzates",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysates",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzate's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysate's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dialyzation ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dialyzation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzation's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dialyze ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dialyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: dialysis"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dialyzer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dialyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialysers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dialyzer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dialyser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diamondize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diamondize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diamondise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diamondizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diamondises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diarrhea ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diarrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "diarrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diarrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "diarrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diarrhea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "diarrhoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diarrheal ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diarrheal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "diarrhoeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diarrhealer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "diarrhoealer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diarrhealest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "diarrhoealest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diarrheic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diarrheic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "diarrhoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diarrheicer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "diarrhoeicer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diarrheicest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "diarrhoeicest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diarrhetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diarrhetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "diarrhoetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diarrheticer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "diarrhoeticer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diarrheticest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "diarrhoeticest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dichotomization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dichotomization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dichotomisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dichotomizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dichotomisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dichotomization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dichotomisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dichotomize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dichotomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dichotomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dichotomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dichotomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dichotomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dichotomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dichotomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dichotomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dickey ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dickey",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dicky",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dickie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dickeys",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dickies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dickey's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dicky's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dickie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diecious ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dioecious",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diecious",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dieciously ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dioeciously",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dieciously",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diene ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diaene",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dieretic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dieretic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diaeretic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dieselization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dieselization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dieselisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dieselizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dieselisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dieselization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dieselisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dieselize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dieselize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dieselise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dieselized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dieselised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dieselizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dieselising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dieselizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dieselises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diestrous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diestrous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dioestrous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diestrual ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diestrual",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dioestrual",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diestrum ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diestrum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dioestrum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diestrus ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diestrus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dioestrus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diestruses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dioestruses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diestrus's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dioestrus's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diethylstilbestrol ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diethylstilbestrol",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diethylstilboestrol",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diethylstilbestrol's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diethylstilboestrol's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dietitian ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dietitian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dietician",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dietitians",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dieticians",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dietitian's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dietician's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "differentialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "differentialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "differentialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "differentializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "differentialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "digitalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "digitalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "digitalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "digitalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitalis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitalises",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2 "),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitalis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "digitization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "digitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "digitize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "digitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "digitizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "digitizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "digitizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "digitiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dike ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dike",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dyke",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diked",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dyked",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diking",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dyking",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dikes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dykes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dike's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dyke's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dyke",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dike",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("lesbian"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dykes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dikes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("lesbian"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dyke's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dike's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("lesbian"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dimensionalization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dimensionalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dimensionalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dimensionalize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dimensionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dimensionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dimensionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dimensionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dimensionalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dimensionalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dimensionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dimensionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dimerization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dimerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dimerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dimerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dimerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dimerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dimerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dimerize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dimerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dimerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dimerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dimerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dimerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dimerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dimerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dimerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diminutivize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diminutivize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diminutivise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diminutivizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diminutivises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dingoes ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dingoes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dingos",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diopter ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diopter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dioptre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diopters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dioptres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diopter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dioptre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diphthongization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diphthongization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diphthongisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diphthongizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diphthongisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diphthongization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diphthongisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diphthongize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diphthongize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diphthongise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diphthongized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diphthongised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diphthongizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diphthongising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diphthongizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diphthongises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diple ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diple",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diploe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diploe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diploae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diploe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diploae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diplomatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diplomatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diplomatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diplomatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diplomatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diplomatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diplomatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "diplomatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diplomatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dis ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diss",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dises",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Improper), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diss's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dissed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dissing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disangularize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disangularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disangularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disangularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disangularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disauthorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disauthorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disauthorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disauthorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disauthorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disauthorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disauthorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disauthorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disauthorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disbowel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disboweled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disbowelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disboweling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disbowelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "discanonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "discanonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discanonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discanonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discanonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discanonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discanonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "discanonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "discanonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discanonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discanonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discanonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "discolor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "discolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "discolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "discoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "discolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discoloreds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "discoloureds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discolored's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "discoloured's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discoloring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "discolouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "discolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "discolorate ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "discolorate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolourate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "discolorated ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "discolorated",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolourated",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "discoloration ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "discoloration",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolouration",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discolorations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolourations",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discoloration's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolouration's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "See note for coloration",
  ],
  },
Cluster {
  header: "discoloredness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "discoloredness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolouredness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "discolorization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "discolorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolourisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolourization",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discolorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolourisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolourizations",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discolorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolourisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolourization's",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "discolorment ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "discolorment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolourment",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discolorments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolourments",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discolorment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discolourment's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "discretization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "discretization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discretisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "discretize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "discretize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discretise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disdenominationalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disdenominationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disdenominationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disdenominationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disdenominationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disdiplomatize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disdiplomatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disdiplomatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disdiplomatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disdiplomatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disembowel ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disemboweled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disembowelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disemboweling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disembowelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disequalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disequalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disequalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disequalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disequalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disequalize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disequalise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disequalizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disequalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disequaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disequalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disequalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disfavor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disfavor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "disfavour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disfavored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "disfavoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disfavoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "disfavouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disfavors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "disfavours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disfavor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "disfavour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disfavorer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disfavorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "disfavourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disfavorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "disfavourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disfavorer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "disfavourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dishabille ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dishabille",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deshabille",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishabilles",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deshabilles",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishabille's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "deshabille's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disharmonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disharmonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disharmonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disharmonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disharmonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disharmonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disharmonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disharmonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disharmonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dishevel ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disheveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dishevelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disheveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dishevelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disheveler ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disheveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disheveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dishonor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dishonor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dishonorable ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dishonorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonorabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonorablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dishonorableness ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dishonorableness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourableness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonorablenesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourablenesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonorableness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourableness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dishonorably ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dishonorably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonorablier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourablier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonorablies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourablies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonorabliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourabliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dishonorary ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dishonorary",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dishonourary",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dishonorer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dishonorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishonorer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishonourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dishumanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dishumanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dishumanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishumanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dishumanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dishumor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dishumor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishumour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishumored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishumoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dishumors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dishumours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disillusionize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disillusionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disillusionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disillusionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disillusionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disillusionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disillusionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disillusionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disillusionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disillusionizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disillusionizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disillusioniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disillusionizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disillusionisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disillusionizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disillusioniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disindividualize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disindividualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disindividualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disindividualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disindividualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disindividualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disindividualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disindividualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disindividualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disindustrialization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disindustrialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disindustrialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disindustrializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disindustrialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disindustrialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disindustrialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disindustrialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disindustrialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disindustrialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disindustrializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disindustrialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disindustrializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disindustrialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disk ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disk",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disc",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "discs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disk's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disc's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disc",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disk",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "discs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disks",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disc's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disk's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
    "Disk vs. disc is a mess, rather than sort it out I will just mark",
    "each as a variant of each other.",
  ],
  },
Cluster {
  header: "disluster ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disluster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dislustre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dismalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dismalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dismalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dismalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dismalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disnaturalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disnaturalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disnaturalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disnaturalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disnaturalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disnaturalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disnaturalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disnaturalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disnaturalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disnaturalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disnaturalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disnaturalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disnaturalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disnaturalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disnaturalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disnaturalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disorganization ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disorganization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disorganisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disorganizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disorganisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disorganization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disorganisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disorganize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disorganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disorganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disorganized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disorganised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disorganizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disorganising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disorganizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disorganises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disorganizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disorganizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disorganiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disorganizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disorganisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disorganizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disorganiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disozonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disozonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disozonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disozonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disozonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dispapalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dispapalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dispapalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dispapalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dispapalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dispatch ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dispatch",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "despatch",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dispatched",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "despatched",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dispatching",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "despatching",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dispatches",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "despatches",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dispatch's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "despatch's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dispauperize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dispauperize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dispauperise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dispauperized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dispauperised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dispauperizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dispauperising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dispauperizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dispauperises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dispersonalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dispersonalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dispersonalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dispersonalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dispersonalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dispopularize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dispopularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dispopularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dispopularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dispopularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disrealize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disrealize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disrealise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disrealizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disrealises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disscepter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disscepter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissceptre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disscepters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissceptres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disscepter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissceptre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disseize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disseize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disseise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disseized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disseised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disseizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disseising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disseizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disseises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disseizee ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disseizee",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disseisee",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disseizin ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disseizin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disseisin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disseizins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disseisins",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disseizin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disseisin's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disseizor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disseizor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disseisor",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disseizors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disseisors",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disseizor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disseisor's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dissensualize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dissensualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissensualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dissensualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissensualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dissocialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dissocialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissocialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dissocialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissocialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dissocializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissocialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dissocializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissocialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dissyllabize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dissyllabize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissyllabise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dissyllabized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dissyllabized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissyllabised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dissyllabizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dissyllabizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissyllabising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dissympathize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dissympathize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissympathise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dissympathizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dissympathises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "distill ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "distill",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "distil",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "distilled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "distilling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "distills",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "distils",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "disutilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "disutilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disutilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "disutilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "disutilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ditsy ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ditsy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ditzy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ditsier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ditzier",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ditziness ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ditziness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ditsiness",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ditzinesses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ditsinesses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ditziness's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ditsiness's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "diversicolored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "diversicolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "diversicoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "divinister ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "divinister",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "divinistre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "divinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "divinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "divinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "divinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "divinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "divinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "divinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "divinize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "divinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "divinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "divinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "divinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "divinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "divinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "divinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "divinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "divisionalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "divisionalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "divisionalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "divisionalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "divisionalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "djellaba ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "djellaba",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "djellabah",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "djellabas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "djellabahes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "djellaba's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "djellabah's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Docete ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Docete",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Docetae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Docete's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Docetae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dockization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dockization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dockisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dockizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dockisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dockization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dockisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dockize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dockize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dockise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dockized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dockised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dockizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dockising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dockizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dockises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "doctorization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "doctorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doctorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "doctorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doctorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "doctorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doctorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "doctorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "doctorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doctorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "doctorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doctorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "doctrinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "doctrinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doctrinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "doctrinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doctrinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "doctrinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doctrinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "doctrinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "doctrinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doctrinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "doctrinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doctrinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "documentarization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "documentarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "documentarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "documentarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "documentarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "documentarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "documentarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "documentarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "documentarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "documentarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "documentarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "documentarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "documentarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "documentarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "documentize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "documentize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "documentise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "documentizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "documentises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Dodonean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Dodonean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Dodonaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Dodonean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Dodonaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "doggone ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "doggone",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doggoned",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::AdjectiveOrAdverb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "doggoner",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doggoneder",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::AdjectiveOrAdverb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "doggonest",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doggonedest",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::AdjectiveOrAdverb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "doggy ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "doggy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doggie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "doggy's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doggie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dogie ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dogie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dogy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dogies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dogie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dogy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dogmatization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dogmatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dogmatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dogmatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dogmatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dogmatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dogmatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dogmatize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dogmatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dogmatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dogmatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dogmatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dogmatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dogmatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dogmatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dogmatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dogmatizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dogmatizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dogmatiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dogmatizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dogmatisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dogmatizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dogmatiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dognap ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dognaped",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dognapped",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dognaping",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dognapping",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "d'oh ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "d'oh",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doh",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "d'ohs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dohs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "d'oh's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doh's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dollarization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dollarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dollarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dollarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dollarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dolomitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dolomitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dolomitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dolomitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dolomitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dolomitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dolomitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dolomitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dolomitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dolomitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dolomitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dolomitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dolomitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dolomitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dolomitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dolomitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dolor ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "dolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "domesticize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "domesticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "domesticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "domesticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "domesticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "domesticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "domesticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "domesticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "domesticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "domine ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "domine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dominae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dopey ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dopey",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dopy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dora ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dora",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doura",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Doricize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Doricize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Doricise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Doricizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Doricises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Doricize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Doricise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "doryline ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "doryline",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dorylinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "doughnut ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "doughnut",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "donut",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "doughnuts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "donuts",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "doughnut's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "donut's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dowel ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "doweled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dowelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "doweling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dowelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "doweler ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "doweler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "doweller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "downdraft ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "downdraft",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "downdraught",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "downdraft's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "downdraught's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "downward ",
  verified: false,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "downward",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "downward",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "downwards",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adverb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dracena ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dracena",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dracaena",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dracenas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dracaenas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dracenaceae ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dracenaceae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dracaenaceae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "draft ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "draft",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draught",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("current of air"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "drafts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughts",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("current of air"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draft's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draught's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("current of air"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draft",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("writing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "drafts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("writing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draft's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("writing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draft",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draught",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: Some("writing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "drafted",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughted",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: Some("writing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "drafting",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughting",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: Some("writing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "drafts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughts",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: Some("writing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draught",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("british: game of checkers"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draughts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("british: game of checkers"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draught's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("british: game of checkers"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "draftboard ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "draughtboard",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draftboard",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("british word"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draughtboards",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draftboards",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("british word"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "draftily ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "draftily",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtily",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draftilier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtilier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draftiliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtiliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "draftiness ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "draftiness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtiness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draftinesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtinesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draftiness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtiness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "draftsman ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "draftsman",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtsman",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draftsmen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtsmen",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draftsman's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtsman's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draughtsman",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("british: checker"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draughtsman's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("british: checker"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draughtsmen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("british: checker"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "draftsmanship ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "draftsmanship",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtsmanship",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draftsmanships",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtsmanships",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draftsmanship's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtsmanship's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "draftsperson ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "draftsperson",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtsperson",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "draftswoman ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "draftswoman",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtswoman",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draftswomen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtswomen",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draftswoman's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtswoman's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "drafty ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "drafty",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughty",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draftier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "draftiest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "draughtiest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dragonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dragonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dragonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dragonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dragonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dragonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dragonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dragonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dragonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dramatizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dramatizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dramatizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dramatizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dramatizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dramatizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dramatization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dramatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dramatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dramatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dramatize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dramatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dramatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dramatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dramatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dramatizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dramatizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dramatizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dramatizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dramatiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dreamed ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dreamed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dreamt",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "drey ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "drier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dryer",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "driest",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dryest",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dryer",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "drier",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "drivel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "driveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "drivelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "driveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "drivelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "driveler ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "driveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "driveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "drivelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "drivellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "driveler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "driveller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "drought ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "drought",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "drouth",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "droughts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "drouthes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "drought's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "drouth's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "droughtiness ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "droughtiness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "drouthiness",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "droughty ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "droughty",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "drouthy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "druggy ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "druggy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "druggie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "druggie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "druggy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "druggies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "druggie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "druggy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dryly ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dryly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "drily",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dualization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dualize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ductilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ductilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ductilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ductilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ductilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "duel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dueled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "duelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dueling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "duelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "duelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "duellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dueler ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dueler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dueller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "duelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "duellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dueler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dueller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "duelist ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "duelist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "duellist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "duelists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "duellists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "duelist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "duellist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "duer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "duer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dure",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dullness ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dullness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dulness",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dullnesses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dulnesses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dullness's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dulness's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dumbfound ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dumbfound",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dumfound",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dumbfounded",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dumfounded",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dumbfounding",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dumfounding",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dumbfounds",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dumfounds",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dumbfoundingly ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dumbfoundingly",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dumfoundingly",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dykey ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dykey",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dikey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("lesbian"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dykier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dikier",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("lesbian"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dykiest",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dikiest",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("lesbian"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dynamize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dynamize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dynamise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dynamized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dynamised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dynamizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dynamising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dynamizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "dynamises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dysesthesia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dysesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dysaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dysesthetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dysesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dysaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dysmenorrhea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dysmenorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dysmenorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dysmenorrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dysmenorrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dysmenorrhea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dysmenorrhoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dysmenorrheal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dysmenorrheal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dysmenorrhoeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dysmenorrheic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dysmenorrheic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dysmenorrhoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dyspnea ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dyspnea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dyspnoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dyspneas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dyspnoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dyspnea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dyspnoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dyspneal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dyspneal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dyspnoeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "dyspneic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dyspneic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "dyspnoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "earlierize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "earlierize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "earlierise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "earlierized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "earlierised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "earlierizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "earlierising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "earlierizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "earlierises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "easternize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "easternize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "easternise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ebionize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ebionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ebionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ebionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ebionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ebionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ebionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ebionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ebionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ebonization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ebonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ebonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ebonize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ebonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ebonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ebonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ebonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ebonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ebonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ebonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ebonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ec ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ec",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aec",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ecclesiasticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ecclesiasticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ecclesiasticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ecclesiasticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ecclesiasticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "echoize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "echoize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "echoise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "echoized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "echoised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "echoizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "echoising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "echoizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "echoises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eclecticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eclecticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eclecticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eclecticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eclecticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ecoid ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ecoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oecoid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "economization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "economization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "economisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "economizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "economisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "economization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "economisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "economize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "economize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "economise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "economized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "economised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "economizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "economising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "economizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "economises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "economizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "economizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "economiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "economizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "economisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "economizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "economiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ecophobia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ecophobia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oecophobia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ecstasize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ecstasize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ecstasise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ecstasized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ecstasised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ecstasizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ecstasising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ecstasizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ecstasises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ecstaticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ecstaticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ecstaticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ecstaticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ecstaticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ectethmoid ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ectethmoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ectoethmoid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ectocelic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ectocelic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ectocoelic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ecumenic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ecumenic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oecumenic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ecumenicalism ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ecumenicalism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oecumenicalism",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ecumenicity ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ecumenicity",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oecumenicity",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ecumenism ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ecumenism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oecumenism",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ecus ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ecus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oecus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "edema ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "edema",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oedema",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "edemas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oedemas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "edema's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oedema's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "edematose ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "edematose",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oedematose",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "edematous ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "edematous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oedematous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Edenization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Edenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Edenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Edenizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Edenisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Edenization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Edenisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Edenize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Edenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Edenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Edenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Edenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Edenize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Edenise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "edicule ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "edicule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aedicule",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "edicule's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aedicule's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "edile ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aedile",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "edile",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aediles",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ediles",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aedile's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "edile's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "edility ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "edility",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aedility",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "editorialization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "editorialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "editorialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "editorializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "editorialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "editorialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "editorialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "editorialize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "editorialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "editorialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "editorialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "editorialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "editorializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "editorialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "editorializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "editorialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "editorializer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "editorializer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "editorialiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "editorializer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "editorialiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eerie ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eerie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eery",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Eetion ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Eetion",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Eaetion",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Eetion's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Eaetion's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "effectualize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "effectualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "effectualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "effectualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "effectualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "effeminatize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "effeminatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "effeminatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "effeminatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "effeminatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "effeminization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "effeminization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "effeminisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "effeminization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "effeminisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "effeminize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "effeminize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "effeminise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "effeminized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "effeminised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "effeminizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "effeminising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "effeminizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "effeminises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eger ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "egre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "egilops ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "egilops",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aegilops",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eglogue ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eglogue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeglogue",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "egoize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "egoize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "egoise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "egoizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "egoises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "egoizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "egoizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "egoiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "egoizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "egoisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "egophony ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "egophony",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aegophony",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "egotize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "egotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "egotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "egotized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "egotised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "egotizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "egotising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "egotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "egotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "egritude ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "egritude",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aegritude",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Egyptianization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Egyptianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Egyptianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Egyptianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Egyptianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Egyptianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Egyptianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Egyptianized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptianised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Egyptianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Egyptianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Egyptianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Egyptianizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptianising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Egyptize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Egyptize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Egyptizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Egyptize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Egyptise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "elasticization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "elasticization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elasticisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "elasticize ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "elasticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elasticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "elasticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elasticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "elasticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elasticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "elasticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elasticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "elasticizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "elasticizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elasticiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "elasticizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elasticisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electricalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electricalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electricalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electricalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electricalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electricize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electricize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electricise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electricizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electricises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electrization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electrization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electrize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electroanesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electroanesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electroanaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electroanesthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electroanaesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electroanesthesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electroanaesthesia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electrocauterization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electrocauterization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrocauterisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrocauterizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrocauterisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrocauterization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrocauterisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electrodialyze ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electrodialyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrodialyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrodialyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrodialyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electrodialyzer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electrodialyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrodialyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrodialyzers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrodialysers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrodialyzer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrodialyser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electrogalvanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electrogalvanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrogalvanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrogalvanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrogalvanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electrohomeopathy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electrohomeopathy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrohomoeopathy",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrohomeopathies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrohomoeopathies",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrohomeopathy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrohomoeopathy's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electrolyzation ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electrolyzation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrolysation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrolyzation's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrolysation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electrolyze ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electrolyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrolyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrolyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrolysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrolyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrolysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrolyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrolyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrolysis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrolyses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2 "),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electrolyzer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electrolyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrolyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrolyzer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrolyser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electromagnetizable ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electromagnetizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electromagnetisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "electrotonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "electrotonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrotonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "electrotonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "electrotonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "elegize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "elegize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elegise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "elegized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elegised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "elegizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elegising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "elegizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elegises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "elementalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "elementalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elementalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "elementalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elementalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eleoblast ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eleoblast",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elaeoblast",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eleolite ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eleolite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elaeolite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eleomargaric ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eleomargaric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elaeomargaric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eleometer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eleometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elaeometer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eleoptene ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eleoptene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elaeoptene",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eleoptene's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "elaeoptene's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Elizabethanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Elizabethanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Elizabethanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Elizabethanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Elizabethanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Elizabethanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Elizabethanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Elizabethville ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Elizabethville",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Elisabethville",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Elizabethville's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Elisabethville's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "embed ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "embed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imbed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "embedded",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imbedded",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "embedding",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imbedding",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "embeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imbeds",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "embed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("<N> journalist"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "embeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("<N> journalist"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "embed's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("<N> journalist"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "embedment ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "embedment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imbedment",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "embedments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imbedments",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emblematicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emblematicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emblematicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emblematicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emblematicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emblematization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emblematization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emblematisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emblematize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emblematize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emblematise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emblematized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emblematised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emblematizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emblematising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emblematizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emblematises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emblemize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emblemize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emblemise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emblemized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emblemised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emblemizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emblemising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emblemizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emblemises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "embolization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "embolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "embolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "embowel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emboweled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "embowelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emboweling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "embowelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emir ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emir",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amir",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ameer",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emirs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amirs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ameers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emir's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "amir's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ameer's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emotionalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emotionalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emotionalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emotionalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emotionalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emotionalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emotionalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emotionalize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emotionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emotionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emotionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emotionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emotionalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emotionalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emotionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emotionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emotionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emotionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emotionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emotionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emotionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "empathize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "empathize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empathise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "empathized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empathised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "empathizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empathising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "empathizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empathises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emperize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emperize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emperise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emperized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emperised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emperizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emperising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emperizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emperises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "empestic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "empestic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "empaestic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emphasize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emphasize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emphasise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emphasized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emphasised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emphasizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emphasising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emphasizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emphasises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emphasizer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emphasizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emphasiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emphasizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emphasisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "employee ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "employee",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "employe",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "employees",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "employes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "employee's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "employe's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emprise ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emprise",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emprize",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emprises",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emprizes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "empyreumatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "empyreumatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empyreumatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "empyreumatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empyreumatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "empyreumatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empyreumatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "empyreumatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empyreumatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aemule",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emuled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aemuled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emuling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aemuling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emules",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aemules",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "emulsionize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "emulsionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emulsionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emulsionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emulsionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emulsionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emulsionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "emulsionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "emulsionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enamel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enameled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "enamelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enameling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "enamelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "enamellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enameler ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enameler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enameller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enamellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enameler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enameller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enamelist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enamelist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enamellist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamelists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enamellists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamelist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enamellist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enamor ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enamor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamoreds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamoureds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamored's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamoured's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enamoredness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enamoredness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamouredness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamorednesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamourednesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamoredness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamouredness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enamorment ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enamorment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamourment",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamorments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamourments",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enamorment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enamourment's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enarbor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enarbor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enarbour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enarbors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enarbours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enarbor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enarbour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "encarnalization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encarnalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encarnalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "encarnalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encarnalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encarnalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encarnalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encarnalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encarnalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encarnalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encarnalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encarnalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "encephalocele ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encephalocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encephalocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enclose ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enclose",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inclose",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enclosed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inclosed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enclosing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inclosing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encloses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incloses",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enclosure ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enclosure",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inclosure",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enclosures",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inclosures",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enclosure's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inclosure's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "encolor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "encolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "encoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encoloring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "encolouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "encolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "encolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "encrust ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encrust",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incrust",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encrusted",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incrusted",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encrusting",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incrusting",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encrusts",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incrusts",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "encumber ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encumber",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incumber",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encumbered",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incumbered",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encumbering",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incumbering",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encumbers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incumbers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "encyclopedia ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encyclopedia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaedia",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encyclopedias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaedias",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encyclopedia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaedia's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "ODE has encyclopEdia first, OED has encyclopAEdia first.  Many",
    "British will prefer the AE spelling.  In the BNC the AE form is",
    "slightly more common (around 6:5 ratio).  Thus I'm not sure how to",
    "encode this.",
  ],
  },
Cluster {
  header: "encyclopediac ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encyclopediac",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaediac",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "See note for encyclopedia",
  ],
  },
Cluster {
  header: "encyclopedial ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encyclopedial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaedial",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "See note for encyclopedia",
  ],
  },
Cluster {
  header: "encyclopedian ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encyclopedian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaedian",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "See note for encyclopedia",
  ],
  },
Cluster {
  header: "encyclopedic ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encyclopedic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaedic",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "See note for encyclopedia",
  ],
  },
Cluster {
  header: "encyclopedical ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encyclopedical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaedical",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "See note for encyclopedia",
  ],
  },
Cluster {
  header: "encyclopedically ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encyclopedically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaedically",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "See note for encyclopedia",
  ],
  },
Cluster {
  header: "encyclopedism ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encyclopedism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaedism",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encyclopedisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaedisms",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "See note for encyclopedia",
  ],
  },
Cluster {
  header: "encyclopedist ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "encyclopedist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaedist",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "encyclopedists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "encyclopaedists",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "See note for encyclopedia",
  ],
  },
Cluster {
  header: "endameba ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "endameba",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "endamoeba",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endamebas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "endamoebas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endamebae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "endamoebae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "endamebiasis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "endamebiasis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "endamoebiasis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "endamebic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "endamebic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "endamoebic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "endeavor ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "endeavor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "endeavour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endeavored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "endeavoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endeavoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "endeavouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endeavors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "endeavours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endeavor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "endeavour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "endeavorer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "endeavorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "endeavourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endeavorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "endeavourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endeavorer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "endeavourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Enders ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Enders",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Endres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "endorse ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "endorse",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indorse",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endorsed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indorsed",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endorsing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indorsing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endorses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indorses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "endorsement ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "endorsement",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indorsement",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endorsements",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indorsements",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endorsement's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indorsement's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "endue ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "endue",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indue",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endued",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indued",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enduing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "induing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "endues",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indues",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eneid ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eneid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeneid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "energization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "energization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "energisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "energize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "energize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "energise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "energized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "energised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "energizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "energising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "energizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "energises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "energizer ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "energizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "energiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "energizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "energisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "energizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "energiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Englishize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Englishize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Englishise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Englishizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Englishises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Englishize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Englishise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "engrandize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "engrandize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "engrandise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "engrandizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "engrandises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "engrandizement ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "engrandizement",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "engrandisement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "engrandizements",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "engrandisements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "engrandizement's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "engrandisement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enharbor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enharbor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enharbour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enhemospore ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enhemospore",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enhaemospore",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enhypostatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enhypostatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enhypostatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enhypostatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enhypostatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enhypostatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enhypostatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enhypostatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enhypostatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enigmatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enigmatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enigmatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enigmatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enigmatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enigmatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enigmatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enigmatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enigmatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enjambment ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enjambment",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enjambement",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enjambments",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enjambements",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enjambment's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enjambement's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enocyte ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enocyte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oenocyte",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enolic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enolic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oenolic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enolization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enolizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enolisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enolization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enolisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enolize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enological ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oenological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enologist ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oenologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oenologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enomania ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enomania",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oenomania",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enroll ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enroll",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enrol",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enrolls",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enrols",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enrollment ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enrollment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enrolment",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enrollments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enrolments",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enrollment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "enrolment's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ensepulcher ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ensepulcher",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ensepulchre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ensepulcherred",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ensepulchred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ensepulcherring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ensepulchring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ensepulchers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ensepulchres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ensepulcher's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ensepulchre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ensorcel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ensorceled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ensorcelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ensorcels",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ensorcells",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "entameba ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "entameba",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "entamoeba",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entamebas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "entamoebas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "entamebae ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "entamebae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "entamoebae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "entamebic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "entamebic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "entamoebic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "entender ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "entender",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "entendre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entenders",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "entendres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enterocele ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enterocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enterocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enteroceles",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enterocoeles",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enthrall ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enthrall",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthral",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthralling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthralled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthralls",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthrals",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthrall",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthral",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inthrall",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("enslave"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthralling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inthralling",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("enslave"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthralled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inthralled",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("enslave"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthralls",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthrals",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inthralls",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("enslave"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enthrallment ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enthrallment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthralment",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthrallments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthralments",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthrallment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthralment's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enthronization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enthronization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthronisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthronizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthronisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthronization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthronisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "enthronize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "enthronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthronized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthronised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthronizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthronising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "enthronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enthronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "entocele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "entocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "entocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "entomologize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "entomologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "entomologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entomologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "entomologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entomologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "entomologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entomologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "entomologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "entre ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "entre",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "entrae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "entrench ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "entrench",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrench",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entrenched",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrenched",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entrenching",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrenching",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entrenches",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrenches",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "entrenchment ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "entrenchment",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrenchment",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entrenchment's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrenchment's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entrenchments",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrenchments",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "entrust ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "entrust",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrust",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entrusted",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrusted",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entrusting",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrusting",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entrusts",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrusts",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "entrustment ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "entrustment",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrustment",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "entrustments",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intrustments",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "envapor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "envapor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "envapour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "envapors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "envapours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "envapor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "envapour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "envenomization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "envenomization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "envenomisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eolic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eolic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeolic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eolipile ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eolipile",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeolipile",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eolipiles",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeolipiles",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eolipile's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeolipile's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eolotropic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eolotropic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeolotropic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eon ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "aeon",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "aeons",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "aeon's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("technical"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("technical"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("technical"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eonian ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eonian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeonian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eonism ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eonism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeonism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eonisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeonisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eonism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeonism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Eopaleozoic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Eopaleozoic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Eopalaeozoic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Eopaleozoic's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Eopalaeozoic's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epaulet ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epaulet",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "epaulette",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epaulets",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "epaulettes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epaulet's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "epaulette's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epenetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epenetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epaenetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ephete ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ephete",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ephetae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epicele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epicele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epicoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epicenter ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epicenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epicentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epicenters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epicentres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epicenter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epicentre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Epicurize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Epicurize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Epicurise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Epicurizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Epicurises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Epicurize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Epicurise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epicurize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epicurise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epicurized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epicurised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epicurizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epicurising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epicurizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epicurises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epidotization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epidotization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epidotisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epidotizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epidotisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epidotized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epidotized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epidotised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epigeal ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epigeal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epigaeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epigean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epigean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epigaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epigeous ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epigeous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epigaeous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epigrammatization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epigrammatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epigrammatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epigrammatize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epigrammatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epigrammatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epigrammatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epigrammatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epigrammatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epigrammatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epigrammatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epigrammatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epigrammatizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epigrammatizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epigrammatiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epilogize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epilogize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epilogise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epilogized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epilogised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epilogizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epilogising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epilogizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epilogises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epilogue ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epilogue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epilog",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epilogues",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epilogs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epilogue's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epilog's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epiloguize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epiloguize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epiloguise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epiloguized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epiloguised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epiloguizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epiloguising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epiloguizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epiloguises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epimerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epimerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epimerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epimerized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epimerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epimerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epimerizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epimerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epimerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epinephrine ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epinephrine",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epinephrin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epinephrines",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epinephrins",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epinephrine's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epinephrin's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epiphanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epiphanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epiphanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epiphanized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epiphanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epiphanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epiphanizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epiphanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epiphanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Episcopalianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Episcopalianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Episcopalianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Episcopalianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Episcopalianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Episcopalianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Episcopalianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "episcopize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "episcopize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "episcopise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "episcopized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "episcopised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "episcopizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "episcopising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "episcopizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "episcopises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epistolize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epistolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epistolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epistolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epistolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epistolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epistolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epistolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epistolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epitaphize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epitaphize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epitaphise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epitaphizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epitaphises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epithetize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epithetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epithetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epithetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epithetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epitomization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epitomization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epitomisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epitomizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epitomisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epitomization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epitomisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epitomize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epitomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epitomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epitomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epitomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epitomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epitomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epitomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epitomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "epitomizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "epitomizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epitomiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epitomizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epitomisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "epitomizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "epitomiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eq ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eq",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeq",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "equal ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "equaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "equaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "equalization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "equalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "equalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "equalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "equalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "equalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "equalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "equalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "equalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "equalizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equalisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "equalizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "equalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "equalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "equalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "equestrianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "equestrianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equestrianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "equestrianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "equestrianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "erbia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "erbia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "erbias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ergotize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ergotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ergotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ergotized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ergotised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ergotizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ergotising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ergotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ergotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "erical ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "erical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aerical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "erke ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "erke",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reke",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ern ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ern",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ren",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ernes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "erns",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rens",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ernest ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ernest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "erogenesis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "erogenesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aerogenesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "erogenic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "erogenic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aerogenic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eros ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eros",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeros",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "erose",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aerose",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eroticization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eroticization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eroticisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eroticize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eroticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eroticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eroticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eroticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eroticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eroticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eroticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eroticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "erses ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "erses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ersh ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ersh",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resh",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reshes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reshoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ert ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ert",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ret",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eruginous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eruginous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeruginous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "erugo ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "erugo",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aerugo",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "erugos",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aerugos",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ervalenta ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ervalenta",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revalenta",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ervalentas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revalentas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "erythrean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "erythrean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "erythraean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "erythremia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "erythremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "erythraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esc ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esc",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesc",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "escalades ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "escalades",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "escaladoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "escalope ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "escalope",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "escollope",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "escalop",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "escalopes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "escollopes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "escalops",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "escalope's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "escollope's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "escalop's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eschynite ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eschynite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeschynite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esculapian ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esculapian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesculapian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esculetin ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esculetin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesculetin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esculin ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esculin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesculin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "esculin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesculin's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Eskimoized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Eskimoized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Eskimoised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Eskimoizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Eskimoiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Eskimoized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Eskimoised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esop ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esop",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aesop",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esophagal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esophagal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oesophagal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esophagean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esophagean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oesophagean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esophagism ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esophagism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oesophagism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esophagismus ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esophagismus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oesophagismus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esophagitis ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esophagitis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oesophagitis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esophagus ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esophagus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oesophagus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "esophagi",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "esophaguses",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oesophagi",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oesophaguses",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Essenize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Essenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Essenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Essenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Essenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Essenize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Essenise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "essentialize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "essentialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "essentialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "essentializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "essentialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esterization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esterization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "esterisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "esterizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "esterisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "esterization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "esterisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esterize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "esterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "esterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "esterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esthesia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "esthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "esthesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesthesia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esthesiogen ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esthesiogen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesthesiogen",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "esthesiogens",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesthesiogens",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esthesiogenic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esthesiogenic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesthesiogenic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esthesis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esthesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesthesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estheses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aestheses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esthetic ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "esthetic",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aesthetic's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "esthetic's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "aesthetics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "esthetics",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esthetical ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esthetical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesthetical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esthetically ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "aesthetically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "esthetically",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esthetician ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esthetician",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aesthetician",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estheticians",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aestheticians",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "esthetician's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aesthetician's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estival ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estival",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aestival",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estivate ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estivate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aestivate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estivated",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aestivated",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estivating",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aestivating",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estivates",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aestivates",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estivation ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estivation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aestivation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estivations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aestivations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estivation's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aestivation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estivator ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estivator",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aestivator",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estivator's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aestivator's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estive ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estive",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aestive",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estradiol ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estradiol",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestradiol",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estradiols",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestradiols",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estradiol's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestradiol's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estral ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estral",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestral",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estrin ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estrin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestrin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estrins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestrins",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estrin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestrin's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estriol ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estriol",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestriol",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estriols",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestriols",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estriol's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestriol's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estrogen ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estrogen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oestrogen",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estrogens",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oestrogens",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estrogen's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oestrogen's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estrogenic ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estrogenic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oestrogenic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estrone ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estrone",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestrone",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estrones",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestrones",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estrone's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestrone's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estrous ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estrous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestrous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estrual ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estrual",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestrual",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estruate ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estruate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestruate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estruation ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estruation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestruation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estrum ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estrum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestrum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estrums",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestrums",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estrus ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estrus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestrus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estruses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestruses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "estrus's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oestrus's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estuate ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estuate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aestuate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estuous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estuous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aestuous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "esture ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "esture",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aesture",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "estus ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "estus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aestus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "et ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "et",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aet",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eternalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eternalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eternalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eternalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eternalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eternalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eternalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eternalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eternalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eternization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eternization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eternizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eternization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eternize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eternize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eternized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eternizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eternizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eternises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etheling ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etheling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aetheling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etherealization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etherealization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherealisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherealizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherealisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherealization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherealisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etherealize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etherealize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherealise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherealized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherealised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherealizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherealising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherealizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherealises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ethered ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ethered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aethered",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etherialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etherialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etherialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etherialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etherialized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etherialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etherializing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etherializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etheric ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etheric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aetheric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etherization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etherization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etherize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etherize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etherizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etherizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etheriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etherisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etherizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etheriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ethicization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ethicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ethicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ethicize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ethicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ethicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ethicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ethicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ethicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ethicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ethicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ethicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ethnicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ethnicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ethnicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ethnicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ethnicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etiogenic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etiogenic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aetiogenic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etiologically ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etiologically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aetiologically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etiologist ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etiologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aetiologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etiologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aetiologist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etiologue ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etiologue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aetiologue",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etiology ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etiology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aetiology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etiology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aetiology's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etiophyllin ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etiophyllin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aetiophyllin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etiotropic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etiotropic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aetiotropic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etiotropically ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etiotropically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "aetiotropically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etymologizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etymologizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etymologisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etymologization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etymologization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etymologisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "etymologize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "etymologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etymologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etymologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etymologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etymologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etymologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "etymologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "etymologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eudemon ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eudemon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eudemons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemons",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eudemonia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eudemonia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemonia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eudemonic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eudemonic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemonic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eudemonics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemonics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eudemonism ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eudemonism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemonism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eudemonisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemonisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eudemonist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eudemonist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemonist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eudemonists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemonists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eudemonist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemonist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eudemonistic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eudemonistic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemonistic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eudemonistical ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eudemonistical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemonistical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eudemonistically ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eudemonistically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemonistically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eudemony ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eudemony",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eudaemony",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "euhemerize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "euhemerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euhemerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euhemerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euhemerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euhemerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euhemerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euhemerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euhemerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eulogization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eulogization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eulogisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eulogizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eulogisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eulogization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eulogisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eulogize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eulogize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eulogise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eulogized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eulogised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eulogizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eulogising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eulogizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eulogises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eulogizer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eulogizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eulogiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eulogizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eulogisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eulogizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eulogiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eunuchize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eunuchize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eunuchise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eunuchized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eunuchised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eunuchizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eunuchising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eunuchizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eunuchises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "euphemization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "euphemization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphemisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euphemization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphemisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "euphemize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "euphemize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphemise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euphemized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphemised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euphemizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphemising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euphemizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphemises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "euphemizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "euphemizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphemiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euphemizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphemisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euphemizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphemiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "euphonization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "euphonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "euphonize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "euphonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euphonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euphonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euphonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "euphuize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "euphuize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphuise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euphuized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphuised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euphuizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphuising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "euphuizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "euphuises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eupnea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eupnea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "eupnoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eupneas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "eupnoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eupnea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "eupnoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eupneic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eupneic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "eupnoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Europeanization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Europeanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Europeanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Europeanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Europeanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Europeanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Europeanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Europeanize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Europeanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Europeanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Europeanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Europeanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Europeanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Europeanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Europeanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Europeanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "europeanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "europeanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "europeanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "europeanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "europeanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "europeanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "europeanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "europeanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Euryale ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Euryale",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Euryalae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Euryale's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Euryalae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "evangelization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "evangelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evangelisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "evangelizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evangelisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "evangelization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evangelisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "evangelize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "evangelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evangelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "evangelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evangelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "evangelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evangelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "evangelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evangelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "evangelizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "evangelizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evangeliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "evangelizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evangelisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "evangelizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evangeliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eventualize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eventualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eventualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eventualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eventualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eventualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eventualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eventualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eventualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "evil ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eviler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eviller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "evilest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evillest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eviternal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eviternal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "aeviternal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "evolutionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "evolutionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evolutionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "evolutionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "evolutionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "excursionize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "excursionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "excursionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "excursionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "excursionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "excursionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "excursionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "excursionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "excursionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "exestuate ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "exestuate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exaestuate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "exhibitionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "exhibitionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exhibitionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "exhibitionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exhibitionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "existentialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "existentialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "existentialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "existentializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "existentialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "exorcise ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "exorcise",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exorcize",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "exorcised",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exorcized",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "exorcising",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exorcizing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "exorcises",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exorcizes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "exorcization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "exorcization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exorcisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "exorcizement ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "exorcizement",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exorcisement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "exorcizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "exorcizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exorciser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "exorcizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exorcisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "exorcizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exorciser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "expediter ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "expediter",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "expeditor",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "expediters",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "expeditors",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "expediter's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "expeditor's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "experimentalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "experimentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "experimentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "experimentalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "experimentalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "experimentalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "experimentalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "experimentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "experimentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "experimentize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "experimentize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "experimentise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "experimentizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "experimentises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "expertize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "expertized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "expertised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "expertizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "expertising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "exsiccate ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "exsiccate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exsiccatae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "extemporization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "extemporization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extemporisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "extemporizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extemporisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "extemporization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extemporisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "extemporize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "extemporize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extemporise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "extemporized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extemporised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "extemporizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extemporising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "extemporizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extemporises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "extemporizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "extemporizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extemporiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "extemporizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extemporisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "extemporizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extemporiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "extendable ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "extendable",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extendible",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "exteriorization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "exteriorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exteriorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "exteriorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exteriorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "exteriorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exteriorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "exteriorize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "exteriorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exteriorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "exteriorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exteriorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "exteriorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exteriorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "exteriorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "exteriorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "externalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "externalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "externalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "externalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "externalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "externalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "externalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "externalize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "externalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "externalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "externalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "externalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "externalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "externalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "externalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "externalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "extol ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "extol",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extoll",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "extols",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extolls",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "extrovert ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "extrovert",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extravert",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "extroverted",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extraverted",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "extroverts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extraverts",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "extrovert's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "extravert's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "extravert is marked as a \"chiefly Psychology\" variant",
  ],
  },
Cluster {
  header: "eying ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eyeing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eying",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fabulize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fabulize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fabulise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fabulized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fabulised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fabulizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fabulising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fabulizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fabulises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "facsimilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "facsimilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "facsimilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "facsimilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "facsimilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "factorization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "factorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "factorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "factorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "factorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "factorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "factorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "factorize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "factorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "factorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "factorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "factorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "factorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "factorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "factorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "factorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "faence ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "faence",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faoence",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "faence's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faoence's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "faerie ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "faerie",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faery",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "faerie's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faery's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fagot ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "faggot",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "faggots",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "faggot's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fagot",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "faggot",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("bundle"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fagots",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "faggots",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("bundle"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fagot's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "faggot's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("bundle"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fagot",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "faggot",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: Some("bundle"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fagots",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "faggots",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: Some("bundle"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fagoted",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "faggoted",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: Some("bundle"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fagoting",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "faggoting",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: Some("bundle"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "faitor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "faitor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faitour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "faitors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faitours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "falafel ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "falafel",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "felafel",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "falafel's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "felafel's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "familiarization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "familiarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "familiarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "familiarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "familiarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "familiarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "familiarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "familiarize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "familiarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "familiarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "familiarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "familiarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "familiarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "familiarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "familiarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "familiarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "familiarizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "familiarizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "familiariser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "familiarizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "familiarisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "familiarizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "familiariser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "familiarizingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "familiarizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "familiarisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fanaticize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fanaticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fanaticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fanaticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fanaticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fanaticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fanaticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fanaticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fanaticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fantasize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fantasize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fantasise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fantasized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fantasised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fantasizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fantasising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fantasizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fantasises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fantasizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fantasizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fantasiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fantasy ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fantasy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phantasy",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fantasied",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phantasied",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fantasying",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phantasying",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fantasies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phantasies",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fantasy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phantasy's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "phantasy marked as archaic or Psych. var. of fantasy by Ox (all versions)",
  ],
  },
Cluster {
  header: "faradization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "faradization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faradisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "faradizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faradisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "faradization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faradisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "faradize ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "faradize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faradise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "faradized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faradised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "faradizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faradising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "faradizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faradises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "faradizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "faradizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faradiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "faradizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faradisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "faradizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faradiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fasciole ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fasciole",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fasciolae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fascisticization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fascisticization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fascisticisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fascisticizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fascisticisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fascisticization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fascisticisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fascisticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fascisticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fascisticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fascisticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fascisticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fascistization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fascistization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fascistisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fascistizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fascistisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fascistization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fascistisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fascistize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fascistize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fascistise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fascistized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fascistised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fascistizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fascistising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fascistizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fascistises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fashionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fashionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fashionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fashionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fashionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fatalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fatalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fatalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fatalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fatalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favor ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favoreds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favoureds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favored's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favoured's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favoring's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouring's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favorable ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favorableness ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favorableness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourableness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorablenesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourablenesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorableness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourableness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favorably ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favorably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorablier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourablier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorabliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourabliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favoredly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favoredly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouredly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favoredlier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouredlier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favoredliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouredliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favoredness ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favoredness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouredness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorednesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourednesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favoredness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouredness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favorer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favoress ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favoress",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouress",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favoringly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favoringly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouringly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favoringlier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouringlier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favoringliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouringliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favorite ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favorite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favoriter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouriter",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorites",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourites",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favoritest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouritest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorite's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourite's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favoritism ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favoritism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouritism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favoritisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouritisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favoritism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favouritism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favorless ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favorless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourlesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "favorlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourlessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "favorlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "favorlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "favourlesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fayer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fayer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fayre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fe ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fecal ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fecal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "faecal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fecalith ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fecalith",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "faecalith",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "feces ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feces",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "faeces",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feces's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "faeces's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fecula ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fecula",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faecula",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "feculence ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feculence",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "faeculence",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Fecunditatis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Fecunditatis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Foecunditatis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Fecunditatis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Foecunditatis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fecundize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fecundize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fecundise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fecundizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fecundises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fedarie ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fedarie",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foedarie",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fedaries",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foedaries",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "federalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "federalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "federalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "federalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "federalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "federalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "federalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "federalize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "federalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "federalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "federalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "federalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "federalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "federalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "federalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "federalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Fedor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Fedor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Faedor",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Fedor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Faedor's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "feer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fere",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "femalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "femalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "femalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "femalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "femalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "feminization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feminization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feminisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feminizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feminisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feminization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feminisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "feminize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feminize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feminise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feminized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feminised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feminizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feminising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feminizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feminises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "femtometer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "femtometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "femtometre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "femtometers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "femtometres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "feralized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ferd ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ferd",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ferie ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ferie",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feriae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ferity ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ferity",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "freity",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ferneau ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ferneau",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Freneau",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ferneau's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Freneau's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ferrel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ferreled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ferrelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ferreling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ferrelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ferritization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ferritization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ferritisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ferritizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ferritisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ferritization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ferritisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fertilizability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fertilizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fertilizable ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fertilizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fertilization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fertilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fertilizational ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fertilizational",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisational",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizationaler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisationaler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizationalest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisationalest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizationals",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisationals",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizational's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisational's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fertilize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fertilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fertilizer ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fertilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertiliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertilisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fertilizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fertiliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fervor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fervor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "fervour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fervors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "fervours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fervor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "fervour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fervorless ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fervorless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "fervourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fervorlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fervorlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "fervourlesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetal ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetal",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetalism ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetalism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetalism",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetation ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetation",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetation's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetation's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetichize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetichize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetichise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetichized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetichised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetichizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetichising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetichizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetichises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "feticidal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feticidal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foeticidal",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "feticide ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feticide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foeticide",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feticides",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foeticides",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feticide's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foeticide's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetid ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetid",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetider",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetider",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetidest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetidest",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetiferous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetiferous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetiferous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetiparous ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetiparous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetiparous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetish ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetish",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetich",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetishes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetiches",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetish's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetich's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetishization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetishization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetishisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetishizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetishisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetishization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetishisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetishize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetishize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetishise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetishized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetishised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetishizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetishising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetishizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fetishises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetor",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetors",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetor's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetoscopy ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetoscopy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetoscopy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetoscopies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetoscopies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "feture ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feture",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foeture",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fetus ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fetus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetus",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetuses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetuses",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fetus's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foetus's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "foetus marked as British-only variant for \"non-technical use\" by Ox ",
  ],
  },
Cluster {
  header: "feudalizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feudalizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feudalisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feudalizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feudalisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feudalizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feudalisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "feudalization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feudalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feudalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feudalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feudalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feudalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feudalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "feudalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feudalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feudalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feudalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feudalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feudalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feudalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "feudalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feudalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "feuter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "feuter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "feutre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiber ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiber",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fibers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiber's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiberboard ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiberboard",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibreboard",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberboards",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibreboards",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberboard's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibreboard's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fibered ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fibered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiberfill ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiberfill",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrefill",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberfills",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrefills",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberfill's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrefill's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiberglass ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiberglass",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibreglass",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberglasses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibreglasses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberglass's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibreglass's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiberization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiberization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrization",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiberize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiberize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrize",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrized",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrizing",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrizes",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiberizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiberizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrizer",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrizers",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiberless ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiberless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibreless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrelesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrelessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiberlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiberlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrelesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiberscope ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiberscope",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrescope",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiberscopes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibrescopes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiberware ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiberware",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fibreware",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fictionalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fictionalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fictionalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fictionalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fictionalize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fictionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fictionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fictionalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fictionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fictionization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fictionization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fictionizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fictionization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fictionize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fictionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fictionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fictionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fictionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fictionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "figurize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "figurize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "figurise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "figurizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "figurises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "filet ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "filet",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fillet",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "filet",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fillets",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "filets",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fillet's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "filet's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fillet",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "filet",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "filleted",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fileted",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "filleting",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fileting",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fillets",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "filets",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "filmize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "filmize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "filmise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "filmizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "filmises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "filterable ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "filterable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "filtrable",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "finalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "finalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "finalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "finalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "finalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "finalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "finalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "finalize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "finalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "finalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "finalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "finalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "finalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "finalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "finalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "finalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Finlandization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Finlandization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Finlandisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Finlandizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Finlandisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiscalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiscalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fiscalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiscalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fiscalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiscalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fiscalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fiscalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fiscalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fiscalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fiscalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fiscalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fjord ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fjord",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fiord",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fjords",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fiords",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fjord's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fiord's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flamboyantize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flamboyantize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flamboyantise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flamboyantizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flamboyantises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flannel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flanneled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "flannelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flanneling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "flannelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flannelette ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flannelette",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flannelet",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flannelette's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flannelet's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flavor ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flavor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavorings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavoring's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavouring's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flavorer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flavorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavorer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flavorful ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flavorful",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourful",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavorfuler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourfuler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavorfulest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourfulest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flavorfully ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flavorfully",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourfully",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavorfullier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourfullier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavorfulliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourfulliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flavorless ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flavorless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavorlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourlesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavorlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourlessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flavorlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flavorlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourlesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flavorous ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flavorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flavourous",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flavorsome ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flavorsome",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavoursome",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavorsomer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavoursomer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavorsomest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavoursomest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flavory ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flavory",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavoury",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavorier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavourier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flavoriest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flavouriest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fledgling ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fledgling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fledgeling",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fledglings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fledgelings",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fledgling's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fledgeling's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Fletcherize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Fletcherize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Fletcherise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Fletcherized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Fletcherised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Fletcherized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Fletcherised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Fletcherizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Fletcherising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Fletcherizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Fletcherises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Fletcherize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Fletcherise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Fletcherizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Fletcherising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flextime ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flextime",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flexitime",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flextimes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flexitimes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flextime's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flexitime's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flier ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "flyer",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fliers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "flyers",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flier's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "flyer's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "floozy ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "floozy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "floozie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "floozy's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "floozie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "floralize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "floralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "floralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "floralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "floralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "floramor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "floramor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "floramour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flotation ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flotation",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "floatation",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flotations",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "floatations",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flotation's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "floatation's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fluidization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fluidization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluidisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fluidizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluidisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fluidization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluidisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fluidize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fluidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fluidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fluidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fluidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fluidizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fluidizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluidiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fluidizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluidisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fluidizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluidiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fluky ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fluky",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flukey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flunky ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flunky",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flunkie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flunkey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flunkies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flunkeys",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flunky's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flunkey's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "flunkie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fluoridization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fluoridization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluoridisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fluoridizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluoridisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fluoridization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluoridisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fluoridize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fluoridize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluoridise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fluoridized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluoridised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fluoridizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluoridising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fluoridizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fluoridises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "flutist ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "flutist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flautist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flutist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flautist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "flutists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "flautists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "focalization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "focalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "focalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "focalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "focalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "focalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "focalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "focalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "focalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "focalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "focalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "focalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "focalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "focalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "focalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "focalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "focus ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "focused",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "focussed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "focusing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "focussing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fogy ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fogy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "fogey",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fogies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "fogeys",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fogy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "fogey's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fondue ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fondue",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fondu",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fondues",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fondus",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fondue's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fondu's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fontanel ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fontanel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fontanelle",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fontanels",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fontanelles",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fontanel's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fontanelle's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "foolhardize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "foolhardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foolhardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "foolhardized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foolhardised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "foolhardizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foolhardising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "foolhardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foolhardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "forb ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "forb",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fourb",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "forche ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "forche",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fourche",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "forecastle ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "forecastle",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fo'c'sle",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "fo'c's'le",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forecastles",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fo'c'sles",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "fo'c's'les",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forecastle's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fo'c'sle's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "fo'c's'le's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "foreignization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "foreignization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foreignisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "foreignizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foreignisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "foreignization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foreignisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "foreignize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "foreignize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foreignise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "foreignizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foreignises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "forejudgment ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "forejudgment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "forejudgement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forejudgments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "forejudgements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forejudgment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "forejudgement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "foreprize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "foreprize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foreprise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "forgather ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "forgather",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "foregather",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forgathered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "foregathered",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forgathering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "foregathering",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forgathers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "foregathers",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "forgo ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "forgo",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "forego",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forwent",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "forewent",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forgone",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foregone",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forgoing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foregoing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forgoes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foregoes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forego",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forewent",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "foregone",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "foregoing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "foregoes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "formalizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "formalizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formalisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "formalization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "formalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "formalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "formalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "formalizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "formalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "forme ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "forme",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "formicide ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "formicide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formicidae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "formularization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "formularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "formularize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "formularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formularizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formularising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "formularizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "formularizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formulariser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formularizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formularisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formularizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formulariser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "formulas ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "formulas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formulae",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("for other uses"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formulas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formulae",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("for some uses"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "formulization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "formulization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formulisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formulizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formulisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formulization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formulisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "formulize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "formulize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formulise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formulized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formulised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formulizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formulising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formulizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formulises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "formulizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "formulizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formuliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "formulizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "formuliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "forswear ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "forswear",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foreswear",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forswore",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foreswore",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forsworn",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foresworn",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forswearing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foreswearing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forswears",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foreswears",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forsworn",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foresworn",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "forumize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "forumize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "forumise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "forumizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "forumises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fosse ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fosse",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fossiled ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fossiled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fossilizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fossilizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fossilizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fossilizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fossilizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fossilizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fossilization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fossilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fossilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fossilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fossilize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fossilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fossilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fossilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fossilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fossule ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fossule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fossulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "foulder ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "foulder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fouldre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fouter ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fouter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foutre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fouters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foutres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "foveole ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "foveole",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "foveolae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fractionalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fractionalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fractionalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fractionalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fractionalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fractionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fractionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fractionalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fractionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fractionization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fractionization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fractionizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fractionization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fractionize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fractionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fractionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fractionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fractionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fractionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fragmentization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fragmentization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fragmentisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fragmentization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fragmentisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fragmentize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fragmentize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fragmentise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fragmentized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fragmentised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fragmentizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fragmentising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fragmentizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fragmentises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fragmentizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fragmentizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fragmentiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frambesia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frambesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "framboesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frambesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "framboesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frambesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "framboesia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Francize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Francize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Francise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Francizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Francises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Francize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Francise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Franklinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Franklinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Franklinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Franklinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Franklinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Franklinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Franklinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fraternization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fraternization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraternisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fraternizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraternisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fraternization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraternisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fraternize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fraternize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraternise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fraternized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraternised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fraternizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraternising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fraternizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraternises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fraternizer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fraternizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraterniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fraternizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraternisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fraternizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraterniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "freebie ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "freebie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "freebee",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "freebies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "freebees",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "freebie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "freebee's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frena ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frena",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraena",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Frenchize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Frenchize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Frenchise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Frenchizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Frenchises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Frenchize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Frenchise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frenetic ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frenetic",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phrenetic",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frenular ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frenular",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraenular",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frenulum ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frenulum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraenulum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frenula",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraenula",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frenulum's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraenulum's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frenum ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frenum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraenum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frenums",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraenums",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frenum's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fraenum's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "friborg ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "friborg",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fribourg",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frictionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frictionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frictionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frictionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frictionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frivol ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frivoled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frivolled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frivoling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frivolling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frivoler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frivoler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frivoller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frivolers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frivollers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frivoler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frivoller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frizado ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frizado",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frisado",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frizette ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frizette",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frisette",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frizettes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frisettes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frizette's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frisette's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frizz ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frizz",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "friz",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frizz's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "friz's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "frowzy ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "frowzy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frowsy",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frowzier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frowsier",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "frowziest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frowsiest",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fryer ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fryer",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frier",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fryers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "friers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fryer's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "frier's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fuehrer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fuehrer",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "f�hrer",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fuehrers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "f�hrers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fuehrer's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "f�hrer's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fuel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fueled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fuelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fueling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fuelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fueler ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fueler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fueller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fuelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fuellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fueler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fueller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fuelizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fuelizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fueliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fuelizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fuelisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fuelizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fueliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fulfill ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fulfill",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "fulfil",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fulfills",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "fulfils",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fulfillment ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fulfillment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fulfilment",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fulfillments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fulfilments",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fulfillment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fulfilment's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "According to Macquarie, \"to fulfill\" is an acceptable variant, but fulfillment is not. (probably similar to install vs instalment in en-GB)",
  ],
  },
Cluster {
  header: "fulgor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fulgor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fulgour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fulgors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fulgours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fulgorous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fulgorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fulgourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fullness ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fullness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fulness",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fullnesses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fulnesses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fullness's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fulness's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "functionalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "functionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "functionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "functionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "functionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "functionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "functionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "functionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "functionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "functionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "funeralize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "funeralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "funeralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "funeralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "funeralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "funnel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "funneled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "funnelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "funneling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "funnelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "funneler ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "funneler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "funneller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "furor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "furor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "furore",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "furors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "furores",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "furor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "furore's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "furcule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "furcule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "furculae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Furie ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Furie",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Furiae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Furie's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Furiae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fuse ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fuse",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fuze",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("detonating device"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fused",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fuzed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("detonating device"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fusing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fuzing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("detonating device"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fuses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fuzes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("detonating device"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fuse's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fuze's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("detonating device"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fuse",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fused",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fusing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fuses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fuse's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fusilier ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fusilier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fusileer",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fusiliers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fusileers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fusilier's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fusileer's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "fustianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "fustianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fustianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fustianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fustianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fustianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fustianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "fustianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "fustianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "futilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "futilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "futilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "futilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "futilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "futurize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "futurize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "futurise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "futurizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "futurises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gabardine ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gabardine",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gabardines",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gabardine's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gaberdine",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gabardine",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gaberdines",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gabardines",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gaberdine's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gabardine's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Gaelicization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Gaelicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gaelicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gaelicizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gaelicisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gaelicization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gaelicisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Gaelicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Gaelicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gaelicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gaelicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gaelicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gaelicize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gaelicise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gaelicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaelicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gaelicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaelicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gaelicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaelicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gaelicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaelicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "galactorrhea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "galactorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "galactorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "galactorrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "galactorrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "galactosemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "galactosemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "galactosaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "galactosemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "galactosaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "galee ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "galee",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galeae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gallantize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gallantize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallantise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gallantizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallantises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "galliardize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "galliardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galliardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Gallicization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Gallicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gallicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gallicizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gallicisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gallicization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gallicisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Gallicize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Gallicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gallicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gallicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gallicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gallicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gallicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gallicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gallicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gallicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gallicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gallicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gallicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Gallicizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Gallicizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Galliciser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gallicizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Galliciser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gallisize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gallisize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallisise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gallisized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallisised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gallisizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallisising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gallisizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallisises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gallize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gallize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gallized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gallizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gallizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gallises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "galosh ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "galosh",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galoshe",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "galosh's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galoshe's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "galvanization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "galvanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galvanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "galvanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galvanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "galvanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galvanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "galvanize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "galvanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galvanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "galvanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galvanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "galvanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galvanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "galvanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galvanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "galvanizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "galvanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galvaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "galvanizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galvanisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "galvanizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "galvaniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gambades ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gambades",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gambadoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gambol ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gamboled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gambolled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gamboling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gambolling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gamy ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gamy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gamey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gargarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gargarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gargarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gargarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gargarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gargarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gargarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gargarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gargarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "garrote ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "garrote",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "garotte",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "garrotte",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "garote",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "garroted",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "garotted",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "garrotted",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "garoted",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "garroting",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "garrotting",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "garotting",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "garoting",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "garrotes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "garottes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "garrottes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "garotes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "garrote's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "garotte's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "garrotte's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "garote's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "garruline ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "garruline",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "garrulinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gas ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gases",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gasses",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gases",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gasoline ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gasoline",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gasolene",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gasolines",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gasolenes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gasoline's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gasolene's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gastrea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gastrea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gastraea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gastreas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gastraeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gastrocele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gastrocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gastrocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gaufer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gaufer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaufre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gaufers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaufres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gauffer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gauffer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gauffre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gauge ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gauge",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gage",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gauge's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gage's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gauged",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaged",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gauging",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaging",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gauges",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gages",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gage",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("archaic; nautical, historical"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gage's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("archaic; nautical, historical"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gaged",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("archaic; nautical, historical"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gaging",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("archaic; nautical, historical"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gages",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("archaic; nautical, historical"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gavel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gaveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gavelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gaveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gavelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gaveler ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gaveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gayety ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gaiety",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gayety",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gaiety's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gayety's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gayly ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gaily",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gayly",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gelatin ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gelatin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatine",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatines",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatine's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gelatinizability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gelatinizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatinizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatinizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gelatinizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gelatinizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatinizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatinizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gelatinization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gelatinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gelatinize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gelatinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gelatinizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gelatinizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatiniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatinizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatinisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gelatinizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelatiniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gelofer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gelofer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gelofre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gemology ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gemology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gemmology",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gemologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gemmologies",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gemology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gemmology's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "genealogize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "genealogize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "genealogise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "genealogized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "genealogised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "genealogizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "genealogising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "genealogizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "genealogises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "generalizability ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "generalizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "generalizable ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "generalizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "generalizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "generalizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "generalizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "generalizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "generalization ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "generalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "generalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "generalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "generalizational ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "generalizational",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalisational",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "generalize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "generalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "generalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "generalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "generalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "generalizer ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "generalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "generalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "generalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "generaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "genialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "genialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "genialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "genialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "genialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "genializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "genialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "genializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "genialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "genteelize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "genteelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "genteelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "genteelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "genteelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "genteelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "genteelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "genteelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "genteelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gentilization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gentilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gentilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gentilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gentilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gentilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gentilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gentilize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gentilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gentilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gentilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gentilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gentilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gentilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gentilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gentilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gentlemanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gentlemanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gentlemanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gentlemanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gentlemanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "geodesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "geodesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geodaesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "geologize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "geologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "geologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "geologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "geologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "geometricize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "geometricize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geometricise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "geometricizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geometricises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "geometrization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "geometrization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geometrisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "geometrizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geometrisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "geometrize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "geometrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geometrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "geometrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geometrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "geometrizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geometrising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "geometrizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geometrises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ger ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gerenuk ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gerenuk",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "greenuk",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Germanization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Germanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Germanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Germanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Germanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "germanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "germanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "germanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "germanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "germanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "germanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Germanize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Germanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Germanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Germanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Germanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Germanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Germanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Germanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Germanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "germanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "germanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "germanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "germanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "germanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "germanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "germanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "germanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Germanizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Germanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Germaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Germanizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Germaniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "germanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "germaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "germanizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "germanisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gerne ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gerne",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grene",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ges ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ges",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Goes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ge",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ge",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ged",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gessed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gessed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gessoed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gesses ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gessoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "geste ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "geste",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gestae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Getae ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Getae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Goetae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Getae's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Goetae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Getz ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Getz",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Goetz",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Getz's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Goetz's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Getzville ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Getzville",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Goetzville",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Getzville's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Goetzville's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gewgaw ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gewgaw",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geegaw",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gewgaws",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geegaws",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gewgaw's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geegaw's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gey ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gey",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goey",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Gheber ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Gheber",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ghebre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ghebers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ghebres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gheber's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ghebre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ghettoization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ghettoization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ghettoisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ghettoizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ghettoisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ghettoization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ghettoisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ghettoize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ghettoize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ghettoise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ghettoized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ghettoised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ghettoizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ghettoising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ghettoizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ghettoises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "giantize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "giantize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "giantise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "giantizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "giantises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gimbal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gimbaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gimballed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gimbaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gimballing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ginkgo ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ginkgo",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gingko",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ginkgoes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ginkgos",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gingkos",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "gingkoes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ginkgo's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gingko's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "girly ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "girly",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "girlie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gizmo ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gizmo",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gismo",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gizmos",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gismos",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gizmo's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gismo's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glacialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glacialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glacialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glacializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glacialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glamorization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glamorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamourization",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "glamorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glamorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glamorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glamorize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glamorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamourize",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "glamorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glamorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamourized",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "glamorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glamorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamourizing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "glamorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glamorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamourizes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "glamorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glamorizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glamorizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamourizer",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "glamoriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glamorizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamourizers",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "glamorisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glamorizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamourizer's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "glamoriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glamorless ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glamorless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "glamourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glamorous ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glamorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamourous",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glamorousness ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glamorousness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamourousness",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glamorousness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamourousness's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glamour ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glamour",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamor",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glamours",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamors",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glamour's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamor's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glamoured",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamored",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glamouring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glamoring",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glauconitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glauconitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glauconitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glauconitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glauconitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glebe ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glebe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glebae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "globalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "globalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "globalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "globalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "globalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "globalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "globalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "globalize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "globalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "globalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "globalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "globalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "globalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "globalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "globalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "globalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "globigerine ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "globigerine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "globigerinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glottalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glottalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glottalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glottalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glottalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glottalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glottalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glottalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glottalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glottalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glucemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glucemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "glucaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glucosemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glucosemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "glucosaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gluing ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gluing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glueing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gluteal ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gluteal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glutaeal",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "OED has glutaeal listed first, ODE doesn't recognize glutaeal",
  ],
  },
Cluster {
  header: "gluteus ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gluteus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glutaeus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glutei",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glutaei",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gluttonize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gluttonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gluttonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gluttonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gluttonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gluttonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gluttonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gluttonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gluttonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glycemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glycemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "glycaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glycemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glycemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "glycaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glycerin ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glycerin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glycerine",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glycerins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glycerines",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glycerin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glycerine's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glycerinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glycerinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glycerinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glycerinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glycerinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glycerolize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glycerolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glycerolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glycerolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glycerolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glycerolized ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glycerolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glycerolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glycogenize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glycogenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glycogenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "glycogenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "glycogenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glycohemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glycohemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "glycohaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "glycosemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "glycosemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "glycosaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Gnosticize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Gnosticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gnosticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gnosticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gnosticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gnosticized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gnosticised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gnosticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gnosticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gnosticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gnosticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gnosticizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gnosticising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gnosticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gnosticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gnosticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gnosticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Gnosticizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Gnosticizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gnosticiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gnosticizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gnosticiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gnosticizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gnosticiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gnosticizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gnosticisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gobbledygook ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gobbledygook",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gobbledegook",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gobbledygooks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gobbledegooks",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gobbledygook's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gobbledegook's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "goddamn ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "goddamn",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goddam",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goddamned",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goddamed",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goddamning",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goddaming",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goddamns",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goddams",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goddamn's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goddam's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "goddize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "goddize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goddise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goddizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goddises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "goiter ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "goiter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goiters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goitres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goiter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goitre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gole ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gole",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goloe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "goliathize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "goliathize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goliathise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goliathized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goliathised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goliathizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goliathising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goliathizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goliathises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "goloshes ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "goloshes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goloshoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gonorrhea ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gonorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gonorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gonorrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gonorrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gonorrhea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gonorrhoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gonorrheal ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gonorrheal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gonorrhoeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gonorrhealer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gonorrhoealer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gonorrhealest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gonorrhoealest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gonorrheic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gonorrheic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gonorrhoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "goodbye ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "goodbye",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goodby",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goodbyes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goodbys",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goodbye's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goodby's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "goody ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "goody",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goodie",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goodies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "goody's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goodie's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Gora ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Gora",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Goura",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gora's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Goura's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gora",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "goura",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "goramies ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "goramies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gouramies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gorgonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gorgonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gorgonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gorgonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gorgonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gorgonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gorgonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gorgonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gorgonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gormandism ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gormandism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gourmandism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gormandisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gourmandisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gormandism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gourmandism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gormandize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gormandize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gormandise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gormandized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gormandised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gormandizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gormandising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gormandizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gormandises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gormandize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gormandise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gormandizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gormandisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gormandizer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gormandizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gormandiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gormandizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gormandisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gormandizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gormandiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gospeler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gospeler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gospeller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gospelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gospellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gospeler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gospeller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gospelize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gospelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gospelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gospelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gospelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gospelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gospelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gospelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gospelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gospellize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gospellize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gospellise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gospellized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gospellised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gospellizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gospellising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gospellizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gospellises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Gothicize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Gothicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gothicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gothicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gothicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gothicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gothicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gothicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gothicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gothicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gothicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gothicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gothicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gothicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gothicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gothicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gothicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Gothicizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Gothicizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gothiciser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gothicizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gothicisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gothicizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Gothiciser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gourmandize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gourmandize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gourmandise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gourmandized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gourmandised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gourmandizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gourmandising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gourmandizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gourmandises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "governmentalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "governmentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "governmentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "governmentalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "governmentalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "governmentalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "governmentalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "governmentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "governmentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Goyesque ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Goyesque",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Goyaesque",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Graecize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Graecize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Graecise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Graecized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Graecised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Graecizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Graecising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1 "),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Graecizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Graecises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Grecise",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Graecise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Grecise's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Graecise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Grecised",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Graecised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Grecised's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Graecised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Grecising",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Graecising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Grecising's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Graecising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Grecize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Grecise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Grecized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Grecised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Grecizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Grecising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grecize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grecise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":4"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grecized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grecised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":4"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grecizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grecising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":4"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grecizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grecises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":4"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grammaticize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "grammaticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grammaticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grammaticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grammaticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grammaticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grammaticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grammaticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grammaticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grams ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("unit of mass"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gram's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gramme's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("unit of mass"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grams",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("unit of mass"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("chickpeas"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gram's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("chickpeas"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grams",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("chickpeas"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "granddad ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "granddad",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grandad",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "granddads",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grandads",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "granddad's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grandad's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grangerization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "grangerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grangerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grangerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grangerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grangerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grangerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grangerize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "grangerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grangerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grangerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grangerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grangerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grangerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grangerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grangerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grangerizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "grangerizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grangeriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grangerizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grangerisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grangerizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grangeriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "granitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "granitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "granitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "granitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "granitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "granitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "granitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "granitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "granitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "granitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "granitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "granitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "granitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "granitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "granitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "granitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "granny ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "granny",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grannie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "granny's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grannie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "granulitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "granulitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "granulitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "granulitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "granulitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "granulize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "granulize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "granulise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "granulizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "granulises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "graphic ",
  verified: false,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "graphic",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "graphical",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "graphic",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "graphical",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: Some("graphical user interface"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "graphitizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "graphitizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "graphitisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "graphitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "graphitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "graphitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "graphitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "graphitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "graphitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "graphitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "graphitize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "graphitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "graphitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "graphitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "graphitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "graphitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "graphitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "graphitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "graphitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gravel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "graveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gravelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "graveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gravelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gray ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gray",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "grey",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grayed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "greyed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grayer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "greyer",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grayest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "greyest",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "graying",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "greying",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grays",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "greys",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gray's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "grey's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gray",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("<N> physical unit"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grays",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("<N> physical unit"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grayish ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "grayish",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "greyish",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grayness ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "grayness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "greyness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grayness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "greyness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grecian ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "grecian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "graecian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Grecianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Grecianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Grecianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Grecianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Grecianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Grecism ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Grecism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Graecism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Grecisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Graecisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Grecism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Graecism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grecism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "graecism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Greekize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Greekize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Greekise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Greekizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Greekises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Greekize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Greekise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "greisenization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "greisenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "greisenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "greisenizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "greisenisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "greisenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "greisenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "greisenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "greisenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "greisenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "greisenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "greisenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "greisenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "greisenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "griffon ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "griffon",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "griffons",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "griffon's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "griffin",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "griffon",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gryphon",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("animal"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "griffins",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "griffons",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gryphons",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("animal"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "griffin's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "griffon's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gryphon's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("animal"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grizard ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "grizard",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grisard",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "grize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "groin ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "groin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "groyne",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("wall"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "groined",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "groyned",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("wall"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "groining",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "groyning",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("wall"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "groin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "groyne's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("wall"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "groins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "groynes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("wall"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "groin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("body part"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "groin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("body part"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "groins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("body part"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grovel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "groveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "grovelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "groveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "grovelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "groveler ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "groveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "groveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grovelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grovellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "groveler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "groveller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grovelingly ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "grovelingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grovellingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grovelinglier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grovellinglier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grovelingliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grovellingliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gruel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "grueled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gruelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grueling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gruelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gruelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gruellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grueling's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gruelling's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "grueler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "grueler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grueller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gruelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gruellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "grueler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "grueller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gruelingly ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gruelingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gruellingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Gueber ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Gueber",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Guebre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Guebers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Guebres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Gueber's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Guebre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "guerrilla ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "guerrilla",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "guerilla",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "guerrillas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "guerillas",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "guerrilla's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "guerilla's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "guizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "guizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "guiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "guizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "guisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gully ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gully",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gulley",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("trench"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gully's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gulley's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("trench"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gullies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gulleys",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("trench"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gunnel ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gunnel",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gunnels",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gunnel's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gunwale",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gunnel",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gunwales",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gunnels",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gunwale's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gunnel's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gutte ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gutte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "guttae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "guttule ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "guttule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "guttulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gutturalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gutturalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gutturalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gutturalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gutturalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gutturalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gutturalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gutturalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gutturalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gutturalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gutturalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gutturalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gutturalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gutturalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gutturalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gutturalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gyne ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gyne",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gynae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecocracy ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecocracy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecocracy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecocracies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecocracies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecocracy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecocracy's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecocrat ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecocrat",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecocrat",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecocrats",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecocrats",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecocrat's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecocrat's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecocratic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecocratic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecocratic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecocraticer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecocraticer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecocraticest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecocraticest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecoid ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecoid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecol ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecol",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecol",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecologic ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecologicer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecologicer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecologicest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecologicest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecologics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecologics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecological ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecologicaler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecologicaler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecologicalest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecologicalest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecologicals",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecologicals",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecological's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecological's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecologist ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecologist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecology ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecology's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecomastia ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecomastia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecomastia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecomastias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecomastias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecomastia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecomastia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecomasty ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecomasty",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecomasty",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gynecomasty's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecomasty's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynecomorphous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynecomorphous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaecomorphous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gyneconitis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gyneconitis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaeconitis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gyneocracy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gyneocracy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaeocracy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gyneolater ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gyneolater",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaeolater",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gyneolatry ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gyneolatry",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynaeolatry",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gynomonecious ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gynomonecious",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gynomonoecious",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gypsy ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gypsy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gipsy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gypsies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gipsies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gypsy's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gipsy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gyrostabilization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gyrostabilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gyrostabilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gyrostabilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gyrostabilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gyrostabilizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gyrostabilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gyrostabiliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gyrostabilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gyrostabilisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gyrostabilizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gyrostabiliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ha ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ha",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hah",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ha",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "habitualize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "habitualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "habitualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "habitualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "habitualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hairdryer ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hairdryer",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hairdrier",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hairdryers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hairdriers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hairdryer's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hairdrier's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hajj ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hajj",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hadj",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hajj's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hadj's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hajji",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hadji",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hajjis",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hadjis",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hajji's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hadji's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hajjs ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hajjs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hadjs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "halloo ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "halloo",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hallo",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "hollo",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hallooing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "halloing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "holloing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "halloos",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hollos",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "hallos",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "halloo's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hallo's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "hollo's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hamletization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hamletization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hamletisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hamletizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hamletisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hamletization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hamletisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hamletize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hamletize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hamletise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hamletizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hamletises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "handsel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "handseled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "handselled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "handseling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "handselling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hankie ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hankie",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hanky",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hankie's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hanky's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hanoverianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hanoverianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hanoverianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hanoverianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hanoverianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hanoverianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hanoverianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hanoverize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hanoverize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hanoverise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hanoverizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hanoverises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hanoverize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hanoverise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hansardize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hansardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hansardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hansardized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hansardised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hansardizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hansardising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hansardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hansardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hansel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hanseled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hanselled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hanseling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hanselling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harbor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harbor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harbored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harboured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harboring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harbors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harbor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harborage ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harborage",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourage",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harborages",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourages",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harborage's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourage's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harborer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harborer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harborers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harborer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harborful ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harborful",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourful",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harborless ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harborless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harborlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourlesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harborlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourlessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harborlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harborlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourlesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harborous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harborous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harbourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harborside ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harborside",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourside",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harborward ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harborward",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "harbourward",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harmonizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harmonizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harmonizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harmonizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harmonizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harmonizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harmonization ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harmonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harmonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harmonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harmonize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harmonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harmonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harmonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harmonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "harmonizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "harmonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmoniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harmonizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmonisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "harmonizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "harmoniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Harmothoe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Harmothoe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Harmothoae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Harmothoe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Harmothoae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Harvardize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Harvardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Harvardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Harvardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Harvardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Harvardize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Harvardise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Harveyize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Harveyize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Harveyise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Harveyizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Harveyises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Harveyize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Harveyise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hashish ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hashish",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hasheesh",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "haschisch",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hashish's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hasheesh's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "haschisch's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hasidean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hasidean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hasidaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hasidean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hasidaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hasmonean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hasmonean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hasmonaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hasmoneans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hasmonaeans",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hatchel ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hatcheled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hatchelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hatcheling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hatchelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hattize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hattize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hattise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hattizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hattises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hattize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hattise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "haussmannization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "haussmannization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "haussmannisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "haussmannizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "haussmannisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "haussmannize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "haussmannize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "haussmannise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "haussmannized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "haussmannised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "haussmannizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "haussmannising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "haussmannizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "haussmannises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "havior ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "havior",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "haviour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "haviors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "haviours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "haviored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "haviored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "havioured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hazardize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hazardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hazardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hazardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hazardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heathenization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heathenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heathenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heathenize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heathenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heathenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "heathenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heathenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "heathenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heathenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "heathenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heathenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heavenize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heavenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heavenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "heavenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heavenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hebraicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hebraicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hebraicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hebraicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hebraicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hebraicize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hebraicise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hebraization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hebraization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hebraisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hebraization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hebraisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hebraization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hebraisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hebraizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hebraisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hebraize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hebraize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hebraise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hebraized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hebraised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hebraizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hebraising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hebraizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hebraises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hebraize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hebraise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hebraized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hebraised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hebraizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hebraising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hebraizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hebraises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hebraizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hebraizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hebraiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hebraizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hebraiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hecatean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hecatean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hecataean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hecatean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hecataean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hectogram ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hectogram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hectogramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hectograms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hectogrammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hectogram's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hectogramme's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hectoliter ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hectoliter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hectolitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hectoliters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hectolitres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hectoliter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hectolitre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hectometer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hectometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hectometre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hectometers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hectometres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hectometer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hectometre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "haed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hegelianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hegelianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hegelianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hegelianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hegelianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hegelianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hegelianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hekatean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hekatean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hekataean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hekatean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hekataean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heloe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "helled ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "helled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "helloed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hellenization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hellenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hellenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hellenizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hellenisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hellenization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hellenisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hellenize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hellenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hellenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hellenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hellenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hellenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hellenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hellenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hellenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hellenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hellenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hellenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hellenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hellenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hellenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hellenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hellenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hellenizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hellenizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Helleniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hellenizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Helleniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemachrome ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemachrome",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemachrome",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemachrome's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemachrome's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemacytometer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemacytometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemacytometer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemacytometer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemacytometer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemad ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemad",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "haemad",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemagglutinate ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemagglutinate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemagglutinate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemagglutinated",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemagglutinated",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemagglutinating",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemagglutinating",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemagglutination ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemagglutination",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemagglutination",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemagglutination's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemagglutination's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemagglutinative ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemagglutinative",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemagglutinative",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemagglutinin ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemagglutinin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemagglutinin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemagogue ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemagogue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemagogue",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemal ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemameba ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemameba",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hemamoeba",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemangioma ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemangioma",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemangioma",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemangiomas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemangiomas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemangioma's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemangioma's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemangiomata ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemangiomata",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemangiomata",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemangiomatosis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemangiomatosis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemangiomatosis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemapophysis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemapophysis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemapophysis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemaspectroscope ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemaspectroscope",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemaspectroscope",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematein ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematein",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematein",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematein's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematein's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematemesis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematemesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematemesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematherm ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematherm",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematherm",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemathermal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemathermal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemathermal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemathermous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemathermous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemathermous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematid ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematin ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematins",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematin's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematinic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematinic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematinic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematinics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematinics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematinic's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematinic's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematite ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematites",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematites",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematite's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematite's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematitic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematitic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematitic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematoblast ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematoblast",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematoblast",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematoblasts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematoblasts",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematoblast's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematoblast's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematobranchiate ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematobranchiate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematobranchiate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematocele ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematocele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematocele's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematocele's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematocrit ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematocrit",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematocrit",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematocrits",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematocrits",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematocrit's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematocrit's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematocryal ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematocryal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematocryal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematocyst ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematocyst",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematocyst",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematocyst's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematocyst's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematocystis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematocystis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematocystis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematocyte ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematocyte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematocyte",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematocyte's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematocyte's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematogenesis ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematogenesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematogenesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematogeneses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematogeneses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematogenesis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematogenesis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematogenous ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematogenous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematogenous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematoid ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematoid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematoidin ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematoidin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematoidin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematologic ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematological ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematologist ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematologist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematology ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematology's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematolysis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematolysis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematolysis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematoma ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematoma",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematoma",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematomas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematomas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematoma's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematoma's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematomata ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematomata",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematomata",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematometer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematometer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematophyte ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematophyte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematophyte",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematophyte's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematophyte's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematopoiesis ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematopoiesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematopoiesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematopoieses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematopoieses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematopoietic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematopoietic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematopoietic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematorrhachis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematorrhachis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematorrhachis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematosepsis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematosepsis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematosepsis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematosin ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematosin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematosin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematosis ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematosis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematosis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematosis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematosis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematothermal ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematothermal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematothermal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematoxylic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematoxylic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematoxylic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematoxylin ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematoxylin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematoxylin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematoxylins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematoxylins",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematoxylin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematoxylin's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematozoal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematozoal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematozoal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematozoic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematozoic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematozoic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematozoon ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematozoon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematozoon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematozoa",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematozoa",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematozoon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematozoon's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematozzoa ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematozzoa",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematozzoa",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hematuria ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hematuria",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematuria",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematurias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematurias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hematuria's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haematuria's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heme ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heme",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haem",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemerythrin ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemerythrin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hemoerythrin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemihypesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemihypesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hemihypoesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemin ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemins",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemin's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemoblast ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemoblast",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoblast",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemoblast's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoblast's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemochromatosis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemochromatosis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemochromatosis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemochromatoses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemochromatoses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemochrome ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemochrome",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemochrome",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemocoel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemocoel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemocoel",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemocoels",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemocoels",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemoconcentration ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemoconcentration",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoconcentration",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemoconcentration's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoconcentration's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemoconia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemoconia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoconia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemocyanin ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemocyanin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemocyanin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemocyanins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemocyanins",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemocyte ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemocyte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemocyte",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemocytes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemocytes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemocyte's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemocyte's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemocytoblast ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemocytoblast",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemocytoblast",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemocytoblast's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemocytoblast's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemocytoblastic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemocytoblastic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemocytoblastic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemocytometer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemocytometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemocytometer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemodialysis ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemodialysis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemodialysis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemodialyses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemodialyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemodialysis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemodialysis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemodilution ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemodilution",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemodilution",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemodynamic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemodynamic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemodynamic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemodynamics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemodynamics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemodynamics's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemodynamics's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemoflagellate ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemoflagellate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoflagellate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemoflagellate's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoflagellate's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemoglobic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemoglobic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoglobic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemoglobin ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemoglobin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoglobin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemoglobin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoglobin's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemoglobinopathy ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemoglobinopathy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoglobinopathy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemoglobinopathies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoglobinopathies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemoglobinous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemoglobinous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoglobinous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemoglobinuria ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemoglobinuria",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoglobinuria",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemoglobinuria's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoglobinuria's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemogram ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemogram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemogram",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemoid ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemolysin ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemolysin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemolysin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemolysin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemolysin's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemolysis ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemolysis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemolysis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemolytic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemolytic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemolytic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemometer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemometer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hemon ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hemon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Haemon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hemon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Haemon's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemophile ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemophile",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemophile",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemophile's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemophile's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemophilia ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemophilia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemophilia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemophilia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemophilia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemophiliac ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemophiliac",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemophiliac",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemophiliacs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemophiliacs",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemophiliac's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemophiliac's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemophilic ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemophilic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemophilic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemopod ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemopod",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemopod",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemopoiesis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemopoiesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemopoiesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemoptysis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemoptysis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoptysis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemoptyses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoptyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemoptysis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemoptysis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemorrhage ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemorrhage",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemorrhage",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemorrhage's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemorrhage's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemorrhaged",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemorrhaged",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemorrhaging",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemorrhaging",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemorrhages",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemorrhages",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemorrhoid ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemorrhoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemorrhoid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemorrhoids",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemorrhoids",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemorrhoidal ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemorrhoidal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemorrhoidal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemorrhoidectomy ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemorrhoidectomy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemorrhoidectomy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemorrhoidectomy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemorrhoidectomy's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemosporid ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemosporid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemosporid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemosporidian ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemosporidian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemosporidian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemostasia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemostasia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemostasia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemostasis ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemostasis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemostasis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemostases",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemostases",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemostasis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemostasis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemostatic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemostatic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemostatic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemothorax ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemothorax",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemothorax",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemothorax's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemothorax's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemotoxic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemotoxic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemotoxic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hemotoxin ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hemotoxin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemotoxin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hemotoxin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "haemotoxin's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heparinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heparinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heparinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hepatization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hepatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hepatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hepatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hepatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hepatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hepatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hepatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hepatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hepatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hepatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hepatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hepatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hepatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hepatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hepatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hepatorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hepatorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hepatorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heraldize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heraldize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heraldise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "heraldizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heraldises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "herborization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "herborization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "herborisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "herborizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "herborisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "herborize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "herborize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "herborise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "herborized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "herborised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "herborizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "herborising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "herborizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "herborises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heredes ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heredes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "haeredes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hereticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hereticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hereticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hereticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hereticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heroinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heroinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heroinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "heroinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heroinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heroization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heroization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heroisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "heroizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heroisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "heroization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heroisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heroize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heroize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heroise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "heroized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heroised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "heroizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heroising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "heroizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heroises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "het ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "het",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "haet",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hets",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "haets",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heterecious ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heterecious",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "heteroecious",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heteric ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heteric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hetaeric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "heterism ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "heterism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hetaerism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hexachlorethane ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hexachlorethane",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hexachloroethane",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hexachlorethanes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hexachloroethanes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hexametrize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hexametrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hexametrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hexametrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hexametrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hexametrizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hexametrising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hexametrizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hexametrises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hexestrol ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hexestrol",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hexoestrol",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hibernicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hibernicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hibernicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hibernicize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hibernicise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hibernicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hibernicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hibernicized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hibernicised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hibernicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hibernicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hibernicizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hibernicising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hibernicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hibernicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hibernicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hibernicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hibernicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hibernicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hibernicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hibernicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hibernization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hibernization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hibernisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hibernizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hibernisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hibernize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hibernize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hibernise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hibernized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hibernised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hibernizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hibernising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hibernizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hibernises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hiccup ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hiccuped",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hiccupped",
        types: &[
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hiccuping",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hiccupping",
        types: &[
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hierarchize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hierarchize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hierarchise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hierarchized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hierarchised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hierarchizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hierarchising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "highfalutin ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "highfalutin",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hifalutin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "highfaluting",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "high_jinks ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "high jinks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hijinks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hijack ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hijack",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "highjack",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hijacked",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "highjacked",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hijacking",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "highjacking",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hijacks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "highjacks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hijack's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "highjack's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hijacker ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hijacker",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "highjacker",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hijackers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "highjackers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hijacker's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "highjacker's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hilloed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hilloed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hilloaed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hinduize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hinduize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hinduise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hinduized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hinduised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hinduized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hinduised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hinduizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hinduising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hinduizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hinduises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hinduize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hinduise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hinduizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hinduising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hippie ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hippie",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hippy",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hippies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hippie's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hippy's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hippy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hippieness ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hippieness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hippiness",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hippienesses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hippinesses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hippieness's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hippiness's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hirsel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hirseled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hirselled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hirseling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hirselling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hispanicization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hispanicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispanicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hispanicizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispanicisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hispanicization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispanicisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hispanicize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hispanicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispanicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hispanicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispanicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hispanicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispanicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hispanicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispanicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hispanicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hispanicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hispanicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hispanicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hispanicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hispanicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hispanicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hispanicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hispaniolize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hispaniolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispaniolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hispaniolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispaniolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hispaniolized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispaniolised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hispaniolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispaniolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hispaniolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispaniolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hispaniolize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispaniolise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hispaniolizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hispaniolising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hispaniolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hispaniolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hispaniolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hispaniolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hispaniolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hispaniolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hispaniolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hispaniolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "historicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "historicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "historicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "historicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "historicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "historicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "historicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "historicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "historicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hmm ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hmm",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hm",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hoagie ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hoagie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hoagy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hoagie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hoagy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hoagies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hollywoodize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hollywoodize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hollywoodise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hollywoodized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hollywoodised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hollywoodized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hollywoodised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hollywoodizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hollywoodising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hollywoodizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hollywoodises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hollywoodize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hollywoodise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hollywoodizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hollywoodising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeoblastic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeoblastic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeoblastic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeobox ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeobox",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeobox",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeochromatic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeochromatic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeochromatic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeochronous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeochronous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeochronous",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeocrystalline ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeocrystalline",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeocrystalline",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeogenic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeogenic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeogenic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeogenous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeogenous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeogenous",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeoid ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeoid",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeokinesis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeokinesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeokinesis",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeomeric ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeomeric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeomeric",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeomerous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeomerous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeomerous",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeomery ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeomery",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeomery",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeomeries",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeomeries",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeomorph ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeomorph",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeomorph",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeomorphs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeomorphs",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeomorph's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeomorph's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeomorphic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeomorphic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeomorphic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeomorphous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeomorphous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeomorphous",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeomorphy ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeomorphy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeomorphy",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeomorphies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeomorphies",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeopath ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeopath",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopath",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeopaths",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopaths",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeopath's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopath's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeopathic ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeopathic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeopathicer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathicer",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeopathicest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathicest",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeopathically ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeopathically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathically",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeopathicallier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathicallier",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeopathicalliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathicalliest",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeopathician ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeopathician",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathician",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeopathicity ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeopathicity",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathicity",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeopathist ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeopathist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathist",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeopathists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathists",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeopathist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathist's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeopathy ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeopathy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathy",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeopathies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathies",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeopathy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopathy's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeophony ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeophony",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeophony",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeoplasia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeoplasia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeoplasia",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeoplastic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeoplastic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeoplastic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeoplasy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeoplasy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeoplasy",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeopolar ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeopolar",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeopolar",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeosis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeosis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeosis",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeoses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeoses",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeostasis ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeostasis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeostasis",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeostases",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeostases",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeostasis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeostasis's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeostatic ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeostatic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeostatic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeoteleuton ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeoteleuton",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeoteleuton",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeoteleutons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeoteleutons",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeothermal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeothermal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeothermal",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeothermic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeothermic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeothermic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeothermous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeothermous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeothermous",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeotic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeotic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeotic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeotype ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeotype",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeotype",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeotypic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeotypic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeotypic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeotypical ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeotypical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeotypical",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeozoic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeozoic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homoeozoic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeward ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeward",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homewards",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adverb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homey ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homey",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homey",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeys",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homey's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homeyness ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homeyness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hominess",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeynesses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hominesses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homeyness's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hominess's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hominization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hominization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hominisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hominized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hominized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hominised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homogenization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homogenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homogenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homogenizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homogenisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homogenization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homogenisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homogenize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homogenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homogenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homogenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homogenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homogenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homogenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homogenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homogenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homogenizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homogenizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homogeniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homogenizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homogenisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homogenizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homogeniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homolog ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homologue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homolog",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homologues",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homologs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homologue's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homolog's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homologization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homologization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homologisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homologize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "homologizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "homologizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homologiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homologizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homologisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "homologizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "homologiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "honky ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "honky",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "honkie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "honkey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honkies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "honkeys",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honky's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "honkie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "honkey's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "honor ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "honor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "honorability ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "honorability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "honorable ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "honorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "honorableness ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "honorableness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourableness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorablenesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourablenesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorableness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourableness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "honorableship ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "honorableship",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourableship",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorableships",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourableships",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorableship's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourableship's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "honorably ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "honorably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorablier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourablier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorablies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourablies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorabliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourabliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "honoree ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "honoree",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honouree",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorees",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourees",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honoree's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honouree's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "honorer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "honorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "honorless ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "honorless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourlesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "honorlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourlessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "honorlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "honorlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "honourlesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hooch ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hooch",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hootch",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hooches",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hootches",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("hut"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hooch's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hootch's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("hut"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hoodlumize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hoodlumize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hoodlumise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hoodlumizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hoodlumises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hooky ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hooky",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hookey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hookies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hookeys",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hooky's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hookey's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hooliganize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hooliganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hooliganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hooliganizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hooliganises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hoosierize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hoosierize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hoosierise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hoosierizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hoosierises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hoosierize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hoosierise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Hooverize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Hooverize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hooverise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hooverizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hooverises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Hooverize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Hooverise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hore ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hore",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "horae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "horizontalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "horizontalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "horizontalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "horizontalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "horizontalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "horizontalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "horizontalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "horizontalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "horizontalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "horizontalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "horizontalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "horizontalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hormonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hormonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hormonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hormonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hormonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "horrorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "horrorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "horrorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "horrorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "horrorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "horsey ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "horsey",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "horsy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hospitalization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hospitalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hospitalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hospitalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hospitalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hospitalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hospitalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hospitalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hospitalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hospitalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hospitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hospitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hospitalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hospitalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hospitalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hospitalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hostilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hostilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hostilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hostilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hostilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hotelization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hotelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hotelisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hotelizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hotelisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hotelization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hotelisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hotelize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hotelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hotelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hotelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hotelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "housel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "houseled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "houselled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "houseling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "houselling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "houselings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "housellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hovel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hoveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hovelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hoveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hovelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hoveler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hoveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hoveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hovelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hovellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hovercraft ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hovercraft",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hovercrafts",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hovercraft",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hucksterize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hucksterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hucksterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hucksterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hucksterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "huer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "huer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hure",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humanitarianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humanitarianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humanitarianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humanitarianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humanitarianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humanization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humanize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humanizer ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humanizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humanisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humanizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humaniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humongous ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humongous",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humungous",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humor ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humoral ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humoral",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humoural",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humorer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humorful ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humorful",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humourful",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "humorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humorless ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humorless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humorlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humourlesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humorlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humourlessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humorlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humorlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humourlesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humorlessness ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humorlessness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humourlessness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humorlessnesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humourlessnesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "humorlessness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humourlessness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humorsome ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humorsome",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humoursome",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "humorsomeness ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "humorsomeness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "humoursomeness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hurrah ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hurrah",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hooray",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hoorah",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "hurray",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hurrahed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hoorayed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hurrayed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hurrahing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hooraying",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hurraying",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hurrahs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hoorays",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hoorahs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "hurrays",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hurrahes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hurrah's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hooray's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hoorah's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "hurray's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hooray",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hurrah",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hurray",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Interjection), archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hurricanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hurricanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hurricanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hurricanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hurricanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "huzzah ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "huzzah",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "huzza",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "huzzahed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "huzzaed",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "huzzah's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "huzza's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "huzzahs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "huzzas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "huzzahing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "huzzaing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyalinization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyalinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyalinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyalinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyalinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyalinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyalinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyalinize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyalinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyalinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyalinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyalinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyalinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyalinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyalinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyalinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hybridizable ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hybridizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hybridizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hybridizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hybridizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hybridizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hybridization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hybridization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hybridizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hybridization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hybridize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hybridize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hybridized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hybridizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hybridizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hybridizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hybridizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hybridizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hybridizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hybridiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydremia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hydraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydremias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hydraemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydremia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hydraemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydremic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydremic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hydraemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydrocele ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydrocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hydrocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydrocephalus ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydrocephalus",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrocephaly",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrocephaluses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrocephalies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrocephalus's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrocephaly's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydrogenization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydrogenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrogenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrogenizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrogenisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrogenization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrogenisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydrogenize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydrogenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrogenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrogenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrogenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrogenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrogenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrogenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrogenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydrolize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydrolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydrolyzable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolysable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolysabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolysables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolysablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolysable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydrolyzate ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolysate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzates",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolysates",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydrolyzation ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolysation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzation's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolysation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydrolyze ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydrolyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrolyses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: hydrolysis"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydrolyzer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydrolyzer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydrolyser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydrorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydrorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hydrorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydroxylization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydroxylization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydroxylisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydroxylizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydroxylisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydroxylization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydroxylisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hydroxylize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hydroxylize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydroxylise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hydroxylizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hydroxylises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyena ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyena",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyaena",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyenas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyena",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "hyaenas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "hyaena",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl"),
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyena's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyaena's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "If hyaena not variant than: _ 2: hyaenas / _V 2: hyaena",
  ],
  },
Cluster {
  header: "hyenic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyenic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyaenic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hygienization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hygienization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hygienisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hygienizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hygienisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hygienization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hygienisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hygienize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hygienize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hygienise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hygienizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hygienises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hymenean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hymenean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hymenaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperbolize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperbolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperbolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperbolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperbolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperbolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperbolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperbolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperbolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypercalcemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypercalcemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypercalcaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypercalcemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypercalcaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypercalcinemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypercalcinemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypercalcinaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypercatharsises ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypercatharsises",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypercatharses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperchloremia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperchloremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperchloraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypercholesterolemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypercholesterolemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypercholesterolaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypercholesterolemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypercholesterolaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypercivilization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypercivilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypercivilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypercivilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypercivilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypercivilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypercivilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypercivilized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypercivilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypercivilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypercivilizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperciviliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypercivilized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypercivilised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypercriticize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypercriticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypercriticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypercriticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypercriticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypercriticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypercriticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypercriticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypercriticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypercryesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypercryesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypercryaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperemia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperemia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperaemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperemphasize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperemphasize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperemphasise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperemphasizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperemphasises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperesthesia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperesthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperaesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperesthesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperaesthesia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperesthete ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperesthete",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperaesthete",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperesthete's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperaesthete's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperesthetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperestheticer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperaestheticer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperestheticest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperaestheticest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperglycemic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperglycemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperglycaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperimmunization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperimmunization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperimmunisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperimmunizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperimmunisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperimmunization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperimmunisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperimmunize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperimmunize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperimmunise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperimmunizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperimmunises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperinsulinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperinsulinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperinsulinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperinsulinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperinsulinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperinsulinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperinsulinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperinsulinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperinsulinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperinsulinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperinsulinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperinsulinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperlipemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperlipemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperlipaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperlipemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperlipemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperlipaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperlipidemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperlipidemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperlipidaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypernatremia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypernatremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypernatraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperoxygenize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperoxygenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperoxygenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperoxygenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperoxygenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperparasitize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperparasitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperparasitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperparasitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperparasitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperpnea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperpnea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperpnoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperpnea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hyperpnoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperrealize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperrealize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperrealise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperrealizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperrealises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypersensitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypersensitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypersensitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypersensitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypersensitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypersensitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypersensitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypersensitize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypersensitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypersensitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypersensitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypersensitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypersensitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypersensitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypersensitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypersensitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperspiritualizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperspiritualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperspiritualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperspiritualizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperspiritualisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperspiritualizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperspiritualising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperthyroidization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperthyroidization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperthyroidisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperthyroidizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperthyroidisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperthyroidization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperthyroidisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyperthyroidize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyperthyroidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperthyroidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyperthyroidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyperthyroidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypervitalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypervitalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypervitalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypervitalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypervitalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypervitalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypervitalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypervitalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypervitalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypervitalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypervitalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypervitalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypesthesia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypesthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypaesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypesthesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypaesthesia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypesthesic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypesthesic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypaesthesic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypethral ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypethral",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypaethral",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyphenization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyphenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyphenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyphenizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyphenisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyphenization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyphenisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyphenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyphenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyphenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyphenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyphenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyphenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyphenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyphenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyphenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypnoidize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypnoidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnoidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnoidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnoidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnoidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnoidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnoidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnoidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypnopedia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypnopedia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypnopaedia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnopedias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypnopaedias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypnotizability ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypnotizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypnotizable ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypnotizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypnotization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypnotization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypnotize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypnotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypnotizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypnotizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypnotizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypnotiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypocenter ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypocenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypocentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypocenters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypocentres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypocenter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypocentre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypogeal ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypogeal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypogaeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypogean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypogean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypogaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypogeic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypogeic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypogaeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypogeous ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypogeous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypogaeous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypogeum ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypogeum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypogaeum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypogea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypogaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypomagnesemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypomagnesemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypomagnesaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypomagnesemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypomagnesaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypomenorrhea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypomenorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypomenorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypomenorrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypomenorrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypophysectomize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypophysectomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypophysectomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypophysectomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypophysectomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypopnea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypopnea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypopnoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypopnea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypopnoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyposensitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyposensitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyposensitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hyposensitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyposensitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hyposensitize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hyposensitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hyposensitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypostasize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypostasize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypostasise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypostasized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypostasised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypostasizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypostasising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypostasizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypostasises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypostatization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypostatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypostatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypostatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypostatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypostatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypostatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypostatize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypostatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypostatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypostatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypostatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypostatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypostatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypostatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypostatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypothesize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypothesize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypothesise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypothesized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypothesised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypothesizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypothesising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypothesizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypothesises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypothesizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypothesizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypothesiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypothesizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypothesisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypothesizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypothesiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypothetize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypothetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypothetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypothetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypothetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypothetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypothetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypothetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hypothetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypoxemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypoxemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypoxaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hypoxemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypoxaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hypoxemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hypoxemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "hypoxaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "hysterectomize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "hysterectomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hysterectomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hysterectomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hysterectomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hysterectomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hysterectomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "hysterectomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "hysterectomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ichneumonized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ichneumonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ichneumonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ichneumonizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ichneumoniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ichneumonized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ichneumonised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ichorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ichorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ichorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ichorrhemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ichorrhemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ichorrhaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "icon ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "icon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ikon",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "icons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ikons",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "icon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ikon's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "icon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ikon",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "icons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ikons",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "icon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ikon's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "iconic ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "iconic",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ikonic",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "iconically ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "iconically",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ikonically",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "iconize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "iconize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iconise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iconized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iconised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iconizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iconising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iconizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iconises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Idea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Idea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Idaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Idea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Idaea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "idealization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "idealization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idealisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idealizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idealisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idealization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idealisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "idealize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "idealize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idealise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idealized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idealised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idealizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idealising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idealizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idealises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "idealizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "idealizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idealiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idealizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idealisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idealizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idealiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Idean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Idean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Idaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Idean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Idaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ideologize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ideologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ideologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ideologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ideologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ideologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ideologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "idiotize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "idiotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idiotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idiotized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idiotised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idiotizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idiotising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idiotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idiotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "idolaster ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "idolaster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolastre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "idolatrize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "idolatrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolatrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idolatrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolatrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idolatrizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolatrising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idolatrizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolatrises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "idolatrizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "idolatrizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolatriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idolatrizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolatriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "idolization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "idolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idolizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idolization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "idolize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "idolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "idolizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "idolizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idoliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idolizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idolisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idolizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idoliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Idumea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Idumea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Idumaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Idumea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Idumaea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Idumean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Idumean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Idumaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Idumeans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Idumaeans",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "idyll ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "idyll",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idyl",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idylls",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idyls",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "idyll's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "idyl's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ier ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ire",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ier's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ire's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ignitable ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ignitable",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ignitible",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "igniter ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "igniter",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ignitor",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "igniters",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ignitors",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "igniter's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ignitor's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ileocecal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ileocecal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ileocaecal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Iliadize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Iliadize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iliadise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Iliadizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iliadises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Iliadize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iliadise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "illegalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "illegalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illegalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "illegalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illegalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "illegalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illegalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "illegalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "illegalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illegalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "illegalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illegalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "illegalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illegalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "illegalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illegalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "illegitimatize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "illegitimatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illegitimatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "illegitimatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illegitimatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "illegitimatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illegitimatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "illegitimatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illegitimatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "illiberalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "illiberalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illiberalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "illiberalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illiberalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "illiberalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illiberalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "illiberalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "illiberalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Illuminize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Illuminize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Illuminise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Illuminizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Illuminises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Imer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Imer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Imre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Imer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Imre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immaterialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immaterialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immaterialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immaterialize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immaterialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immaterialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immaterialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immaterialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immaterializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immaterialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immaterializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immaterialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immiserization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immiserization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immiserisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immiserizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immiserisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immiserize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immiserize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immiserise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immiserized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immiserised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immiserizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immiserising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immiserizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immiserises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immobilization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immobilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immobilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immobilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immobilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immobilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immobilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immobilize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immobilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immobilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immobilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immobilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immobilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immobilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immobilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immobilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immobilizer ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immobilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immobiliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immobilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immobilisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immoralize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immoralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immoralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immoralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immoralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immoralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immoralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immoralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immoralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immortalizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immortalizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immortalizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immortalizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immortalizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immortalizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immortalization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immortalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immortalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immortalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immortalize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immortalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immortalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immortalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immortalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immortalizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immortalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immortalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immortalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immortaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immunization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immunization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immunisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immunizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immunisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immunization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immunisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immunize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immunize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immunise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immunized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immunised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immunizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immunising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immunizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immunises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "immunizer ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "immunizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immuniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "immunizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "immuniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "impactionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "impactionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "impactionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "impactionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "impactionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "impanel ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "impanel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empanel",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "impaneled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empaneled",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "impanelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empanelled",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "impaneling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empaneling",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "impanelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empanelling",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "impanels",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empanels",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "impanelment ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "impanelment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "empanelment",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "impasted ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "impasted",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "impastoed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "imperialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "imperialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imperialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "imperializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imperialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "imperialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imperialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "imperialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "imperialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imperialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "imperialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imperialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "imperializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imperialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "imperializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imperialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "imperil ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "imperiled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imperilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "imperiling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imperilling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "imperscriptible ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "imperscriptible",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imprescriptible",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "impersonalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "impersonalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "impersonalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "impersonalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "impersonalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "impersonalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "impersonalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "impersonalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "impersonalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "impersonalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "impersonalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "impersonalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "impersonalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "impersonalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "impersonalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "impersonalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "impostor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "impostor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imposter",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "impostors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imposters",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "impostor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "imposter's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "improviser ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "improviser",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "improvisor",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "improvisers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "improvisors",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "improviser's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "improvisor's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "improvizatorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "improvizatorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "improvisatorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "improvisatorize",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "improvizatorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "improvisatorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "improvisatorizes",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "incarnalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "incarnalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incarnalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "incarnalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "incarnalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incarnalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "incarnalizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "incarnalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incarnalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "incenter ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "incenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "incenters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incentres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "incentivization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "incentivization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incentivisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "incentivizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incentivisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "incentivize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "incentivize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incentivise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "incentivized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incentivised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "incentivizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incentivising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "incentivizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incentivises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "incerate ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "incerate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "increate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "incognizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "incognizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incognisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "incognizance ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "incognizance",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incognisance",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "incognizances",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incognisances",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "incognizance's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incognisance's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "incognizant ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "incognizant",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "incognisant",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Indianization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Indianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Indianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Indianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Indianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Indianize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Indianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Indianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Indianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Indianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Indianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Indianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Indianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Indianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "indigene ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "indigene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indigenae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "indigenization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "indigenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indigenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "indigenizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indigenisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "indigenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "indigenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indigenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "indigenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indigenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "indigenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indigenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "indigenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indigenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "individualization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "individualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "individualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "individualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "individualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "individualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "individualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "individualize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "individualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "individualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "individualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "individualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "individualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "individualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "individualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "individualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "individualizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "individualizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "individualiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "individualizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "individualisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "individualizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "individualiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "individualizingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "individualizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "individualisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "indraft ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "indraft",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indraught",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "indrafts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indraughts",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "indraft's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "indraught's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "industrialization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "industrialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "industrialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "industrializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "industrialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "industrialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "industrialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "industrialize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "industrialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "industrialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "industrialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "industrialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "industrializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "industrialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "industrializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "industrialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "inesthetic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "inesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "infamize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "infamize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infamise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "infamized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infamised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "infamizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infamising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "infamizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infamises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "infamonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "infamonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infamonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "infamonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infamonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "infamonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infamonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "infamonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infamonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "infere ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "infere",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infree",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "inferiorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "inferiorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inferiorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inferiorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inferiorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "infernalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "infernalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infernalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "infernalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infernalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "infidelize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "infidelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infidelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "infidelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infidelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "infinitize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "infinitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infinitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "infinitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "infinitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "inflection ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "inflection",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "inflexion",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inflection's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "inflexion's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inflections",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "inflexions",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "informalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "informalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "informalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "informalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "informalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ingveonic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ingveonic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ingvaeonic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ingveonic's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ingvaeonic's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ingweonic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ingweonic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ingwaeonic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ingweonic's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ingwaeonic's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "inhumanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "inhumanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inhumanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inhumanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inhumanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "initial ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "initialed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "initialled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "initialing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "initialling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "initialer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "initialer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initialler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "initialers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initiallers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "initialer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initialler's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "initializable ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "initializable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initialisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "initialization ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "initialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "initializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "initialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "initialize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "initialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "initialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "initializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "initializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "initializer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "initializer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initialiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "initializers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "initialisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "inorganization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "inorganization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inorganisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inorganizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inorganisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "inorganized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "inorganized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inorganised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "inquire ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "inquire",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enquire",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inquired",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enquired",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inquiring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enquiring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inquires",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enquires",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "Enquire is more common in British English, but inquire may still be",
    "used when meaning \"make a formal investigation\" -- Ox",
    "In addition oxford has marks inquire as the preferred variant.",
  ],
  },
Cluster {
  header: "inquiry ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "inquiry",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enquiry",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inquiries",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enquiries",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inquiry's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enquiry's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "See note for inquire ",
  ],
  },
Cluster {
  header: "insignia ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "insignia",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insigne",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "insignia",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insignias",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "insignes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl"),
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "insignia's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insigne's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "insolubilization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "insolubilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insolubilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "insolubilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insolubilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "insolubilize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "insolubilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insolubilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "insolubilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insolubilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "insolubilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insolubilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "insolubilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insolubilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "install ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "install",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "instal",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "installs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "instals",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "installment ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "installment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "instalment",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "installments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "instalments",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "installment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "instalment's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "instill ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "instill",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "instil",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "instills",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "instils",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "instituter ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "instituter",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "institutor",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "instituters",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "institutors",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "instituter's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "institutor's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "institutionalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "institutionalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "institutionalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "institutionalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "institutionalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "institutionalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "institutionalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "institutionalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "institutionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "institutionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "institutionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "institutionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "institutionalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "institutionalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "institutionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "institutionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "institutionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "institutionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "institutionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "institutionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "institutionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "instrumentalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "instrumentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "instrumentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "instrumentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "instrumentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "insularize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "insularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "insularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "insurrectionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "insurrectionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insurrectionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "insurrectionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insurrectionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "insurrectionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insurrectionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "insurrectionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "insurrectionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "integralization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "integralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "integralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "integralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "integralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "integralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "integralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "integralize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "integralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "integralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "integralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "integralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "intellectualization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "intellectualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intellectualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intellectualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intellectualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intellectualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intellectualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "intellectualize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "intellectualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intellectualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intellectualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intellectualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intellectualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intellectualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intellectualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intellectualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "intellectualizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "intellectualizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intellectualiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intellectualizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intellectualisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intellectualizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intellectualiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "intercivilization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "intercivilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intercivilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intercivilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intercivilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intercivilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intercivilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "intercolonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "intercolonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intercolonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intercolonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intercolonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intercolonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intercolonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "intercrystallization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "intercrystallization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intercrystallisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intercrystallizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intercrystallisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intercrystallization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intercrystallisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "intercrystallize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "intercrystallize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intercrystallise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intercrystallizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intercrystallises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "interhemal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "interhemal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "interhaemal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "interhybridize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "interhybridize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interhybridise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "interhybridizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interhybridises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "interiorization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "interiorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interiorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "interiorize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "interiorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interiorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "interiorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interiorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "interiorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interiorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "interiorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interiorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "interjectionalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "interjectionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interjectionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "interjectionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interjectionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "interjectionalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interjectionalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "interjectionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interjectionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "interjectionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "interjectionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interjectionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "interjectionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interjectionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "interjudgment ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "interjudgment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interjudgement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "interjudgments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interjudgements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "interjudgment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interjudgement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "intern ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "intern",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interne",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intern",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interne",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "interns",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intern's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interne's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "internalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "internalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "internalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "internalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "internalize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "internalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "internalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "internalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "internalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "internationalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "internationalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internationalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "internationalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internationalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "internationalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internationalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "internationalize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "internationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "internationalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internationalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "internationalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internationalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "internationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "internment ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "internment",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internement",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "internments",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "internements",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "internship ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "internship",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interneship",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "internships",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interneships",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "interorganizational ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "interorganizational",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "interorganisational",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "intime ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "intime",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intimae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "intraorganization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "intraorganization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intraorganisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intraorganizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intraorganisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "intraorganization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "intraorganisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "inure ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "inure",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enure",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inured",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enured",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inuring",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enuring",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inures",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "enures",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "invigor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "invigor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "invigour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "inward ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "inward",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "inward",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "inwards",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adverb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "iodization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "iodization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iodisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iodization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iodisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "iodize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "iodize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iodise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iodized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iodised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iodizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iodising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iodizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iodises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "iodizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "iodizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iodiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iodizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iodisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iodizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iodiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ionicization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ionicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ionicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ionicizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ionicisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ionicization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ionicisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ionicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ionicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ionicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ionicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ionicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ionicized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ionicised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ionicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ionicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ionicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ionicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ionicize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ionicise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ionicizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ionicising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ionizable ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ionizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ionization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ionization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ionize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ionizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ionizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ioniser",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionisers",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ioniser's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ionizion ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ionizion",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionision",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ionizions",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ionisions",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Iphinoe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Iphinoe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iphinoae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Iphinoe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iphinoae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ipomea ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ipomea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ipomoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Iranize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Iranize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iranise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Iranizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iranises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Iranize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iranise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Iricize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Iricize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iricise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Iricize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iricise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Iricized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Iricized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iricised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Iricized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iricised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Iricizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Iricizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iricising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Iricizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Iricising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "iridectomize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "iridectomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iridectomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "iridectomized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "iridectomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iridectomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "iridectomizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "iridectomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iridectomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "iridization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "iridization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iridisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iridizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iridisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "iridize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "iridize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iridise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iridized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iridised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iridizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iridising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "iridizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "iridises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Irishize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Irishize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Irishise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Irishized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Irishised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Irishized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Irishised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Irishizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Irishising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Irishizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Irishises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Irishize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Irishise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Irishizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Irishising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ironize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ironize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ironise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ironized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ironised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ironizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ironising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ironizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ironises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "irrationalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "irrationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "irrationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "irrationalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "irrationalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "irrationalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "irrationalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "irrationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "irrationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "irrealizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "irrealizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "irrealisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "irrecognizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "irrecognizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "irrecognisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "irregularize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "irregularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "irregularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "irregularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "irregularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ischemia ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ischemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ischaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ischemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ischaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ischemia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ischaemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ischemic ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ischemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ischaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Islamicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Islamicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Islamicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Islamicized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamicised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Islamicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Islamicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Islamicize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamicise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Islamicizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamicising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Islamization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Islamization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Islamizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Islamization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Islamize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Islamize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Islamized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Islamizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Islamizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Islamises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "isochronization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "isochronization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isochronisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "isochronize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "isochronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isochronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "isochronized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isochronised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "isochronizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isochronising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "isochronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isochronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "isoimmunization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "isoimmunization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isoimmunisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "isoimmunizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isoimmunisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "isoimmunization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isoimmunisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "isoimmunize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "isoimmunize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isoimmunise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "isoimmunizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isoimmunises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "isomerization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "isomerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isomerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "isomerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isomerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "isomerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isomerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "isomerize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "isomerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isomerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "isomerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isomerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "isomerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isomerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "isomerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "isomerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Israelitize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Israelitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Israelitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Israelitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Israelitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Israelitize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Israelitise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Italianization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Italianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Italianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Italianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Italianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Italianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Italianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Italianize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Italianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Italianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Italianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Italianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Italianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Italianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Italianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Italianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "italianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "italianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "italianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "italianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "italianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "italianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "italianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "italianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Italianizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Italianizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Italianiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Italianizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Italianisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Italianizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Italianiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "italicization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "italicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "italicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "italicizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "italicisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "italicization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "italicisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "italicize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "italicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "italicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "italicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "italicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "italicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "italicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "italicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "italicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "itemization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "itemization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "itemisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "itemizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "itemisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "itemization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "itemisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "itemize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "itemize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "itemise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "itemized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "itemised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "itemizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "itemising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "itemizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "itemises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "itemizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "itemizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "itemiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "itemizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "itemisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "itemizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "itemiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Iturean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Iturean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ituraean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Itureans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ituraeans",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "izing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "izing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jacobean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jacobean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jacobaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Jacobinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Jacobinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jacobinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jacobinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jacobinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Jacobinize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Jacobinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jacobinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jacobinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jacobinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jacobinized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jacobinised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jacobinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jacobinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jacobinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jacobinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jacobinize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jacobinise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jacobinizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jacobinising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jail ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jail",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaol",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jails",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaols",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jail's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaol's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jailed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaoled",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jailing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaoling",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jailbird ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jailbird",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaolbird",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jailbird's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaolbird's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jailbirds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaolbirds",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jailbreak ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jailbreak",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaolbreak",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jailbreak's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaolbreak's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jailbreaks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gaolbreaks",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jailer ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jailer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jailor",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gaoler",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jailers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jailors",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gaolers",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jailer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jailor's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gaoler's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "janizary ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "janizary",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "janisary",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "janizaries",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "janisaries",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Japanization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Japanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Japanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Japanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Japanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Japanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Japanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Japanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Japanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Japanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Japanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Japanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Japanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Japanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Japanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Japanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jargonization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jargonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jargonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jargonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jargonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jargonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jargonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jargonize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jargonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jargonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jargonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jargonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jargonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jargonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jargonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jargonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jasmine ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jasmine",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jessamine",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jasmines",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jessamines",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jasmine's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jessamine's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jasperize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jasperize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jasperise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jasperized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jasperised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jasperizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jasperising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jasperizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jasperises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jeez ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jeez",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "geez",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jeopardization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jeopardization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jeopardisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jeopardize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jeopardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jeopardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jeopardized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jeopardised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jeopardizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jeopardising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jeopardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jeopardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jerrycan ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jerrycan",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jerrican",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jerry can",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jerrycans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jerricans",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jerry cans",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jerrycan's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jerrican's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jerry can's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Jesuitization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Jesuitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jesuitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jesuitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jesuitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Jesuitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Jesuitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jesuitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jesuitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jesuitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jesuitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jesuitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jesuitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jesuitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jewel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jeweled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jewelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jeweling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jewelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jeweler ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jeweler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jeweller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jewelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jewellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jeweler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jeweller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jewelry ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jewelry",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jewellery",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jewelry's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jewellery's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jibe ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jibe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("agree"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jibed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("agree"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jibing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("agree"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jibes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("agree"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gibe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jibe",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("insult"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gibed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jibed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("insult"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gibing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jibing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("insult"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gibes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jibes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("insult"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gibe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jibe's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("insult"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jibe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gybe",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("sailing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jibed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gybed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("sailing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jibing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gybing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("sailing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jibes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gybes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("sailing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jibe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gybe's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("sailing"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jihad ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jihad",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jehad",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jihads",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jehads",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jihad's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jehad's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jinni ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jinni",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jinn",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "djinni",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "djinn",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jinn",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jinns",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "djinns",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "djinn",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl"),
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jinni's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jinn's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "djinni's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "djinn's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jinrikisha ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jinrikisha",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jinricksha",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jinriksha",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jinrikishas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jinrickshas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jinrikshas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jinrikisha's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jinricksha's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "jinriksha's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jokey ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jokey",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "joky",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Jonathanization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Jonathanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jonathanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jonathanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jonathanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Jonathanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Jonathanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "journalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "journalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "journalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "journalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "journalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "journalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "journalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "journalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "journalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "journalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "journalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "journalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "journalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "journalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "journalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "journalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "journalizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "journalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "journaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "journalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "journalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "journalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "journaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jovialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jovialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jovialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jovializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jovialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Judaization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Judaization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Judaisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Judaization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Judaisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Judaize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Judaize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Judaise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Judaized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Judaised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Judaizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Judaising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Judaizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Judaizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Judaiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Judaizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Judaiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "judaizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "judaiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Judean ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Judean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Judaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Judeans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Judaeans",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Judeophobia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Judeophobia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Judaeophobia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Judeophobia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Judaeophobia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "judgment ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "judgment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "judgement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "judgments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "judgements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "judgment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "judgement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "judgmental ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "judgmental",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "judgemental",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "judgmentaler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "judgementaler",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "judgmentalest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "judgementalest",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "judicialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "judicialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "judicialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "judicializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "judicialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jujitsu ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jujitsu",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jujutsu",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "jiujitsu",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jujitsus",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jujutsus",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "jiujitsus",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jujitsu's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jiujitsu's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "jujutsu's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Julide ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Julide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Julidae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Julide's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Julidae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "jumboize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "jumboize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jumboise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jumboized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jumboised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jumboizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jumboising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "jumboizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "jumboises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "junketeer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "junketeer",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "junketer",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "junketeers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "junketers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "junketeer's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "junketer's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "juvenilize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "juvenilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "juvenilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "juvenilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "juvenilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kabbalah ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kabbalah",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cabala",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kabbala",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kabala",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "cabbala",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "cabbalah",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kane ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kane",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kanae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kaolinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kaolinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kaolinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kaolinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kaolinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kaolinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kaolinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kaolinize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kaolinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kaolinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kaolinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kaolinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kaolinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kaolinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kaolinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kaolinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kebab ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kebab",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kabob",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kebob",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kebabs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kabobs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kebobs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kebab's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kebob's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kabob's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ked ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ked",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kaed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Keizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Keizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Keiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Keizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Keiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kellia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kellia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "koellia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kembed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kembed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kemboed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kennel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kenneled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kennelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kenneling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kennelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Keppel ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Keppel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Koeppel",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Keppel's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Koeppel's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "keratinization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "keratinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "keratinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "keratinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "keratinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "keratinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "keratinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "keratinize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "keratinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "keratinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "keratinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "keratinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "keratinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "keratinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "keratinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "keratinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Kerin ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Kerin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Krein",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Kerin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Krein's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kernel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kerneled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kernelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kerneling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kernelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kerosene ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kerosene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kerosine",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kerosenes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kerosines",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kerosene's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kerosine's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ketchup ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ketchup",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catsup",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "catchup",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ketchups",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catsups",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "catchups",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ketchup's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "catsup's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "catchup's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ketonemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ketonemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ketonaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ketonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ketonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ketonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ketonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ketonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ketonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ketonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ketonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ketonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ketonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ketonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ketonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kiddie ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kiddie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kiddy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kiddie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kiddy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kidnap ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kidnapped",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kidnaped",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kidnapping",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kidnaping",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kidnappee ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kidnappee",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kidnapee",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kidnappees",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kidnapees",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kidnappee's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kidnapee's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kidnapper ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kidnapper",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kidnaper",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kidnappers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kidnapers",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kidnapper's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kidnaper's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kilogram ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kilogram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kilogramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kilograms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kilogrammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kilogram's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kilogramme's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kiloliter ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kiloliter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "kilolitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kiloliters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "kilolitres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kiloliter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "kilolitre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kilometer ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kilometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kilometre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kilometers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kilometres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kilometer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kilometre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kindergartner ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kindergartner",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kindergartener",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kindergartners",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kindergarteners",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kindergartner's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kindergartener's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kinesthesia ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kinesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kinesthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kinesthesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaesthesia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kinesthesis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kinesthesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaesthesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kinestheses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaestheses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kinesthesis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaesthesis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kinesthetic ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kinesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kinestheticer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaestheticer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kinestheticest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaestheticest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kinesthetics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaesthetics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kinesthetically ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kinesthetically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaesthetically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kinestheticallier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaestheticallier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kinestheticalliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "kinaestheticalliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kissogram ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kissogram",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kissagram",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kissograms",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kissagrams",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kissogram's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kissagram's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "knickknack ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "knickknack",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nicknack",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "knickknacks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nicknacks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "knickknack's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nicknack's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "knockwurst ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "knockwurst",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "knackwurst",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "knockwursts",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "knackwursts",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "knockwurst's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "knackwurst's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kooky ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kooky",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kookie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kopeck ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kopeck",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kopek",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "copeck",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kopecks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kopeks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "copecks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kopeck's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kopek's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "copeck's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Kossean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Kossean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Kossaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Kossean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Kossaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kumquat ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kumquat",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cumquat",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kumquats",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cumquats",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kumquat's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "cumquat's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kyanize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kyanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kyanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kyanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kyanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kyanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kyanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "kyanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kyanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "kyles ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "kyles",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "kyloes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "label ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "labelable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labelable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labellable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "labeler ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labeler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labeller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labeler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labeller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "labialization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "labialize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "labiate ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labiate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labiatae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "labilization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "labilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "labiovelarization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labiovelarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labiovelarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labiovelarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labiovelarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "labiovelarize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labiovelarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labiovelarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "labiovelarized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labiovelarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labiovelarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "labiovelarizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labiovelarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labiovelarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "labor ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "labor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "laboured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laboring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laboring's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labouring's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "labored's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "laboured's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laborability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laborability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laborable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laborable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laborage ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laborage",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourage",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laboredly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laboredly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labouredly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laboredlier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labouredlier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laboredliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labouredliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laboredness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laboredness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labouredness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborednesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourednesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laboredness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labouredness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laborer ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laborer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laboress ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laboress",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labouress",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laborhood ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laborhood",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourhood",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborhoods",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourhoods",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborhood's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourhood's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laboringly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laboringly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labouringly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laboringlier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labouringlier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laboringliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labouringliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Laborism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Laborism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "Labourism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Laborism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "Labourism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laborist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laborist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labourist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labourists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labourist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Laborite ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Laborite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Labourite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Laborites",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Labourites",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Laborite's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Labourite's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labourite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborites",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labourites",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborite's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "labourite's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laborless ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laborless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourlesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laborlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourlessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laborlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laborlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "labourlesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laborsaving ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laborsaving",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "laboursaving",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laborsome ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laborsome",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "laboursome",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laborsomely ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laborsomely",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "laboursomely",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lachrymal ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lachrymal",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lacrimal",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lackluster ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lackluster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lacklustre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lacklusterrer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lacklustrer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lacklusterrest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lacklustrest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lacklusters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lacklustres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lackluster's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lacklustre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laconize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laconize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laconise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laconized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laconised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laconizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laconising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laconizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laconises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lactonized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lactonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lactonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lagniappe ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lagniappe",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lagnappe",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lagniappes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lagnappes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lagniappe's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lagnappe's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laicization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laicizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laicisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laicization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laicisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laicize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lairize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lairize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lairise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lairized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lairised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lairizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lairising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lairizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lairises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lambaste ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lambaste",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lambast",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lambastes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lambasts",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laminarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laminarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laminarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laminarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laminarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laminarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laminarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laminarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laminarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Lander ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Lander",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Landre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Lander's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Landre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Laothoe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Laothoe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Laothoae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Laothoe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Laothoae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lapeled ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lapeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lapelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lasagna ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lasagna",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lasagne",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lasagnas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lasagnes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lasagna's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lasagne's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laster ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lastre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "latentize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "latentize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "latentise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "latentizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "latentises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lateralization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lateralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lateralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lateralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lateralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lateralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lateralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lateralize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lateralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lateralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lateralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lateralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laterization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laterization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laterisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laterizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laterisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laterization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laterisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Latinization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Latinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Latinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Latinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Latinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Latinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Latinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Latinize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Latinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Latinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Latinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Latinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Latinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Latinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Latinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Latinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "latinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "latinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Latinizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Latinizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Latiniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Latinizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Latinisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "launderette ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "launderette",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laundrette",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "launderettes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laundrettes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "launderette's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laundrette's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laure ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laure",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laurae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "laurel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "laureled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laurelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "laureling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laurelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lavalier_microphone ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lavalier microphone",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lavaliere microphone",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lavalier microphones",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lavaliere microphones",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lavalier microphone's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lavaliere microphone's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lavolted ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lavolted",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lavoltaed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leaped ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leaped",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "leapt",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "learned ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "learned",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "learnt",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leatherize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leatherize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leatherise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leatherizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leatherises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lefty ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lefty",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leftie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lefty's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leftie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "legalization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "legalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "legalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "legalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "legging ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "legging",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leggin",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leggings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leggins",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legging's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leggin's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "legitimatize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "legitimatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legitimatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legitimatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legitimatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legitimatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legitimatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legitimatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legitimatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "legitimization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "legitimization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legitimisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legitimizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legitimisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legitimization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legitimisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "legitimize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "legitimize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legitimise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legitimized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legitimised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legitimizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legitimising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "legitimizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legitimises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "legitimizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "legitimizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "legitimiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leguminose ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leguminose",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leguminosae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lemmatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lemmatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lemmatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lemmatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lemmatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lemmatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lemmatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lemmatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lemmatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leporide ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leporide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leporidae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leptospire ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leptospire",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leptospirae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Lernean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Lernean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Lernaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Lernean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Lernaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lernean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lernaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lesses ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lethalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lethalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lethalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lethalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lethalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lethargize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lethargize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lethargise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lethargized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lethargised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lethargizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lethargising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lethargizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lethargises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leucemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leucemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leucaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leucemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leucaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leucemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leucemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leucaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leuchemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leuchemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leuchaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leucocythemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leucocythemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leucocythaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leucocythemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leucocythemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leucocythaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leucorrhea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leucorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leucorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leucorrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leucorrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leucorrhea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leucorrhoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leucorrheal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leucorrheal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leucorrhoeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Leucothoe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Leucothoe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Leucothoae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Leucothoe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Leucothoae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leukemia ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leukemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "leukaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leukemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "leukaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leukemia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "leukaemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leukemogenesis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leukemogenesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "leukaemogenesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leukemogeneses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "leukaemogeneses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leukorrhea ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leukorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "leukorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "leukorrheal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "leukorrheal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "leukorrhoeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "level ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "level",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "levelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "levelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "levellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leveler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "leveller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "levelest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "levellest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "levelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "leveling's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "levelling's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levigate ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levigate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevigate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "levigated",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevigated",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "levigating",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevigating",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "levigates",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevigates",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levo ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levo",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevo",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levoduction ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levoduction",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevoduction",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levogyrate ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levogyrate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevogyrate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levogyre ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levogyre",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevogyre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levogyrous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levogyrous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevogyrous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levolactic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levolactic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevolactic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levorotation ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levorotation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevorotation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "levorotations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevorotations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "levorotation's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevorotation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levorotatory ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levorotatory",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevorotatory",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levotartaric ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levotartaric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevotartaric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levoversion ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levoversion",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevoversion",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levulin ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levulin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevulin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "levulose ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "levulose",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevulose",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "levuloses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevuloses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "levulose's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "laevulose's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lexicalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lexicalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lexicalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lexicalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lexicalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lexiconize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lexiconize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lexiconise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lexiconizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lexiconises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "libel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "libeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "libeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "libelant ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "libelant",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libellant",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "libelants",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libellants",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "libelant's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libellant's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "libelee ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "libelee",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libellee",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "libelees",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libellees",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "libelee's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libellee's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "libeler ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "libeler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libeller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "libelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "libeler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libeller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "libelous ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "libelous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "libellous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "libelouser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "libellouser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "libelousest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "libellousest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "libelously ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "libelously",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libellously",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "libelouslier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libellouslier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "libelousliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "libellousliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "liberalization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "liberalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liberalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liberalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liberalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liberalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liberalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "liberalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "liberalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liberalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liberalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liberalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liberalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liberalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liberalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liberalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "liberalizer ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "liberalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liberaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liberalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liberalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liberalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liberaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "license ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "license",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "licence",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "licenses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "licences",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "license's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "licence's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "license",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "licence",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "licensed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "licenced",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "licensing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "licencing",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "licenses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "licences",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "licenseless ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "licenseless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "licenceless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "licenselesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "licenselesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "licencelesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lichenization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lichenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lichenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lichenizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lichenisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lichenization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lichenisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lichenize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lichenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lichenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lichenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lichenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lichenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lichenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lichenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lichenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "licorice ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "licorice",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquorice",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "licorice's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquorice's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lignitize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lignitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lignitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lignitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lignitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lignose ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lignose",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lignosae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ligule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ligule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ligulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "likability ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "likability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "likeability",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "likabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "likeabilities",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "likability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "likeability's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "likable ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "likable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "likeable",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "likableness ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "likableness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "likeableness",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "likablenesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "likeablenesses",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "likableness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "likeableness's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Lilliputianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Lilliputianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Lilliputianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Lilliputianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Lilliputianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Lilliputianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Lilliputianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "linchpin ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "linchpin",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lynchpin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "linchpins",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lynchpins",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "linchpin's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lynchpin's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "linearizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "linearizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linearisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "linearization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "linearization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linearisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "linearizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linearisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "linearization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linearisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "linearize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "linearize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linearise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "linearized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linearised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "linearizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linearising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "linearizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linearises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "linearizion ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "linearizion",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linearision",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "linenize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "linenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "linenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "linenizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "linenizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lineniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "linenizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linenisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "linge ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "linge",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lingoe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lingualize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lingualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lingualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lingualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lingualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "linguine ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "linguine",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linguini",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "linguines",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linguinis",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "linguine's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "linguini's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lionizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lionizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lionisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lionizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lionisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lionizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lionisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lionization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lionization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lionisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lionizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lionisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lionization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lionisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lionize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lionizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lionizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lioniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lionizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lionisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lionizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lioniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lipemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lipemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lipaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lipemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lipemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lipaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lipoidemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lipoidemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lipoidaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "liquefier ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "liquefier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquifier",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liquefiers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquifiers",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "liquefy ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "liquefy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquify",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liquefied",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquified",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liquefying",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquifying",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liquefies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquifies",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "liquidization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "liquidization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquidisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "liquidize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "liquidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liquidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liquidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liquidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "liquidizer ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "liquidizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquidiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liquidizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquidisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liquidizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liquidiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lissome ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lissome",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lissom",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lissomely ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lissomely",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lissomly",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lissomeness ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lissomeness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lissomness",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lissomenesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lissomnesses",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Listerize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Listerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Listerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Listerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Listerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Listerized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Listerised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Listerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Listerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Listerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Listerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Listerize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Listerise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Listerizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Listerising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "litchi ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "litchi",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lychee",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lichee",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "litchis",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lychees",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lichees",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "litchi's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lychee's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lichee's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "liter ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "liter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "litre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "litres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "liter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "litre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "literalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "literalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "literalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "literalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "literalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "literalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "literalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "literalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "literalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "literalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "literalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "literalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "literalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "literalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "literalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "literalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "literalizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "literalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "literaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "literalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "literalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "literalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "literaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lithed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lithed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lithoed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lithemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lithemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lithaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lithemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lithemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lithaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lithographize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lithographize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lithographise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lithographizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lithographises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lithopedion ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lithopedion",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lithopaedion",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lithopedium ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lithopedium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lithopaedium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lithophyse ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lithophyse",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lithophysae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lithotritize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lithotritize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lithotritise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lithotritized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lithotritised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lithotritizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lithotritising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lithotritizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lithotritises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "livable ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "livable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liveable",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "livableness ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "livableness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "liveableness",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "loathe ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "loathe",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("detest"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "loathes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("detest"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "loathed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("detest"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "loathing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("detest"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "loath",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loth",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "loathe",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: Some("reluctant"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lobotomize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lobotomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lobotomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lobotomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lobotomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lobotomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lobotomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lobotomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lobotomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "localizable ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "localizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "localizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "localizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "localizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "localizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "localization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "localization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "localizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "localization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "localize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "localize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "localized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "localizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "localizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "localizer ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "localizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "localizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "localizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "localiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "locator ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "locator",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "locater",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "locators",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "locaters",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "locator's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "locater's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lodestar ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lodestar",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loadstar",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lodestars",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loadstars",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lodestar's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loadstar's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lodestone ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lodestone",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loadstone",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lodestones",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loadstones",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lodestone's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loadstone's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lodgment ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lodgment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lodgement",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lodgments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lodgements",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lodgment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lodgement's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lodicule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lodicule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lodiculae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "loed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "loed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "looed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "logicalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "logicalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "logicalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "logicalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "logicalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "logicalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "logicalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "logicalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "logicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "logicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "logicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "logicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "logicize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "logicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "logicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "logicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "logicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "logicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "logicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "logicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "logicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "logopedic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "logopedic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "logopaedic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "logopedics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "logopaedics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "logopedics's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "logopaedics's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "logorrhea ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "logorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "logorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "logorrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "logorrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "logorrhea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "logorrhoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lollipop ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lollipop",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lollypop",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lollipops",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lollypops",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lollipop's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lollypop's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lollygag ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lollygag",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lallygag",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lollygagged",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lallygagged",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lollygagging",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lallygagging",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lollygags",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lallygags",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Londonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Londonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Londonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Londonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Londonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Londonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Londonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Londonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Londonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Londonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Londonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Londonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Londonized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Londonised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Londonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Londonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Londonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Londonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Londonize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Londonise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Londonizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Londonising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "longeval ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "longeval",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "longaeval",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "longevous ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "longevous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "longaevous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "loony ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "loony",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "looney",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "loonier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "looneyier",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "loonies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "looneyies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "looneys",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "loony's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "looney's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lordy ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lordy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lourdy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lorings ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lorings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lory ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lory",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loury",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "louter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "louter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loutre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "louver ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "louver",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "louvre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "louvered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "louvred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "louvers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "louvres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "louver's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "louvre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lovability ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lovability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lovaebility",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lovable ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lovable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loveable",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lovableness ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lovableness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loveableness",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lovably ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lovably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loveably",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "love ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "love",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "luv",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "loves",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "luvs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "loyalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "loyalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loyalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "loyalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "loyalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lumbarization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lumbarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lumbarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lumbarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lumbarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lumbarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lumbarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lunatize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lunatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lunatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lunatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lunatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lunule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lunule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lunulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lupine ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lupine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lupin",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("plant"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lupines",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lupins",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("plant"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lupine's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lupin's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("plant"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lupine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: Some("wolfish"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "luster ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "luster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lustre",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lustered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lustred",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lustering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lustring",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "luster's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lustre's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lustering's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lustring's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lusters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lustres",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lusterless ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lusterless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lustreless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lusterlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lustrelesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lusterlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lustrelessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lusterware ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lusterware",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lustreware",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lusterwares",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lustrewares",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lusterware's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lustreware's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "luteinization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "luteinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "luteinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "luteinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "luteinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "luteinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "luteinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "luteinize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "luteinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "luteinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "luteinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "luteinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "luteinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "luteinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "luteinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "luteinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Lutheranize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Lutheranize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Lutheranise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Lutheranizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Lutheranises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Lutheranize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Lutheranise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Lutheranizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Lutheranizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Lutheraniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Lutheranizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Lutheranisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Lutheranizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Lutheraniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lutrine ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lutrine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lutrinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "luvvy ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "luvvy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "luvvie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lovey",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "luvvie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "loveys",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "luvvies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lymphedema ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lymphedema",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "lymphoedema",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lymphemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lymphemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lymphaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lyophilization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lyophilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyophilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lyophilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyophilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lyophilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyophilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lyophilize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lyophilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyophilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lyophilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyophilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lyophilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyophilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lyophilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyophilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lyophilizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lyophilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyophiliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lyricization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lyricization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyricisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lyricization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyricisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lyricize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lyricize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyricise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lyricized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyricised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lyricizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyricising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lyricizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lyricises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lysogenization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lysogenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lysogenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "lysogenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "lysogenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lysogenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lysogenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lysogenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lysogenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lysogenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "lysogenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "lysogenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mac ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mac",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mack",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "macs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mac's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mack's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "macadamization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "macadamization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macadamisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "macadamizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macadamisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "macadamization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macadamisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "macadamize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "macadamize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macadamise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "macadamized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macadamised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "macadamizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macadamising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "macadamizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macadamises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "macarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "macarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "macarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "macarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "macarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Maccabean ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Maccabean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Maccabaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "machinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "machinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "machinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "machinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "machinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "machinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "machinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "machinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "machinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "machinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "machinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "machinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mackintosh ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mackintosh",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macintosh",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mackintoshes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macintoshes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mackintosh's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macintosh's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "macrander ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "macrander",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "macrandre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "macule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "macule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maculae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Madera ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Madera",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Madoera",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Madera's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Madoera's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maderization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maderization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maderisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maderizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maderisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maderize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maderize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maderise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maderized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maderised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maderizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maderising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maderizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maderises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maaed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "magicalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "magicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "magicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "magnetizability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "magnetizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "magnetizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "magnetizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "magnetizable ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "magnetizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "magnetizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "magnetizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "magnetization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "magnetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "magnetizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "magnetization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "magnetize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "magnetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "magnetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "magnetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "magnetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "magnetizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "magnetizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "magnetizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "magnetizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "magnetiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maharajah ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maharajah",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maharaja",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maharajah's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maharaja's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maharajahs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maharajas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maharani ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maharani",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maharanee",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maharanis",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maharanees",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maharani's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maharanee's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mah-jongg ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mah-jongg",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mahjong",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mah-jonggs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mahjongs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mah-jongg's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mahjong's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mahoganize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mahoganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mahoganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mahoganized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mahoganised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mahoganizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mahoganising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mahoganizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mahoganises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maiger ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maiger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maigre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mainor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mainor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mainour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mainors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mainours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mainprize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mainprize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mainprise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maisters ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maisters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maistres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Maize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Maize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Maise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Maize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Maise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "majolica ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "majolica",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maiolica",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "majolicas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maiolicas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "majolica's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maiolica's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "majorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "majorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "majorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "majorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "majorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "majuscule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "majuscule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "majusculae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Malayize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Malayize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Malayise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Malayizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Malayises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Malayize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Malayise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "malleablize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "malleablize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "malleablise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "malleablizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "malleablises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "malodor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "malodor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "malodour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "malodors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "malodours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "malodor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "malodour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "malodorousness ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "malodorousness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "malodourousness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mamma ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mamma",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("gland"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mammae",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mammas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("gland"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mamma's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("gland"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mama",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "momma",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mamma",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("mother"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mamas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mommas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mammas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("mother"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mama's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "momma's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mamma's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("mother"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mandarinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mandarinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mandarinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mandarinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mandarinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Mandean ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Mandean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Mandaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mandeans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Mandaeans",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mandean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Mandaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mandoer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mandoer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mandore",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mandrel ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mandrel",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mandril",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mandrels",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mandrils",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mandrel's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mandril's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maneuver ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maneuver",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "manoeuvre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuver",
        types: &[
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maneuvered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "manoeuvred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuvered",
        types: &[
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maneuvering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "manoeuvring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuvering",
        types: &[
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maneuverings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "manoeuvrings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuverings",
        types: &[
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maneuvers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "manoeuvres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuvers",
        types: &[
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maneuver's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "manoeuvre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuver's",
        types: &[
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maneuverability ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maneuverability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuvrability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maneuverabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuvrabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maneuverability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuvrability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maneuverable ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maneuverable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuvrable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maneuverabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuvrabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maneuverablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuvrablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maneuverer ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maneuverer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuvrer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maneuverers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuvrers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maneuverer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manoeuvrer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Manhattanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Manhattanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Manhattanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Manhattanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Manhattanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Manicheanism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Manicheanism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Manichaeanism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Manicheanism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Manichaeanism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Manicheans ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Manicheans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Manichaeans",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Manicheism ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Manicheism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Manichaeism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Manicheisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Manichaeisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Manicheism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Manichaeism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Manicheus ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Manicheus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Manichaeus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Manicheus's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Manichaeus's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "manikin ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "manikin",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mannikin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "manikins",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mannikins",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "manikin's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mannikin's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "manila ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "manila",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manilla",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":lower"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "manila's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "manilla's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":lower"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Manila",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Manilla",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":upper"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Manila's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Manilla's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":upper"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mannerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mannerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mannerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mannerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mannerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mantelpiece ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mantelpiece",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mantlepiece",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mantelpieces",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mantlepieces",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mantes ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mantes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mantoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marabou ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marabou",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marabout",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marabous",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marabouts",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marabou's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marabout's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marabout",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marabouts",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marabout's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marbleize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marbleize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marbleise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marbleized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marbleised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marbleizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marbleising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marbleizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marbleises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "margarite ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "margarite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "margaritae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marginalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marginalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marginalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marginalize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marginalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marginalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marginalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marginalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marginalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marginalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marginalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marginalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marijuana ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marijuana",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marihuana",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marijuanas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marihuanas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marijuana's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marihuana's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marlinespike ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marlinespike",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marlinspike",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marlinespikes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marlinspikes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marlinespike's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marlinspike's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marmarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marmarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marmarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marmarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marmarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marmarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marmarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marmarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marmarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marmelize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marmelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marmelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marmelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marmelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marmelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marmelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marmelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marmelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marshal ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marshaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marshalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marshaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marshalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marsupialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marsupialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marsupialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marsupializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marsupialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marsupialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marsupialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marsupialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marsupialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marsupialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marsupialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marsupialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marsupializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marsupialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marsupializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marsupialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Marte ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Marte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Martae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Marte's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Martae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "martialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "martialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "martializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "martialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "martialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "martialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "martializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "martyrization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "martyrization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martyrisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "martyrizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martyrisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "martyrization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martyrisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "martyrize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "martyrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martyrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "martyrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martyrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "martyrizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martyrising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "martyrizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martyrises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "martyrizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "martyrizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martyriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "martyrizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "martyrisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marvel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marvelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marvelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marveler ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marvelous ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marvelous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marvellous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marvelouser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marvellouser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marvelousest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marvellousest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marvelously ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marvelously",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marvellously",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marvelouslier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marvellouslier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marvelousliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marvellousliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "marvelousness ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "marvelousness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marvellousness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marvelousnesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marvellousnesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "marvelousness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "marvellousness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "masculinization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "masculinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "masculinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "masculinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "masculinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "masculinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "masculinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "masculinize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "masculinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "masculinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "masculinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "masculinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "masculinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "masculinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "masculinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "masculinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mat ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "matte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "mat",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "matt",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: Some("without lustre"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mat",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "matte",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "matt",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("make matte "),
  comment: Some("Ox Amer thinks its matte, M-W: mat"),
  },
  Entry {
    variants: &[
      Variant {
        word: "mats",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mattes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "matts",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("make matte"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "matted",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("make matte"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "matting",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("make matte"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mat",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise, most uses"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mats",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise, most uses"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mat's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise, most uses"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "matted",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise, most uses"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "matting",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise, most uses"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "matte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise, some uses"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mattes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise, some uses"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "matte's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise, some uses"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "matted",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise, some uses"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "matting",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise, some uses"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Matt",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("Matthew"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "materialization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "materialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "materialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "materializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "materialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "materialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "materialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "materialize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "materialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "materialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "materialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "materialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "materializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "materialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "materializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "materialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "materializer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "materializer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "materialiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "materializers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "materialisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "materializer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "materialiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maternalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maternalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maternalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maternalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maternalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maternalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maternalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maternalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maternalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maters ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "matres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "math ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "math",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maths",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mathematicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mathematicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mathematicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mathematicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mathematicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mathematicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mathematicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mathematicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mathematicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mathematization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mathematization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mathematisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mathematizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mathematisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mathematization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mathematisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mathematize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mathematize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mathematise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mathematized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mathematised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mathematizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mathematising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mathematizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mathematises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "matronize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "matronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "matronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "matronized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "matronised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "matronizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "matronising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "matronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "matronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Matthean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Matthean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Matthaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maudlinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maudlinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maudlinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maudlinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maudlinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mauger ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mauger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maugre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maven ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maven",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mavin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mavens",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mavins",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maven's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mavin's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maximization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maximization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maximisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maximizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maximisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maximization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maximisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maximize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maximize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maximise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maximized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maximised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maximizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maximising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maximizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maximises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "maximizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maximizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maximiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maximizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maximisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maximizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maximiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mazurka ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mazurka",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mazourka",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mazurkas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mazourkas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mazurka's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mazourka's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meager ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meager",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "meagre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meagerrer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "meagrer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meagerrest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "meagrest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meagerly ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meagerly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "meagrely",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meagerness ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meagerness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "meagreness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meagernesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "meagrenesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meagerness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "meagreness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meandrine ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meandrine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maeandrine",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meandriniform ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meandriniform",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maeandriniform",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meanie ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meanie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "meany",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meanies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meanie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "meany's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meany",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meanies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meany's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mechanicalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mechanicalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanicalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mechanicalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanicalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mechanicalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanicalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mechanicalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mechanicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mechanicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mechanizable ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mechanizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mechanization ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mechanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mechanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mechanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mechanize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mechanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mechanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mechanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mechanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mechanizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mechanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mechanizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechanisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mechanizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mechaniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meck ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meck",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moeck",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Med ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Med",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Moed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Med's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Moed's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "medal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "medaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "medalist ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "medalist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medallist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medalists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medallists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medalist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medallist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "medalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "medalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "medialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "medialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "medialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "medialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mediatization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mediatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mediatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mediatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mediatize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mediatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mediatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mediatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mediatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "medicalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "medicalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medicalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medicalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medicalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "medicalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "medicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medicalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medicalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medicalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medicalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "medicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "medieval ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "medieval",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediaeval",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medievals",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediaevals",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medieval's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediaeval's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "medievalism ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "medievalism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediaevalism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medievalisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediaevalisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "medievalism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediaevalism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "medievally ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "medievally",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediaevally",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Mediterraneanization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Mediterraneanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mediterraneanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mediterraneanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mediterraneanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mediterraneanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mediterraneanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Mediterraneanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Mediterraneanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mediterraneanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mediterraneanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mediterraneanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mediterraneanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mediterraneanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mediumization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mediumization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediumisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mediumizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediumisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mediumization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediumisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mediumize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mediumize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediumise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mediumizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mediumises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "megameter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "megameter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "megametre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "megameters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "megametres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "megbote ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "megbote",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "maegbote",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "megerg ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "megerg",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "megaerg",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melanemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melanemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "melanaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melanemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melanemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "melanaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melanization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "melanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "melanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "melanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "meloe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melena ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melena",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melaena",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melenic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melenic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melaenic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Melie ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Melie",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Meliae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Melie's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Meliae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meline ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meline",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melitemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melitemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melitaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melithemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melithemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melithaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mellon ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mellon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moellon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melodization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melodization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melodisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melodize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melodize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melodise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "melodized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melodised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "melodizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melodising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "melodizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melodises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melodizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melodizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melodiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "melodizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melodiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melodramatization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melodramatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melodramatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "melodramatize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "melodramatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melodramatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "melodramatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melodramatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "melodramatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melodramatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "melodramatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "melodramatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "memorialization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "memorialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "memorialize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "memorialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "memorializer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "memorializer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorialiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorializers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorialisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorializer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorialiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "memorizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "memorizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "memorization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "memorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "memorize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "memorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "memorizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "memorizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memoriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memorisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "memorizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "memoriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "menad ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maenad",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "menad",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maenads",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "menads",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "maenad's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "menad's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "menadic ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "maenadic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "menadic",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Mendelize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Mendelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mendelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mendelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mendelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mendelize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mendelise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meningorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meningorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "meningorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "menorrhea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "menorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "menorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "menorrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "menorrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "menorrheic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "menorrheic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "menorrhoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mense ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mense",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mensae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mentalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mentalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mentalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mentalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mentalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mentalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mentalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mentalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meow ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meow",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "miaow",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meowed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "miaowed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meowing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "miaowing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meows",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "miaows",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meow's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "miaow's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mephitine ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mephitine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mephitinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Mera ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Mera",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Maera",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mera's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Maera's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mercerization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mercerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mercerize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mercerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mercerizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mercerizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "merceriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercerizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercerisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercerizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "merceriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "merchandise ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "merchandise",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "merchandize",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "merchandised",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "merchandized",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "merchandising",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "merchandizing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "merchandises",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "merchandizes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "merchandiser ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "merchandiser",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "merchandizer",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "merchandisers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "merchandizers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "merchandiser's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "merchandizer's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mercurialization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mercurialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercurialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercurializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercurialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercurialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercurialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mercurialize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mercurialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercurialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercurialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercurialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercurializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercurialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercurializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercurialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mercurize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mercurize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercurise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercurized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercurised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercurizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercurising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mercurizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mercurises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Meroe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Meroe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Meroae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Meroe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Meroae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mesmerizability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mesmerizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesmerizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesmerizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mesmerizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mesmerizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesmerizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesmerizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mesmerization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mesmerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesmerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesmerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mesmerize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mesmerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesmerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesmerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesmerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mesmerizer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mesmerizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmeriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesmerizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmerisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesmerizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesmeriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mesoglea ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mesoglea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesogloea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesogleas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesogloeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mesogleal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mesogleal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesogloeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mesprize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mesprize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesprise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesprizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesprises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mesquite ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mesquite",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesquit",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesquites",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesquits",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mesquite's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mesquit's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "ox: possible austrian variant",
  ],
  },
Cluster {
  header: "metabolizable ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metabolizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metabolisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metabolizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metabolisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metabolizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metabolisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metabolize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metabolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metabolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metabolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metabolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metabolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metabolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metabolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metabolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metacenter ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metacenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metacentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metacenters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metacentres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metacenter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metacentre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metagrabolize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metagrabolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metagrabolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metagrabolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metagrabolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metagrabolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metagrabolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metagrabolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metagrabolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metagrobolize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metagrobolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metagrobolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metagrobolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metagrobolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metagrobolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metagrobolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metagrobolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metagrobolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metal ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "metalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "metalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metalization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallization",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallizations",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallization's",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallize",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallized",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallizing",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metallizes",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metamerization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metamerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metamerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metamerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metamerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metamerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metamerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metamerized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metamerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metamerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metamerizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metameriseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metamerized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metamerised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metaphonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metaphonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metaphonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metaphonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metaphonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metaphorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metaphorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metaphorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metaphorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metaphorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metaphysicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metaphysicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metaphysicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metaphysicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metaphysicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metastasize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metastasize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metastasise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metastasized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metastasised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metastasizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metastasising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metastasizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metastasises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metathesize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metathesize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metathesise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metathesized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metathesised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metathesizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metathesising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metathesizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metathesises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meteorization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meteorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "meteorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meteorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "meteorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meteorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "meteorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meteorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meteorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "meteorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meteorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "meteorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meter ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("measuring device"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("measuring device"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("measuring device"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("measuring device"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("measuring device"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "metre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("metric meter, rhythmic pattern"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "metres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("metric meter, rhythmic pattern"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "meter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "metre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("metric meter, rhythmic pattern"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metergram ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metergram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metregram",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "meterless ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "meterless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metreless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metership ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metership",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metreship",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metestrus ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metestrus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metoestrus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metestrus's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metoestrus's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "methadone ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "methadone",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methadon",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "methadones",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methadons",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "methadone's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methadon's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "methemoglobin ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "methemoglobin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "methaemoglobin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "methodization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "methodization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methodisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "methodizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methodisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "methodization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methodisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "methodize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "methodize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methodise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "methodized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methodised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "methodizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methodising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "methodizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methodises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "methodizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "methodizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methodiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "methodizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methodisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "methodizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "methodiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metope ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metope",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metopae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metricize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metricize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metricise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metricized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metricised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metricizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metricising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metricizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metricises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metrize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metropolitanization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metropolitanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metropolitanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "metropolitanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "metropolitanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metropolitanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metropolitanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metropolitanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metropolitanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metropolitanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "metropolitanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "metropolitanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Mexicanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Mexicanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mexicanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mexicanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mexicanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mexicanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mexicanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "micelle ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "micelle",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micellae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "micresthete ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "micresthete",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micraesthete",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "microcolorimeter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "microcolorimeter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microcolourimeter",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "microcolorimetric ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "microcolorimetric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microcolourimetric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "microcolorimetrically ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "microcolorimetrically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microcolourimetrically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "microcolorimetry ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "microcolorimetry",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microcolourimetry",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "microliter ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "microliter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microlitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "microliters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microlitres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "microliter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microlitre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "micrometer ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "micrometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("gauge"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micrometers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("gauge"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micrometer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("gauge"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micrometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "micrometre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("metric meter"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micrometers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "micrometres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("metric meter"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micrometer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "micrometre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("metric meter"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "micromillimeter ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "micromillimeter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micromillimetre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micromillimeters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micromillimetres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micromillimeter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micromillimetre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "microminiaturization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "microminiaturization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microminiaturisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "microminiaturizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microminiaturisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "microminiaturization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microminiaturisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "microminiaturize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "microminiaturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microminiaturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "microminiaturized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microminiaturised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "microminiaturizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microminiaturising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "microminiaturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microminiaturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "microminiaturizer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "microminiaturizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microminiaturiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "microminiaturizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microminiaturisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "micronization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "micronization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micronisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micronizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micronisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micronization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micronisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "micronize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "micronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "micropaleontologist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "micropaleontologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "micropalaeontologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micropaleontologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "micropalaeontologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "micropaleontology ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "micropaleontology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "micropalaeontology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micropaleontologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "micropalaeontologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micropaleontology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "micropalaeontology's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "micropolarization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "micropolarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micropolarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micropolarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micropolarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micropolarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "micropolarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "microscopize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "microscopize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microscopise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "microscopizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microscopises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "microspheric ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "microspheric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microsphaeric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Midlandize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Midlandize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Midlandise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Midlandizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Midlandises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Midlandize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Midlandise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "midsize ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "midsize",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "midsized",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "midwife ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "midwifed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "midwived",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "midwifing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "midwiving",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "migniardize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "migniardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "migniardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "militarization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "militarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "militarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "militarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "militarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "militarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "militarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "militarize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "militarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "militarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "militarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "militarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "militarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "militarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "militarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "militarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "milligram ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "milligram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "milligramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "milligrams",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "milligrammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "milligram's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "milligramme's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "milliliter ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "milliliter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "millilitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "milliliters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "millilitres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "milliliter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "millilitre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "millimeter ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "millimeter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "millimetre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "millimeters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "millimetres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "millimeter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "millimetre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "millionaire ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "millionaire",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "millionnaire",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "millionaires",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "millionnaires",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "millionaire's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "millionnaire's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "millionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "millionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "millionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "millionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "millionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "millipede ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "millipede",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "millepede",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "millipedes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "millepedes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "millipede's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "millepede's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "milometer ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "milometer",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mileometer",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "milometers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mileometers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "milometer's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mileometer's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Miltonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Miltonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Miltonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Miltonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Miltonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Miltonized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Miltonised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Miltonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Miltonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Miltonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Miltonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Miltonize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Miltonise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Miltonizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Miltonising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mineralizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mineralizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mineralizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mineralizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mineralization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mineralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mineralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mineralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mineralize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mineralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mineralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mineralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mineralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mineralizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mineralizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineraliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mineralizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mineralizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineraliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mineralogize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mineralogize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralogise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mineralogized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralogised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mineralogizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralogising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mineralogizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mineralogises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "miniaturization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "miniaturization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miniaturisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miniaturizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miniaturisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miniaturization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miniaturisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "miniaturize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "miniaturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miniaturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miniaturized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miniaturised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miniaturizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miniaturising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miniaturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miniaturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "minibuses ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "minibuses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "minibusses",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "minimization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "minimization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "minimisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "minimizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "minimisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "minimization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "minimisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "minimize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "minimize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "minimise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "minimized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "minimised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "minimizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "minimising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "minimizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "minimises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "minimizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "minimizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "minimiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "minimizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "minimisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "minimizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "minimiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "minuscule ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "minuscule",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miniscule",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "minuscules",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miniscules",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "minuscule's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miniscule's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "miraculize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "miraculize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miraculise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miraculized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miraculised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miraculizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miraculising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miraculizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miraculises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mirrorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mirrorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mirrorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mirrorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mirrorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misadvize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misadvize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misadvise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misalphabetize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misalphabetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misalphabetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misalphabetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misalphabetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misanthropize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misanthropize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misanthropise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misanthropized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misanthropised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misanthropizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misanthropising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misanthropizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misanthropises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misauthorization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misauthorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misauthorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misauthorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misauthorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misauthorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misauthorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misauthorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misauthorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misauthorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misauthorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misauthorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misbaptize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misbaptize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misbaptise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misbaptizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misbaptises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misbehavior ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misbehavior",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "misbehaviour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misbehaviors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "misbehaviours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misbehavior's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "misbehaviour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "miscanonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "miscanonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miscanonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miscanonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miscanonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mischaracterization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mischaracterization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mischaracterisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mischaracterizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mischaracterisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mischaracterization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mischaracterisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mischaracterize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mischaracterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mischaracterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mischaracterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mischaracterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "miscolor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "miscolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "miscolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miscolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "miscoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miscoloring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "miscolouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miscolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "miscolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miscolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "miscolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "miscoloration ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "miscoloration",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "miscolouration",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misdemeanor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misdemeanor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "misdemeanour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misdemeanors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "misdemeanours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misdemeanor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "misdemeanour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misemphasize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misemphasize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misemphasise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misemphasizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misemphasises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misjudgment ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misjudgment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "misjudgement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misjudgments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "misjudgements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misjudgment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "misjudgement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mislabel ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mislabeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mislabelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mislabeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mislabelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mislabor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mislabor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mislabour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mislabored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mislaboured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mislaboring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mislabouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mislabors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mislabours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mislabor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mislabour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misopedia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misopedia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "misopaedia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misopedism ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misopedism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "misopaedism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misopedist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misopedist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "misopaedist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misopedist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "misopaedist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misorganization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misorganization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misorganisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misorganizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misorganisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misorganization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misorganisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misorganize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misorganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misorganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misorganizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misorganises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misprizal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misprizal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misprisal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misprize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misprize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misprise",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misprized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misprised",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misprizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misprising",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misprizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misprises",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misprizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misprizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mispriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misrealize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misrealize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misrealise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misrealizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misrealises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misrecognize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misrecognize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misrecognise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "misrecognizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misrecognises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "misrouting ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "misrouting",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "misrouteing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "missilery ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "missilery",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missilry",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "missilery's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missilry's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "missionarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "missionarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missionarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "missionarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missionarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "missionarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missionarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "missionarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missionarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "missionization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "missionization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missionisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "missionize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "missionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "missionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "missionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "missionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "missionizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "missionizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missioniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "missionizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missionisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "missus ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "missus",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missis",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "missuses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missises",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "missus's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "missis's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "miter ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "miter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mitered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mitred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mitering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mitring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mitres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mitre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "miterer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "miterer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mitrer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miterer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mitrer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "miterflower ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "miterflower",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mitreflower",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "miterwort ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "miterwort",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mitrewort",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miterworts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mitreworts",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "miterwort's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mitrewort's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mithridatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mithridatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mithridatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mithridatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mithridatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mithridatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mithridatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mithridatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mithridatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mobilizable ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mobilizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mobilizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mobilizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mobilizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mobilizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mobilization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mobilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mobilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mobilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mobilize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mobilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mobilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mobilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mobilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mobilizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mobilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobiliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mobilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobilisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mobilizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mobiliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "modalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "modalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "model ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "modeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modeler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modeller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modeler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modeller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modeling's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modelling's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "modernizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "modernizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modernisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modernizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modernisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modernizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modernisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "modernization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "modernization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modernisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modernizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modernisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modernization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modernisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "modernize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "modernize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modernise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modernized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modernised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modernizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modernising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modernizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modernises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "modernizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "modernizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moderniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modernizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modernisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modernizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moderniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "modularization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "modularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "modularize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "modularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "modularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "modularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "moggy ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "moggy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "moggie",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moggy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "moggie's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Mohammedanization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Mohammedanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mohammedanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mohammedanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mohammedanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mohammedanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mohammedanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Mohammedanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Mohammedanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mohammedanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mohammedanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mohammedanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mohammedanized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mohammedanised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mohammedanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mohammedanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mohammedanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mohammedanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mohammedanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mohammedanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mohammedanizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mohammedanising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "moisturization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "moisturization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moisturisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "moisturize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "moisturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moisturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moisturized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moisturised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moisturizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moisturising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moisturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moisturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "moisturizer ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "moisturizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moisturiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moisturizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moisturisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moisturizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moisturiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "molarization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "molarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "molarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "molarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mold ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mold",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mould",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molded",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "moulded",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molding",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "moulding",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mold's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mould's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moldings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mouldings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molding's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "moulding's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "moulds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "molder ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "molder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "moulder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molder's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "moulder's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moldered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mouldered",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moldering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mouldering",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molders",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "moulders",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "moldy ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "moldy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mouldy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moldier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mouldier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moldiest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "mouldiest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "molluscan ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "molluscan",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "molluskan",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mollusk ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mollusk",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mollusc",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mollusks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "molluscs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mollusk's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mollusc's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Molochize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Molochize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Molochise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Molochizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Molochises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Molochize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Molochise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molochize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "molochise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molochized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "molochised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molochizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "molochising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molochizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "molochises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "molt ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "molt",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "moult",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molt's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "moult's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molted",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "moulted",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molting",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "moulting",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "molts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "moults",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mom ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mom",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("mother"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mums",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("mother"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mom's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mum's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("mother"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: Some("silent"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mommy ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mommy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mommie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mommy's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mommie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monarchize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monarchize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monarchise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monarchized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monarchised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monarchizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monarchising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monarchizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monarchises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monarchizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monarchizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monarchiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monarchizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monarchisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monasticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monasticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monasticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monasticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monasticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monecian ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monecian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monoecian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monecious ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monoecious",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monecious",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monestrous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monestrous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monoestrous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monetization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monetizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monetisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monetization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monetisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monetize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "moneyed ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "moneyed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monied",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Mongolize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Mongolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mongolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mongolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mongolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mongolized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mongolised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mongolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mongolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mongolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mongolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mongolize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mongolise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Mongolizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Mongolising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mongrelization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mongrelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mongrelisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mongrelizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mongrelisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mongrelization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mongrelisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mongrelize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mongrelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mongrelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mongrelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mongrelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mongrelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mongrelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mongrelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mongrelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mongrelizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mongrelizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mongreliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "moniker ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "moniker",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monicker",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monikers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monickers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moniker's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monicker's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monochordize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monochordize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monochordise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monochordizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monochordises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monologist ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monologuist",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monologuists",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monologuist's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monologize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monologue ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monologue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monolog",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monologues",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monologs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monologue's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monolog's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monologuize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monologuize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monologuise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monologuized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monologuised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monologuizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monologuising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monologuizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monologuises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monometalism ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monometalism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monometallism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monometalist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monometalist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monometallist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monophthongize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monophthongize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monophthongise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monophthongized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monophthongised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monophthongizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monophthongising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monophthongizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monophthongises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monopolizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monopolizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopolisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monopolizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopolisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monopolizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopolisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monopolization ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monopolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monopolizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopolisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monopolization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopolisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monopolize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monopolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monopolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monopolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monopolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monopolizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monopolizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopoliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monopolizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopolisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monopolizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monopoliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monotonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monotonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monotonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monotonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monotonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monumentalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monumentalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monumentalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monumentalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monumentalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monumentalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monumentalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "monumentalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "monumentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monumentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monumentalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monumentalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monumentalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monumentalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "monumentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "monumentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "moralization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "moralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "moralize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "moralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "moralizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "moralizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moraliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moralizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moralisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moralizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moraliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "moralizingly ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "moralizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moralisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "moralizinglies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moralisinglies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Moravianized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Moravianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Moravianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Moravianizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Moravianiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Moravianized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Moravianised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "morbidize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "morbidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morbidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "morbidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morbidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Morea ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Morea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Moraea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Morea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Moraea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "morne ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "morne",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mourne",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "morningly ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "morningly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mourningly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "morphinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "morphinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morphinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "morphinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morphinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "morphinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morphinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "morphinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "morphinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morphinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "morphinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morphinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "morsel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "morseled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morselled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "morseling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morselling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "morselization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "morselization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morselisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "morselizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morselisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "morselization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morselisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "morselize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "morselize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morselise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "morselizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morselises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mortalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mortalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mortalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mortalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mortalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mortalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mortalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mortalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mortalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mortarize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mortarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mortarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mortarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mortarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mortgagor ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mortgagor",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mortgager",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mortgagors",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mortgagers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mortgagor's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mortgager's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mortise ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mortise",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mortice",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mortised",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morticed",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mortising",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morticing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mortises",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mortices",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mortise's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mortice's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "morule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "morule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "morulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Moslemize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Moslemize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Moslemise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Moslemizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Moslemises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Moslemize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Moslemise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "motorization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "motorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "motorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "motorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "motorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "motorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "motorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "motorize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "motorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "motorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "motorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "motorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "motorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "motorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "motorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "motorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mousy ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mousy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mousey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "movability ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "movability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moveability",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "movable ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "movable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moveable",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "movables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moveables",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "movable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moveable's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "movableness ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "movableness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moveableness",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "movablenesses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moveablenesses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "movably ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "movably",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moveably",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mucorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mucorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "mucorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mucose ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mucose",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mucosae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mujahideen ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mujahideen",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mujahedin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "mujahedeen",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mujahideen's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mujahedin's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "mujahedeen's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "multicolor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "multicolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "multicolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "multicolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "multicolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "multicolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "multicolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "multicolored ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "multicolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "multicoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "multicoloredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "multicolouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "multicoloreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "multicoloureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "multicolorous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "multicolorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "multicolourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "multifibered ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "multifibered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "multifibred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "multifibereds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "multifibreds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "multifibered's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "multifibred's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "multileveled ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "multileveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "multilevelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "multiplexer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "multiplexer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "multiplexor",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "multiplexers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "multiplexors",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "multiplexer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "multiplexor's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "municipalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "municipalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "municipalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "municipalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "municipalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "municipalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "municipalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "municipalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "municipalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "municipalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "municipalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "municipalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "municipalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "municipalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "municipalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "municipalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "municipalizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "municipalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "municipaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "municipalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "municipalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "murena ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "murena",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "muraena",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "murenas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "muraenas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "muscularize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "muscularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "muscularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "muscularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "muscularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "museumize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "museumize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "museumise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "museumizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "museumises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "musicalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "musicalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "musicalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "musicalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "musicalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "musicalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "musicalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "musicalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "musicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "musicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "musicalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "musicalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "musicalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "musicalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "musicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "musicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Muslim ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Muslim",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Moslem",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Muslim's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Moslem's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mustache ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mustache",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "moustache",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mustaches",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "moustaches",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mustache's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "moustache's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mustached ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mustached",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moustached",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mustachio ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mustachio",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moustachio",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mustachio's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moustachio's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mustachioed ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mustachioed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "moustachioed",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mutagenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mutagenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mutagenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mutagenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mutagenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mutagenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mutagenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mutagenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mutagenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mutualization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mutualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mutualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mutualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mutualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mutualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mutualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mutualize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mutualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mutualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mutualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mutualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mutualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mutualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mutualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mutualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mycohemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mycohemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "mycohaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myelinization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myelinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "myelinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "myelinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "myelinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "myelinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "myelinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myelocele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myelocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "myelocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myelocythemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myelocythemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "myelocythaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mylonitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mylonitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mylonitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mylonitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mylonitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mylonitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mylonitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mylonitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mylonitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mylonitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mylonitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mylonitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mylonitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mylonitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myna ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myna",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "mynah",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mynas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "mynahes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "myna's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "mynah's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myocele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "myocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myohematin ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myohematin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "myohaematin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myrialiter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myrialiter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "myrialitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myriameter ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myriameter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "myriametre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mysticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mysticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mysticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mysticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mysticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mythicization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mythicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mythicize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mythicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mythicizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mythicizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythiciser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythicizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythicisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythicizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythiciser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mythize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mythize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mythologization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mythologization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythologisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythologizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythologisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythologization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythologisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mythologize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mythologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mythologizer ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mythologizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythologiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythologizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythologisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "mythologizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythologiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mythopeic ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mythopeic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythopoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mythopeist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mythopeist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythopoeist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mythopoetize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mythopoetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythopoetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mythopoetized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mythopoetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythopoetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "mythopoetizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "mythopoetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "mythopoetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myxameba ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myxameba",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "myxamoeba",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myxedema ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myxedema",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "myxoedema",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "myxedemas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "myxoedemas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "myxedema's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "myxoedema's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myxedematous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myxedematous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "myxoedematous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myxedemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myxedemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "myxoedemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "myxemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "myxemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "myxaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "naan ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "naan",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nan",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naans",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nans",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naan's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nan's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Nabateans ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Nabateans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Nabataeans",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Nabathean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Nabathean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Nabathaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Nabathean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Nabathaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nakedize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nakedize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nakedise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nakedizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nakedises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nanization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nanogram ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nanogram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nanogramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nanograms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nanogrammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nanometer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nanometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "nanometre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nanometers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "nanometres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nanometer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "nanometre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "napea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "napea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "napaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "naphthalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "naphthalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naphthalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naphthalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naphthalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naphthalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naphthalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naphthalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naphthalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Napoleonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Napoleonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Napoleonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Napoleonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Napoleonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Napoleonize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Napoleonise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "narc ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "narc",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nark",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "narcs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "narks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "narc's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nark's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nark",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2 -- british"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "narked",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2 -- british"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "narking",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2 -- british"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "narks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2 -- british"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nark's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2 -- british"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "narcotization ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "narcotization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "narcotisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "narcotizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "narcotisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "narcotization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "narcotisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "narcotize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "narcotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "narcotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "narcotized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "narcotised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "narcotizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "narcotising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "narcotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "narcotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nasalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nasalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nasalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nasalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nasalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nasalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nasalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nasalize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nasalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nasalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nasalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nasalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nasalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nasalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nasalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nasalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nationalization ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nationalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nationalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nationalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nationalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nationalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nationalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nationalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nationalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nationalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nationalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nationalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nationalizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nationalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nationaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nationalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nationalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nationalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nationaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "naturalization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "naturalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naturalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naturalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naturalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naturalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naturalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "naturalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "naturalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naturalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naturalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naturalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naturalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naturalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naturalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naturalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "naturalizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "naturalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naturaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naturalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naturalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naturalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naturaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "naturize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "naturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "naught ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "naught",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nought",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("digit zero"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naughts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noughts",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("digit zero"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naught's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nought's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("digit zero"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naught",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nought",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: Some("worthless"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naught",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("<N> nothing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naughts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("<N> nothing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "naught's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("<N> nothing"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nebularization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nebularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nebularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nebularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nebularize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nebularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nebularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nebulization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nebulization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebulisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nebulizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebulisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nebulization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebulisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nebulize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nebulize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebulise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nebulized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebulised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nebulizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebulising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nebulizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebulises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nebulizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nebulizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebuliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nebulizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebulisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nebulizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nebuliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "necremia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "necremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "necraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "necrotize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "necrotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "necrotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "necrotized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "necrotised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "necrotizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "necrotising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "necrotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "necrotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nectarize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nectarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nectarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nectarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nectarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nectarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nectarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nectarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nectarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neebor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neebor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neebour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neencephalon ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neencephalon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neoencephalon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nere",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "negligee ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "negligee",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neglig",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "negligees",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "negligs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "negligee's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neglig's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Negritize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Negritize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negritise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Negritize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negritise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Negritized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Negritized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negritised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Negritized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negritised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Negritizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Negritizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negritising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Negritizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negritising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Negroization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Negroization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negroisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Negroizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negroisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Negroization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negroisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Negroize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Negroize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negroise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Negroized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negroised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Negroized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negroised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Negroizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negroising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Negroizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negroises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Negroize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negroise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Negroizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Negroising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neighbor ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neighbor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighbored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighboured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighboring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighbor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighbors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neighborer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neighborer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neighboress ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neighboress",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbouress",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neighborhood ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neighborhood",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourhood",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborhoods",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourhoods",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborhood's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourhood's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neighborless ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neighborless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "neighbourlesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourlessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neighborlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neighborlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourlesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neighborlike ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neighborlike",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourlike",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborlikes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourlikes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborlike's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourlike's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neighborliness ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neighborliness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourliness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborlinesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourlinesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborliness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourliness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neighborly ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neighborly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborlier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourlier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neighborship ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neighborship",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourship",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborships",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourships",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neighborship's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "neighbourship's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Neogea ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Neogea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Neogaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Neogea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Neogaea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Neogeal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Neogeal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Neogaeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Neogeal's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Neogaeal's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Neogean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Neogean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Neogaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Neogeic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Neogeic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Neogaeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Neogeic's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Neogaeic's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neologization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neologization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neologisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neologize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neopaganize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neopaganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neopaganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neopaganized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neopaganised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neopaganizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neopaganising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neopaganizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neopaganises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neoterize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neoterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neoterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neoterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neoterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neoterizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neoterising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neoterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neoterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nephrectomize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nephrectomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nephrectomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nephrectomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nephrectomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nephrectomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nephrectomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nephrocele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nephrocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nephrocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nephrotomize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nephrotomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nephrotomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nesslerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nesslerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nesslerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nesslerized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nesslerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nesslerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nesslerizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nesslerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nesslerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "net ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "net",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nett",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nets",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "netts",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "net's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nett's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neurepithelium ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neurepithelium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neuroepithelium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neurocele ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neurocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neurocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neurocelian ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neurocelian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neurocoelian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neuronal ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neuronal",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neuronic",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neuroticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neuroticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neuroticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neuroticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neuroticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neutralization ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neutralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neutralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neutralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neutralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neutralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neutralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neutralize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neutralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neutralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neutralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neutralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neutralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neutralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neutralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neutralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neutralizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neutralizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neutraliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neutralizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neutralisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "neutralizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "neutraliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "neves ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "neves",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naeves",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nevoid ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nevoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "naevoid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Newmanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Newmanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Newmanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Newmanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Newmanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Newmanized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Newmanised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Newmanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Newmanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Newmanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Newmanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Newmanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Newmanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Newmanizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Newmanising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "newspaperized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "newspaperized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "newspaperised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "newspaperizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "newspaperiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "newspaperized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "newspaperised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nickel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nickeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nickelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nickeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nickelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nickelization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nickelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nickelisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nickelizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nickelisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nickelization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nickelisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nickelize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nickelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nickelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nickelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nickelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nickelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nickelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nickelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nickelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nicotinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nicotinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nicotinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nicotinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nicotinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nicotinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nicotinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nicotinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nicotinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nielled ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nielled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nielloed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "niger ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "niger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nigre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "niggardize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "niggardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "niggardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "niggardized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "niggardised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "niggardizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "niggardising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "niggardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "niggardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "night ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "night",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nite",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nights",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nites",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "night's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nite's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "Ox labels nite as an informal spelling",
  ],
  },
Cluster {
  header: "nightie ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nightie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nighty",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nightie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nighty's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nightingalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nightingalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nightingalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nightingalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nightingalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Nipponize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Nipponize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Nipponise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Nipponizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Nipponises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Nipponize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Nipponise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "niter ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "niter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "niters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "niter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nitridization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nitridization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitridisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nitridizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitridisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nitridization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitridisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nitridize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nitridize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitridise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nitridizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitridises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nitrogenization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nitrogenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitrogenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nitrogenizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitrogenisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nitrogenization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitrogenisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nitrogenize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nitrogenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitrogenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nitrogenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitrogenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nitrogenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitrogenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nitrogenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nitrogenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nitroglycerin ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nitroglycerin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "nitroglycerine",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nitroglycerins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "nitroglycerines",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nitroglycerin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "nitroglycerine's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nodalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nodalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nodalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nodalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nodalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nodalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nodalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nodalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nodalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nodulize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nodulize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nodulise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nodulizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nodulises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Noemon ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Noemon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Noaemon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Noemon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Noaemon's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nomadization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nomadization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nomadisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nomadizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nomadisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nomadization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nomadisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nomadize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nomadize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nomadise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nomadized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nomadised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nomadizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nomadising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nomadizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nomadises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nominalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nominalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nominalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nominalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nominalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nominalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nominalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nominalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nominalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nominalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nominalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nominalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nominalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nominalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonacknowledgment ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonacknowledgment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonacknowledgement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonacknowledgments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonacknowledgements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonacknowledgment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonacknowledgement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonanemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonanemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "nonanaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonanesthetized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonanesthetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonanaesthetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonanaesthetized",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonapostatizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonapostatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonapostatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonapostatizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonapostatisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonapostatizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonapostatising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncanonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncanonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncanonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncanonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncanonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncanonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncanonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncartelized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncartelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncartelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncartelizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncarteliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncartelized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncartelised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncatechizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncatechizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncatechisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncatechizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncatechisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncatechizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncatechisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncivilized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncivilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncivilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncivilizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonciviliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncivilized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncivilised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncolor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncoloring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncolorings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncolorability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncolorability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncolorabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncolorability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncolorable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncolorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncolorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncolorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncolorableness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncolorableness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourableness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncolorableness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourableness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncolorably ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncolorably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncolorablies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourablies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncolorer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncolorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncolorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncolorfast ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncolorfast",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "noncolourfast",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncrystallizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncrystallizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncrystallisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncrystallizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncrystallisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncrystallizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncrystallisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncrystallized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncrystallized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncrystallised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncrystallizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncrystalliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncrystallized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncrystallised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noncrystallizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noncrystallizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncrystallising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncrystallizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncrystallisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "noncrystallizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noncrystallising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nondemobilization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nondemobilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nondemobilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nondemobilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nondemobilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nondemobilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nondemobilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nondialyzing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nondialyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nondialysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nondialyzings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nondialysings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nondialyzing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nondialysing's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nondimensionalize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nondimensionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nondimensionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nondimensionalized ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nondimensionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nondimensionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonecumenic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonecumenic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonoecumenic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonecumenical ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonecumenical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonoecumenical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonencyclopedic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonencyclopedic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonencyclopaedic",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "See note for encyclopedia",
  ],
  },
Cluster {
  header: "nonesthetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonesthetical ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonesthetical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonaesthetical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonesthetically ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonesthetically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonaesthetically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonfavorite ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonfavorite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "nonfavourite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonfavorites",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "nonfavourites",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonfavorite's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "nonfavourite's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonflavored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonflavored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "nonflavoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonfulfillment ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonfulfillment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonfulfilment",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonfulfillments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonfulfilments",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonfulfillment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonfulfilment's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nongalvanized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nongalvanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nongalvanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nongalvanizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nongalvaniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nongalvanized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nongalvanised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nongelatinizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nongelatinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nongelatinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nongelatinizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nongelatinisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nongelatinizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nongelatinising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonhydrolyzable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonhydrolyzable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonhydrolysable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonhydrolyzables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonhydrolysables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonhydrolyzable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonhydrolysable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonimmunized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonimmunized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonimmunised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonimmunizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonimmuniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonimmunized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonimmunised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonionized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonionizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonioniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonionized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonionised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonionizing ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonionizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonionisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonionizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonionising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonlocalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonlocalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonlocalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonlocalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonlocaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonlocalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonlocalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonmagnetizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonmagnetizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonmagnetisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonmagnetizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonmagnetisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonmagnetizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonmagnetisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonnitrogenized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonnitrogenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonnitrogenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonnitrogenizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonnitrogeniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonnitrogenized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonnitrogenised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonorganization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonorganization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonorganisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonorganizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonorganisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonorganization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonorganisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonoxidizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonoxidizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonoxidisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonoxidizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonoxidisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonoxidizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonoxidisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonoxidizing ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonoxidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonoxidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonoxidizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonoxidisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonoxidizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonoxidising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonparlor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonparlor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "nonparlour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonparlors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "nonparlours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonparlor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "nonparlour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonpenalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonpenalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonpenalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonpenalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonpenaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonpenalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonpenalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonphosphorized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonphosphorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonphosphorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonphosphorizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonphosphoriseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonphosphorized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonphosphorised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonpolarizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonpolarizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonpolarisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonpolarizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonpolarisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonpolarizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonpolarisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonpolarized ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonpolarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonpolarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonpolarizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonpolarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonpolarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonpolarizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonpolarisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonpolarizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonpolarising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonrationalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonrationalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonrationalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonrationalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonrationaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonrationalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonrationalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonrealization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonrealization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonrealisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonrealizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonrealisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonrealization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonrealisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonrecognized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonrecognized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonrecognised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonrecognizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonrecogniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonrecognized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonrecognised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonschematized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonschematized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonschematised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonschematizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonschematiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonschematized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonschematised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonsensitized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonsensitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonsensitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonsensitizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonsensitiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonsensitized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonsensitised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonspecialized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonspecialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonspecialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonspecializeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonspecialiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonspecialized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonspecialised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonstandardized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonstandardized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonstandardised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonstandardizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonstandardiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonstandardized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonstandardised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonstylized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonstylized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonstylised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonstylizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonstyliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonstylized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonstylised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonsuccor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonsuccor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonsuccour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonsympathizer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonsympathizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonsympathiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonsympathizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonsympathisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonsympathizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonsympathiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonsynthesized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonsynthesized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonsynthesised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonsynthesizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonsynthesiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonsynthesized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonsynthesised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nontemporizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nontemporizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nontemporising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nontemporizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nontemporisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nontemporizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nontemporising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonutilized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonutilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonutilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonutilizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonutiliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonutilized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonutilised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonvisualized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonvisualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonvisualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonvisualizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonvisualiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonvisualized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonvisualised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonvolatilized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonvolatilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonvolatilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonvolatilizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonvolatiliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonvolatilized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonvolatilised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nonvulcanizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nonvulcanizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonvulcanisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonvulcanizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonvulcanisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nonvulcanizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nonvulcanisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "norice ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "norice",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nourice",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "normalizable ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "normalizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normalisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "normalization ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "normalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "normalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "normalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "normalize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "normalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "normalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "normalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "normalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "normalizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "normalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "normalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "normalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Normanization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Normanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Normanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Normanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Normanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Normanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Normanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Normanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Normanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Normanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Normanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Normanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Normanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Normanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Normanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Normanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "normanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "normanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "normanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "normanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "normanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Normanizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Normanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Normaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Normanizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Normanisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Normanizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Normaniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Northernize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Northernize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Northernise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Northernize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Northernise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "northernize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "northernise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "northernized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "northernised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "northernizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "northernising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "northernizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "northernises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nosize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nosize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nosise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nosohemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nosohemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "nosohaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nosy ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nosy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nosey",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nosies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noseys",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nosy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nosey's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "notarization ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "notarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "notarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "notarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "notarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "notarize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "notarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "notarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "notarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "notarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "notarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "notarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "notarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "notarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "noter ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "noter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "notre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nothingize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nothingize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nothingise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nothingizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nothingises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Notogea ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Notogea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Notogaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Notogea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Notogaea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nounize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nounize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nounise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nounizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nounises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "novelization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "novelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "novelisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "novelizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "novelisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "novelization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "novelisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "novelize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "novelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "novelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "novelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "novelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "novelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "novelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "novelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "novelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "novelizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "novelizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noveliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "novelizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "novelisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "novelizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "noveliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "novene ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "novene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "novenae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nuclearization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nuclearization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nuclearisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nuclearizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nuclearisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nuclearize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nuclearize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nuclearise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nuclearized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nuclearised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nuclearizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nuclearising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nuclearizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nuclearises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "numskull ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "numskull",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "numbskull",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "numskulls",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "numbskulls",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "numskull's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "numbskull's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nuptialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nuptialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nuptialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "nuptializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nuptialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "nympheum ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "nympheum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nymphaeum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Obe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Obe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Oboe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Obe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Oboe's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "obelize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "obelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "obelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "obelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "obelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "obelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "obelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "obelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "obelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "objectivize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "objectivize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "objectivise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "objectivized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "objectivised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "objectivizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "objectivising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "objectivizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "objectivises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "objectization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "objectization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "objectisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "objectizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "objectisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "objectization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "objectisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "objectize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "objectize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "objectise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "objectizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "objectises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oblivionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oblivionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oblivionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oblivionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oblivionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Occidentalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Occidentalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Occidentalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Occidentalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Occidentalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Occidentalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Occidentalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Occidentalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Occidentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Occidentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Occidentalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Occidentalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Occidentalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Occidentalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Occidentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Occidentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "occidentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "occidentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "occidentalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "occidentalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "occidentalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "occidentalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "occidentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "occidentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ocher ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ocher",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ochre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ochered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ochred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ochering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ochring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ocherrer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ochrer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ocherrest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ochrest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ochers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ochres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ocher's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ochre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ocherish ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ocherish",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ochreish",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ocherous ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ocherous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ochreous",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ocherouser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ochreouser",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ocherousest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ochreousest",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ochery ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ochery",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ochrey",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "octet ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "octet",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "octette",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "octets",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "octettes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "octet's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "octette's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ocyrrhoe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ocyrrhoe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ocyrrhoae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ocyrrhoe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ocyrrhoae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "odor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "odor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "odour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "odors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "odours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "odor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "odour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "odored ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "odored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "odoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "odorful ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "odorful",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "odourful",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "odorfuller",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "odourfuller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "odorfullest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "odourfullest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "odorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "odorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "odorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "odorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "odorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "odorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "odorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "odorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "odorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "odorizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "odorizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "odoriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "odorless ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "odorless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "odourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "odorlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "odourlesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "odorlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "odourlessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "odorlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "odorlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "odourlesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "offense ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "offense",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "offence",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "offenses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "offences",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "offense's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "offence's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "offenseless ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "offenseless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "offenceless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "offenselesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "offencelesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "offenselessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "offencelessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "offenselesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "offenselesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "offencelesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "offenselessly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "offenselessly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "offencelessly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "officialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "officialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "officialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "officializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "officialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "officialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "officialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "officialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "officialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "officialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "officializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "officialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "OK ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "OK",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "okay",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "OK's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "okay's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "OKed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "okayed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "OKing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "okaying",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "OKs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "okays",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "OK",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("Oklahoma"),
  comment: None,
  },
  ],
  notes: &[
    "OK is generally the preferred form, but it is very difficult to",
    "misspell one for the other and in some contexts okay is preferred,",
    "so I am marking these as equal variants.  Note that the Canadian",
    "Ox. has okay as the primary form.",
  ],
  },
Cluster {
  header: "oleomargarine ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oleomargarine",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oleomargarin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oleomargarines",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oleomargarins",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oleomargarine's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oleomargarin's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oligemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oligemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oligaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oligocythemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oligocythemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oligocythaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "olpe ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "olpe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "olpae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Olympianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Olympianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Olympianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Olympianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Olympianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Olympianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Olympianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "omber ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "omber",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ombre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ombers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ombres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "omber's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ombre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "omelet ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "omelet",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "omelette",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "omelets",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "omelettes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "omelet's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "omelette's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oneyer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oneyer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oneyre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oneyers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oneyres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "onionized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "onionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "onionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "onionizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "onioniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "onionized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "onionised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ontologize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ontologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ontologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "onward ",
  verified: false,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "onward",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "onward",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "onwards",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adverb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ooglea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ooglea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oogloea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oophorectomize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oophorectomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oophorectomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oophorectomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oophorectomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oophorectomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oophorectomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oophorectomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oophorectomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "opaled ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "opaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "opalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "opalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "opalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "opalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "opalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "opalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "opalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "opalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "operationalization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "operationalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "operationalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "operationalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "operationalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "operationalize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "operationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "operationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "operationalized ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "operationalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "operationalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "operatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "operatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "operatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "operatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "operatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "operatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "operatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "operatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "operatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "opiniaster ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "opiniaster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "opiniastre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "opiniater ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "opiniater",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "opiniatre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "optimalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "optimalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "optimalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "optimalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "optimalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "optimalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "optimalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "optimalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "optimization ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "optimization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "optimizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "optimization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "optimize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "optimize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "optimized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "optimizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "optimizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "optimizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "optimizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "optimizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "optimizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optimiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "optionalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "optionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "optionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "optionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oralization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oralize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orang ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orang",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ourang",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orangs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ourangs",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orangize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orangize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orangise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orangizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orangises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orangutan ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orangutan",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orangutang",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orangutans",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orangutangs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orangutan's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orangutang's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oratorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oratorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oratorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oratorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oratorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orchester ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orchester",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orchestre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "organdy ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "organdy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organdie",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organdy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organdie's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "organizability ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "organizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "organizable ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "organizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "organization ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "organization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "organizational ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "organizational",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisational",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizationaler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisationaler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizationalest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisationalest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizationals",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisationals",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizational's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisational's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "organizationally ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "organizationally",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisationally",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizationallier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisationallier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizationalliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisationalliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "organizationist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "organizationist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisationist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizationists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisationists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizationist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisationist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "organize ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "organize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "organizer ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "organizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "organizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "organiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orientalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orientalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orientalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orientalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orientalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orientalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orientalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Orientalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Orientalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Orientalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Orientalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Orientalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Orientalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Orientalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orientalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orientalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orientalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orientalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orientalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orientalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orientalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orientalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orientization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orientization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orientisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orientizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orientisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orientization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orientisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orientize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orientize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orientise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orientizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orientises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ornamentalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ornamentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ornamentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ornamentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ornamentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orology ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ourology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ourologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orphanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orphanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orphanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orphanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orphanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orsel ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orsel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oursel",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthocenter ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthocenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthocentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthocenters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthocentres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthocenter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthocentre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthogonalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthogonalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthogonalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthogonalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthogonalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthogonalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthogonalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthogonalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthogonalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthogonalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthogonalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthogonalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthogonalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthogonalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthographize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthographize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthographise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthographized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthographized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthographised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthographizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthographizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthographising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthopedia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthopedia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthopedic ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthopedic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthopedicer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedicer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthopedicest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedicest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthopedics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthopedic's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedic's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthopedics's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedics's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthopedical ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthopedical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthopedically ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthopedically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthopedicallier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedicallier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthopedicalliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedicalliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthopedist ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthopedist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthopedists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthopedist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthopedy ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthopedy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "orthopedies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "orthopaedies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthopnea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthopnea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthopnoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "orthopneic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "orthopneic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "orthopnoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "osteopedion ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "osteopedion",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "osteopaedion",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ostracizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ostracizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostracisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ostracizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostracisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ostracizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostracisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ostracization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ostracization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostracisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ostracizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostracisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ostracization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostracisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ostracize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ostracize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostracise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ostracized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostracised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ostracizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostracising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ostracizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostracises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ostracizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ostracizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostraciser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ostracizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostracisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ostracizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ostraciser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "othematoma ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "othematoma",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "othaematoma",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "otorrhea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "otorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "otorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ottomanization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ottomanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ottomanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ottomanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ottomanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ottomanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ottomanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ottomanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ottomanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ottomanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ottomanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ottomanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ottomanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ottomanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outcavil ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outcaviled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "outcavilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outcaviling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "outcavilling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outclamor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outclamor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outclamour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outclamors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outclamours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outclamor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outclamour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outerness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outerness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "outreness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outhumor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outhumor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outhumour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outhumored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outhumoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outhumoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outhumouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outhumors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outhumours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outhumor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outhumour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outhyperbolize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outhyperbolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "outhyperbolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outhyperbolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "outhyperbolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outlabor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outlabor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outlabour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outlabors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outlabours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outlabor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outlabour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outluster ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outluster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "outlustre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outmaneuver ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outmaneuver",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "outmanoeuvre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outmaneuvered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "outmanoeuvred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outmaneuvering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "outmanoeuvring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outmaneuvers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "outmanoeuvres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outrival ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outrivaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "outrivalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outrivaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "outrivalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outsavor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outsavor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outsavour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outsavoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outsavouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outsavors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outsavours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outsavor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outsavour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outsize ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outsize",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "outsized",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outsplendor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outsplendor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outsplendour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outsplendors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outsplendours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outsplendor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "outsplendour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "outtyrannize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "outtyrannize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "outtyrannise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "outtyrannizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "outtyrannises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ovalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ovalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ovalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ovalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ovalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ovalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ovalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ovalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ovalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ovalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ovalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ovalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ovariectomized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ovariectomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ovariectomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overagonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overagonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overagonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overagonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overagonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overanalyze ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overanalyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overanalyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overanalyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overanalyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overbrutalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overbrutalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overbrutalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overbrutalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overbrutalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overcapitalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overcapitalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcapitalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcapitalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcapitalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcapitalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcapitalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overcapitalize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overcapitalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcapitalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcapitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcapitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcapitalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcapitalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcapitalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcapitalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overcentralization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overcentralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcentralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcentralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcentralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcentralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcentralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overcentralize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overcentralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcentralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcentralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcentralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overcivilization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overcivilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcivilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcivilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcivilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcivilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcivilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overcivilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overcivilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcivilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcivilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcivilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overclamor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overclamor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overclamour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overclamors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overclamours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overclamor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overclamour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overcolor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overcolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overcolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overcolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcoloring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overcolouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overcolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overcoloration ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overcoloration",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcolouration",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcoloration's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcolouration's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overcriticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overcriticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcriticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overcriticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overcriticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overdoctrinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overdoctrinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overdoctrinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overdoctrinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overdoctrinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overdramatize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overdramatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overdramatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overdramatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overdramatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overdramatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overdramatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overdramatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overdramatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overemotionalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overemotionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overemotionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overemotionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overemotionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overemphasize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overemphasize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overemphasise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overemphasized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overemphasised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overemphasizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overemphasising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overemphasizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overemphasises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overemphasizer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overemphasizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overemphasiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overemphasizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overemphasisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overfavor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overfavor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overfavour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overfavors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overfavours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overfavor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overfavour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overfavorable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overfavorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overfavourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overfavorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overfavourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overfavorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overfavourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overfavorably ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overfavorably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overfavourably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overfavorablies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overfavourablies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overfertilization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overfertilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overfertilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overgeneralize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overgeneralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overgeneralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overgeneralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overgeneralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overgeneralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overgeneralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overgeneralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overgeneralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overhonor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overhonor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overhonour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overhonors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overhonours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overhonor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "overhonour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overhumanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overhumanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overhumanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overhumanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overhumanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overindustrialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overindustrialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overindustrialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overindustrializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overindustrialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overindustrialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overindustrialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overindustrialize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overindustrialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overindustrialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overindustrializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overindustrialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overjudgment ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overjudgment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overjudgement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overjudgments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overjudgements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overjudgment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overjudgement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overlabor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overlabor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overlabour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overlabored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overlaboured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overlaboring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overlabouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overlabors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overlabours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overlabor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overlabour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overnationalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overnationalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overnationalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overnationalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overnationalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overnationalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overnationalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overrapturize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overrapturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overrapturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overrapturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overrapturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overrationalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overrationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overrationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overrationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overrationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oversentimentalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oversentimentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oversentimentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oversentimentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oversentimentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oversize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oversize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oversized",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overspecialization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overspecialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overspecialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overspecializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overspecialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overspecialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overspecialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overspecialize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overspecialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overspecialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overspecialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overspecialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overspecializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overspecialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overspecializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overspecialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oversystematize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oversystematize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oversystematise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oversystematizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oversystematises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overunionized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overunionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overunionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overunionizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overunioniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overunionized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overunionised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overurbanization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overurbanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overurbanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overurbanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overurbanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overurbanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overurbanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overutilization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overutilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overutilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "overwomanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "overwomanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overwomanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "overwomanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "overwomanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ower ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ower",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "owre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "owercome ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "owercome",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "owrecome",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "owerword ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "owerword",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "owreword",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxalemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxalemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oxalaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxidizability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxidizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxidizable ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxidizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxidization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxidization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxidize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxidizement ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxidizement",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidizements",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidizement's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxidizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxidizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxidizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxidiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxygenizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxygenizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygenisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxygenizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygenisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxygenizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygenisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxygenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxygenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxygenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxygenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxygenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxygenizement ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxygenizement",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygenisement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxygenizements",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygenisements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxygenizement's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygenisement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxygenizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxygenizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygeniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxygenizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygenisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxygenizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "oxygeniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxyhematin ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxyhematin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oxyhaematin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "oxyhemoglobin ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "oxyhemoglobin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oxyhaemoglobin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "oxyhemoglobins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "oxyhaemoglobins",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ozena ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ozena",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ozaena",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ozena",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ozoena",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ozonization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ozonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ozonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ozonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ozonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ozonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ozonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ozonize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ozonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ozonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ozonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ozonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ozonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ozonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ozonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ozonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ozonizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ozonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ozoniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ozonizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ozonisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ozonizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ozoniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pachyemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pachyemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pachyaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pachyhemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pachyhemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pachyhaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "packetization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "packetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "packetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "packetize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "packetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "packetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "packetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "packetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "packetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "packetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "packetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "packetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "packetizer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "packetizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "packetiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "packetizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "packetisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "packetizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "packetiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paeanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paeanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paeanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paeanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paeanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paganization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paganization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paganisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paganizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paganisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paganization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paganisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paganize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paganized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paganised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paganizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paganising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paganizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paganises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paganizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paganizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paganiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paganizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paganisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paganizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paganiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pagine ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pagine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paginae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pajama ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pajama",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pyjama",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pajamas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pyjamas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pajama's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pyjama's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pajamas's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pyjamas's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pajamaed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pajamaed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyjamased",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palame ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palame",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palamae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palatalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palatalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palatalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palatalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palatalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palatalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palatalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palatalize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palatalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palatalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palatalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palatalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palatalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palatalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palatalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palatalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palatization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palatize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleanthropic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleanthropic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeanthropic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Palearctic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Palearctic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Palaearctic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleencephalon ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleencephalon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeencephalon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleencephalon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeencephalon's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoencephalon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoencephalon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoencephalon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoencephalon's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleentomology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleentomology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palaeentomology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleethnologic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleethnologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeethnologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoethnologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoethnologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleethnological ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleethnological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeethnological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoethnological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoethnological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleethnologist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleethnologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeethnologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoethnologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoethnologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleethnology ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleethnology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeethnology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoethnology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoethnology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleichthyologist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleichthyologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeichthyologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleichthyology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleichthyology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeichthyology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoalchemical ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoalchemical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoalchemical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoanthropic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoanthropic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoanthropic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoanthropography ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoanthropography",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoanthropography",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoanthropological ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoanthropological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoanthropological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoanthropologist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoanthropologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoanthropologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoanthropologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoanthropologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoanthropology ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoanthropology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoanthropology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoanthropologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoanthropologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Paleoanthropus ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Paleoanthropus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Palaeoanthropus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Paleoanthropus's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Palaeoanthropus's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoatavism ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoatavism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoatavism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoatavistic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoatavistic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoatavistic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleobiogeography ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleobiogeography",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobiogeography",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleobiologic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleobiologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobiologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleobiological ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleobiological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobiological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleobiologist ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleobiologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobiologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleobiologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobiologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleobiologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobiologist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleobiology ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleobiology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobiology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleobiologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobiologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleobiology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobiology's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleobotanic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleobotanic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobotanic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleobotanical ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleobotanical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobotanical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleobotanically ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleobotanically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobotanically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleobotanist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleobotanist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobotanist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleobotanists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobotanists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleobotany ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleobotany",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobotany",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleobotanies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobotanies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleobotany's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeobotany's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoceanography ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoceanography",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoceanography",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleochorology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleochorology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeochorology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoclimatic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoclimatic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoclimatic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoclimatologic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoclimatologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoclimatologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoclimatological ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoclimatological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoclimatological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoclimatologist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoclimatologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoclimatologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoclimatologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoclimatologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoclimatologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoclimatologist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoclimatology ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoclimatology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoclimatology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoclimatologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoclimatologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoclimatology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoclimatology's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Paleoconcha ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Paleoconcha",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Palaeoconcha",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Paleoconcha's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Palaeoconcha's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleocosmic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleocosmic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeocosmic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleocosmology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleocosmology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeocosmology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleocrystal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleocrystal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeocrystal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleocrystallic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleocrystallic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeocrystallic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleocrystalline ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleocrystalline",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeocrystalline",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleocrystic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleocrystic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeocrystic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleocyclic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleocyclic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeocyclic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleodendrologic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleodendrologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeodendrologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleodendrological ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleodendrological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeodendrological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleodendrologically ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleodendrologically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeodendrologically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleodendrologist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleodendrologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeodendrologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleodendrology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleodendrology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeodendrology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoecologic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoecologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoecologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoecological ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoecological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoecological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoecologist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoecologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoecologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoecologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoecologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoecologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoecologist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoecology ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoecology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoecology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoecologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoecologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoecology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoecology's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoentomologic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoentomologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoentomologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoentomological ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoentomological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoentomological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoentomologist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoentomologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoentomologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoentomologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoentomologist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoeremology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoeremology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoeremology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoethnic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoethnic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoethnic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoethnography ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoethnography",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoethnography",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleofauna ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleofauna",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeofauna",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleogenesis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleogenesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeogenesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleogenesis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeogenesis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleogenetic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleogenetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeogenetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleogeographer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleogeographer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeogeographer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleogeographers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeogeographers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleogeographic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleogeographic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeogeographic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleogeographical ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleogeographical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeogeographical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleogeographically ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleogeographically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeogeographically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleogeography ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleogeography",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeogeography",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleogeographies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeogeographies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleogeology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleogeology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeogeology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoglaciology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoglaciology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoglaciology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoglyph ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoglyph",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoglyph",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleograph ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleograph",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeograph",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleographic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleographic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeographic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleographical ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleographical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeographical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleographically ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleographically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeographically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleographist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleographist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeographist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoherpetologist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoherpetologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoherpetologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoherpetology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoherpetology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoherpetology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleohistology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleohistology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeohistology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleohydrography ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleohydrography",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeohydrography",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoichthyology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoichthyology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoichthyology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleolatry ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleolatry",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeolatry",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleolimnology ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleolimnology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeolimnology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleolith ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleolith",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeolith",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleoliths",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoliths",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleolith's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeolith's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleolithic ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleolithic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeolithic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Paleolithic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Palaeolithic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleolithical ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleolithical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeolithical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleolithist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleolithist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeolithist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleolithoid ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleolithoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeolithoid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleolithy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleolithy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeolithy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleological ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleologist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeologist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleomagnetism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleomagnetism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeomagnetism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleomagnetisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeomagnetisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleomagnetism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeomagnetism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleometallic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleometallic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeometallic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleometeorological ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleometeorological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeometeorological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleometeorology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleometeorology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeometeorology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleontographic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleontographic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontographic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleontographical ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleontographical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontographical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleontography ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleontography",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontography",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleontographies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontographies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleontography's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontography's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleontol ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleontol",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontol",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleontologic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleontologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleontological ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleontological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleontologically ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleontologically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontologically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleontologist ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleontologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleontologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontologist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleontologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleontology ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleontology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleontology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeontology's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleopathologic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleopathologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeopathologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleopathological ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleopathological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeopathological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleopathologist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleopathologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeopathologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleopathologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeopathologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleopathology ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleopathology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeopathology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleopathologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeopathologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleopedology ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleopedology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeopedology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleophysiography ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleophysiography",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeophysiography",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleophysiology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleophysiology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeophysiology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleophytic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleophytic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeophytic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleophytological ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleophytological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeophytological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleophytologist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleophytologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeophytologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleophytology ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleophytology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeophytology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleoplain ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleoplain",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeoplain",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleopotamology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleopotamology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeopotamology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleopsychic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleopsychic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeopsychic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleopsychological ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleopsychological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeopsychological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleopsychology ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleopsychology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeopsychology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleornithological ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleornithological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeornithological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleornithology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleornithology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeornithology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleostriatal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleostriatal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeostriatal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleostriatum ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleostriatum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeostriatum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleostylic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleostylic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeostylic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleostyly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleostyly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeostyly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleotechnic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleotechnic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeotechnic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleothalamus ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleothalamus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeothalamus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Paleotropical ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Paleotropical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Palaeotropical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Paleotropical's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "Palaeotropical's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleovolcanic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleovolcanic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeovolcanic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleozoic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleozoic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeozoic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleozoologic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleozoologic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeozoologic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleozoological ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleozoological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeozoological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleozoologist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleozoologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeozoologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleozoologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeozoologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleozoologist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeozoologist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paleozoology ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paleozoology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeozoology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleozoologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeozoologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paleozoology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaeozoology's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palestra ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palestra",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaestra",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palestras",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaestras",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palestra's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaestra's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palestrae ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palestrae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaestrae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palestral ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palestral",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaestral",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palestrian ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palestrian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaestrian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palestric ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palestric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaestric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paletiology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paletiology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "palaetiology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palladiumize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palladiumize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palladiumise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palladiumizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palladiumises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palletization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palletization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palletisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palletizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palletisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palletization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palletisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palletize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palletize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palletise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palletized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palletised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palletizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palletising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palletizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palletises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "palletizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "palletizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palletiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palletizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palletisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "palletizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "palletiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pamperize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pamperize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pamperise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pamperizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pamperises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pamphletize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pamphletize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pamphletise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pamphletizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pamphletises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "panderize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "panderize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panderise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "panderizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panderises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pandora ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pandora",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pandoura",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "panegyrize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "panegyrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panegyrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "panegyrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panegyrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "panegyrizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panegyrising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "panegyrizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panegyrises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "panel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paneled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paneling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "panelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paneling's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panelling's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "panelist ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "panelist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panellist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "panelists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panellists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "panelist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panellist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "panelization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "panelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panellisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panellization",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "panelized ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "panelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "panesthesia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "panesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "panaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "panesthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "panaesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "panesthetic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "panesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "panaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pangea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pangea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pangaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pannier ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pannier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panier",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "panniers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paniers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pannier's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panier's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "panomphean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "panomphean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panomphaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pantheonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pantheonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pantheonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pantheonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pantheonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pantheonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pantheonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pantheonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pantheonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pantheonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pantheonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pantheonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pantie ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pantie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panty",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pantie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "panty's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "papalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "papalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "papalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "papalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "papalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "papalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "papalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "papalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "papalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "papalizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "papalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "papalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "papilledema ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "papilledema",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "papilloedema",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "papule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "papule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parabolization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parabolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parabolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parabolization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parabolisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parabolize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parabolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parabolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parabolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parabolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parabolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parabolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parabolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parabolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paraffinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paraffinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paraffinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paraffinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paraffinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paragraphize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paragraphize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paragraphise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paragraphizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paragraphises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parakeet ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parakeet",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parrakeet",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parakeets",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parrakeets",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parakeet's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parrakeet's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parallelization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parallelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parallelisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parallelizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parallelisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parallelization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parallelisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parallelize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parallelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parallelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parallelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parallelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parallelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parallelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parallelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parallelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parallelizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parallelizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralleliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parallelizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parallelisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paralogize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paralogize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralogise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralogized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralogised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralogizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralogising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralogizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralogises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paralyzation ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paralyzation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralysation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralyzations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralysations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralyzation's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralysation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paralyze ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paralyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralyses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: paralysis"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paralyzedly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paralyzedly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralysedly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralyzedlies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralysedlies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paralyzer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paralyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralyzers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralysers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralyzer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralyser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paralyzingly ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paralyzingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralysingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paralyzinglies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paralysinglies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Paramecium ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Paramecium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Paramoecium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Paramecium's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Paramoecium's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parameterizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parameterizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parameterisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parameterization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parameterization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parameterisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parameterizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parameterisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parameterization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parameterisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parameterize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parameterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parameterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parameterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parameterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parameterizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parameterising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parameterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parameterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parametrizable ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parametrizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parametrisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parametrization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parametrization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parametrisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parametrizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parametrisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parametrization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parametrisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parametrize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parametrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parametrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parametrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parametrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parametrizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parametrising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parametrizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parametrises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paraselene ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paraselene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paraselenae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parasitemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parasitemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "parasitaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parasitization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parasitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parasitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parasitize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parasitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parasitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parasitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parasitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parasitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parasitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parasitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parasitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parathyroidectomize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parathyroidectomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parathyroidectomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parathyroidectomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parathyroidectomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parathyroidectomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parathyroidectomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parathyroidectomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parathyroidectomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paratragedia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paratragedia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paratragoedia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parcel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parceled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parcelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parceling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parcelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parceling's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parcelling's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parceler ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parceler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parceller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parchmentize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parchmentize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parchmentise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parchmentized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parchmentised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parchmentizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parchmentising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parchmentizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parchmentises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parecious ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parecious",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paroecious",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pareciously ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pareciously",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paroeciously",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pareciousness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pareciousness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paroeciousness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parecism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parecism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paroecism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parecy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parecy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paroecy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parenesis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parenesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paraenesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pareneses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paraeneses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parenetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parenetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paraenetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parenetical ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parenetical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paraenetical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parenthesization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parenthesization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parenthesisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parenthesize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parenthesize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parenthesise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parenthesized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parenthesised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parenthesizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parenthesising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parenthesizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parenthesises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paresthesia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paresthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paraesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paresthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paraesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paresthesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paraesthesia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paresthetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paresthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paraesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parfocalization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parfocalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parfocalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parfocalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parfocalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parfocalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Parisianization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Parisianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Parisianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Parisianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Parisianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Parisianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Parisianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Parisianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Parisianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Parisianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Parisianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Parisianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Parisianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Parisianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parlor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parlor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "parlour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parlors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "parlours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parlor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "parlour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parlorish ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parlorish",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parlourish",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parlormaid ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parlormaid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "parlourmaid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parmigiana ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parmigiana",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parmigiano",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parochialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parochialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parochialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parochializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parochialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parochialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parochialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parochialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parochialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parochialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parochialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parochialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parochializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parochialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parochializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parochialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parrotize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parrotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parrotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parrotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parrotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "parsonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "parsonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parsonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "parsonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "parsonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "partialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "partialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "partialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "partialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "partialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "partializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "partialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "partializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "partialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "particolored ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "particolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "particoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "particoloredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "particolouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "particoloreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "particoloureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "particularization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "particularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "particularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "particularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "particularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "particularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "particularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "particularize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "particularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "particularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "particularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "particularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "particularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "particularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "particularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "particularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "particularizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "particularizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "particulariser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "particularizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "particulariser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "partisan ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "partisan",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "partizan",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "partisans",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "partizans",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "partisan's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "partizan's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "partisan",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "partizan",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "partisans",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "partizans",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "partisan's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "partizan's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "partisanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "partisanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "partisanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "partisanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "partisanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pasiphae ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pasiphae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pasiphaae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pasiphae's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pasiphaae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "passades ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "passades",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "passadoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "passivization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "passivization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "passivisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "passivize ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "passivize",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "passivise",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "passivized",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "passivised",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "passivizing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "passivising",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "passivizes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "passivises",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pasteurization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pasteurization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pasteurisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pasteurizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pasteurisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pasteurization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pasteurisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pasteurize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pasteurize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pasteurise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pasteurized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pasteurised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pasteurizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pasteurising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pasteurizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pasteurises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pasteurizers ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pasteurizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pasteurisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pasteurizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pasteuriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pasteurizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pasteurisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pasteurizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pasteuriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pastoralize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pastoralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pastoralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pastoralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pastoralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pastorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pastorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pastorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pastorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pastorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "paternalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "paternalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paternalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paternalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paternalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "patine ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "patine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patined",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patinaed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "patrialization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "patrialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patrialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patrializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patrialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "patrialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "patrialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patrialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patrialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patrialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patrializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patrialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patrializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patrialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "patrices ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "patrices",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patricoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "patronizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "patronizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "patronization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "patronization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "patronize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "patronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "patronizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "patronizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patroniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patroniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "patronizingly ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "patronizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizinglier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisinglier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patronizingliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patronisingliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "patte ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "patte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pattae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "patternize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "patternize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patternise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "patternizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "patternises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Paulinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Paulinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Paulinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Paulinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Paulinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Paulinize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Paulinise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pauperization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pauperization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pauperisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pauperizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pauperisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pauperization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pauperisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pauperize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pauperize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pauperise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pauperized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pauperised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pauperizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pauperising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pauperizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pauperises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pauperizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pauperizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pauperiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pauperizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pauperisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pauperizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pauperiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pavior ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pavior",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Paviour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pavior's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Paviour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pavior",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paviour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "paviors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paviours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pavior's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paviour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pawpaw ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pawpaw",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papaw",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pawpaws",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papaws",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pawpaw's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "papaw's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Paynize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Paynize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Paynise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Paynizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Paynises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Paynize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Paynise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pe ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pearlization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pearlization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pearlisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pearlize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pearlize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pearlise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pearlized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pearlised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pearlizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pearlising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pearlizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pearlises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peasantize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peasantize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peasantise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peasantizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peasantises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pectization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pectization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pectisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pectizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pectisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pectization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pectisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pectize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pectize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pectise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pectized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pectised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pectizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pectising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pectizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pectises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peculiarize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peculiarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peculiarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peculiarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peculiarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peculiarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peculiarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peculiarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peculiarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedagogism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedagogism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedagogism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedagogue ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedagogue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedagog",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedagogues",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedagogs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedagogue's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedagog's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedal ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pedalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedaler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pedaller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedalers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pedallers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedaler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pedaller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pedalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedanticize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedanticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedanticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedanticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedanticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedanticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedanticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedanticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedanticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedantize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedantize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedantise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedantized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedantised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedantizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedantising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedantizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedantises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedatrophia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedatrophia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedatrophia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedatrophy ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedatrophy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedatrophy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peddler ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peddler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedlar",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peddlers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedlars",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peddler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedlar's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pederast ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pederast",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paederast",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pederast's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paederast's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pederasts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paederasts",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pederastic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pederastic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paederastic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pederastically ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pederastically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paederastically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedestal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedestaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedestalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedestaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedestalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedestrianization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedestrianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedestrianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedestrianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedestrianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedestrianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedestrianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedestrianize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedestrianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedestrianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedestrianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedestrianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedestrianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedestrianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedestrianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pedestrianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pediatric ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pediatric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paediatric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pediatrics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paediatrics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pediatrician ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pediatrician",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pediatrist",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "paediatrician",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pediatricians",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pediatrists",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "paediatricians",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pediatrician's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "pediatrist's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "paediatrician's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pediatry ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pediatry",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paediatry",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedobaptism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedobaptism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedobaptism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedobaptism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedobaptism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedobaptist ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedobaptist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedobaptist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedobaptist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedobaptist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedodontic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedodontic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedodontic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedodontics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedodontics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedogenesis ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedogenesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedogenesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedogeneses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedogeneses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedogenetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedogenetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedogenetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedogenic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedogenic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedogenic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedological ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedologist ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedologist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedologist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedologists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedologists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedology ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paedology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paedologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedology's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paedology's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedometrical ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedometrical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedometrical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedomorphic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedomorphic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedomorphic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedomorphism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedomorphism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedomorphism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedomorphisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedomorphisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedomorphosis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedomorphosis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedomorphosis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedomorphoses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedomorphoses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedophile ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedophile",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paedophile",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedophiles",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paedophiles",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedophilia ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedophilia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paedophilia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedophilias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paedophilias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedophiliac ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedophiliac",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paedophiliac",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pedophiliacs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paedophiliacs",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedophilic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedophilic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "paedophilic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedotribe ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedotribe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedotribe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedotrophic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedotrophic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedotrophic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedotrophist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedotrophist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedotrophist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pedotrophy ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pedotrophy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "paedotrophy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pekingese ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pekingese",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pekinese",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pekingese's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pekinese's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pelletization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pelletization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pelletisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pelletizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pelletisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pelletization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pelletisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pelletize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pelletize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pelletise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pelletized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pelletised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pelletizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pelletising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pelletizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pelletises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pelletizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pelletizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pelletiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pelorized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pelorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pelorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pemmicanization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pemmicanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pemmicanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pemmicanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pemmicanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pemmicanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pemmicanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pemmicanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pemmicanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pemmicanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pemmicanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pemmicanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pemmicanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pemmicanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "penalizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "penalizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penalisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penalizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penalisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penalizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penalisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penalizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penalisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penalizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penalisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "penalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "penalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "penalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "penalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pencil ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "penciled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pencilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penciling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pencilling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pencilings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pencillings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "penciler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "penciler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penciller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pencilers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pencillers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penciler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penciller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "penelopize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "penelopize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penelopise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penelopized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penelopised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penelopizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penelopising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penelopizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penelopises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "penlight ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "penlight",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penlite",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penlights",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penlites",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "penlight's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "penlite's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "penne ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "penne",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pennae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pense ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pense",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pensae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peptizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peptizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peptization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peptization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peptize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peptize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peptizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peptizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peptonemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peptonemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptonaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peptonization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peptonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peptonize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peptonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peptonizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peptonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptoniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peptonizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peptoniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peracid ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peracid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preacid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peracidity ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peracidity",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preacidity",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peract ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peract",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preact",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peracute ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peracute",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preacute",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perambulatory ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perambulatory",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preambulatory",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perces ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perces",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preces",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perchlorethane ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perchlorethane",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perchloroethane",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perchlorethylene ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perchlorethylene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perchloroethylene",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perchloric ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perchloric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prechloric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perclose ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perclose",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preclose",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "percompound ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "percompound",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precompound",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "percurrent ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "percurrent",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precurrent",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "percussionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "percussionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "percussionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "percussionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "percussionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perdicine ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perdicine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perdicinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perdiligent ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perdiligent",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prediligent",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perdy ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perdy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "predy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perennialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perennialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perennialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "perennializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perennialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pereon ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pereon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peraeon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "perea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peraea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pereopod ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pereopod",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peraeopod",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pereopods",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peraeopods",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perfectivize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perfectivize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perfectivise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "perfectivizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perfectivises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perfervor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perfervor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perfervour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "perfervors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perfervours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "perfervor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perfervour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perfet ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perfet",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefet",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perfoliation ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perfoliation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefoliation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "perfoliations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefoliations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prefoliation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praefoliation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "performant ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "performant",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preformant",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "performative ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "performative",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preformative",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perhazard ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perhazard",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prehazard",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pericecal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pericecal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pericaecal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pericecitis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pericecitis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pericaecitis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pericenter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pericenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pericentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perichete ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perichete",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perichaete",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pericope ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pericope",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pericopae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "periesophageal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "periesophageal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perioesophageal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peril ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "periled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "periling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "perilling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "periodicalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "periodicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "periodicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "periodicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "periodicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "periodization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "periodization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "periodisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "periodizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "periodisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "periodization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "periodisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "periodize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "periodize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "periodise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "periodized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "periodised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "periodizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "periodising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "periodizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "periodises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perjink ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perjink",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prejink",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perlection ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perlection",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelection",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prelection",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praelection",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prelection's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praelection's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perlingual ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perlingual",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelingual",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perlingually ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perlingually",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelingually",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "permissable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "permissable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "premissable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "permixture ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "permixture",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "premixture",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pernasal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pernasal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prenasal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peroral ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peroral",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preoral",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perorally ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perorally",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preorally",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perotic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perotic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preotic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peroxidize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peroxidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peroxidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peroxidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peroxidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peroxidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peroxidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peroxidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peroxidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "peroxidizement ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "peroxidizement",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peroxidisement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peroxidizements",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peroxidisements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "peroxidizement's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "peroxidisement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perperfect ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perperfect",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preperfect",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Perse ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Perse",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Perse's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "perse",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prese",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Persianization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Persianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Persianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Persianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persianized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persianised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persianizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persianising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Persicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Persicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persicized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persicised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persicize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persicise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Persicizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Persicising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "persolve ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "persolve",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "presolve",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "personalization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "personalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "personalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "personalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "personalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "personalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "personalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "personalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "personalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "personization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "personization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "personizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "personization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "personize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "personize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "personized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "personizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "personizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "personises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perst ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perst",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Peruvianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Peruvianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Peruvianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Peruvianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Peruvianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Peruvianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Peruvianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perv ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perv",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prev",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perve ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perve",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preve",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "perved",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preved",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "perving",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preving",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "perves",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preves",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perviable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perviable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "previable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "perviousness ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "perviousness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "previousness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "perviousnesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "previousnesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "petaled ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "petaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "petalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "petling ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "petling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poetling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Petrarchize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Petrarchize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Petrarchise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Petrarchized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Petrarchised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Petrarchized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Petrarchised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Petrarchizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Petrarchising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Petrarchizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Petrarchises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Petrarchize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Petrarchise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Petrarchizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Petrarchising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "petrolization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "petrolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "petrolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "petrolizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "petrolisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "petrolization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "petrolisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "petrolize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "petrolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "petrolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "petrolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "petrolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phacochere ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phacochere",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phacochoere",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phacocherine ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phacocherine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phacochoerine",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Phaenna ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Phaenna",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Phaaenna",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Phaenna's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Phaaenna's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phagedena ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phagedena",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phagedaena",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phagedenas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phagedaenas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phagedenic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phagedenic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phagedaenic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phagedenical ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phagedenical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phagedaenical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phagedenous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phagedenous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phagedaenous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phagocytize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phagocytize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phagocytise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phagocytizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phagocytises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phalanges ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phalanges",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: phalange"),
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phalanxes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phalanges",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: phalanx"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phantomize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phantomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phantomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phantomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phantomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phantomizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phantomizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phantomiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phantomizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phantomisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pharisean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pharisean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pharisaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pharisean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pharisaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pharmacopeian ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pharmacopeian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pharmacopoeian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pharmacopoeia ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pharmacopoeia",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pharmacopeia",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pharmacopoeias",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pharmacopeias",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pharmacopoeia's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pharmacopeia's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pharmacopoeial ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pharmacopoeial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pharmacopeial",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phelonion ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phelonion",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaelonion",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phelonions",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaelonions",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenicia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenicia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phoenicia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenicopter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenicopter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phoenicopter",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenogam ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenogam",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaenogam",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenogams",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaenogams",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenogenesis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenogenesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaenogenesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenogenetic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenogenetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaenogenetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenolization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenolizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenolisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenolization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenolisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenolize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenological ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaenological",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenology ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaenology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaenologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenomenalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenomenalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenomenalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenomenalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenomenalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenomenalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenomenalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenomenalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenomenalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenomenalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenomenalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenomenalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenomenalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenomenalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenomenalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenomenalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenomenism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenomenism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaenomenism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenomenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenomenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenomenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenomenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenomenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenomenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenomenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phenomenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phenomenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phenozygous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phenozygous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaenozygous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pheophyl ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pheophyl",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaeophyl",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pheophyll ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pheophyll",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaeophyll",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pheophytin ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pheophytin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phaeophytin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phiale ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phiale",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phialae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "philanthropize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "philanthropize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philanthropise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philanthropized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philanthropised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philanthropizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philanthropising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philanthropizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philanthropises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Philippize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Philippize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philippise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Philippized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philippised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Philippized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philippised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Philippizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philippising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Philippizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philippises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Philippize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philippise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Philippizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philippising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Philistinize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Philistinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philistinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Philistinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philistinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Philistinized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philistinised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Philistinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philistinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Philistinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philistinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Philistinize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philistinise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Philistinizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Philistinising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "philopena ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "philopena",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philopoena",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philopenas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philopoenas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "philosophization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "philosophization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philosophisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philosophizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philosophisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philosophization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philosophisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "philosophize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "philosophize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philosophise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philosophized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philosophised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philosophizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philosophising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philosophizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philosophises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "philosophizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "philosophizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philosophiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philosophizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philosophisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philosophizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philosophiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "philter ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "philter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philtre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philtre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "philters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "philtres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phlebotomization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phlebotomization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phlebotomisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phlebotomizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phlebotomisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phlebotomization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phlebotomisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phlebotomize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phlebotomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phlebotomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phlebotomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phlebotomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phlebotomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phlebotomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phlebotomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phlebotomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phleum ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phleum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phloeum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phlyctena ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phlyctena",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phlyctaena",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phlyctenae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phlyctaenae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phlyctena's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phlyctaena's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phlyctenula ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phlyctenula",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phlyctaenula",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Phocean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Phocean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Phocaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phonemicization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phonemicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phonemicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phonemicizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phonemicisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phonemicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phonemicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phonemicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phonemicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phonemicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phonemicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phonemicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phonemicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phonemicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phoneticization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phoneticization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phoneticisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phoneticizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phoneticisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phoneticization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phoneticisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phoneticize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phoneticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phoneticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phoneticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phoneticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phoneticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phoneticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phoneticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phoneticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phonetization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phonetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phonetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phonetizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phonetisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phonetize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phonetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phonetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phonetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phonetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phonetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phonetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phonetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phonetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phony ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phony",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "phoney",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phonies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "phoneys",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phony's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "phoney's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phonier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phoniest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phonied",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phoneyed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("<V>"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phonying",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phoneying",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("<V>"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phosphatization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phosphatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phosphatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phosphatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phosphatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phosphatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phosphatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phosphatize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phosphatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phosphatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phosphatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phosphatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phosphatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phosphatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phosphatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phosphatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phosphorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phosphorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phosphorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phosphorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phosphorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phosphorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phosphorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phosphorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phosphorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photesthesis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photesthesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photaesthesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photesthesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photoesthesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photocatalyzer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photocatalyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photocatalyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photocatalyzers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photocatalysers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photocatalyzer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photocatalyser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photoesthetic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photoesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photoaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photographize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photographize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photographise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photographizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photographises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photoionization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photoionization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photoionisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photoionizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photoionisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photoionization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photoionisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photoisomerization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photoisomerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photoisomerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photoisomerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photoisomerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photoisomerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photoisomerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photolabeled ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photolabeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photolabelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photolabeler ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photolabeler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photolabeller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photolabeling ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photolabeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photolabelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photolyzable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photolyzable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photolysable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photolyze ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photolyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photolyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photolyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photolysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photolyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photolysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photolyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photolyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photopolymerization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photopolymerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photopolymerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photopolymerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photopolymerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photopolymerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photopolymerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photosensitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photosensitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosensitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photosensitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosensitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photosensitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosensitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photosensitize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photosensitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosensitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photosensitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosensitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photosensitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosensitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photosensitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosensitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photosensitizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photosensitizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosensitiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photosensitizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosensitisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photosensitizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosensitiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "photosynthesize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "photosynthesize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosynthesise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photosynthesized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosynthesised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photosynthesizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosynthesising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "photosynthesizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "photosynthesises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phrenologize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phrenologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phrenologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phrenologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phrenologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phrenologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phrenologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "phrenologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phrenologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Phylactolema ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Phylactolema",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Phylactolaema",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Phylactolemata",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Phylactolaemata",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Phylactolema's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Phylactolaema's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Phylactolemata's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Phylactolaemata's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phytoecology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phytoecology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phytooecology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "phytohemagglutinin ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "phytohemagglutinin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "phytohaemagglutinin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "piarhemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "piarhemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "piarhaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "piaster ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "piaster",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "piastre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "piasters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "piastres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "piaster's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "piastre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "picogram ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "picogram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "picogramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "picograms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "picogrammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "picogram's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "picogramme's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "picometer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "picometer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "picometre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "picometers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "picometres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "picometer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "picometre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pictorialization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pictorialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pictorialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pictorializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pictorialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pictorialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pictorialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pictorialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pictorialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pictorialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pictorialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pictorialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pictorializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pictorialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pictorializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pictorialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "picturization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "picturization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "picturisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "picturizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "picturisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "picturization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "picturisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "picturize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "picturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "picturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "picturized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "picturised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "picturizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "picturising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "picturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "picturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pidginization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pidginization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pidginisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pidginizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pidginisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pidginization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pidginisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pidginize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pidginize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pidginise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pidginized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pidginised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pidginizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pidginising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pidginizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pidginises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "piggyback ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "piggyback",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pickaback",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "piggybacking",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pickabacking",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "piggybacks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pickabacks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "piggyback's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pickaback's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pigmean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pigmean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pigmaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pigmentize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pigmentize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pigmentise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pigmentizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pigmentises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pilaf ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pilaf",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pilaw",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "pilau",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "pilaff",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pilafs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pilaws",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "pilaus",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "pilaffs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pilaf's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pilaw's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "pilau's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "pilaff's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pilgrimize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pilgrimize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pilgrimise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pilgrimized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pilgrimised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pilgrimizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pilgrimising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pilgrimizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pilgrimises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pillarize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pillarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pillarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pillarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pillarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pillorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pillorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pillorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pillorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pillorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pillorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pillorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pillorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pillorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pillowbeer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pillowbeer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pillowbere",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pindarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pindarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pindarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pindarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pindarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pindarized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pindarised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pindarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pindarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pindarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pindarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pindarize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pindarise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pindarizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pindarising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "piney ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "piney",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "piny",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pinkie ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pinkie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pinkies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pinkie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pinkie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pinky",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pinkies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pinkie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pinky's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pinnule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pinnule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pinnulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pinochle ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pinochle",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pinocle",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pinochle's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pinocle's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pinyon ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pinyon",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pi�on",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pinyons",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pi�ons",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pinyon's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pi�on's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "piratize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "piratize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "piratise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "piratizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "piratises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "piroshki ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "piroshki",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pirozhki",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "piroshki's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pirozhki's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "piscine ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "piscine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "piscinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pistol ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pistoled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pistolled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pistoling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pistolling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pixie ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pixie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pixy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pixie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pixy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pixieish ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pixieish",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pixyish",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pizes ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pizzazz ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pizzazz",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pizazz",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pizzazzes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "pizazzes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pizzazz's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pizazz's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Placean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Placean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Placaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Placean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Placaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "placebos ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "placebos",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "placeboes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plage ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plage",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plagae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plagiarization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plagiarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plagiarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plagiarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plagiarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plagiarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plagiarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plagiarize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plagiarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plagiarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plagiarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plagiarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plagiarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plagiarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plagiarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plagiarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plagiarizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plagiarizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plagiariser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plagiarizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plagiarisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plagiarizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plagiariser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Plante ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Plante",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Plantae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Plante's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Plantae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plantule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plantule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plantulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plasmapheresis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plasmapheresis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasmaphaeresis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plasmolyze ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plasmolyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasmolyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plasmolyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasmolysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plasmolyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasmolysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plasmolyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasmolyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plasticization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plasticization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasticisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plasticizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasticisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plasticization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasticisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plasticize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plasticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plasticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plasticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plasticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plasticizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plasticizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasticiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plasticizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasticisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plasticizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plasticiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "platea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "platea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plataea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "platinization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "platinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "platinize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "platinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "platitudinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "platitudinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platitudinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platitudinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platitudinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platitudinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platitudinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "platitudinize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "platitudinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platitudinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platitudinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platitudinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platitudinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platitudinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platitudinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platitudinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "platitudinizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "platitudinizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platitudiniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platitudinizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platitudiniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Platonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Platonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Platonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Platonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Platonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Platonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Platonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Platonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Platonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Platonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Platonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Platonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Platonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Platonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "platonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Platonizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Platonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Platoniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Platonizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Platoniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "platycelian ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "platycelian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platycoelian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "platycelous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "platycelous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "platycoelous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plebeianization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plebeianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plebeianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plebeianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plebeianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plebeianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plebeianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plebeianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plebeianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plebeianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plebeianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plebeianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plebeianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plebeianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pleuroperitoneal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pleuroperitoneal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pleuroperitonaeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plow ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plow",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plough",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plowed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ploughed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plowing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ploughing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plow's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plough's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plowing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ploughing's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plows",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ploughs",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plowable ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plowable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ploughable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plowboy ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plowboy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ploughboy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plowboys",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ploughboys",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plowboy's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ploughboy's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plower ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plower",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plougher",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plowers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ploughers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plower's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plougher's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plowhead ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plowhead",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ploughhead",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plowman ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plowman",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ploughman",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plowmen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ploughmen",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plowman's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ploughman's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plowshare ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plowshare",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ploughshare",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plowshares",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ploughshares",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "plowshare's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ploughshare's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "plumule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "plumule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "plumulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pluralization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pluralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pluralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pluralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pluralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pluralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pluralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pluralize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pluralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pluralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pluralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pluralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pluralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pluralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pluralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pluralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pluralizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pluralizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pluraliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pluralizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pluralisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pluralizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pluraliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "poeticization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "poeticization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poeticisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "poeticize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "poeticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poeticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poeticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poeticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poeticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poeticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poeticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poeticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "poetization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "poetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poetizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poetisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poetization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poetisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "poetize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "poetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "poetizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "poetizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poetiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poetizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poetisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poetizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poetiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pogromize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pogromize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pogromise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pogromizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pogromises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pokey ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pokey",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pokeys",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pokey's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poky",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pokey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pokier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pokiest",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polarizability ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polarizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polarizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polarizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polarizable ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polarizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polarizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polarizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polarization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polarize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polarizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polarizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polariser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polarizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polarisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polarizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polariser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polemicize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polemicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polemicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polemicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polemicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polemicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polemicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polemicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polemicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polemize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polemize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polemise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polemized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polemised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polemizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polemising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polemizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polemises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "policize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "policize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "policise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "policizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "policises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "policizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "policizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "policiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "policizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "policisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "poliencephalitis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "poliencephalitis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polioencephalitis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poliencephalitis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polioencephalitis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "poliencephalomyelitis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "poliencephalomyelitis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polioencephalomyelitis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "politicalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "politicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "politicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "politicization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "politicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "politicizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politicisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "politicization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politicisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "politicize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "politicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "politicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "politicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "politicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "politicizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "politicizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politiciser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "politicizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politicisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "politize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "politize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "politizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "politises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pollack ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pollack",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pollock",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pollack's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pollock's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pollenizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pollenizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polleniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pollinize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pollinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pollinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pollinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pollinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pollinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pollinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pollinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pollinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pollinizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pollinizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polliniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polliwog ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polliwog",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pollywog",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polliwogs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pollywogs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polliwog's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pollywog's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pollock ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pollock",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("name"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Polonization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Polonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Polonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Polonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Polonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polychetous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polychetous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polychaetous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polychromatize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polychromatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polychromatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polychromatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polychromatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polychromize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polychromize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polychromise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polychromizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polychromises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polycythemia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polycythemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "polycythaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polycythemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "polycythaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polycythemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polycythemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "polycythaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polydemonism ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polydemonism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polydaemonism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polydemonisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polydaemonisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polydemonist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polydemonist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polydaemonist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polydemonist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polydaemonist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polyemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polyemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polyaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polyemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polyemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polyaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polyestrous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polyestrous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polyoestrous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polygamize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polygamize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polygamise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polygamizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polygamises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polyhemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polyhemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polyhaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polyhemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polyhemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polyhaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polymerization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polymerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polymerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polymerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polymerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polymerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polymerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polymerize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polymerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polymerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polymerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polymerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polymerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polymerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polymerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polymerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polypnea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polypnea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polypnoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polypneic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polypneic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polypnoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polysulphurization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polysulphurization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polysulphurisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polysulphurizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polysulphurisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polysulphurization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polysulphurisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "polytetrafluorethylene ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "polytetrafluorethylene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polytetrafluoroethylene",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "polytetrafluorethylenes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "polytetrafluoroethylenes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pomerium ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pomerium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pomoerium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pomme ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pomme",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pommae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pommel ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pommeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pommelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pommeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pommelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "popliteal ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "popliteal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poplitaeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "popularization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "popularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "popularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "popularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "popularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "popularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "popularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "popularize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "popularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "popularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "popularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "popularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "popularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "popularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "popularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "popularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "popularizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "popularizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "populariser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "popularizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "popularisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "popularizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "populariser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "porcelainization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "porcelainization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porcelainisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "porcelainizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porcelainisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "porcelainization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porcelainisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "porcelainize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "porcelainize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porcelainise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "porcelainized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porcelainised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "porcelainizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porcelainising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "porcelainizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porcelainises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "porcellanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "porcellanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porcellanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "porcellanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porcellanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "porcellanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porcellanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "porcellanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porcellanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "porer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "porer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "porers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "poringly ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "poringly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pouringly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "porphyrization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "porphyrization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porphyrisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "porphyrization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "porphyrisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "porte ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "porte",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "portae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "portionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "portionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "portionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "portionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "portionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "positivize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "positivize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "positivise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "positivizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "positivises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "postcecal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "postcecal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "postcaecal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "postcenal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "postcenal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "postcoenal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "posterize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "posterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "posterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "posterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "posterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "postfetal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "postfetal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "postfoetal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "postilion ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "postilion",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "postillion",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "postilions",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "postillions",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "postilion's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "postillion's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "postmedieval ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "postmedieval",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "postmediaeval",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "postsynchronization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "postsynchronization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "postsynchronisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "posturize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "posturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "posturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "posturized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "posturised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "posturizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "posturising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "posturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "posturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "potentialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "potentialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "potentialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "potentializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "potentialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "potentialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "potentialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "potentialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "potentialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "potentialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "potentializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "potentialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "potentize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "potentize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "potentise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "potentized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "potentised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "potentizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "potentising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "potentizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "potentises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pouder ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pouder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poudre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pouders",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poudres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pouf ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pouffe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pouf",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pouffes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "poufs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pouffe's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pouf's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pouf",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poufs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pouf's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "poulder ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "poulder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pouldre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "poulders",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pouldres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "powderization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "powderization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "powderisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "powderizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "powderisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "powderization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "powderisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "powderize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "powderize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "powderise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "powderizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "powderises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "powderizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "powderizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "powderiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "powderizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "powderisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "powellize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "powellize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "powellise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "powellized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "powellised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "powellizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "powellising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "powellizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "powellises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "practice ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "practice",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "practise",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "practices",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "practises",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "practice's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "practise's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "practice",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "practise",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "practiced",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "practised",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "practicing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "practising",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "practices",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "practises",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "practicer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "practicer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "practiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "practicers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "practisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "practicer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "practiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pragmatization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pragmatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pragmatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pragmatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pragmatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pragmatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pragmatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pragmatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pragmatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pragmatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pragmatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pragmatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pragmatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pragmatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pragmatizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pragmatizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pragmatiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pragmatizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pragmatisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preabdomen ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preabdomen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praeabdomen",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preacetabular ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preacetabular",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praeacetabular",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preacherize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preacherize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preacherise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preacherizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preacherises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preacknowledgment ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preacknowledgment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preacknowledgement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preacknowledgments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preacknowledgements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preacknowledgment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preacknowledgement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preanal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preanal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praeanal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preanesthetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preanesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preanaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preanesthetics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preanaesthetics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preanesthetic's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preanaesthetic's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preapprize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preapprize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preapprise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preapprized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preapprized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preapprised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preapprizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preapprizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preapprising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preauthorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preauthorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preauthorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prebaptize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prebaptize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prebaptise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prebaptizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prebaptises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precancelation ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precancelation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precancellation",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precava ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precava",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praecava",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precavae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praecavae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precipe ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precipe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praecipe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precipes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praecipes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precisionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precisionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precisionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precisionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precisionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precivilization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precivilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precivilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precivilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precivilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precivilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precivilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precocial ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precocial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praecocial",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precolor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "precolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "precoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precoloring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "precolouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "precolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precolorings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "precolourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "precolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precolorable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precolorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "precolourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precolorabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "precolourabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precolorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "precolourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precolorablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "precolourablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precolorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "precolourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precoloration ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precoloration",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precolouration",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precolorations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precolourations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precoloration's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precolouration's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preconization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preconization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preconisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preconizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preconisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preconization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preconisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preconize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preconize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preconise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preconized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preconised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preconizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preconising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preconizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preconises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preconizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preconizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preconiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preconizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preconisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preconizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preconiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precoracoid ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precoracoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praecoracoid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precordial ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precordial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praecordial",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precordium ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precordium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praecordium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precordia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praecordia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precornu ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precornu",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praecornu",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precox ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precox",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praecox",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precriticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precriticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precriticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "precriticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "precriticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "precuneus ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "precuneus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praecuneus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "predial ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "predial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praedial",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "predials",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praedials",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "predialist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "predialist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praedialist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prediality ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prediality",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praediality",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preemptor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preemptor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preaemptor",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preemptor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preaemptor's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preesophageal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preesophageal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praeesophageal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preesophageal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preoesophageal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preestival ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preestival",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preaestival",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preexistent ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preexistent",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preaexistent",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prefavor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prefavor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefavour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prefavors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefavours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prefavor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefavour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prefavorable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prefavorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefavourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prefavorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefavourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prefavorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefavourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prefavorably ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prefavorably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefavourably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prefavorablies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefavourablies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prefavorite ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prefavorite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefavourite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prefavorites",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefavourites",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prefavorite's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefavourite's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prefectorial ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prefectorial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praefectorial",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prefertilization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prefertilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefertilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prefertilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefertilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prefertilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefertilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prefertilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prefertilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefertilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prefertilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prefertilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preflavor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preflavor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preflavour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preflavors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preflavours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preflavor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preflavour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preflavoring ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preflavoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preflavouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preflavorings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preflavourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prefloration ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prefloration",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praefloration",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pregalvanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pregalvanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pregalvanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pregalvanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pregalvanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prehallux ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prehallux",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praehallux",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prehumor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prehumor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prehumour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prehumors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prehumours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prehumor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prehumour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preinitialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preinitialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preinitialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preinitialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preinitialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preinitializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preinitialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preinitializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preinitialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prejudgment ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prejudgment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "prejudgement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prejudgments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "prejudgements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prejudgment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "prejudgement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prelabor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prelabor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelabour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prelabors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelabours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prelabor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelabour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prelabrum ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prelabrum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praelabrum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prelatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prelatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prelatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prelatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prelatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prelect ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prelect",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praelect",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prelected",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praelected",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prelecting",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praelecting",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prelects",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praelects",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prelector ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prelector",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praelector",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prelectorship ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prelectorship",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praelectorship",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prelectress ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prelectress",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praelectress",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prelocalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prelocalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelocalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prelocalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelocalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prelocalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prelocalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preludium ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preludium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "praeludium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preludize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preludize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preludise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preludizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preludises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "premaxilla ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "premaxilla",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praemaxilla",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "premial ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "premial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proemial",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "premillennialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "premillennialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "premillennialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "premillennialized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "premillennialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "premillennialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "premillennializing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "premillennializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "premillennialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "premise ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "premise",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "premiss",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "premises",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "premisses",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "premise's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "premiss's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "premise",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "premonopolize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "premonopolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "premonopolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "premonopolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "premonopolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "premunire ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "premunire",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praemunire",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prenarial ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prenarial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praenarial",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preneural ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preneural",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praeneural",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prenomen ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "praenomen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prenomen",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "praenomens",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prenomens",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "praenomen's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prenomen's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prenomina ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prenomina",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praenomina",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prenominal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prenominal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praenominal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preoffense ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preoffense",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preoffence",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preoffenses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preoffences",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preoffense's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preoffence's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preoperculum ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preoperculum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praeoperculum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preorganization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preorganization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preorganisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preorganizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preorganisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preorganization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preorganisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preorganize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preorganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preorganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preorganizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preorganises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preoxidize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preoxidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preoxidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preoxidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preoxidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preoxidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preoxidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preoxidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preoxidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prepaleolithic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prepaleolithic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prepalaeolithic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prepositor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prepositor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praepositor",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prepositure ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prepositure",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praepositure",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preposter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preposter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praeposter",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prepostor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prepostor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praepostor",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prepostors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praepostors",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prepostor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praepostor's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preppy ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preppy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preppie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preppy's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preppie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preprogram ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preprogram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preprogramme",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prepubis ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prepubis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praepubis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prerealization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prerealization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prerealisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prerealizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prerealisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prerealization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prerealisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prerealize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prerealize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prerealise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prerealizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prerealises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prerecognize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prerecognize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prerecognise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prerecognizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prerecognises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Presbyterianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Presbyterianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Presbyterianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Presbyterianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Presbyterianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Presbyterianized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Presbyterianised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Presbyterianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Presbyterianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Presbyterianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Presbyterianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Presbyterianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Presbyterianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Presbyterianizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Presbyterianising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prescutum ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prescutum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praescutum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "presidium ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "presidium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praesidium",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "presidiums",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "presidia",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "praesidiums",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: Some("M-w has presidiums as a variant, but Ox doesn't say anything about the variant form"),
  },
  Entry {
    variants: &[
      Variant {
        word: "presidium's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praesidium's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prespecialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prespecialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prespecialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prespecializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prespecialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "presphenoid ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "presphenoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praesphenoid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "presplendor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "presplendor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "presplendour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "presplendors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "presplendours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "presplendor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "presplendour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pressurization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pressurization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pressurisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pressurizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pressurisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pressurization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pressurisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pressurize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pressurize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pressurise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pressurized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pressurised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pressurizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pressurising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pressurizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pressurises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pressurizer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pressurizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pressuriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pressurizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pressurisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prestandardization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prestandardization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prestandardisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prestandardizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prestandardisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prestandardization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prestandardisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prestandardize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prestandardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prestandardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prestandardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prestandardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "presternal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "presternal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praesternal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "presternum ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "presternum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praesternum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prestomium ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prestomium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praestomium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "presympathize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "presympathize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "presympathise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "presympathizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "presympathises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "presystolic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "presystolic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praesystolic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pretaxation ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pretaxation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praetaxation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pretense ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pretense",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pretence",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pretenses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pretences",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pretense's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pretence's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pretensive ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pretensive",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pretencive",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preterit ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preterit",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "preterite",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preterits",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "preterites",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preterit's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "preterite's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pretexta ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pretexta",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praetexta",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pretextae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praetextae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pretexta's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praetexta's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pretorial ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pretorial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praetorial",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pretorians ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pretorians",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Praetorians",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pretorium ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pretorium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praetorium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pretorius ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pretorius",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Praetorius",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pretorius's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Praetorius's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pretorship ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pretorship",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praetorship",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preutilizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preutilizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preutilisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preutilizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preutilisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preutilizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preutilisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preutilization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preutilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preutilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preutilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preutilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preutilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preutilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preutilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preutilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preutilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preutilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preutilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preventable ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preventable",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "preventible",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "preview ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "preview",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prevue",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("movie preview"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "previews",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prevues",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("movie preview"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preview's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prevue's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("movie preview"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preview",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "previews",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "preview's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "previze ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "previze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "previse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prezygapophysis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prezygapophysis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "praezygapophysis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pricey ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pricey",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pricy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prier ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pryer",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "priers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pryers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prier's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pryer's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "primeval ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "primeval",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "primaeval",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "primevally ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "primevally",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "primaevally",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prioritization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prioritization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prioritisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prioritizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prioritisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prioritization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prioritisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prioritize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prioritize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prioritise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prioritized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prioritised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prioritizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prioritising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prioritizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prioritises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prioritizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prioritisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prioritizer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prioritizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prioritiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prioritizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prioritisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "priorization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "priorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "priorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "priorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "priorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "priorize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "priorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "priorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "priorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "priorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "priorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "priorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "priorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "priorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "privatization ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "privatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "privatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "privatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "privatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "privatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "privatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "privatize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "privatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "privatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "privatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "privatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "privatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "privatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "privatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "privatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "privatizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "privatizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "privatiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "privatizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "privatisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("reward"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("reward"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("reward"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("reward"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("reward"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("reward"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("reward"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("otherwise"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "priser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("otherwise"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("otherwise"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("otherwise"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proadvertizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proadvertizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proadvertising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "problemize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "problemize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "problemise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "problemizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "problemises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "processionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "processionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "processionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "processionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "processionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Procrusteanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Procrusteanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Procrusteanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Procrusteanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Procrusteanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Procrusteanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Procrusteanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proctodeal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proctodeal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proctodaeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proctodeum ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proctodeum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proctodaeum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proctodea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proctodaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proctodeums",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proctodaeums",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proctodeum's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proctodaeum's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proctorization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proctorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proctorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proctorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proctorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proctorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proctorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proctorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proctorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proctorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proctorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proctorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proctorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proctorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proctorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proctorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prodigalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prodigalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prodigalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prodigalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prodigalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prodigalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prodigalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prodigalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prodigalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "productize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "productize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "productise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "productized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "productised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "productizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "productising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "productizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "productises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "productizer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "productizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "productiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "productizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "productisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proemium ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proemium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prooemium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "profanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "profanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "profanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "profanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "profanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "profer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "profer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "profre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "professionalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "professionalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "professionalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "professionalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "professionalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "professionalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "professionalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "professionalize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "professionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "professionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "professionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "professionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "professionalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "professionalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "professionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "professionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "professionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "professionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "professionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "professionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "professionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "program ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "program",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "programme",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "programs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "programmes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "program's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "programme's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "program",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "programme",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "programmed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "programed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "programmed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "programming",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "programing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "programming",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "programs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "programmes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "program",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("computer program"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "programmed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "programed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("computer program"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "programming",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "programing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("computer program"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "programs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("computer program"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "program's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("computer program"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "programmer ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "programmer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "programer",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "programmers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "programers",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "programmer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "programer's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proindustrialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proindustrialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proindustrialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "projectization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "projectization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "projectisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "projectizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "projectisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proletarianization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proletarianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proletarianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proletarianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proletarianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proletarianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proletarianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proletarianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proletarianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proletarianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proletarianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proletarianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proletarianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proletarianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proletarianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proletarianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proletarize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proletarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proletarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proletarized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proletarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proletarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proletarizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proletarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proletarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prologize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prologue ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prologue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prolog",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prologues",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prologs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prologue's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prolog's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prologuize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prologuize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prologuise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prologuized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prologuised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prologuizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prologuising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prologuizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prologuises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prologuizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prologuizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prologuiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prologuizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prologuisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prologuizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prologuiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prolusionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prolusionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prolusionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prolusionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prolusionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pronominalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pronominalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pronominalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pronominalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pronominalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pronominalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "propagandize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "propagandize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propagandise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "propagandized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propagandised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "propagandizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propagandising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "propagandizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propagandises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "propellant ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "propellant",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propellent",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "propellants",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propellents",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "propellant's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propellent's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prophetize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prophetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prophetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prophetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prophetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "propositionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "propositionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propositionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "propositionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propositionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "propretor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "propretor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propraetor",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "propretors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propraetors",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "propretor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propraetor's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "propretorial ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "propretorial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propraetorial",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "propretorian ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "propretorian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propraetorian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "propylitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "propylitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propylitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "propylitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propylitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "propylitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "propylitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propylitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "propylitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propylitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "propylitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propylitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "propylitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "propylitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proselytization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proselytization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proselytisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proselytizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proselytisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proselytization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proselytisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proselytize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proselytize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proselytise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proselytized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proselytised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proselytizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proselytising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proselytizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proselytises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proselytizer ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proselytizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proselytiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proselytizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proselytisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proselytizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proselytiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proseuche ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proseuche",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proseuchae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prosocele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prosocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prosocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prosopopeia ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prosopopoeia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prosopopeia",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prosopopoeias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prosopopeias",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prosopopeial ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prosopopeial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prosopopoeial",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prostatorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prostatorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prostatorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "protectionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "protectionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protectionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "protectionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protectionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proteide ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proteide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proteidae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Protestantize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Protestantize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Protestantise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Protestantized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Protestantised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Protestantized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Protestantised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Protestantizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Protestantising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Protestantizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Protestantises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Protestantize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Protestantise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Protestantizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Protestantising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "protester ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "protester",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protestor",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "protesters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protestors",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "protester's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protestor's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Prothoenor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Prothoenor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Prothoaenor",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Prothoenor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Prothoaenor's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "protoarcheology ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "protoarcheology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protoarchaeology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "protocol ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "protocoled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protocolled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "protocoling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protocolling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "protocolization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "protocolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protocolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "protocolizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protocolisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "protocolization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protocolisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "protocolize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "protocolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protocolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "protocolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protocolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "protocolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protocolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "protocolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "protocolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proverbialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proverbialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proverbialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proverbialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proverbialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proverbializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proverbialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proverbializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proverbialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "proverbize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "proverbize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proverbise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "proverbizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "proverbises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "provincialization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "provincialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "provincialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "provincializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "provincialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "provincialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "provincialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "provincialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "provincialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "provincialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "provincialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "provincialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "provincializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "provincialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "provincializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "provincialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Prussianization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Prussianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Prussianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Prussianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Prussianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Prussianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Prussianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prussianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prussianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prussianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prussianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Prussianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Prussianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Prussianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Prussianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Prussianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Prussianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Prussianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Prussianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Prussianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prussianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prussianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prussianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prussianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prussianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prussianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prussianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prussianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Prussianizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Prussianizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Prussianiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Prussianizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Prussianisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Prussianizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Prussianiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prussianizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prussianiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prussianizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "prussianiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "psalmodize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "psalmodize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psalmodise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "psalmodized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psalmodised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "psalmodizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psalmodising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "psalmodizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psalmodises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pseudelephant ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pseudelephant",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudoelephant",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pseudembryo ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pseudembryo",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudoembryo",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pseudembryonic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pseudembryonic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudoembryonic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pseudesthesia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pseudesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pseudhemal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pseudhemal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudhaemal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pseudoanemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pseudoanemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudoanaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pseudoanemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudoanaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pseudoanemia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudoanaemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pseudoanemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pseudoanemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudoanaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pseudoanemics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudoanaemics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pseudocele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pseudocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pseudocelom ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pseudocelom",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudocoelom",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pseudoedema ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pseudoedema",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudooedema",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pseudoedemas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudooedemas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pseudoedema's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudooedema's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pseudographize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pseudographize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudographise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pseudographizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudographises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pseudolunule ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pseudolunule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pseudolunulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "psychoanalyze ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "psychoanalyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psychoanalyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "psychoanalyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psychoanalysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "psychoanalyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psychoanalysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "psychoanalyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psychoanalyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "psychoanalyses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: psychoanalysis"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "psychoanalyzer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "psychoanalyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psychoanalyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "psychoanalyzers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psychoanalysers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "psychoanalyzer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psychoanalyser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "psychologization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "psychologization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psychologisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "psychologize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "psychologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psychologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "psychologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psychologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "psychologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psychologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "psychologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "psychologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ptolemean ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ptolemean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ptolemaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ptolemean's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ptolemaean's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ptyalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ptyalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ptyalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ptyalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ptyalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ptyalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ptyalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ptyalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ptyalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "publice ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "publice",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "publicae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "publicize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "publicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "publicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "publicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "publicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "publicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "publicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "publicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "publicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "puebloization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "puebloization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "puebloisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "puebloizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "puebloisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "puebloization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "puebloisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "puebloize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "puebloize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "puebloise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "puebloizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "puebloises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "puers ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "puers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pures",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pulicide ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pulicide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulicidae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pullmanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pullmanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pullmanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pullmanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pullmanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pullmanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pullmanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pulpitize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pulpitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulpitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pulpitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulpitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pulverizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pulverizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pulverizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pulverizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pulverizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pulverizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pulverization ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pulverization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pulverizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pulverization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pulverize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pulverize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pulverized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pulverizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pulverizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pulverizer ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pulverizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulveriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pulverizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulverisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pulverizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulveriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pulvilized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pulvilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pulvilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pummel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pummeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pummelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pummeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pummelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pummeler ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pummeler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pummeller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pumpkin ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pumpkin",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "punkin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pumpkins",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "punkins",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pumpkin's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "punkin's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pupilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pupilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pupilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pupilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pupilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "puppetize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "puppetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "puppetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "puppetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "puppetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Puritanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Puritanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Puritanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Puritanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Puritanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Puritanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Puritanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "puritanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "puritanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "puritanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "puritanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "puritanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "puritanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "puritanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "puritanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Puritanizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Puritanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Puritaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Puritanizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Puritanisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Puritanizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Puritaniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyemia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyemia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyaemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pygmean ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pygmean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pygmaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pygmy ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pygmy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pigmy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pygmies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pigmies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pygmy's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pigmy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyorrhea ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyorrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyorrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyorrhea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyorrhoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyorrheal ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyorrheal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyorrhoeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyorrhealer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyorrhoealer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyorrhealest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyorrhoealest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyorrheic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyorrheic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyorrhoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyramidize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyramidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyramidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyramidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyramidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pyrenean ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pyrenean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pyrenaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pyreneus ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pyreneus",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pyrenaeus",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pyreneus's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pyrenaeus's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyridinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyridinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyridinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyridinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyridinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyritization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyritization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyritisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyritizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyritisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyritization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyritisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyritize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyritize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyritise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyritized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyritised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyritizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyritising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyritizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyritises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyrolyzable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyrolyzable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyrolysable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyrolyzate ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyrolyzate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyrolysate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyrolyzates",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyrolysates",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyrolyzate's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyrolysate's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyrolyze ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyrolyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyrolyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyrolyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyrolysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyrolyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyrolysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyrolyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyrolyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pyrolyze's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyrolyse's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pyrolyzer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pyrolyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pyrolyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Pythagoreanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Pythagoreanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pythagoreanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pythagoreanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pythagoreanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Pythagoreanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Pythagoreanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pythonine ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pythonine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pythoninae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "pythonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "pythonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pythonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "pythonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "pythonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quadraphonic ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quadraphonic",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quadriphonic",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "quadrophonic",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quadraphonics",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quadriphonics",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "quadrophonics",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quahog ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quahog",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quahaug",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quahogs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quahaugs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quahog's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quahaug's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Quakerization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Quakerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Quakerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Quakerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Quakerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Quakerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Quakerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Quakerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Quakerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Quakerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Quakerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Quakerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Quakerize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Quakerise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quantization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quantization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quantisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quantizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quantisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quantization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quantisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quantize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quantize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quantise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quantized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quantised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quantizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quantising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quantizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quantises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quantizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quantizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quantiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quantizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quantisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quantizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quantiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quarrel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quarreled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quarrelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quarreling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quarrelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quarreler ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quarreler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quarreller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quarrelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quarrellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quarreler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quarreller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quarrelous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quarrelous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quarrellous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quarterization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quarterization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quarterisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quarterizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quarterisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quarterization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quarterisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quartet ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quartet",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quartette",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quartets",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quartettes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quartet's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quartette's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quaters ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quaters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quatres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "que ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "que",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "querele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "querele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "querelae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "questionary ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "questionary",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quaestionary",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "questionaries",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quaestionaries",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "questor ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quaestor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "questor",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quaestors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "questors",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quaestor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "questor's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "questorial ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "questorial",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quaestorial",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "questorship ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "questorship",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quaestorship",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "questorship's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quaestorship's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "queuing ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "queuing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "queueing",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quininize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quininize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quininise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quininizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quininises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quintessentialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quintessentialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quintessentialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quintessentialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quintessentialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quintessentializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quintessentialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quintessentializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quintessentialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "quintet ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "quintet",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quintette",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quintets",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quintettes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "quintet's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "quintette's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "raccoon ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "raccoon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racoon",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "raccoon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racoon's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "raccoons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racoons",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "racemization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "racemization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racemisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racemizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racemisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racemization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racemisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "racemize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "racemize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racemise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racemized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racemised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racemizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racemising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racemizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racemises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "racialization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "racialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "racialize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "racialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "racket ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "racket",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racquet",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("game, related implement"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rackets",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racquets",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("game, related implement"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racket's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "racquet's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some("game, related implement"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racket",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racketed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racketing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rackets",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "racket's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Radborne ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Radborne",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Radbourne",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Radborne's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Radbourne's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "radialization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "radialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "radialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "radialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "radicalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "radicalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radicalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radicalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radicalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radicalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radicalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "radicalize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "radicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radicalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radicalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radicalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radicalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "radiopasteurization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "radiopasteurization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiopasteurisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "radiosensitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "radiosensitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiosensitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radiosensitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiosensitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radiosensitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiosensitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radiosensitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiosensitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "radiosterilization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "radiosterilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiosterilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "radiosterilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "radiosterilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiosterilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radiosterilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiosterilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radiosterilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiosterilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radiosterilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiosterilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "radiumization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "radiumization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiumisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radiumizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiumisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radiumization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiumisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "radiumize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "radiumize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiumise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "radiumizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "radiumises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rajah ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rajah",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "raja",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rajah's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "raja's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rajahs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rajas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Rame ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Rame",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ramae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Rame's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Ramae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ramekin ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ramekin",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ramequin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ramekins",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ramequins",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ramekin's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ramequin's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rampier ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rampier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rampire",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rancor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rancor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rancour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rancors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rancours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rancor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rancour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "randomization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "randomization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "randomisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "randomizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "randomisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "randomization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "randomisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "randomize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "randomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "randomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "randomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "randomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "randomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "randomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "randomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "randomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "randomizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "randomizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "randomiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "randomizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "randomisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "randomizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "randomiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ranee ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ranee",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rani",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ranees",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ranis",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ranee's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rani's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "raphe ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "raphe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "raphae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "raphide ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "raphide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "raphidae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rapturize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rapturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rapturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rapturized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rapturised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rapturizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rapturising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rapturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rapturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rascalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rascalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rascalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rascalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rascalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rationalizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rationalizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationalisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rationalizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationalisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rationalizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationalisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rationalization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rationalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rationalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rationalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rationalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rationalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rationalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rationalizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rationalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rationalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rationalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rationaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ratite ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ratite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ratitae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ratlines ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ratlines",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ratlins",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ratline's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ratlin's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ratline",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ratlin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("o: usually plural"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ravel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "raveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ravelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "raveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ravelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "raveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ravelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ravelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ravellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "raveling's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "ravelling's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "raveler ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "raveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "raveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ravelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ravellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "raveler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "raveller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reacclimatization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reacclimatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reacclimatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reacclimatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reacclimatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reacclimatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reacclimatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reacclimatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reacclimatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reacclimatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reacclimatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reacclimatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reacclimatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reacclimatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reacclimatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reacclimatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reactualize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reactualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reactualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reactualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reactualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "readvertize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "readvertize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "readvertise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "readvertized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "readvertised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "readvertizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "readvertising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "realisticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "realisticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realisticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "realizability ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "realizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "realizable ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "realizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "realizableness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "realizableness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisableness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizableness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisableness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "realizably ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "realizably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizablies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisablies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "realization ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "realization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "realize ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "realize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "realizer ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "realizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "realizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "realizingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "realizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "realisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reanalyze ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reanalyze",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reanalyse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reanalyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reanalysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reanalyzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reanalysing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reanalyzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reanalyses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reanalyses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: reanalysis"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reanalyzer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reanalyzer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reanalyser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reanalyzers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reanalysers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reanimalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reanimalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reanimalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reanimalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reanimalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reapologize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reapologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reapologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reapologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reapologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reauthorization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reauthorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reauthorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reauthorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reauthorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reauthorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reauthorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reauthorize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reauthorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reauthorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reauthorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reauthorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rebaptization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rebaptization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebaptisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rebaptizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebaptisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rebaptization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebaptisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rebaptize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rebaptize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebaptise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rebaptized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebaptised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rebaptizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebaptising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rebaptizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebaptises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rebourbonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rebourbonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebourbonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rebourbonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebourbonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rebrutalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rebrutalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebrutalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rebrutalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rebrutalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recanalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recanalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recanalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recapitalization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recapitalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recapitalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recapitalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recapitalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recapitalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recapitalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recapitalize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recapitalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recapitalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recapitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recapitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recapitalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recapitalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recapitalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recapitalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recarbonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recarbonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recarbonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recarbonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recarbonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recarbonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recarbonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recarbonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recarbonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recarbonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recarbonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recarbonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recarbonizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recarbonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recarboniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recarbonizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recarbonisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recategorized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recategorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recategorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recausticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recausticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recausticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recausticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recausticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recentralization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recentralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recentralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recentralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recentralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recentralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recentralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recentralize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recentralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recentralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recentralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recentralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recercele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recercele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recercelae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reciprocalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reciprocalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reciprocalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reciprocalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reciprocalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recivilization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recivilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recivilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recivilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recivilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recivilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recivilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recivilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recivilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recivilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recivilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recivilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recognizability ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recognizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recognizable ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recognizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recognizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recognizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recognizably ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recognizably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recognizance ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recognizance",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisance",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recognizances",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisances",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recognizance's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisance's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recognize ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recognize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recognized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recognizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recognizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recognizedly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recognizedly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisedly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recognizedlies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisedlies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recognizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recognizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recogniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recognizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recognizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recogniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recognizingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recognizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recognizinglies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recognisinglies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recolonization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recolonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recolonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recolonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recolonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recolonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recolonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recolonize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recolonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recolonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recolonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recolonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recolonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recolonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recolonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recolonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recolor ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "recolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "recoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recoloring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "recolouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "recolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "recolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recoloration ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recoloration",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recolouration",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recolorations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recolourations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recoloration's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recolouration's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reconceptualization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reconceptualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reconceptualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reconceptualizing ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reconceptualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reconceptualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reconnoiter ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reconnoiter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "reconnoitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reconnoitered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "reconnoitred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reconnoitering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "reconnoitring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reconnoiters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "reconnoitres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reconnoiter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "reconnoitre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reconnoiterer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reconnoiterer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reconnoitrer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reconnoiterers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reconnoitrers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reconnoiterer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reconnoitrer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recriticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recriticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recriticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recriticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recriticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recrystallization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recrystallization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recrystallisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recrystallizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recrystallisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recrystallization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recrystallisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "recrystallize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "recrystallize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recrystallise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recrystallized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recrystallised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recrystallizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recrystallising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "recrystallizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "recrystallises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "redial ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "redialed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "redialled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "redialing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "redialling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "redialer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "redialer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "redialler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "redisseize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "redisseize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "redisseise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "redisseizin ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "redisseizin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "redisseisin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reedify ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reedify",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reaedify",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reedified",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reaedified",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reedifying",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reaedifying",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reedifies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reaedifies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reemphasize ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reemphasize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reemphasise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reemphasized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reemphasised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reemphasizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reemphasising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reemphasizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reemphasises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reemphasizer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reemphasizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reemphasiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reemphasizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reemphasisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reentrance ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reentrance",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reaentrance",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reentrance's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reaentrance's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reexportation ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reexportation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reaexportation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reexportation's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reaexportation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reexporter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reexporter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reaexporter",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reexporter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reaexporter's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "refavor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "refavor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "refavour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "refavors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "refavours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "refavor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "refavour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "refertilization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "refertilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "refertilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "refertilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "refertilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "refertilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "refertilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "refertilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "refertilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "refertilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "refertilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "refertilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reflectorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reflectorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reflectorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reflectorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reflectorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reflectorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reflectorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reflectorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reflectorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reflectorize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reflectorise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reforestization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reforestization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reforestisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reforestization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reforestisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reforestize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reforestize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reforestise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reforestizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reforestises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reformades ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reformades",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reformadoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "refuel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "refueled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "refuelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "refueling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "refuelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "regalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "regalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "regalvanization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "regalvanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regalvanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regalvanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regalvanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regalvanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regalvanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "regalvanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "regalvanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regalvanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regalvanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regalvanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reges ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reges",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "regionalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "regionalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regionalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regionalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regionalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regionalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regionalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "regionalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "regionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regionalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regionalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "regularization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "regularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "regularize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "regularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "regularizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "regularizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regulariser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regularizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regularisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regularizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regulariser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "regulize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "regulize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regulise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regulized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regulised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regulizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regulising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "regulizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "regulises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reharmonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reharmonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reharmonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reharmonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reharmonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reharmonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reharmonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reharmonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Rehm ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Rehm",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Roehm",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Rehm's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Roehm's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rehonor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rehonor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rehonour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rehonors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rehonours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rehonor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rehonour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rehumanization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rehumanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rehumanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rehumanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rehumanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rehumanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rehumanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rehumanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rehumanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rehumanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rehybridize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rehybridize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rehybridise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rehybridizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rehybridises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reindustrialization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reindustrialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reindustrialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reindustrializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reindustrialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reindustrialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reindustrialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reindustrialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reindustrialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reindustrialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reindustrializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reindustrialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reindustrializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reindustrialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reinforce ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reinforce",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reenforce",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reinforced",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reenforced",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reinforcing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reenforcing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reinforces",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reenforces",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reinforceable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reinforceable",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reenforceable",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reinitialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reinitialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reinitialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reinitialize ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reinitialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reinitialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reinitialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reinitialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reinitializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reinitialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reinitializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reinitialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reitemize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reitemize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reitemise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reitemizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reitemises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rejuvenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rejuvenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rejuvenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rejuvenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rejuvenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rejuvenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rejuvenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rejuvenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rejuvenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "relabel ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "relabeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "relabelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "relabeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "relabelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "relabeler ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "relabeler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "relabeller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "relabelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "relabellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "relativization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "relativization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "relativisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "relativizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "relativisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "relativization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "relativisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "relativize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "relativize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "relativise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "relativized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "relativised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "relativizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "relativising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "relativizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "relativises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "religionize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "religionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "religionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "religionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "religionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "religionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "religionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "religionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "religionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remagnetization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remagnetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remagnetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remagnetizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remagnetisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remagnetization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remagnetisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remagnetize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remagnetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remagnetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remagnetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remagnetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rematerialize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rematerialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rematerialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rematerialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rematerialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rematerializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rematerialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rememorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rememorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rememorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rememorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rememorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remilitarization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remilitarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remilitarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remilitarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remilitarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remilitarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remilitarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remilitarize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remilitarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remilitarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remilitarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remilitarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remilitarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remilitarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remilitarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remilitarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remineralization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remineralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remineralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remineralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remineralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remineralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remineralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remineralize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remineralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remineralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remineralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remineralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remineralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remineralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remineralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remineralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remobilization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remobilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remobilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remobilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remobilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remobilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remobilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remobilize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remobilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remobilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remobilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remobilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remodel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remodeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remodelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remodeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remodelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remold ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remold",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "remould",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remolded",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "remoulded",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remolding",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "remoulding",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remolds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "remoulds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remonetization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remonetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remonetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remonetizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remonetisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remonetization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remonetisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remonetize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remonetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remonetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remonetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remonetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remonetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remonetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remonetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remonetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remoralization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remoralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remoralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remoralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remoralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "remoralize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "remoralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remoralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remoralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remoralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remoralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remoralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "remoralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "remoralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "renationalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "renationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "renationalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renationalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "renationalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renationalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "renationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rencounter ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rencounter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rencontre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Renferd ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Renferd",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Renfred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Renferd's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Renfred's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "renormalization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "renormalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renormalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "renormalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renormalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "renormalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renormalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "renormalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "renormalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renormalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "renormalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renormalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "renormalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renormalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "renormalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "renormalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reobjectivization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reobjectivization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reobjectivisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reobjectivizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reobjectivisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reobjectivization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reobjectivisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reobjectivize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reobjectivize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reobjectivise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reobjectivizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reobjectivises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reorganization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reorganization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reorganizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reorganization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reorganizational ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reorganizational",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganisational",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reorganizationist ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reorganizationist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganisationist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reorganizationists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganisationists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reorganizationist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganisationist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reorganize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reorganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reorganized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reorganizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reorganizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reorganizer ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reorganizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reorganizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reorganizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reorganiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reoxidize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reoxidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reoxidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reoxidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reoxidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reoxidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reoxidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reoxidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reoxidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reoxygenize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reoxygenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reoxygenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reoxygenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reoxygenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "repaganization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "repaganization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repaganisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "repaganizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repaganisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "repaganization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repaganisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "repaganize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "repaganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repaganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "repaganizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repaganises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "repaganizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "repaganizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repaganiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "repaganizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repaganisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "repatronize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "repatronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repatronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "repatronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repatronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "repellent ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "repellent",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repellant",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "repellents",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repellants",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "repellent's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repellant's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "repersonalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "repersonalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repersonalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rephosphorization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rephosphorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rephosphorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rephosphorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rephosphorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rephosphorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rephosphorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rephosphorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rephosphorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rephosphorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rephosphorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rephosphorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "repolymerization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "repolymerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repolymerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "repolymerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repolymerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "repolymerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repolymerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "repolymerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "repolymerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repolymerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "repolymerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repolymerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "repopularize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "repopularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repopularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reprivatization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reprivatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reprivatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reprivatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reprivatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reprivatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reprivatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reprivatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reprivatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reprivatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reprivatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reprivatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reprivatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reprivatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reprivatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reprivatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reprized ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reprized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reprised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reprogram ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reprogram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reprogramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reprograms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reprogrammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "republicanization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "republicanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "republicanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "republicanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "republicanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "republicanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "republicanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "republicanize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "republicanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "republicanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "republicanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "republicanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "republicanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "republicanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "republicanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "republicanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "republicanizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "republicanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "republicaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "republicanizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "republicanisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "republicanizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "republicaniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "repulverize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "repulverize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repulverise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "repulverizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "repulverises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reroyalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reroyalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reroyalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reroyalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reroyalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reseize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reseize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reseise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reseizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reseizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reseiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "resensitization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "resensitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resensitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resensitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resensitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resensitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resensitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "resensitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "resensitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resensitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resensitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resensitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "resepulcher ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "resepulcher",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resepulchre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resepulchers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resepulchres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "resinize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "resinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "resolemnize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "resolemnize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resolemnise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resolemnizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resolemnises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "respectabilize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "respectabilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "respectabilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "respectabilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "respectabilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "respectabilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "respectabilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "respectabilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "respectabilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "restandardization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "restandardization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "restandardisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "restandardizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "restandardisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "restandardization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "restandardisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "restandardize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "restandardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "restandardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "restandardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "restandardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "restaurateur ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "restaurateur",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "restauranteur",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "restaurateurs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "restauranteurs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "restaurateur's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "restauranteur's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "resterilize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "resterilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resterilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resterilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resterilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "restigmatize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "restigmatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "restigmatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "restigmatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "restigmatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "resurrectionize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "resurrectionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resurrectionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resurrectionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resurrectionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resurrectionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resurrectionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resurrectionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resurrectionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "resymbolization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "resymbolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resymbolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resymbolizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resymbolisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resymbolization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resymbolisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "resymbolize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "resymbolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resymbolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resymbolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resymbolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "resynchronization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "resynchronization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resynchronisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resynchronizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resynchronisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resynchronization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resynchronisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "resynchronize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "resynchronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resynchronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resynchronized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resynchronised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resynchronizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resynchronising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resynchronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resynchronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "resynthesize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "resynthesize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resynthesise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "resynthesizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "resynthesises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "retinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "retinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "retinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "retinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "retinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "retinule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "retinule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "retinulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "retranquilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "retranquilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "retranquilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "retranquilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "retranquilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "retrocecal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "retrocecal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "retrocaecal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reutilization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reutilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reutilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reutilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reutilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reutilize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reutilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reutilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reutilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reutilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reutilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reutilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reutilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reutilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revalorization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revalorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revalorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revalorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revalorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revalorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revalorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revalorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revalorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revalorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revalorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revalorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revalorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revalorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revalorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revalorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revaporization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revaporization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revapourisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revaporizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revapourisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revaporization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revapourisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "see note for \"vaporize\"",
  ],
  },
Cluster {
  header: "revaporize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revaporize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revaporise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revaporizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revaporises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "see note for \"vaporize\"",
  ],
  },
Cluster {
  header: "revel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revelationize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revelationize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revelationise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revelationizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revelationises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reveler ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reveler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "reveller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "reverie ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "reverie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revery",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "reverie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revery's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revigor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revigor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revigour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revisualization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revisualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revisualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revisualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revisualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revisualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revisualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revisualize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revisualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revisualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revisualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revisualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revitalization ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revitalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revitalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revitalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revitalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revitalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revitalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revitalize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revitalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revitalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revitalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revitalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revitalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revitalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revitalizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revitalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revitaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revitalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revitalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revivalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revivalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revivalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revivalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revivalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revocable ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revocable",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revokable",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revolatilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revolatilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revolatilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revolatilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revolatilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revolutionize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revolutionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revolutionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revolutionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revolutionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revolutionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revolutionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revolutionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revolutionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revolutionizement ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revolutionizement",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revolutionisement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revolutionizements",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revolutionisements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revolutionizement's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revolutionisement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "revolutionizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "revolutionizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revolutioniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revolutionizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revolutionisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "revolutionizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "revolutioniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Rexford ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Rexford",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Rexfourd",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Rexford's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Rexfourd's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhaphe ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhaphe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhaphae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhapsodize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhapsodize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhapsodise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhapsodized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhapsodised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhapsodizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhapsodising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhapsodizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhapsodises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rheadine ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rheadine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhoeadine",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhebosis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhebosis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhaebosis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Rhetian ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Rhetian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Rhaetian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Rhetian's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Rhaetian's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Rhetic ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Rhetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Rhaetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhetorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhetorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhetorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhetorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhetorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhetorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhetorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhetorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhetorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rheumatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rheumatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rheumatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rheumatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rheumatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rheumatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rheumatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rheumatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rheumatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhinocele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhinocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhinocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhinocelian ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhinocelian",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhinocoelian",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhinorrhea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhinorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhinorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhinorrheal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhinorrheal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhinorrhoeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhodanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhodanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhodanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhodanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhodanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhodanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhodanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhodanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhodanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhotacize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhotacize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhotacise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhotacized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhotacised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhotacizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhotacising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhotacizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhotacises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhythmicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhythmicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhythmicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhythmicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhythmicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhythmizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhythmizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhythmisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhythmizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhythmisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhythmizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhythmisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhythmization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhythmization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhythmisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhythmizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhythmisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhythmization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhythmisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rhythmize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rhythmize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhythmise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhythmized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhythmised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhythmizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhythmising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rhythmizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rhythmises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rickshaw ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rickshaw",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ricksha",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rickshaws",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rickshas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rickshaw's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ricksha's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ridiculize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ridiculize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ridiculise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ridiculizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ridiculises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rigidize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rigidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rigidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rigidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rigidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rigidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rigidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rigidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rigidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rigmarole ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rigmarole",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rigamarole",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rigmaroles",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rigamaroles",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rigmarole's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rigamarole's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rigor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rigor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rigour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rigors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rigours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rigor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rigour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rigorism ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rigorism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rigourism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rigorism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rigourism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rigorist ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rigorist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rigourist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rigorist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rigourist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rigoristic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rigoristic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rigouristic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "riotize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "riotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "riotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "riotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "riotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "riposte ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "riposte",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ripost",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ripostes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "riposts",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "riposte's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ripost's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ritornelle ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ritornelle",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ritournelle",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ritornelles",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ritournelles",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ritualization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ritualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ritualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ritualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ritualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ritualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ritualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ritualize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ritualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ritualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ritualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ritualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ritualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ritualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ritualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ritualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rival ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rivaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rivalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rivaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rivalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rivaless ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rivaless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rivalless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rivalesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rivallesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rivaless's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rivalless's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rivalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rivalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rivalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rivalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rivalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rivalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rivalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rivalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rivalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "robotization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "robotization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "robotisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "robotizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "robotisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "robotization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "robotisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "robotize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "robotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "robotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "robotized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "robotised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "robotizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "robotising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "robotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "robotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "roed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "roed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rooed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Roentgen ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Roentgen",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "R�ntgen",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "roentgenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "roentgenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "roentgenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "roer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "roer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rore",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Romania ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Romania",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Roumania",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Romanias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Roumanias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Romania's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Roumania's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Romanization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Romanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Romanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Romanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Romanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Romanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Romanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Romanize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Romanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Romanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Romanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Romanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Romanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Romanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Romanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Romanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "romanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "romanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "romanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "romanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "romanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "romanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "romanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "romanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Romanizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Romanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Romaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Romanizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Romanisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Romanizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Romaniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "romanticization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "romanticization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "romanticisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "romanticizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "romanticisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "romanticization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "romanticisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "romanticize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "romanticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "romanticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "romanticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "romanticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "romanticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "romanticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "romanticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "romanticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ropy ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ropy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ropey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ropier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ropiest",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "routinization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "routinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "routinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "routinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "routinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "routinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "routinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "routinize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "routinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "routinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "routinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "routinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "routinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "routinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "routinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "routinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rowel ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "roweled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "rowelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "roweling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "rowelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "royalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "royalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "royalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "royalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "royalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "royalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "royalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "royalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "royalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "royalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "royalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "royalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "royalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "royalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "royalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "royalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rubberization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rubberization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rubberisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rubberize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rubberize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rubberise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rubberized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rubberised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rubberizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rubberising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rubberizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rubberises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ruble ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ruble",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rouble",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rubles",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "roubles",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ruble's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rouble's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rubricize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rubricize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rubricise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rubricizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rubricises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ruffianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ruffianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruffianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ruffianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruffianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ruggedization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ruggedization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruggedisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ruggedizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruggedisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ruggedization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruggedisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ruggedize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ruggedize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruggedise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ruggedized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruggedised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ruggedizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruggedising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ruggedizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruggedises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rumor ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rumor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rumour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rumored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rumoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rumoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rumouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rumors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rumours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rumor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rumour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rumorer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rumorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rumourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rumorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rumourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rumorer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rumourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rumormonger ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rumormonger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rumourmonger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rumormongers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rumourmongers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rumormonger's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "rumourmonger's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ruralization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ruralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ruralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ruralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ruralize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ruralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ruralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ruralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ruralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ruralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Russianization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Russianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Russianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Russianizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Russianisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Russianization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Russianisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "russianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "russianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Russianize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Russianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Russianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Russianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Russianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Russianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Russianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Russianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Russianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "russianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "russianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "rusticize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "rusticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rusticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rusticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rusticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rusticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rusticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "rusticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "rusticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Sabbathize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Sabbathize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sabbathise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Sabbathizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sabbathises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Sabbathize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sabbathise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sabbatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sabbatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabbatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sabbatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabbatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sabbatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabbatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sabbatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabbatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Sabean ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Sabean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sabaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Sabeans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sabaeans",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "saber ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "saber",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabre",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sabered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabred",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sabering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabring",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sabers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabres",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saber's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabre's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "saberbill ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "saberbill",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabrebill",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "saberlike ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "saberlike",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabrelike",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saberlikes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabrelikes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saberlike's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabrelike's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sabertooth ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sabertooth",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sabretooth",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "saccharization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "saccharization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saccharisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saccharizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saccharisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "saccharize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "saccharize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saccharise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saccharized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saccharised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saccharizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saccharising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saccharizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saccharises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sacerdotalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sacerdotalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacerdotalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sacerdotalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacerdotalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sacerdotalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacerdotalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sacerdotalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacerdotalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sacralization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sacralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sacralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sacralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sacralize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sacralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sacralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sacralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sacralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sacramentize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sacramentize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacramentise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sacramentizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sacramentises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Safier ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Safier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Safire",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Safier's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Safire's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sailorizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sailorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sailorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sailorizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sailorisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sailorizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sailorising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sake ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sake",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sakes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sake's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sake",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saki",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sakes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sakis",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sake's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saki's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "salable ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "salable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saleable",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "salabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saleabler",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "salablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saleablest",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "salinization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "salinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "salinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "salinize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "salinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "salinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "salinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "salinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "saltiers ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "saltiers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saltires",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "saltierwise ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "saltierwise",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saltirewise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "saltpeter ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "saltpeter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saltpetre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saltpeters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saltpetres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saltpeter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saltpetre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sanctuarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sanctuarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanctuarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sanctuarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanctuarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sanctuarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanctuarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sanctuarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanctuarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sandal ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sandaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sandalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sandaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sandalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Sanferd ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Sanferd",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sanfred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Sanferd's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sanfred's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Sanforize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Sanforize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sanforise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Sanforizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sanforising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Sanforizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sanforises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sanitization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sanitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sanitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sanitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sanitize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sanitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sanitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sanitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sanitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sanitizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sanitizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanitiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sanitizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanitisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sanitizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sanitiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Sanskritize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Sanskritize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sanskritise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Sanskritize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sanskritise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sapientize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sapientize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sapientise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sapientizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sapientises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sapor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sapor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sapour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sapors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sapours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sapremia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sapremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sapraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sapremias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sapraemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sapremia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sapraemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sapremic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sapremic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sapraemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sarcine ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sarcine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sarcinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sari ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sari",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saree",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saris",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sarees",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sari's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saree's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Sassanide ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Sassanide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sassanidae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "satanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "satanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "satanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "satellitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "satellitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satellitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "satellitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satellitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "satellitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satellitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "satellitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satellitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "satinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "satinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "satinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "satirizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "satirizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satirisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "satirizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satirisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "satirizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satirisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "satirization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "satirization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satirisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "satirize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "satirize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satirise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "satirized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satirised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "satirizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satirising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "satirizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satirises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "satirizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "satirizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satiriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "satirizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satirisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "satirizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "satiriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sative ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sative",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sativae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savable ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savable",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saveable",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savagize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savagize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "savagise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savagizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "savagises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Savannah ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Savannah",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savanna",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "savannah",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savannas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "savannahes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savanna's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "savannah's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savior ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savior",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "saviour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saviors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "saviours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savior's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "saviour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savioress ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savioress",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saviouress",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "saviorhood ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "saviorhood",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "saviourhood",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saviorhoods",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "saviourhoods",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saviorhood's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "saviourhood's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "saviorship ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "saviorship",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "saviourship",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saviorships",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "saviourships",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "saviorship's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "saviourship's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savorer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savorer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savorily ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savorily",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savourily",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savorilies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savourilies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savoriness ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savoriness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savouriness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savorinesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savourinesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savoriness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savouriness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savoringly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savoringly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savouringly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savoringlier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savouringlier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savoringlies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savouringlies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savoringliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savouringliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savorless ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savorless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savorlesser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savourlesser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savorlessest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savourlessest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savorlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savorlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savourlesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savorly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savorly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savourly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savorous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "savourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savorouser",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "savourouser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savorousest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "savourousest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "savory ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "savory",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savoury",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savorier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savourier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savories",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savouries",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savoriest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savouriest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "savory's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "savoury's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Saxonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Saxonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Saxonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Saxonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Saxonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Saxonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Saxonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Saxonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Saxonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Saxonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Saxonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Saxonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Saxonized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Saxonised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Saxonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Saxonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Saxonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Saxonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Saxonize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Saxonise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Saxonizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Saxonising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scalawag ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scalawag",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scallywag",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scalawags",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scallywags",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scalawag's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scallywag's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scallop ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scallop",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scollop",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scalloped",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scolloped",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scalloping",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scolloping",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scallops",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scollops",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scallop's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scollop's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scandal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scandaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scandalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scandaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scandalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scandalization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scandalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scandalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scandalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scandalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scandalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scandalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scandalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scandalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scandalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scandalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scandalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scandalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scandalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scandalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scandalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scandalizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scandalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scandaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scandalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scandalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scandalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scandaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scaped ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scaped",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scapaed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scary ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scary",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scarey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scena ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scena",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scaena",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scenarioization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scenarioization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scenarioisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scenarioizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scenarioisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scenarioization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scenarioisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scenarioize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scenarioize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scenarioise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scenarioizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scenarioises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scenarization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scenarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scenarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scenarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scenarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scenarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scenarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scenarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scenarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scenarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scenarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scenarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scenarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scenarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scenarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scenarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scepter ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scepter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sceptre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sceptered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sceptred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sceptering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sceptring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scepters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sceptres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scepter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sceptre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scepterdom ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scepterdom",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sceptredom",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scepterless ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scepterless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sceptreless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scepterlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scepterlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sceptrelesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Scevor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Scevor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Scevour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Scevor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Scevour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schedulize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schedulize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schedulise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schedulizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schedulises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schemata ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schemata",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schemas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schematization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schematization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schematisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schematizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schematisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schematization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schematisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schematize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schematize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schematise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schematized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schematised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schematizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schematising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schematizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schematises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schematizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schematizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schematiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schematizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schematisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schillerization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schillerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schillerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schillerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schillerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schillerize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schillerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schillerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schillerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schillerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schillerizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schillerising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schillerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schillerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schismatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schismatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schismatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schismatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schismatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schismatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schismatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schismatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schismatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schistocelia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schistocelia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schistocoelia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schlemiel ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schlemiel",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shlemiel",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schlemiels",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shlemiels",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schlemiel's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shlemiel's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schlep ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schlep",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schlepp",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shlep",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "shlepp",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schlepped",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shlepped",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schlepping",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shlepping",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schleps",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schlepps",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shleps",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "shlepps",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schlep's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schlepp's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shlepp's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "shlep's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schlock ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schlock",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schlocky",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shlock",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "shlocky",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schmaltz ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schmaltz",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schmalz",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "shmaltz",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schmaltzes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schmalzes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "shmaltzes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schmaltz's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schmalz's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "shmaltz's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schmaltzy ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schmaltzy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schmalzy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schmo ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schmo",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schmoe",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schmo's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schmoe's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "schnapps ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "schnapps",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schnaps",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "schnapps's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schnaps's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scientize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scientize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scientise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scientized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scientised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scientizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scientising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scientizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scientises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sclere ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sclere",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sclerae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sclerotization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sclerotization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sclerotisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sclerotization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sclerotisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sclerotize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sclerotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sclerotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sclerotized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sclerotised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sclerotizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sclerotising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sclerotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sclerotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scorse ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scorse",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scourse",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scorsed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scoursed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scorsing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scoursing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scorses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scourses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Scotticize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Scotticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Scotticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Scotticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Scotticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Scotticized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Scotticised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Scotticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Scotticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Scotticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Scotticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Scotticize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Scotticise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Scotticizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Scotticising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scripturalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scripturalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scripturalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scripturalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scripturalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scrod ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scrod",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schrod",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scrods",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schrods",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scrod's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schrod's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scrunchy ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scrunchy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrunchie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scrunchy's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrunchie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scrutinization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scrutinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrutinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scrutinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrutinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scrutinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrutinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scrutinize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scrutinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrutinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scrutinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrutinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scrutinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrutinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scrutinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrutinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scrutinizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scrutinizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrutiniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scrutinizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrutinisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scrutinizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrutiniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "scrutinizingly ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "scrutinizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrutinisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "scrutinizinglies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "scrutinisinglies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "se ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "se",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "se",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "seborrhea ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "seborrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seborrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "seborrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seborrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "seborrhea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seborrhoea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "seborrheic ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "seborrheic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seborrhoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sectarianization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sectarianization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectarianisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sectarianize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sectarianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectarianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectarianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectarianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectarianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectarianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectarianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectarianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sectionalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sectionalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectionalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectionalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sectionalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sectionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectionalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sectionization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sectionization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectionizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sectionize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sectionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sectorization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sectorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sectorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sectorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sectorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sectorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "secularization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "secularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "secularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "secularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "secularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "secularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "secularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "secularize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "secularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "secularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "secularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "secularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "secularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "secularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "secularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "secularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "secularizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "secularizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seculariser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "secularizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "secularisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "secularizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seculariser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "seculum ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "seculum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saeculum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "seculums",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saeculums",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "securitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "securitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "securitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "securitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "securitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "securitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "securitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "securitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "securitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "securitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "securitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "securitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "securitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "securitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Seed ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Seed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Saeed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Seed's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Saeed's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "seizable ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "seizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "seizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "seizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "seizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "seizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "seizin ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "seizin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seisin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "seizins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seisins",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "seizin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seisin's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "seizings ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "seizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "selle ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "selle",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sellae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "selvage ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "selvage",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "selvedge",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "selvages",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "selvedges",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "selvage's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "selvedge's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Semenov ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Semenov",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semaenov",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Semenov's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semaenov's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semianesthetic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semianesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semianaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semicarbonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semicarbonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semicarbonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semicarbonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semicarbonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semicivilization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semicivilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semicivilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semicivilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semicivilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semicivilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semicivilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semicivilized ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semicivilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semicivilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semicivilizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semiciviliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semicivilized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semicivilised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semifossilized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semifossilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semifossilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semifossilizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semifossiliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semifossilized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semifossilised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semihonor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semihonor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semihonour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semihonors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semihonours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semihonor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semihonour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semihumanized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semihumanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semihumanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semihumanizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semihumaniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semihumanized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semihumanised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semimercerized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semimercerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semimercerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semimercerizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semimerceriseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semimercerized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semimercerised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semimineralized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semimineralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semimineralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semimineralizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semimineraliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semimineralized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semimineralised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "seminarize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "seminarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seminarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "seminarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seminarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "seminationalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "seminationalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seminationalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "seminationalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seminationalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "seminationalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "seminationalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semiorganized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semiorganized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semiorganised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semiorganizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semiorganiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semiorganized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semiorganised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semioxidized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semioxidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semioxidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semioxidizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semioxidiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semioxidized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semioxidised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semioxygenized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semioxygenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semioxygenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semioxygenizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semioxygeniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semioxygenized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semioxygenised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semiprofessionalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semiprofessionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semiprofessionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semiprofessionalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semiprofessionaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semiprofessionalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semiprofessionalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semite ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semitae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Semiticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Semiticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semiticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Semiticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semiticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Semiticize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semiticise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Semitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Semitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Semitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Semitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Semitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Semitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Semitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Semitized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semitised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Semitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Semitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Semitize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semitise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Semitizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Semitising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semivulcanized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semivulcanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semivulcanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semivulcanizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semivulcaniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "semivulcanized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "semivulcanised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "semper ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "semper",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sempre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "senilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "senilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "senilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "senilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "senilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sensationalize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sensationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensationalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensationalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensationalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensationalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sensitization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sensitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sensitize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sensitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sensitizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sensitizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensitiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensitizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensitisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensitizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensitiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sensize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sensize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sensualization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sensualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sensualize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sensualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sensualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sensualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sentimentalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sentimentalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sentimentalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sentimentalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sentimentalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sentimentalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sentimentalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sentimentalize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sentimentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sentimentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sentimentalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sentimentalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sentimentalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sentimentalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sentimentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sentimentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sentimentalizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sentimentalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sentimentaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sentimentalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sentimentalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sentinel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sentineled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sentinelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sentineling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sentinelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sepaled ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sepaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sepalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sepaledder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sepalledder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sepaleddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sepalleddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Septembrizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Septembrizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Septembriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Septembrizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Septembrisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Septembrizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Septembriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "septemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "septemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "septaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "septet ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "septet",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "septette",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "septet's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "septette's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "septets",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "septettes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "septicemia ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "septicemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "septicaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "septicemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "septicaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "septicemia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "septicaemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "septicemic ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "septicemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "septicaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "septicization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "septicization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "septicisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "septicizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "septicisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "septicization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "septicisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "septicolored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "septicolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "septicoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sepulcher ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sepulcher",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sepulchre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sepulchered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sepulchred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sepulchering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sepulchring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sepulchers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sepulchres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sepulcher's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sepulchre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sepulchralize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sepulchralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sepulchralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sepulchralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sepulchralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sequentialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sequentialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sequentialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sequentialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sequentialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sequentializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sequentialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sequentializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sequentialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "serape ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "serape",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sarape",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serapes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sarapes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serape's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sarape's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "serenize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "serenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "serializability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "serializability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serialisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "serializable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "serializable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serialisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "serialization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "serialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "serialize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "serialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sericitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sericitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sericitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sericitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sericitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sermonize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sermonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sermonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sermonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sermonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sermonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sermonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sermonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sermonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sermonizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sermonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sermoniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sermonizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sermonisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sermonizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sermoniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "serose ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "serose",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serosae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "serpentinization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "serpentinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serpentinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serpentinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serpentinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serpentinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serpentinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "serpentinize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "serpentinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serpentinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serpentinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serpentinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serpentinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serpentinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serpentinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serpentinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "serpentize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "serpentize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serpentise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serpentized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serpentised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serpentizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serpentising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "serpentizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serpentises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "serre ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "serre",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "serrae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "servilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "servilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "servilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "servilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "servilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "seta ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "seta",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "saeta",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "setule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "setule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "setulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "severalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "severalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "severalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "severalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "severalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "severization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "severization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "severisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "severizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "severisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "severization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "severisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "severize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "severize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "severise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "severizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "severises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sextet ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sextet",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sextette",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sextet's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sextette's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sextets",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sextettes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sexualization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sexualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sexualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sexualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sexualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sexualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sexualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sexualize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sexualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sexualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sexualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sexualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sexualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sexualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sexualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sexualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Shakespearize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Shakespearize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Shakespearise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Shakespearizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Shakespearises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Shakespearize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Shakespearise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "shard ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "shard",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sherd",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("pottery fragments"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shards",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sherds",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("pottery fragments"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shard's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sherd's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("pottery fragments"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "shareable ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "shareable",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sharable",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sharpie ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sharpie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sharpy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sharpie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sharpy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sheik ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sheikh",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sheik",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shaykh",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shaikh",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("Arab chief"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sheikhs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sheiks",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shaykhs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shaikhs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("Arab chief"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sheikh's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sheik's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shaykh's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shaikh's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("Arab chief"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sheik",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sheikh",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sheiks",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sheikhs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sheik's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sheikh's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sheikdom ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sheikdom",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sheikhdom",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sheikdoms",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sheikhdoms",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sheikdom's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sheikhdom's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "shellac ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "shellac",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shellack",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shellacs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shellacks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shellac's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shellack's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "shepherdize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "shepherdize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shepherdise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shepherdizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shepherdises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sher ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sher",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shoer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sherardize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sherardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sherardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sherardized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sherardised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sherardizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sherardising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sherardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sherardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sherbet ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sherbet",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sherbert",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sherbets",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sherberts",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sherbet's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sherbert's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Shere ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Shere",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Shree",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Shere's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Shree's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shere",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shree",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "shillelagh ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "shillelagh",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shillalah",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shillelagh's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shillalah's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "shillelaghes ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "shillelaghes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shillalahes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Shintoize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Shintoize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Shintoise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Shintoizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Shintoises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Shintoize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Shintoise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "shit ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "shit",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shat",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: Some("past tense"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Shkoder ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Shkoder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Shkodaer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Shkoder's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Shkodaer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "shorty ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "shorty",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shortie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shorty's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shortie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "shovel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "shoveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shovelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shoveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shovelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "shoveler ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "shoveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shoveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shovelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shovellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shoveler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "shoveller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "shrivel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "shriveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shrivelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shriveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shrivelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "shtick ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "shtick",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shtik",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "schtick",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shticks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "shtiks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "schticks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "shtick's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "schtick's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "shtik's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sialorrhea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sialorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sialorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "siderealize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "siderealize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "siderealise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "siderealizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "siderealises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "signal ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "signaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "signalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "signaler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "signaller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "signalers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "signallers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "signaler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "signaller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "signaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "signalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "signalization ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "signalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "signalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "signalize ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "signalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "signalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "signalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "signalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "signalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "signalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "signalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "signalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "siled ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "siled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "siloed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "silicatization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "silicatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silicatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "silicatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silicatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "silicatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silicatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "siliceous ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "siliceous",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silicious",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "silicidize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "silicidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silicidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "silicidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silicidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "siliconize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "siliconize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "siliconise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "siliconizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "siliconises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "silicule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "silicule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "siliculae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "silique ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "silique",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "siliquae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "silverize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "silverize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silverise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "silverized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silverised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "silverizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silverising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "silverizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silverises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "silverizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "silverizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silveriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "silverizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silverisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "silverizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silveriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "similarize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "similarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "similarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "similarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "similarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "similize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "similize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "similise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "similized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "similised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "similizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "similising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "similizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "similises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "simonize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "simonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "simonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Sine ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Sine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Sine's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sinae's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "singularization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "singularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "singularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "singularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "singularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "singularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "singularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "singularize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "singularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "singularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "singularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "singularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "singularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "singularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "singularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "singularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sinicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sinicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sinicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sinicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sinicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sinicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sinicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sinicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sinicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "siphon ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "siphon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphon",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "siphoned",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphoned",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "siphoning",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphoning",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "siphons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphons",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "siphon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphon's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "siphonless ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "siphonless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphonless",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "siphonlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "siphonlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphonlesses",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "siphonlike ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "siphonlike",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphonlike",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "siphonlikes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphonlikes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "siphonlike's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphonlike's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "siphonophore ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "siphonophore",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphonophore",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "siphonophore's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphonophore's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "siphonostele ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "siphonostele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphonostele",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "siphonostele's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphonostele's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "siphonostelic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "siphonostelic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphonostelic",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "siphonostely ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "siphonostely",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphonostely",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sirenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sirenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sirenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sirenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sirenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sirenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sirenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sirenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sirenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sirree ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sirree",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "siree",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sirrees",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sirees",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sirree's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "siree's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sisterize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sisterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sisterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sisterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sisterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sistern ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sistern",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sistren",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sizable ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sizeable",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sizableness ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sizableness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sizeableness",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sizably ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sizably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sizeably",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sizel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sizel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sisel",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skeer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skeer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skere",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skeletonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skeletonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skeletonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skeletonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skeletonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skeletonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skeletonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skeletonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skeletonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skeletonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skeletonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skeletonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skeletonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skeletonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skeletonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skeletonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skeletonizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skeletonizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skeletoniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skeletonizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skeletonisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skeletonizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skeletoniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skeptic ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skeptic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "sceptic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skepticer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "scepticer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skepticest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "scepticest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skeptics",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "sceptics",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skeptic's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "sceptic's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skeptical ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skeptical",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "sceptical",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skepticaler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "scepticaler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skepticalest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "scepticalest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skeptically ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skeptically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "sceptically",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skepticism ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skepticism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "scepticism",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skepticisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "scepticisms",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skepticism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "scepticism's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skepticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skepticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skepticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skepticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skepticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skillful ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skillful",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "skilful",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skillfully ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skillfully",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "skilfully",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skillfulness ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skillfulness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "skilfulness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skillfulness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "skilfulness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skulduggery ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skulduggery",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skullduggery",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skulduggery's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skullduggery's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "skyer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "skyer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skyre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "skyers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "skyres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Slavicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Slavicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavicize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavicise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Slavization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Slavization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Slavize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Slavize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Slavonicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Slavonicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavonicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavonicized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonicised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavonicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavonicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavonicize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonicise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavonicizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonicising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Slavonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Slavonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavonized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavonize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Slavonizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Slavonising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sle ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sle",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "slae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "slenderize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "slenderize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "slenderise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "slenderized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "slenderised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "slenderizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "slenderising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "slenderizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "slenderises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sley ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "slier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "slyer",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "slyest",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sloganize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sloganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sloganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sloganized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sloganised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sloganizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sloganising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sloganizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sloganises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sloganizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sloganisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sluggardize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sluggardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sluggardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sluggardized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sluggardised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sluggardizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sluggardising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sluggardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sluggardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "slumberous ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "slumberous",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "slumbrous",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "slyly ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "slyly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "slily",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "smidgen ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "smidgen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "smidgeon",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "smidge",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "smidgin",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "smidgens",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "smidgeons",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "smidges",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "smidgins",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "smidgen's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "smidgeon's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "smidge's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "smidgin's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "smolder ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "smolder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "smoulder",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "smoldered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "smouldered",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "smoldering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "smouldering",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "smolders",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "smoulders",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "smolder's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "smoulder's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "smoothie ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "smoothie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "smoothy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "smoothie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "smoothy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "snivel ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sniveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "snivelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sniveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "snivelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "snivelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "snivellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sniveler ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sniveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sniveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "snivelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "snivellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sniveler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sniveller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "snobsniveling ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "snobsniveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "snobsnivelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "snorkel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "snorkeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "snorkelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "snorkeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "snorkelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "snowplow ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "snowplow",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "snowplough",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "snowplow's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "snowplough's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "snowplows",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "snowploughs",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "snowshed ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "snowshed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "snowshoed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "snuffcolored ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "snuffcolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "snuffcoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sol ",
  verified: false,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "so",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sol",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "so",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sols",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sos",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sol's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "so's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sol",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sols",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sol's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "soberize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "soberize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soberise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soberized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soberised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soberizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soberising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soberizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soberises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sobriquet ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sobriquet",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soubriquet",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sobriquets",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soubriquets",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sobriquet's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soubriquet's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "socialization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "socialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "socialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "socializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "socialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "socialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "socialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "socialize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "socialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "socialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "socialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "socialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "socializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "socialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "socializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "socialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "socializer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "socializer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "socialiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "socializer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "socialiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sockdologizing ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sockdologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sockdologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Socratize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Socratize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Socratise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Socratized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Socratised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Socratized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Socratised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Socratizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Socratising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Socratizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Socratises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Socratize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Socratise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Socratizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Socratising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sodomize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sodomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sodomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sodomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sodomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sodomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sodomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sodomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sodomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "softy ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "softy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "softie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "softy's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "softie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "soignee ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "soignee",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "solarization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "solarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "solarize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "solarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "soldierize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "soldierize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soldierise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soldierizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soldierises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "solecize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "solecize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solecise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solecized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solecised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solecizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solecising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solecizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solecises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "solemnization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "solemnization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solemnisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solemnizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solemnisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solemnization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solemnisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "solemnize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "solemnize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solemnise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solemnized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solemnised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solemnizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solemnising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solemnizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solemnises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "solemnizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "solemnizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solemniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solemnizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solemnisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solemnizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solemniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "soliloquization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "soliloquization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soliloquisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "soliloquize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "soliloquize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soliloquise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soliloquized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soliloquised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soliloquizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soliloquising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soliloquizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soliloquises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soliloquizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soliloquisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soliloquizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soliloquising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "soliloquizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "soliloquizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soliloquiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soliloquizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soliloquisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soliloquizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soliloquiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "soliloquizingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "soliloquizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soliloquisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "solmization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "solmization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solmisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solmizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solmisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solmization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solmisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "solonization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "solonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "solubilization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "solubilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solubilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solubilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solubilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solubilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solubilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "solubilize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "solubilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solubilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solubilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solubilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solubilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solubilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solubilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solubilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "solutize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "solutize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solutise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solutizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solutises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "solutizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "solutizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solutiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solutizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solutisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "solutizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "solutiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "somber ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "somber",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sombre",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "somberer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sombrer",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "somberest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sombrest",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "somberish ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "somberish",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sombreish",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "somberly ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "somberly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sombrely",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "somberness ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "somberness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sombreness",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sombernesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sombrenesses",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "somberness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sombreness's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "somesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "somesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "somaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "somesthesis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "somesthesis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "somaesthesis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "somniloquize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "somniloquize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "somniloquise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "somniloquized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "somniloquised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "somniloquizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "somniloquising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "somniloquizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "somniloquises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sonantized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sonantized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sonantised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sonantizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sonantiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sonantized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sonantised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sonnetization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sonnetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sonnetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sonnetization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sonnetisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sonnetize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sonnetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sonnetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sonnetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sonnetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sonnetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sonnetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sonnetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sonnetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sorbitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sorbitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sorbitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sorbitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sorbitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sorbitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sorbitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sorbitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sorbitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sorbitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sorbitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sorbitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sorbitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sorbitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sord ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sord",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sourd",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sordine ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sordine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sourdine",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sordines",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sourdines",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sororize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sororize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sororise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sororized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sororised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sororizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sororising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sororizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sororises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sory ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sory",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soury",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "souffled ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "souffled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soufflaed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "soulter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "soulter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soultre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "southernize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "southernize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "southernise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "southernized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "southernised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "southernizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "southernising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "southernizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "southernises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Sovietization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Sovietization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sovietisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Sovietization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sovietisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sovietization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sovietisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sovietizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sovietisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sovietization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sovietisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Sovietize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Sovietize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sovietise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Sovietized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sovietised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Sovietizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Sovietising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sovietize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sovietise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sovietized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sovietised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sovietizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sovietising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sovietizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sovietises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "soy ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "soy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("sauce"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soya",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("soybean"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soys",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soyas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("soybean"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "soy's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "soya's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("soybean"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spacey ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spacey",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spacy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spanemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spanemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spanaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spanemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spanemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spanaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Spaniardization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Spaniardization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spaniardisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Spaniardizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spaniardisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Spaniardization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spaniardisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Spaniardize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Spaniardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spaniardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Spaniardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spaniardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Spaniardize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spaniardise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spaniolize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spaniolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spaniolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spaniolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spaniolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spaniolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spaniolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spaniolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spaniolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Spanishize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Spanishize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spanishise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Spanishizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spanishises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Spanishize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spanishise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spanopnea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spanopnea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spanopnoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sparers ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sparers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sparres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Spartanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Spartanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spartanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Spartanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spartanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Spartanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spartanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spatialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spatialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spatialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spatializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spatialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spatialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spatialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spatialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spatialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spatialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spatializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spatialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "specialization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "specialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "specialize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "specialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "specializer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "specializer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specialiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specializers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specialisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specializer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specialiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "specialty ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "specialty",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "speciality",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specialties",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specialities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specialty's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "speciality's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "specificize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "specificize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specificise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specificizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specificises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "specimenize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "specimenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specimenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specimenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "specimenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "specter ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "specter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spectre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spectres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spectre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spectered ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spectered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spectred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "specterlike ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "specterlike",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spectrelike",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specterlikes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spectrelikes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "specterlike's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spectrelike's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spectrocolorimetry ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spectrocolorimetry",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spectrocolourimetry",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spectrocolorimetry's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spectrocolourimetry's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "speer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "speer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spere",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "speer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spere's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spelean ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spelean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spelaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "speleological ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "speleological",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spelaeological",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "speleothem ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "speleothem",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spelaeothem",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "speleothems",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spelaeothems",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spelled ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spelled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "spelt",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spelt",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("<N> species of wheat"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spermatorrhea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spermatorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spermatorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spermatorrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spermatorrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "speronares ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "speronares",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "speronaroes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sphenethmoid ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sphenethmoid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sphenoethmoid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sphenethmoidal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sphenethmoidal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sphenoethmoidal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spherocrystal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spherocrystal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sphaerocrystal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spheroidization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spheroidization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spheroidisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spheroidizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spheroidisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spheroidize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spheroidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spheroidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spheroidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spheroidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spheroidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spheroidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spheroidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spheroidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spherosome ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spherosome",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sphaerosome",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spic ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spic",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spick",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "spik",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spics",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spicks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "spiks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spic's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spick's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "spik's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Spiers ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Spiers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spires",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Spiers's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Spires's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spilled ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spilled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "spilt",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spinule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spinule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spinulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spiral ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spiraled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "spiralled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spiraling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "spiralling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spiralization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spiralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spiralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spiralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spiralize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spiralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spiralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spirea ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spirea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiraea",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spireas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiraeas",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spirea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiraea's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spiritize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spiritize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiritise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spiritizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiritises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spiritualization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spiritualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiritualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spiritualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiritualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spiritualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiritualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spiritualize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spiritualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiritualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spiritualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiritualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spiritualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiritualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spiritualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiritualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spiritualizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spiritualizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiritualiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spiritualizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiritualisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spiritualizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spiritualiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spirochetal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spirochetal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spirochaetal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spirochete ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spirochete",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spirochaete",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spirochetes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spirochaetes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spirochete's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spirochaete's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spirochetosis ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spirochetosis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spirochaetosis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spirochetoses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spirochaetoses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spirochetosis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spirochaetosis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spirochetotic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spirochetotic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spirochaetotic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "splanchnocele ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "splanchnocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "splanchnocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "splendor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "splendor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "splendour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "splendors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "splendours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "splendor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "splendour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "splendorproof ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "splendorproof",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "splendourproof",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "splenectomized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "splenectomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "splenectomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "splenization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "splenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "splenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "splenizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "splenisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spoiled ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spoiled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "spoilt",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spumoni ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spumoni",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spumone",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spumonis",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spumones",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "spumoni's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spumone's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "spyer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "spyer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spyre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "squame ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "squame",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "squamae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "squamule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "squamule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "squamulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Squier ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Squier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Squire",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Squier's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Squire's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "squirrel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "squirreled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "squirrelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "squirreling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "squirrelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stabilizable ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stabilizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabilisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stabilization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stabilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stabilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stabilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stabilizator ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stabilizator",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabilisator",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stabilizators",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabilisators",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stabilize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stabilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stabilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stabilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stabilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stabilizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stabilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabiliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stabilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabilisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stabilizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stabiliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stageyness ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stageyness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stagy ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stagy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stagey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stallionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stallionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stallionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stallionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stallionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stalwartize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stalwartize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stalwartise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stalwartizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stalwartises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stammelcolor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stammelcolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stammelcolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "standardizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "standardizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "standardizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "standardizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "standardization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "standardization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "standardizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "standardization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "standardize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "standardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "standardized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "standardizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "standardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "standardizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "standardizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "standardizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "standardizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "standardiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stanzes ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stanzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stanzoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stapedectomized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stapedectomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stapedectomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "staphyledema ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "staphyledema",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "staphyloedema",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "statisticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "statisticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "statisticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "statisticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "statisticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stearrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stearrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stearrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "steatorrhea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "steatorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "steatorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "steatorrheas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "steatorrhoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stele ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stelae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stencil ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stenciled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stencilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stenciling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stencilling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stenciler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stenciler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stenciller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stencilers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stencillers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stenciler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stenciller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stencilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stencilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stencilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stenopeic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stenopeic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stenopaeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stentor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stentor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stentour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stentors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stentours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stercoremia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stercoremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stercoraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stere ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stere",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stree",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sterilizability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sterilizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sterilizable ",
  verified: true,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sterilizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sterilization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sterilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sterilize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sterilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sterilizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sterilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "steriliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sterilisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sterilizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "steriliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stert ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stert",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stret",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stigmatization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stigmatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stigmatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stigmatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stigmatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stigmatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stigmatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stigmatize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stigmatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stigmatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stigmatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stigmatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stigmatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stigmatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stigmatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stigmatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stigmatizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stigmatizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stigmatiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stigmatizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stigmatisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stigmatizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stigmatiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stilbestrol ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stilbestrol",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stilboestrol",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stilbestrols",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stilboestrols",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stilbestrol's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stilboestrol's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stimuli ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stimuli",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stimuluses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stipule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stipule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stipulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stockinette ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stockinette",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stockinet",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stockinettes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stockinets",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stockinette's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stockinet's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stogie ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stogie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stogy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stogie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stogy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stomatodeum ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stomatodeum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stomatodaeum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stomodea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stomodea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stomodaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stomodeal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stomodeal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stomodaeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stomodeum ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stomodeum",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stomodaeum",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stomodeums",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stomodaeums",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stomodeum's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stomodaeum's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stony ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stony",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stoney",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stoor",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "storier ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "storier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stourier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "story ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "story",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("book"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stories",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("book"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "story's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("book"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "storied",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("book"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "storying",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("book"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "story",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "storey",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("level of building"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stories",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "storeys",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("level of building"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "story's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "storey's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("level of building"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stower ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stower",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stowre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stowers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stowres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "straitjacket ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "straitjacket",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "straightjacket",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "straitjacketed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "straightjacketed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "straitjacketing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "straightjacketing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "straitjackets",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "straightjackets",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "straitjacket's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "straightjacket's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "strata ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "strata",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stratums",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "strobile ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "strobile",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "strobilae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "strobilization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "strobilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "strobilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "strobilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "strobilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "strobilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "strobilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "structuralization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "structuralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "structuralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "structuralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "structuralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "structuralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "structuralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "structuralize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "structuralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "structuralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "structuralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "structuralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "strychninization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "strychninization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "strychninisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "strychninizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "strychninisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "strychninization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "strychninisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "strychninize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "strychninize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "strychninise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "strychninizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "strychninises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stumor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stumor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stumour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sty ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sty",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sties",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sty's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sty",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stied",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stying",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sties",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sty",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stye",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sties",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "styes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sty's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stye's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stylization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stylization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stylisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stylizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stylisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stylization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stylisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stylize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stylize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stylise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stylized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stylised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stylizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stylising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stylizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stylises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stylizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stylizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "styliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stylizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stylisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stylizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "styliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stylopized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stylopized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stylopised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "stymie ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "stymie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stymy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stymied",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stymyed",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "stymie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "stymy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subarmor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subarmor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subarmour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subarmors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subarmours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subarmor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subarmour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subcaliber ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subcaliber",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subcalibre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subcategorizing ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subcategorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subcategorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subcenter ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subcenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subcentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subduer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subduer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subdure",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "suberate ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "suberate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subaerate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "suberization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "suberization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suberisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "suberizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suberisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "suberization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suberisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "suberize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "suberize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suberise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "suberized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suberised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "suberizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suberising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "suberizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suberises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subesophageal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subesophageal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suboesophageal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subetheric ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subetheric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subaetheric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subflavor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subflavor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subflavour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subflavors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subflavours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subflavor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subflavour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subjectivization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subjectivization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subjectivisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subjectivizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subjectivisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subjectivization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subjectivisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subjectivize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subjectivize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subjectivise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subjectivized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subjectivised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subjectivizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subjectivising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subjectivizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subjectivises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sublimize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sublimize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sublimise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sublimized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sublimised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sublimizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sublimising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sublimizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sublimises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subminiaturization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subminiaturization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subminiaturisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subminiaturization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subminiaturisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subminiaturize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subminiaturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subminiaturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subminiaturized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subminiaturised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subminiaturizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subminiaturising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subminiaturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subminiaturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subpoena ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subpoena",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subpena",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subpoenaed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subpenaed",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subpoenaing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subpenaing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subpoenas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subpenas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subpoena's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subpena's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subpulverizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subpulverizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subpulveriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subpulverizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subpulverisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subpulverizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subpulveriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subsidizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subsidizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subsidizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subsidizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subsidization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subsidization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subsidizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subsidization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subsidize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subsidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subsidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subsidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subsidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subsidizer ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subsidizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subsidizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subsidizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subsidiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subspecialize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subspecialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subspecialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subspecializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subspecialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subspecialize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subspecialise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subspecialty ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subspecialty",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subspeciality",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subspecialties",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subspecialities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subspecialty's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subspeciality's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "substandardize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "substandardize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "substandardise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "substandardizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "substandardises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "substantialize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "substantialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "substantialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "substantialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "substantialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "substantializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "substantialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "substantializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "substantialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "substantivize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "substantivize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "substantivise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "substantivized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "substantivised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "substantivizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "substantivising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "substantivizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "substantivises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "substerilization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "substerilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "substerilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subterraneanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subterraneanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subterraneanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subterraneanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subterraneanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subtilization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subtilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subtilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subtilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subtilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subtilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subtilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subtilize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subtilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subtilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subtilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subtilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subtilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subtilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subtilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subtilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subtilizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subtilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subtiliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subtilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subtilisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subtotal ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subtotaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subtotalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subtotaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subtotalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "suburbanization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "suburbanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suburbanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "suburbanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suburbanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "suburbanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suburbanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "suburbanize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "suburbanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suburbanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "suburbanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suburbanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "suburbanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suburbanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "suburbanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suburbanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subvitalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subvitalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subvitalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subvitalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subvitalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "subvitalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "subvitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subvitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subvitalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subvitalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subvitalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subvitaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subvitalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subvitaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "subvitalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "subvitalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "succor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "succor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "succour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "succored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "succoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "succoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "succouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "succors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "succours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "succor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "succour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "succorable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "succorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "succourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "succorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "succourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "succorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "succourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "succorer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "succorer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "succourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "succorers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "succourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "succorer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "succourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "succorful ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "succorful",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "succourful",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "succorless ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "succorless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "succourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "succorlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "succorlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "succourlesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "succorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "succorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "succorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "succube ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "succube",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "succubae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "suers ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "suers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sures",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "suggestionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "suggestionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suggestionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "suggestionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "suggestionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sulcalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sulcalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulcalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulcalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulcalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulcalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulcalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulcalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulcalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sulfate ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sulfate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sulphate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfate's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sulphate's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfates",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sulphates",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sulfatize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sulfatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulfatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulfatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sulfide ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sulfide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sulphide",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfide's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sulphide's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfides",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sulphides",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sulfur ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sulfur",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "sulphur",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfured",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "sulphured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfuring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "sulphuring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfurs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "sulphurs",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfur's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "sulphur's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfur",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("Chemistry"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfured",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("Chemistry"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfuring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("Chemistry"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfurs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("Chemistry"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfur's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("Chemistry"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sulfuric ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sulfuric",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sulphuric",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sulfurization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sulfurization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurization",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfurizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurizations",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfurization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurization's",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sulfurize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sulfurize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurize",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfurized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurized",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfurizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurizing",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sulfurizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sulphurizes",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sulfurous ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sulfurous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sulphurous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sultanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sultanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sultanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sultanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sultanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sumac ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sumac",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sumach",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sumacs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sumaches",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sumac's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sumach's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "summarizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "summarizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summarisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "summarization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "summarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "summarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "summarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "summarize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "summarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "summarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "summarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "summarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "summarizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "summarizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summariser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "summarizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summarisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "summarizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summariser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "summerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "summerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "summerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "summerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "superacknowledgment ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "superacknowledgment",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superacknowledgement",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superacknowledgments",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superacknowledgements",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superacknowledgment's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superacknowledgement's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "supercanonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "supercanonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercanonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supercanonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercanonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supercanonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercanonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "supercarbonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "supercarbonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercarbonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supercarbonizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercarbonisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supercarbonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercarbonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "supercarbonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "supercarbonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercarbonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supercarbonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercarbonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "supercivilization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "supercivilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercivilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supercivilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercivilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supercivilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercivilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "supercivilized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "supercivilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercivilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supercivilizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superciviliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supercivilized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supercivilised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "superemphasize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "superemphasize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superemphasise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superemphasizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superemphasises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "superfetation ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "superfetation",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superfoetation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superfetations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superfoetations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "superficialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "superficialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superficialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superficialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superficialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superficializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superficialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superficializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superficialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "superhumanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "superhumanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superhumanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superhumanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superhumanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superhumanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superhumanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superhumanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superhumanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "supernaturalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "supernaturalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supernaturalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supernaturalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supernaturalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supernaturalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supernaturalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supernaturalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supernaturalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "superorganization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "superorganization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superorganisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superorganizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superorganisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superorganization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superorganisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "superorganize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "superorganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superorganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superorganizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superorganises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "supersensitization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "supersensitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supersensitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supersensitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supersensitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supersensitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supersensitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "supersensitize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "supersensitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supersensitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "supersensitized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "supersensitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supersensitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "supersensitizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "supersensitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supersensitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "superspecialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "superspecialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superspecialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "superspecializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "superspecialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "supersubtilized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "supersubtilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supersubtilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supersubtilizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supersubtiliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supersubtilized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supersubtilised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "supersulphurize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "supersulphurize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supersulphurise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "supersulphurizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "supersulphurises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "surgerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "surgerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "surgerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "surgerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "surgerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "surprizal ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "surprizal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "surprisal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sursize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sursize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sursise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "surveil ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "surveil",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "surveille",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "ODE lists surveille as a variant, m-w doesn't only acknowledges",
    "\"surveil\"",
  ],
  },
Cluster {
  header: "swab ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "swab",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "swob",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "swabbed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "swobbed",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "swabbing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "swobbing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "swabs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "swobs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "swab's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "swob's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sweepstakes ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sweepstakes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sweepstake",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sweep-stake",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sweepstakes's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "sweepstake's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sweep-stake's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sweetbrier ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sweetbrier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sweetbriar",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sweetbriers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sweetbriars",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sweetbrier's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sweetbriar's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "swivel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "swiveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "swivelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "swiveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "swivelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sycophantize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sycophantize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sycophantise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sycophantized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sycophantised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sycophantizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sycophantising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sycophantizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sycophantises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syllabize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syllabize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syllabise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syllabized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syllabised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syllabizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syllabising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syllabizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syllabises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syllogization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syllogization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syllogisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syllogizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syllogisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syllogize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syllogize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syllogise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syllogized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syllogised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syllogizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syllogising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syllogizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syllogises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syllogizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syllogizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syllogiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syllogizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syllogisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sylvan ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sylvan",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "silvan",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sylvanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sylvanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sylvanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sylvanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sylvanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sylviine ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sylviine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sylviinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symbol ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symboled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symbolled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symboling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symbolling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symbolization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symbolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symbolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symbolizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symbolisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symbolization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symbolisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symbolize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symbolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symbolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symbolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symbolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symbolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symbolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symbolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symbolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symbolizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symbolizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symboliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symbolizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symbolisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symbolizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symboliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symmetrically ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symmetrically",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symmetricly",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symmetricalness ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symmetricalness",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symmetricness",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symmetricalnesses",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symmetricnesses",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symmetrization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symmetrization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symmetrisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symmetrizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symmetrisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symmetrization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symmetrisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symmetrize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symmetrize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symmetrise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symmetrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symmetrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symmetrizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symmetrising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symmetrizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symmetrises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sympathize ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sympathize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sympathise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sympathized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sympathised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sympathizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sympathising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sympathizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sympathises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sympathizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sympathisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sympathizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sympathising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sympathizer ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sympathizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sympathiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sympathizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sympathisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sympathizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sympathiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sympathizingly ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sympathizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sympathisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sympathizinglier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sympathisinglier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sympathizingliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sympathisingliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symphonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symphonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symphonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symphonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symphonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symphonize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symphonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symphonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symphonized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symphonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symphonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symphonizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symphonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symphonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symptomatize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symptomatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symptomatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symptomatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symptomatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symptomatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symptomatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symptomatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symptomatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "symptomize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "symptomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symptomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "symptomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "symptomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synagogue ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synagogue",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synagog",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synagogues",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synagogs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synagogue's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synagog's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synalepha ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synalepha",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synaloepha",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synalephas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synaloephas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synalephe ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synalephe",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synaloephe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "sync ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sync",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synch",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synced",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synched",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syncing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synching",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syncs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synches",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sync's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synch's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synchronizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synchronizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchronisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synchronizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchronisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synchronizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchronisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synchronization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synchronization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchronisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synchronizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchronisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synchronization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchronisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synchronize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synchronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synchronized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchronised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synchronizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchronising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synchronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synchronizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchronising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synchronizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synchronizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchroniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synchronizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchronisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synchronizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synchroniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syncretize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syncretize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syncretise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syncretized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syncretised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syncretizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syncretising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syncretizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syncretises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syndicalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syndicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syndicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syndicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syndicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synecious ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synecious",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synoecious",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synecology ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synecology",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synoecology",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synecologies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synoecologies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syneresis ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syneresis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synaeresis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synereses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synaereses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syneresis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synaeresis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synergize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synergize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synergise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synergized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synergised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synergizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synergising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synergizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synergises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synesthesia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "synaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synesthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "synaesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synesthesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "synaesthesia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synesthetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "synaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synestheticer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "synaestheticer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synestheticest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "synaestheticest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synetic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synoetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synoecize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synoecize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synoecise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synoecized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synoecised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synoecizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synoecising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synoecizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synoecises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synonymize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synonymize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synonymise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synonymized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synonymised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synonymizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synonymising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synonymizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synonymises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synopsize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synopsize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synopsise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synopsized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synopsised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synopsizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synopsising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synopsizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synopsises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synthesization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synthesization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthesisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synthesizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthesisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synthesization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthesisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synthesize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synthesize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthesise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthetize",
        types: &[
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synthesized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthesised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthetized",
        types: &[
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synthesizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthesising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthetizing",
        types: &[
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synthesizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthesises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthetizes",
        types: &[
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synthesizer ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synthesizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthesiser",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synthesizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthesisers",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synthesizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthesiser's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synthetization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synthetization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthetisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "synthetizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "synthetizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthetiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synthetizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthetisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "synthetizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "synthetiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syntonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syntonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syntonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syntonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syntonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syntonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syntonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syntonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syntonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syntonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syntonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syntonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syntonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syntonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syphilization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syphilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syphilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syphilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syphilize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syphilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syphilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syphilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syphilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Syrianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Syrianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Syrianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Syrianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Syrianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Syrianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Syrianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syringocele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syringocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "syringocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "syrup ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "syrup",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sirup",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syrups",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sirups",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "syrup's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "sirup's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "systematization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "systematization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systematisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systematizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systematisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systematization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systematisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "systematize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "systematize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systematise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systematized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systematised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systematizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systematising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systematizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systematises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "systematizer ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "systematizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systematiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systematizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systematisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systematizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systematiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "systemizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "systemizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systemizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systemizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systemizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systemizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "systemization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "systemization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systemizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systemization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "systemize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "systemize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systemized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systemizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systemizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "systemizer ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "systemizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systemizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "systemizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "systemiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "taboo ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "taboo",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabu",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tabooed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabued",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tabooing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabuing",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "taboos",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabus",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "taboo's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabu's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tabored ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tabored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taboured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "taboring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "taborer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "taborer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "taborers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "taborer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "taboret ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "taboret",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabouret",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "taborets",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabourets",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "taboret's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabouret's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "taborin ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "taborin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabourin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "taborins",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabourins",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "taborine ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "taborine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabourine",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tabularization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tabularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tabularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tabularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tabularize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tabularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tabularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tabularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tabularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tabule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tabule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tabulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tachypnea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tachypnea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tachypnoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tachypneas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tachypnoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tachypneic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tachypneic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tachypnoeic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "taffetized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "taffetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taffetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tailorization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tailorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tailorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tailorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tailorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tailorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tailorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tailorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tailorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tailorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tailorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tailorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Talmudization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Talmudization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Talmudisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Talmudizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Talmudisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Talmudization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Talmudisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Talmudize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Talmudize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Talmudise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Talmudizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Talmudises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Talmudize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Talmudise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tamable ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tamable",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tameable",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tambura ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tambura",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tamboura",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tamburas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tambouras",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tambura's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tamboura's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Tammanyize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Tammanyize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tammanyise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tammanyizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tammanyises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tammanyize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tammanyise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tanalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tanalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tanalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tandemize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tandemize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tandemise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tandemizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tandemises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tantalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tantalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tantalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tantalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tantalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tantalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tantalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tantalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tantalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tantalizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tantalizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tantalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tantalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tantalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tantalizingly ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tantalizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tantalizinglier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalisinglier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tantalizinglies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalisinglies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tantalizingliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalisingliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tantalizingness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tantalizingness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalisingness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tantalizingnesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tantalisingnesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tariffize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tariffize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tariffise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tariffizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tariffises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tartarization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tartarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tartarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tartarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tartarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tartarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tartarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tartarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tartarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tartarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tartarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tartarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tartarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tartarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tartarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tartarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tassel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tassel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tasseled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tasselled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tasseling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tasselling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tassels",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tassells",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tautologize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tautologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tautologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tautologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tautologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tautologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tautologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tautologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tautologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tavernize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tavernize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tavernise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tavernizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tavernises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "taxidermize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "taxidermize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taxidermise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "taxidermized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taxidermised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "taxidermizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taxidermising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "taxidermizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taxidermises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Taylorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Taylorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Taylorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Taylorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Taylorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Taylorize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Taylorise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "te ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "te",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "teasel ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "teasel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teazle",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "teazel",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teasels",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teazles",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "teazels",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teasel's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teazle's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "teazel's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teaseling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teaselling",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(":ing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teaselings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teasellings",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(":ing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teaseling's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teaselling's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: Some(":ing"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teasel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teaseled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teaselled",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teaseling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teaselling",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teasels",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "teaseler ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "teaseler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teaseller",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teaselers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teasellers",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teaseler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teaseller's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Tebilize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Tebilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tebilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tebilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tebilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tebilized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tebilised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tebilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tebilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tebilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tebilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tebilize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tebilise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tebilizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tebilising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "technicalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "technicalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "technicalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "technicalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "technicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "technicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "technicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "technicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "technicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "technicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "technicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "technicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "technicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "technicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "technicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "technicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "technicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Technicolor ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Technicolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Technicolour",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Improper), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("trademark"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Technicolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Technicolours",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Improper), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("trademark"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Technicolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Technicolour's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Improper), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Improper), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("trademark"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "technicolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "technicolour",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "technicolored ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "technicolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "technicoloured",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "technologize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "technologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "technologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "teer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "teer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teaer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "teetotal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "teetotaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teetotalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teetotaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teetotalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "teetotaler ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "teetotaler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teetotaller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teetotalers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teetotallers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teetotaler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teetotaller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tegu ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tegu",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taegu",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tele ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "telae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "telepathize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "telepathize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "telepathise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "telepathized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "telepathised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "telepathizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "telepathising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "telepathizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "telepathises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "telesthesia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "telesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "telaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "telesthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "telaesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "telesthesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "telaesthesia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "telesthetic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "telesthetic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "telaesthetic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "telestheticer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "telaestheticer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "telestheticest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "telaestheticest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tellurize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tellurize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tellurise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tellurized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tellurised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tellurizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tellurising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tellurizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tellurises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tels ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tels",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taels",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "templize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "templize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "templise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "templizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "templises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tempo ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tempos",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tempi",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "temporalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "temporalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "temporization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "temporization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "temporize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "temporize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "temporizer ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "temporizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "temporizingly ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "temporizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporizinglier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporisinglier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "temporizingliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "temporisingliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tenderization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tenderization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenderisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tenderizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenderisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tenderization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenderisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tenderize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tenderize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenderise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tenderized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenderised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tenderizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenderising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tenderizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenderises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tenderizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tenderizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenderiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tenderizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenderisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tenderizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenderiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tendinitis ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tendinitis",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tendonitis",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tendinitises",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tendonitises",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tendinitis's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tendonitis's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tendriled ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tendriled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tendrilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tenementization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tenementization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenementisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tenementizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenementisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tenementization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenementisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tenementize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tenementize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenementise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tenementizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenementises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tenia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tenia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taenia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tenias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taenias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tenia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taenia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "teniasis ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "teniasis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taeniasis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teniases",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taeniases",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teniasis's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taeniasis's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "teniacidal ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "teniacidal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taeniacidal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "teniacide ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "teniacide",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taeniacide",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teniacides",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taeniacides",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teniacide's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taeniacide's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "teniae ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "teniae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taeniae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "teniafuge ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "teniafuge",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taeniafuge",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teniafuges",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taeniafuges",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teniafuge's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taeniafuge's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tenioid ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tenioid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "taenioid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tenthmeter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tenthmeter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tenthmetre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tepee ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tepee",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teepee",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tipi",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tepees",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teepees",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tipis",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tepee's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teepee's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tipi's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "terf ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "terf",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tref",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "terma ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "terma",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trema",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "termatic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "termatic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trematic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "terminalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "terminalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terminalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "terminalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terminalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "terminalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terminalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "terminalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "terminalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terminalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "terminalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terminaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "terminalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terminalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ternize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ternize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ternise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ternizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ternises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "terre ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "terre",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terrae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "terrestrialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "terrestrialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terrestrialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "terrestrializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terrestrialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "territorialization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "territorialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "territorialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "territorializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "territorialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "territorialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "territorialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "territorialize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "territorialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "territorialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "territorialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "territorialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "territorializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "territorialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "territorializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "territorialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "terrorization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "terrorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terrorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "terrorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terrorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "terrorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terrorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "terrorize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "terrorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terrorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "terrorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terrorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "terrorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terrorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "terrorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terrorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "terrorizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "terrorizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terroriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "terrorizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terrorisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "terrorizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "terroriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "terts ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "terts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trets",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "teste ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "teste",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "testae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "testimonialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "testimonialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "testimonialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "testimonializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "testimonialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "testimonialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "testimonialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "testimonialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "testimonialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "testimonialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "testimonialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "testimonialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "testimonializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "testimonialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "testimonializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "testimonialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "testimonializer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "testimonializer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "testimonialiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "testimonializers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "testimonialisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tetanization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tetanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tetanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tetanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tetanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tetanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tetanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tetanize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tetanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tetanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tetanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tetanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tetanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tetanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tetanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tetanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tetrachlorethylene ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tetrachlorethylene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tetrachloroethylene",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tetrachlorethylenes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tetrachloroethylenes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Teutonization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Teutonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Teutonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Teutonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Teutonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Teutonize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Teutonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Teutonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Teutonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Teutonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Teutonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Teutonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Teutonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Teutonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "teutonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "teutonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "texturize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "texturize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "texturise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "texturized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "texturised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "texturizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "texturising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "texturizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "texturises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thalamocele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thalamocele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thalamocoele",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thalassemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thalassemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thalassaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thalassemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thalassaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thalassemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thalassemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thalassaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theater ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theater",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "theatre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theaters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "theatres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theater's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "theatre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theatergoer ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theatergoer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatregoer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theatergoers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatregoers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theatergoer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatregoer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theatergoing ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theatergoing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatregoing",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theatergoings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatregoings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theatergoing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatregoing's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theaterless ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theaterless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatreless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theaterlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theaterlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatrelesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theaterlike ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theaterlike",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatrelike",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theaterlikes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatrelikes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theaterlike's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatrelike's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theatricalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theatricalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatricalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theatricalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatricalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theatricalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatricalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theatricalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theatricalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatricalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theatricalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatricalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theatricalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatricalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theatricalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatricalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theatricize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theatricize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatricise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theatricized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatricised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theatricizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatricising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theatricizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theatricises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theologization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theologization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theologisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theologizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theologisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theologization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theologisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theologize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theologize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theologise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theologized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theologised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theologizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theologises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theologizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theologizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theologiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theologizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theologisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theologizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theologiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Theone ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Theone",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Theonoe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Theone's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Theonoe's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theorization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theorize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theorizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theorizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theoriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theorizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theorisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theorizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theoriser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "theosophize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "theosophize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theosophise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theosophized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theosophised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theosophizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theosophising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "theosophizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "theosophises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "therap ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "therap",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "threap",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thereness ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thereness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "threeness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "therenesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "threenesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thermalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thermalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thermalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thermalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thermalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thermalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thermalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thermalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "therme ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "therme",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thermesthesia ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thermesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermaesthesia",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thermesthesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermaesthesia's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thermoanesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thermoanesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermoanaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thermoanesthesias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermoanaesthesias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thermoanesthesia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermoanaesthesia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thermometerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thermometerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermometerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thermometerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermometerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thermopolymerization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thermopolymerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermopolymerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thermopolymerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermopolymerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thermopolymerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermopolymerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thermosiphon ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thermosiphon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermosyphon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thermosiphons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermosyphons",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thermosiphon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thermosyphon's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thesmothete ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thesmothete",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thesmothetae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thiamine ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thiamine",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thiamin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thiamines",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thiamins",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thiamine's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thiamin's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thralldom ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thralldom",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "thraldom",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thralldoms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "thraldoms",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thralldom's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "thraldom's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thronize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thronize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thronise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thronizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thronises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "through ",
  verified: true,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "through",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thru",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thruway ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thruway",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "throughway",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thruways",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "throughways",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thruway's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "throughway's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thymectomize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thymectomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thymectomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thyroidectomized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thyroidectomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thyroidectomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "thyroidization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "thyroidization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thyroidisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thyroidizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thyroidisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "thyroidization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "thyroidisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tidbit ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tidbit",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "titbit",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tidbits",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "titbits",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tidbit's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "titbit's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Tillford ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Tillford",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tillfourd",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tillford's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tillfourd's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Timonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Timonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Timonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Timonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Timonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Timonized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Timonised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Timonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Timonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Timonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Timonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Timonize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Timonise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Timonizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Timonising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Timor ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Timor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Timour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Timor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Timour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tinsel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tinseled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tinselled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tinseling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tinselling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tinselier ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tinselier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tinsellier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tinseliest ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tinseliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tinselliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tire ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tire",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("exhausted"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tires",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("exhausted"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tire's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("exhausted"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tired",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("exhausted"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tiring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("exhausted"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tire",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("wheel"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tires",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("whell"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tire's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("wheel"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "titer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "titer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "titre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "titers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "titres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "titer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "titre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "titivate ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "titivate",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tittivate",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "titivated",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tittivated",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "titivating",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tittivating",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "titivates",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tittivates",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "titivation ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "titivation",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tittivation",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "titivations",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tittivations",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "titivation's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tittivation's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tittup ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tittuped",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tittupped",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tittuping",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tittupping",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tittupy ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tittupy",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tittuppy",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "toffee ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "toffee",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "toffy",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "toffees",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "toffies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "toffee's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "toffy's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tog ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "togged",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "toged",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "togging",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "toging",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ton ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ton",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ton's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tonne",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("metric ton"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tonne's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("metric ton"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tonicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tonicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tonicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tonicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tonicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "topee ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "topee",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "topi",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("hat"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "topees",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "topis",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("hat"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "topee's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "topi's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("hat"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "topi",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "topis",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl: topi"),
  description: Some("antelope "),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "topesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "topesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "topaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "toret ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "toret",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "touret",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tormentor ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tormentor",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tormenter",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tormentors",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tormenters",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tormentor's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tormenter's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "torporize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "torporize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "torporise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "torporizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "torporises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Toryize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Toryize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Toryise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Toryizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Toryises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Toryize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Toryise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "total ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "totaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totaling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "totaler ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "totaler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totaller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totalers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totallers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totaler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totaller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "totalitarianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "totalitarianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalitarianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "totalization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "totalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "totalizator ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "totalizator",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalisator",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totalizators",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalisators",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totalizator's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalisator's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "totalize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "totalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "totalizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "totalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "totemization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "totemization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totemisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totemizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totemisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "totemization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "totemisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tourize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tourize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tourise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tourizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tourises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "toward ",
  verified: false,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "toward",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "toward",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "towards",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Preposition), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "towel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "toweled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "towelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "toweling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "towelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "towelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "towellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "toweling's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "towelling's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "toxanemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "toxanemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "toxanaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "toxemia ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "toxemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "toxaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "toxemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "toxaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "toxemia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "toxaemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "toxemic ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "toxemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "toxaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "toxicemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "toxicemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "toxicaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "toxicohemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "toxicohemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "toxicohaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "toxihemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "toxihemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "toxihaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "toxinemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "toxinemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "toxinaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trabeate ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trabeate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trabeatae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trabecule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trabecule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trabeculae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tractorization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tractorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tractorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tractorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tractorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tractorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tractorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tractorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tractorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tractorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tractorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tractorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "traditionalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "traditionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traditionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traditionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traditionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traditionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traditionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "traditionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "traditionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traditionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traditionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traditionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tragicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tragicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tragicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tragicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tragicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tragicolored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tragicolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tragicoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "traitorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "traitorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traitorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traitorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traitorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trammel ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trammeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "trammelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trammeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "trammelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trammeler ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trammeler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trammeller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trammelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trammellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trammeler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trammeller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tranquility ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tranquility",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tranquillity",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tranquillities",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquility's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tranquillity's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tranquilization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tranquilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tranquilize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tranquilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillize",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillized",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillizing",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillizes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquilizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillizings",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: Some("normally an adj."),
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquilizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillizing's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: Some("normally an adj."),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tranquilizer ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tranquilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillizer",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquilliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillizers",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquilizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillizer's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquilliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tranquilizingly ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tranquilizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquilisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tranquillization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tranquillization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquillizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tranquillization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tranquillizingly ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tranquillizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tranquillisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "transcendentalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "transcendentalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transcendentalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transcendentalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transcendentalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "transcendentalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "transcendentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transcendentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transcendentalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transcendentalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transcendentalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transcendentalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transcendentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transcendentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "transcolor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "transcolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transcolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transcolorrer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transcolourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transcolorrest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transcolourest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "transcoloration ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "transcoloration",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transcolouration",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transcolorations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transcolourations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transcoloration's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transcolouration's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "transgender ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "transgender",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transgendered",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Eq), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "transistorization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "transistorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transistorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transistorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transistorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transistorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transistorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "transistorize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "transistorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transistorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transistorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transistorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transistorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transistorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transistorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transistorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "transparentize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "transparentize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transparentise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "transparentizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "transparentises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "traumatization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "traumatization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traumatisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traumatizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traumatisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traumatization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traumatisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "traumatize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "traumatize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traumatise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traumatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traumatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traumatizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traumatising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traumatizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traumatises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "travel ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "traveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "travelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "travelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "travellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traveler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "traveller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "travelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "travelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "travellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "traveling's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "travelling's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "travelogue ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "travelogue",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "travelog",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "travelogues",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "travelogs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "travelogue's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "travelog's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trial ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trialed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trialled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trialing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trialling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trialization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "triangularization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "triangularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "triangularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "triangularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "triangularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "triangularize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "triangularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "triangularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "triangularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "triangularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "triangularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "triangularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "triangularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "triangularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trichinization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trichinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trichinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trichinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trichinize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trichinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trichinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trichinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trichinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trichlorethylene ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trichlorethylene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichloroethylene",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trichlorethylenes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichloroethylenes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trichotomize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trichotomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichotomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trichotomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichotomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trichotomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichotomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trichotomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trichotomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tricolor ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tricolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tricolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tricolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tricolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tricolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tricolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tricolored ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tricolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tricoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "triecious ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "triecious",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trioecious",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trieciously ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trieciously",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trioeciously",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "triene ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "triene",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "triaene",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trifluoride ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trifluoride",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trifluouride",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trillionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trillionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trillionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trillionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trillionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trimerization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trimerization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trimerisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trimerizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trimerisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trimerization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trimerisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tripylean ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tripylean",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tripylaean",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trivialization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trivialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trivialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trivializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trivialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trivialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trivialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trivialize ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trivialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trivialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trivialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trivialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trivializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trivialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trivializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trivialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trolley ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trolley",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trolly",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trolleyed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trollied",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trolleying",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trollying",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trolleys",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trollies",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trolley's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trolly's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tropeolin ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tropeolin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tropaeolin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tropeolin's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tropaeolin's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tropicalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tropicalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tropicalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tropicalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tropicalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tropicalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tropicalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tropicalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tropicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tropicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tropicalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tropicalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tropicalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tropicalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tropicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tropicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trowel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "troweled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trowelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "troweling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trowelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "troweler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "troweler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "troweller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trowelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trowellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "troweler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "troweller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Trubenize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Trubenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Trubenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Trubenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Trubenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Trubenized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Trubenised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Trubenizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Trubenising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Trubenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Trubenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Trubenize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Trubenise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Trubenizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Trubenising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "trypsinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "trypsinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trypsinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trypsinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trypsinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "trypsinize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "trypsinise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tsoris ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tsoris",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tsouris",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tubercularization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tubercularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tubercularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tubercularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tubercularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tubercularize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tubercularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tubercularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tubercularized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tubercularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tubercularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tubercularizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tubercularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tubercularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tuberculinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tuberculinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tuberculinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tuberculinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tuberculinize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tuberculinize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculinise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tuberculinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tuberculinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tuberculinizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculinises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tuberculization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tuberculization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tuberculizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tuberculize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tuberculize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tuberculized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tuberculizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tuberculizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberculises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tuberization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tuberization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tuberizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tuberization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tuberize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tuberize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tuberizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tuberises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tubulization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tubulization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tubulisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tubulizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tubulisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tubulization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tubulisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tularemia ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tularemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tularaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tularemias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tularaemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tularemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tularemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tularaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tumbrel ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tumbrel",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tumbril",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tumbrels",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tumbrils",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tumbrel's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tumbril's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tumor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tumor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tumour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tumors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tumours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tumor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "tumour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tumored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tumored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tumoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tunnel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tunneled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tunnelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tunneling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tunnelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tunnelings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tunnellings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tunneler ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tunneler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tunneller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tunnelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tunnellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tunneler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tunneller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "turdine ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "turdine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "turdinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Turkicize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Turkicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Turkicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Turkicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Turkicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Turkicized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Turkicised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Turkicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Turkicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Turkicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Turkicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Turkicize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Turkicise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Turkicizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Turkicising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Turkize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Turkize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Turkise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Turkizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Turkises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Turkize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Turkise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Tuscanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Tuscanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tuscanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tuscanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tuscanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tuscanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tuscanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tutele ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tutele",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tutelae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tutorization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tutorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tutorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tutorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tutorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tutorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tutorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tutorize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tutorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tutorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tutorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tutorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tutorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tutorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tutorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tutorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "twier ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "twier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "twire",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "twiers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "twires",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tyke ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tyke",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tike",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tykes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tikes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tyke's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tike's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Tylerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Tylerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tylerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tylerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tylerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Tylerize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Tylerise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "typhemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "typhemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "typhaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "typhemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "typhoemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "typhoemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "typhoaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "typhlenteritis ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "typhlenteritis",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "typhloenteritis",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tyrannize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tyrannize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyrannise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tyrannized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyrannised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tyrannizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyrannising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tyrannizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyrannises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tyrannizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyrannisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tyrannizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyrannising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tyrannizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tyrannizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyranniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tyrannizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyrannisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tyrannizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyranniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tyrannizingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tyrannizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyrannisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tyrannizinglier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyrannisinglier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tyrannizingliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tyrannisingliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "tyro ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "tyro",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tiro",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tyreo",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tyros",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tiros",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tyreos",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "tyro's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "tiro's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "tyreo's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ukulele ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ukulele",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ukelele",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ukuleles",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ukeleles",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ukulele's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ukelele's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ultracentralizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ultracentralizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ultracentraliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ultracentralizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ultracentralisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ultracentralizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ultracentraliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ultrahonorable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ultrahonorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ultrahonourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ultrahonorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ultrahonourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ultrahonorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ultrahonourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ultraspecialization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ultraspecialization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ultraspecialisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ultraspecializations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ultraspecialisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ultraspecialization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ultraspecialisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ultrastandardization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ultrastandardization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ultrastandardisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ultrastandardizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ultrastandardisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ultrastandardization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ultrastandardisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unacclimatized ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unacclimatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unacclimatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unagonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unagonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unagonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unalcoholized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unalcoholized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unalcoholised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unalcoholizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unalcoholisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unalcoholizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unalcoholiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unalcoholizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unalcoholiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unalcoholized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unalcoholised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unalphabetized ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unalphabetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unalphabetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unamortization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unamortization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unamortisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unamortizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unamortisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unamortization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unamortisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unamortized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unamortized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unamortised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unanalyzable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unanalyzable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unanalysable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unanalyzed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unanalyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unanalysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unanatomizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unanatomizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unanatomisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unanatomized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unanatomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unanatomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unanemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unanemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unanaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unanimalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unanimalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unanimalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unanimalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unanimaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unanimalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unanimalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unantagonizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unantagonizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unantagonisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unantagonizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unantagonisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unantagonizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unantagonisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unantagonizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unantagonisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unantagonizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unantagonisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unantagonized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unantagonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unantagonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unantagonizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unantagonisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unantagonizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unantagoniseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unantagonizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unantagonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unantagonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unantagonizinger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unantagonisinger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unantagonizingest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unantagonisingest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unapologizing ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unapologizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unapologising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unapologizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unapologising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unapostatized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unapostatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unapostatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unapostatizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unapostatiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unapostatized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unapostatised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unappetizing ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unappetizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unappetising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unappetizinger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unappetisinger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unappetizingest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unappetisingest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unappetizingly ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unappetizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unappetisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unappetizinglier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unappetisinglier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unappetizingliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unappetisingliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unapprized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unapprized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unapprised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unarbored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unarbored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unarboured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unarmored ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unarmored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unarmoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unarmoredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unarmouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unarmoreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unarmoureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unauthorize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unauthorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unauthorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unauthorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unauthorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unauthorized ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unauthorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unauthorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unauthorizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unauthorisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unauthorizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unauthoriseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unauthorizedly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unauthorizedly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unauthorisedly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unauthorizedness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unauthorizedness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unauthorisedness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unbaptize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unbaptize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbaptise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unbaptized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbaptised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unbaptizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbaptising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unbaptizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbaptises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unbarbarize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unbarbarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbarbarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unbarbarized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unbarbarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbarbarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unbarbarizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unbarbarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbarbarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unbarricaded ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unbarricaded",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbarricadoed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unbastardized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unbastardized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbastardised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unbastardizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbastardisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unbastardizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbastardiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unbeknownst ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unbeknownst",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "unbeknown",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unbrutalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unbrutalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbrutalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unbrutalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbrutalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unbrutalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbrutalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unbrutalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbrutalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unbrutize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unbrutize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbrutise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unbrutized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbrutised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unbrutizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbrutising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unbrutizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unbrutises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncanceled ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncanceled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncancelled",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncanceledder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncancelledder",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncanceleddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncancelleddest",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncandor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncandor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncandour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncanonization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncanonization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncanonisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncanonization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncanonisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncanonize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncanonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncanonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncanonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncanonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncanonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncanonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncanonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncanonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncantonized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncantonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncantonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncantonizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncantoniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncantonized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncantonised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncapitalized ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncapitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncapitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncapitalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncapitalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncapitalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncapitaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncaramelized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncaramelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncaramelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncatechized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncatechized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncatechised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncatechizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncatechiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncatechized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncatechised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncatechizedness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncatechizedness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncatechisedness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncategorized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncategorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncategorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncategorizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncategorisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncategorizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncategoriseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncatholicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncatholicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncatholicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncatholicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncatholicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncatholicizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncatholicising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncatholicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncatholicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncauterized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncauterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncauterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncauterizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncauteriseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncauterized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncauterised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncelestialized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncelestialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncelestialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncelestializeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncelestialiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncelestialized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncelestialised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncenter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncenter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncentre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncentralized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncentralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncentralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncharacterized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncharacterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncharacterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncharacterizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncharacterisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncharacterizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncharacteriseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unchastizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unchastizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unchastisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unchastized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unchastized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unchastised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unchloridized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unchloridized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unchloridised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unchloridizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unchloridiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unchloridized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unchloridised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unchristianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unchristianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unchristianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unchristianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unchristianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unchristianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unchristianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unchristianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unchristianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncircularized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncircularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncircularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncircularizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncircularisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncircularizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncirculariseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncircularizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncirculariseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncircularized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncircularised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncivilizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncivilizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncivilisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncivilizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncivilisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncivilizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncivilisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncivilizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncivilisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncivilizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncivilisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncivilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncivilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncivilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncivilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncivilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncivilized ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncivilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncivilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncivilizedly ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncivilizedly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncivilisedly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncivilizedness ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncivilizedness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncivilisedness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unclericalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unclericalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unclericalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unclericalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unclericalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncognizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncognizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncognisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncolonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncolonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncolonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncolonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncolonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncolorable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncolorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncolorabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolourabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncolorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncolorablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolourablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncolorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncolorably ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncolorably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolourably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncolorablier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolourablier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncolorablies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolourablies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncolorabliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolourabliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncolored ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "uncoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncoloredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "uncolouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncoloreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "uncoloureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncoloreds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "uncoloureds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncoloredly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncoloredly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolouredly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncoloredlier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolouredlier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncoloredliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolouredliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncoloredness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncoloredness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolouredness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncolorednesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolourednesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncoloredness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncolouredness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unconcerted ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unconcerted",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unconcreted",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unconventionalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unconventionalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unconventionalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unconventionalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unconventionalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncriticizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncriticizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncriticisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncriticizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncriticisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncriticizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncriticisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncriticizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncriticisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncriticizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncriticisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncriticizably ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncriticizably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncriticisably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncriticized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncriticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncriticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncriticizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncriticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncriticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncriticizingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncriticizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncriticisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncrystallizability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncrystallizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncrystallisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncrystallizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncrystallisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncrystallizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncrystallisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncrystallizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncrystallizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncrystallisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncrystallizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncrystallisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncrystallizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncrystallisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncrystallizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncrystallisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncrystallizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncrystallisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncrystallized ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncrystallized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncrystallised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uncurricularized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uncurricularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncurricularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncurricularizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncurriculariseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uncurricularized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uncurricularised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undefense ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undefense",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undefence",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undefenses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undefences",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undefense's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undefence's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undemagnetizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undemagnetizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undemagnetisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undemagnetizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undemagnetisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undemagnetizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undemagnetisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undemocratization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undemocratization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undemocratisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undemocratization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undemocratisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undemocratize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undemocratize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undemocratise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undemocratized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undemocratised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undemocratizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undemocratising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undemocratizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undemocratises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undenominationalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undenominationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undenominationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undenominationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undenominationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undercapitalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undercapitalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercapitalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undercapitalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercapitalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undercapitalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercapitalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undercapitalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undercapitalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercapitalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undercapitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercapitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undercapitalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercapitalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undercapitalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercapitalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undercolor ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undercolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undercolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undercoloring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercolouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undercolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undercolorings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercolourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undercolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undercolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "underemphasize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "underemphasize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underemphasise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underemphasized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underemphasised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underemphasizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underemphasising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underemphasizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underemphasises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undergoer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undergoer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undergore",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "underlaborer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "underlaborer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underlabourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underlaborers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underlabourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underlaborer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underlabourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "underorganization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "underorganization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underorganisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underorganizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underorganisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underorganization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underorganisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "underoxidize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "underoxidize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underoxidise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underoxidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underoxidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underoxidizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underoxidising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underoxidizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underoxidises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "underprize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "underprize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underprise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underprized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underprised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underprizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underprising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underprizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underprises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "underrealize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "underrealize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underrealise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underrealized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underrealised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underrealizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underrealising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underrealizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underrealises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undersavior ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undersavior",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undersaviour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undersaviors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undersaviours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undersavior's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undersaviour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undersized ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undersized",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undersize",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "underutilization ",
  verified: true,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "underutilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underutilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underutilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underutilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underutilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underutilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "underutilize ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "underutilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underutilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underutilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underutilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underutilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underutilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "underutilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "underutilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undervitalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undervitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undervitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undervitalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undervitaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undervitalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undervitalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undialyzed ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undialyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undialysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undialyzeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undialyseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undialyzed's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undialysed's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undiphthongize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undiphthongize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undiphthongise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undiphthongizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undiphthongises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undiscolored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undiscolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undiscoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undiscoloredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undiscolouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undiscoloreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undiscoloureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undiscoloreds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undiscoloureds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undiscolored's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undiscoloured's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undishonored ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undishonored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undishonoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undisorganized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undisorganized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undisorganised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undramatizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undramatizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undramatisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undramatizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undramatisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undramatizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undramatisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undramatizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undramatisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undramatizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undramatisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undramatized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undramatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undramatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undramatizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undramatiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undramatized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undramatised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undreamed ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undreamed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undreamt",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "undualize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "undualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "undualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "undualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uneconomizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uneconomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uneconomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unenamored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unenamored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unenamoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unenamoredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unenamouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unenamoreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unenamoureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unenamoreds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unenamoureds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unenamored's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unenamoured's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unendeavored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unendeavored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unendeavoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unenergized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unenergized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unenergised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unenergizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unenergiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unenergized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unenergised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unepitomized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unepitomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unepitomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unepitomizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unepitomisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unepitomizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unepitomiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unequaled ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unequaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unequalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unequaledder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unequalledder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unequaleddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unequalleddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unequalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unequalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unequalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unequalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unequalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unequalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unequalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unequalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unequalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uneulogized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uneulogized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uneulogised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uneulogizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uneulogisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uneulogizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uneulogiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uneulogizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uneulogiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uneulogized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uneulogised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unevangelized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unevangelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unevangelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unevangelizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unevangelisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unevangelizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unevangeliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unevangelizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unevangeliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unevangelized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unevangelised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfamiliarized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfamiliarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfamiliarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfamiliarizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfamiliarisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfamiliarizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfamiliariseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfavorable ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfavorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavorabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavorablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfavorableness ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfavorableness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourableness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavorablenesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourablenesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavorableness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourableness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfavorably ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfavorably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavorablier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourablier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavorabliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourabliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfavored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfavored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavoredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavoreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavoureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavored's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavoured's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfavoring ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfavoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavoringer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavouringer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavoringest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavouringest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavorings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavoring's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavouring's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfavorite ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfavorite",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourite",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavoriter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavouriter",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavorites",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourites",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavoritest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavouritest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfavorite's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unfavourite's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfeminize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfeminize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfeminise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfeminized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfeminized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfeminised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfeminizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfeminizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfeminising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfertilizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfertilizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfertilisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfertilizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfertilisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfertilizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfertilisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfertilizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfertilisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfertilizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfertilisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfertilized ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfertilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfertilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfertilizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfertilisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfertilizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfertiliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfertilizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfertilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfertilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfeudalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfeudalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfeudalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfeudalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfeudalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfeudalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfeudalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfeudalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfeudalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unflavored ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unflavored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unflavoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unflavoredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unflavouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unflavoreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "unflavoureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unflavorous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unflavorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unflavourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfocused ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfocused",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfocussed",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unformalized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unformalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unformalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unformalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unformalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unformalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unformaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfossilized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfossilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfossilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfossilizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfossilisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfossilizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfossiliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfossilizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfossiliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfossilized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfossilised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfraternized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfraternized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfraternised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unfraternizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unfraternizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfraternising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfraternizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfraternisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unfraternizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unfraternising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ungalvanized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ungalvanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungalvanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungalvanizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungalvaniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungalvanized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungalvanised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ungelatinizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ungelatinizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungelatinisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungelatinizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungelatinisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungelatinizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungelatinisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ungelatinized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ungelatinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungelatinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungelatinizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungelatiniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungelatinized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungelatinised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ungeneralized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ungeneralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungeneralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungeneralizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungeneralisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungeneralizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungeneraliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ungeneralizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ungeneralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungeneralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ungentilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ungentilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungentilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungentilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungentilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ungentlemanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ungentlemanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungentlemanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungentlemanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungentlemanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unglamorous ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unglamorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unglamourous",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unglamorously ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unglamorously",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unglamourously",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Possible), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ungospelized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ungospelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungospelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungospelizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungospeliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungospelized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungospelised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ungraphitized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ungraphitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungraphitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungraphitizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungraphitiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ungraphitized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ungraphitised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unharbor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unharbor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unharbour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unharbor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unharbour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unharbored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unharboured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unharmonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unharmonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unharmonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unharmonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unharmonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unharmonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unharmonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unharmonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unharmonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unheroize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unheroize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unheroise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unheroizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unheroises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unhonorable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unhonorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhonourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhonorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhonourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhonorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhonourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unhonorably ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unhonorably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhonourably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhonorablies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhonourablies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unhonored ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unhonored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhonoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhonoredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhonouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhonoreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhonoureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unhouseled ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unhouseled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhouselled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unhumanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unhumanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhumanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhumanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhumanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhumanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhumanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhumanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhumanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unhumored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unhumored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhumoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhumoredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhumouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhumoreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhumoureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unhumorous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unhumorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhumourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unhumorously ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unhumorously",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhumourously",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unhydrolyzed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unhydrolyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhydrolysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhydrolyzeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhydrolyseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhydrolyzed's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhydrolysed's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unhypnotizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unhypnotizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhypnotisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhypnotizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhypnotisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhypnotizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhypnotisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhypnotizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhypnotisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhypnotizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhypnotisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unhypnotize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unhypnotize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhypnotise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhypnotized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhypnotised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhypnotizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhypnotising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unhypnotizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unhypnotises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unicolor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unicolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unicolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unicolorate ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unicolorate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unicolourate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unicolored ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unicolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unicoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unicolorous ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unicolorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unicolourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unidealized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unidealized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unidealised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unidealizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unidealisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unidealizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unidealiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unidolized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unidolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unidolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unidolizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unidolisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unidolizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unidoliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unidolizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unidoliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unidolized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unidolised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uniformization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uniformization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uniformisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uniformizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uniformisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uniformization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uniformisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uniformize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uniformize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uniformise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uniformized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uniformised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uniformizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uniformising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uniformizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uniformises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unilateralization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unilateralization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unilateralisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unilateralizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unilateralisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unilateralization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unilateralisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unilateralize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unilateralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unilateralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unilateralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unilateralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unimmortalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unimmortalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unimmortalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unimmortalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unimmortalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unimmortalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unimmortalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unimmortalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unimmunized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unimmunized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unimmunised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unindividualize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unindividualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unindividualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unindividualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unindividualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unindividualized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unindividualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unindividualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unindustrialized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unindustrialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unindustrialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unindustrializeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unindustrialiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unindustrialized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unindustrialised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uninitializable ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uninitializable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uninitialisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uninitialized ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uninitialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uninitialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unionization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unionization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unionisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unionizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unionisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unionization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unionisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unionize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unionizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unionising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unionizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unionizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unioniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unionizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unionisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unitalicized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unitalicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unitalicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Unitarianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Unitarianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Unitarianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Unitarianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Unitarianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Unitarianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Unitarianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unitemized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unitemized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unitemised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unitization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unitisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unitize ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "universalization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "universalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "universalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "universalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "universalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "universalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "universalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "universalize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "universalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "universalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "universalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "universalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "universalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "universalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "universalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "universalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "universalizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "universalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "universaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "universalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "universalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "universalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "universaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unjeopardized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unjeopardized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unjeopardised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unjournalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unjournalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unjournalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unkennel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unkenneled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unkennelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unkenneling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unkennelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unlabeled ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unlabeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlabeledder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabelledder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlabeleddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabelleddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unlabialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unlabialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlabialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlabializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlabializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unlaborable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unlaborable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlaborables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlaborable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unlabored ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unlabored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlaboured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlaboredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlaboreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlaboureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlabored's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlaboured's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unlaboring ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unlaboring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlaboringer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabouringer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlaboringest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabouringest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlaborings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlaboring's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlabouring's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unlearned ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unlearned",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "unlearnt",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unlegalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unlegalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlegalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlegalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlegalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlegalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlegaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unlevel ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unleveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlevelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unleveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlevelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unliberalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unliberalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unliberalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unliberalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unliberalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unliberalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unliberaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unlionized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unlionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unliteralized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unliteralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unliteralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unlocalizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unlocalizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlocalisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlocalizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlocalisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlocalizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlocalisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlocalizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlocalisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlocalizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlocalisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unlocalize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unlocalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlocalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlocalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlocalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlocalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlocalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unlocalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unlocalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmacadamized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmacadamized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmacadamised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmagnetized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmagnetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmagnetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmagnetizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmagnetisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmagnetizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmagnetiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmagnetizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmagnetiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmagnetized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmagnetised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmaterialized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmaterialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmaterialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmechanize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmechanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmechanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmechanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmechanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmechanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmechanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmechanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmechanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmediatized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmediatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmediatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmediatizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmediatiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmediatized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmediatised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmedieval ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmedieval",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmediaeval",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmelodized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmelodized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmelodised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmemorialized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmemorialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmemorialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmemorializedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmemorialisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmemorializeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmemorialiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmemorializeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmemorialiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmemorialized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmemorialised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmemorized ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmemorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmemorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmercerized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmercerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmercerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmercerizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmerceriseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmercerized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmercerised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmesmerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmesmerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmesmerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmesmerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmesmerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmesmerized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmesmerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmesmerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmesmerizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmesmerisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmesmerizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmesmeriseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmetalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmetalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmetallised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmetallized",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmetalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmetalliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmetallizeds",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmetalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmetallised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmetallized's",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmethodized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmethodized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmethodised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmethodizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmethodisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmethodizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmethodiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmethodizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmethodiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmethodized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmethodised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmethodizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmethodizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmethodising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmethodizinger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmethodisinger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmethodizingest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmethodisingest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmethodizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmethodisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmethodizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmethodising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmilitarized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmilitarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmilitarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmineralized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmineralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmineralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmineralizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmineralisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmineralizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmineraliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmineralizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmineraliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmineralized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmineralised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unminimized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unminimized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unminimised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unminimizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unminimisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unminimizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unminimiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unminimizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unminimizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unminimising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmissionized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmissionized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmissionised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmissionizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmissioniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmissionized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmissionised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmiter ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmiter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmitre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmiters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmitres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmobilized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmobilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmobilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmobilizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmobilisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmobilizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmobiliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmobilizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmobiliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmobilized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmobilised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmodernize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmodernize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmodernise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmodernizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmodernises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmodernized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmodernized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmodernised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmodernizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmodernisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmodernizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmoderniseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmonopolize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmonopolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmonopolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmonopolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmonopolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmonopolizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmonopolisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmonopolizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmonopoliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmonopolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmonopolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmonopolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmonopolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmonopolizinger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmonopolisinger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmonopolizingest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmonopolisingest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmonopolizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmonopolisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmoralize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmoralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmoralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmoralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmoralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmoralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmoralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmoralizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmoralises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmoralizinger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmoralisinger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmoralizingest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmoralisingest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmoralizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmoralisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmotorized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmotorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmotorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmotorizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmotorisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmotorizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmotoriseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmunicipalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmunicipalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmunicipalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmunicipalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmunicipalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmunicipalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmunicipaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmunicipalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmunicipaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmunicipalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmunicipalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmutualized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmutualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmutualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmutualizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmutualisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmutualizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmutualiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmutualizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmutualiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmutualized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmutualised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unmysticize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unmysticize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmysticise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmysticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmysticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmysticizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmysticising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unmysticizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unmysticises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unnationalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unnationalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnationalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnationalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnationalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnationalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnationaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unnaturalizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unnaturalizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnaturalisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnaturalizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnaturalisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnaturalizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnaturalisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unnaturalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unnaturalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnaturalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnaturalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnaturalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnaturalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnaturalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnaturalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnaturalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unneighbored ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unneighbored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneighboured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unneighborlike ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unneighborlike",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneighbourlike",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unneighborlikes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneighbourlikes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unneighborlike's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneighbourlike's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unneighborliness ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unneighborliness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneighbourliness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unneighborlinesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneighbourlinesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unneighborliness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneighbourliness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unneighborly ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unneighborly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneighbourly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unneighborlier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneighbourlier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unneighborliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneighbourliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unneutralize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unneutralize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneutralise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unneutralized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unneutralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneutralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unneutralizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unneutralizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unneutralising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unnitrogenized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unnitrogenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnitrogenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnitrogenizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnitrogenisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnitrogenizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnitrogeniseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnitrogenizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnitrogeniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnitrogenized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnitrogenised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unnoncolorables ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unnoncolorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnoncolourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unnoncolorably ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unnoncolorably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnoncolourably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnoncolorablies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnoncolourablies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unnormalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unnormalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnormalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnormalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnormalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnormalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnormalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnormalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnormaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnormalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnormalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnormalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnormalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnormalizinger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnormalisinger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unnormalizingest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unnormalisingest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unoptimize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unoptimize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoptimise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unoptimized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoptimised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unoptimizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoptimising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unoptimizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoptimises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unorganizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unorganizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unorganisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unorganizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unorganisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unorganizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unorganisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unorganizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unorganisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unorganizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unorganisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unorganize ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unorganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unorganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unorganized ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unorganized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unorganised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unorganizedly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unorganizedly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unorganisedly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unorganizedness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unorganizedness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unorganisedness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unoxidizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unoxidizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoxidisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unoxidizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoxidisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unoxidizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoxidisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unoxidizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoxidisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unoxidizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoxidisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unoxidized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unoxidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoxidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unoxidizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoxidisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unoxidizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoxidiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unoxygenized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unoxygenized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoxygenised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unoxygenizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoxygeniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unoxygenized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unoxygenised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpaganize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpaganize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpaganise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpaganizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpaganises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpalisaded ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpalisaded",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpalisadoed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpanegyrized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpanegyrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpanegyrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unparagonized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unparagonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparagonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparagonizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparagoniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparagonized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparagonised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unparalleled ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unparalleled",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparallelled",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unparalyzed ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unparalyzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparalysed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparalyzedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparalysedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparalyzeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparalyseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparalyzeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparalyseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparalyzed's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparalysed's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unparameterized ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unparameterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparameterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unparametrized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unparametrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparametrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unparceled ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unparceled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparcelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparceledder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparcelledder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparceleddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparcelleddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unparenthesized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unparenthesized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparenthesised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unparticularized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unparticularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparticularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparticularizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparticularisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparticularizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparticulariseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparticularizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparticulariseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparticularized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparticularised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unparticularizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unparticularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparticularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparticularizinger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparticularisinger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparticularizingest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparticularisingest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparticularizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparticularisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unparticularizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unparticularising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpartizan ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpartizan",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpartisan",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpasteurized ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpasteurized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpasteurised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpatronizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpatronizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpatronisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpatronizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpatronisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpatronizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpatronisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpatronizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpatronisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpatronizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpatronisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpatronized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpatronized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpatronised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpatronizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpatronizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpatronising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpatronizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpatronising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpauperized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpauperized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpauperised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpauperizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpauperiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpauperized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpauperised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpenalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpenalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpenalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpenalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpenalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpenalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpenaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unperceptively ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unperceptively",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpreceptively",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpersonalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpersonalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpersonalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpersonalizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpersonalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpersonalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unphilosophize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unphilosophize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unphilosophise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unphilosophizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unphilosophises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unphilosophized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unphilosophized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unphilosophised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unphosphatized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unphosphatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unphosphatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unphosphatizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unphosphatisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unphosphatizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unphosphatiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unphosphatizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unphosphatiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unphosphatized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unphosphatised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpictorialize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpictorialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpictorialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpictorialized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpictorialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpictorialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpictorializing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpictorializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpictorialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unplagiarized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unplagiarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unplagiarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unplagiarizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unplagiarisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unplagiarizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unplagiariseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unplagiarizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unplagiariseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unplagiarized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unplagiarised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpluralized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpluralized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpluralised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpoeticized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpoeticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpoeticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpoeticizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpoeticisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpoeticizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpoeticiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpoeticizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpoeticiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpoeticized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpoeticised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpoetize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpoetize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpoetise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpoetizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpoetises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpoetized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpoetized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpoetised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpolarizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpolarizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolarisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpolarizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolarisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpolarizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolarisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpolarized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpolarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpolarizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolarisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpolarizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolariseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpolarizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolariseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpolarized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolarised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpolymerized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpolymerized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolymerised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpolymerizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolymerisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpolymerizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolymeriseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpolymerizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolymeriseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpolymerized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpolymerised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpopularize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpopularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpopularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpopularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpopularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpopularized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpopularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpopularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpracticed ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpracticed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "unpractised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpracticedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "unpractisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpracticeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "unpractiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpressurized ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpressurized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpressurised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unprotestantize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unprotestantize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unprotestantise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unprotestantized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unprotestantised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unprotestantizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unprotestantising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unprotestantizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unprotestantises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpublicized ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpublicized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpublicised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpulverize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpulverize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpulverise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpulverizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpulverises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unpulverized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unpulverized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpulverised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpulverizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpulverisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unpulverizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unpulveriseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unquantized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unquantized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unquantised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unradicalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unradicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unradicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unradicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unradicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrancored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrancored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrancoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrancoredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrancouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrancoreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrancoureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrancoreds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrancoureds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrancored's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrancoured's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrancorous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrancorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrancourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrationalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrationalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrationalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrationalizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrationalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrationalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unravel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unraveled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unravelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unraveling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unravelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unraveler ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unraveler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unraveller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unravelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unravellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unraveler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unraveller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrealizable ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrealizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrealisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrealizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrealisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrealizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrealisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrealize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrealize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrealise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrealized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrealised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrealizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrealising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrealizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrealises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrealizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrealisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrecognizable ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrecognizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrecognisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrecognizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrecognisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrecognizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrecognisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrecognizably ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrecognizably",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrecognisably",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrecognized ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrecognized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrecognised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrecognizing ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrecognizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrecognising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrecognizingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrecognizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrecognisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unreconnoitered ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unreconnoitered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unreconnoitred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unreconnoiteredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unreconnoitredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unreconnoitereddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unreconnoitreddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unreconnoitereds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unreconnoitreds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unreconnoitered's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unreconnoitred's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unregularized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unregularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unregularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unreorganized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unreorganized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unreorganised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unreorganizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unreorganisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unreorganizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unreorganiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrevelationize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrevelationize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrevelationise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrevelationizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrevelationises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrivaled ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrivaled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrivalled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrivaledder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrivalledder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrivaleddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrivalleddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unromanized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unromanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unromanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unromanticized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unromanticized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unromanticised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unromanticizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unromanticisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unromanticizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unromanticiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unromanticizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unromanticiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unromanticized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unromanticised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unroyalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unroyalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unroyalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unroyalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unroyaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unroyalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unroyalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unrumored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unrumored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrumoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrumoredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrumouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unrumoreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unrumoureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsabered ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsabered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsabred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsaberedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsabredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsabereddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsabreddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsatirizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsatirizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsatirisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsatirize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsatirize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsatirise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsatirizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsatirises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsatirized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsatirized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsatirised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsatirizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsatirisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsatirizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsatiriseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsavored ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsavored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsavoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsavoredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsavouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsavoreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsavoureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsavoredly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsavoredly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsavouredly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsavoredness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsavoredness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsavouredness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsavorily ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsavorily",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsavourily",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsavorilier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsavourilier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsavorilies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsavourilies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsavoriliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsavouriliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsavoriness ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsavoriness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsavouriness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsavorinesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsavourinesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsavoriness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsavouriness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsavory ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsavory",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "unsavoury",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsavorier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "unsavourier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsavories",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "unsavouries",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsavoriest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "unsavouriest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsavory's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "unsavoury's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unscandalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unscandalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscandalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unscandalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscandalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unscandalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unscandalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscandalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unscandalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscandalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unscandalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscandaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unscepter ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unscepter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsceptre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unscepters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsceptres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unscepter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsceptre's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsceptered ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsceptered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsceptred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unscepteredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsceptredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsceptereddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsceptreddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unschematized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unschematized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unschematised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unschematizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unschematisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unschematizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unschematiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unschematizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unschematiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unschematized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unschematised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unscored ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unscored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unscoring ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unscoring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unscrutinized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unscrutinized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscrutinised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unscrutinizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscrutinisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unscrutinizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscrutiniseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unscrutinizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unscrutinizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscrutinising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unscrutinizinger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscrutinisinger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unscrutinizingest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscrutinisingest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unscrutinizingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unscrutinizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscrutinisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unscrutinizinglier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscrutinisinglier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unscrutinizingliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unscrutinisingliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsectarianize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsectarianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsectarianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsectarianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsectarianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsectionalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsectionalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsectionalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsecularize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsecularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsecularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsecularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsecularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsecularized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsecularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsecularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsecularizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsecularisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsecularizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unseculariseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsensitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsensitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsensitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsensitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsensitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsensitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsensitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsensitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsensitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsensualize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsensualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsensualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsensualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsensualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsensualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsensualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsensualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsensualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsentimentalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsentimentalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsentimentalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsentimentalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsentimentalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsentimentalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsentimentalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsentimentalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsepulcher ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsepulcher",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsepulchre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsepulchers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsepulchres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsepulchered ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsepulchered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsepulchred",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unserialized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unserialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unserialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unshakable ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unshakable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "unshakeable",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unshed ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unshed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unshoed",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsignalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsignalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsignalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsignalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsignalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsignalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsignaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsignalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsignaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsignalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsignalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsiphon ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsiphon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsyphon",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsiphons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsyphons",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsiphon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsyphon's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsocialized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsocialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsocialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsocializedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsocialisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsocializeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsocialiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsocializing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsocializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsocialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsolemnize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsolemnize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsolemnise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsolemnizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsolemnises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsolemnized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsolemnized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsolemnised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsolemnizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsolemnisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsolemnizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsolemniseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsomber ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsomber",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsombre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsomberly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsomberly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsombrely",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsomberness ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsomberness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsombreness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsomberness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsombreness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unspecialized ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unspecialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspecialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unspecializedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspecialisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unspecializeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspecialiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unspecializing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unspecializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspecialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unspecializinger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspecialisinger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unspecializingest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspecialisingest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unspecterlike ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unspecterlike",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspectrelike",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unspecterlikes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspectrelikes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unspecterlike's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspectrelike's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unspiritualize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unspiritualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspiritualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unspiritualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspiritualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unspiritualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspiritualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unspiritualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspiritualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsplendorous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsplendorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsplendourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsplendorously ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsplendorously",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsplendourously",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unspoiled ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unspoiled",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unspoilt",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unstabilized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unstabilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstabilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unstabilizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unstabilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstabilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unstandardizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unstandardizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstandardisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unstandardized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unstandardized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstandardised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unstandardizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstandardisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unstandardizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstandardiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsterilized ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsterilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsterilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unstigmatized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unstigmatized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstigmatised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unstigmatizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstigmatisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unstigmatizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstigmatiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unstigmatizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstigmatiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unstigmatized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstigmatised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unstoicize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unstoicize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstoicise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unstoicizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unstoicises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsubsidized ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsubsidized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsubsidised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsubstantialize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsubstantialize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsubstantialise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsubstantialized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsubstantialised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsubstantializing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsubstantialising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsubstantializes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsubstantialises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsuccorable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsuccorable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsuccourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsuccorables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsuccourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsuccorable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsuccourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsuccored ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsuccored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsuccoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsulphurized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsulphurized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsulphurised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsulphurizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsulphuriseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsulphurized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsulphurised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsummarizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsummarizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsummarisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsummarized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsummarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsummarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsummarizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsummarisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsummarizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsummariseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsupernaturalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsupernaturalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsupernaturalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsupernaturalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsupernaturalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsupernaturalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsupernaturalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsupernaturalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsymbolized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsymbolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsymbolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsymbolizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsymbolisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsymbolizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsymboliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsymmetrized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsymmetrized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsymmetrised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsymmetrizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsymmetriseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsymmetrized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsymmetrised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsympathizability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsympathizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsympathizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsympathizability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsympathizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsympathizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsympathizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsympathizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsympathized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsympathized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsympathizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsympathizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsympathizing ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsympathizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsympathizinger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisinger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsympathizingest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisingest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsympathizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsympathizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsympathizingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsympathizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsympathizinglier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisinglier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsympathizingliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsympathisingliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsynchronized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsynchronized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsynchronised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsynchronizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsynchronisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsynchronizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsynchroniseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsynthesized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsynthesized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsynthesised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsynthesizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsynthesisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsynthesizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsynthesiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsystematized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsystematized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsystematised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsystematizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsystematisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsystematizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsystematiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsystematizedly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsystematizedly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsystematisedly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsystematizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsystematizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsystematising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsystematizinger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsystematisinger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsystematizingest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsystematisingest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unsystemizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unsystemizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsystemisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsystemizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsystemisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unsystemizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unsystemisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "untantalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "untantalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untantalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untantalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untantalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untantalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untantaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "untantalizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "untantalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untantalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untantalizinger",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untantalisinger",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untantalizingest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untantalisingest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untantalizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untantalisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untantalizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untantalising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "untartarized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "untartarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untartarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untartarizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untartariseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untartarized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untartarised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "untechnicalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "untechnicalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untechnicalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untechnicalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untechnicalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "untemporizing ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "untemporizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untemporising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untemporizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untemporisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untemporizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untemporising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unterrorized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unterrorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unterrorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "untheorizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "untheorizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untheorisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untheorizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untheorisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untheorizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untheorisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "untrammeled ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "untrammeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untrammelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untrammeledder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untrammelledder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "untrammeleddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untrammelleddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "untranquilized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "untranquilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untranquilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "untranquillize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "untranquillize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untranquillise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "untranquillized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "untranquillized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untranquillised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "untyrannized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "untyrannized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "untyrannised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unutilizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unutilizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unutilisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unutilizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unutilisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unutilizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unutilisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unutilized ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unutilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unutilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unvaporized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unvaporized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvaporised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvaporizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvaporiseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvaporized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvaporised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "see note for \"vaporize\"",
  ],
  },
Cluster {
  header: "unvectorizable ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unvectorizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvectorisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unverbalized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unverbalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unverbalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unvictimized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unvictimized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvictimised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unvisualized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unvisualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvisualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvisualizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvisualisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvisualizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvisualiseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unvitalized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unvitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvitalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvitaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvitalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvitalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unvitriolized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unvitriolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvitriolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvitriolizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvitrioliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvitriolized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvitriolised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unvocalized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unvocalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvocalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvocalizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvocalisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvocalizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvocaliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvocalizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvocaliseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvocalized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvocalised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unvolatilize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unvolatilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvolatilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvolatilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvolatilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unvolatilized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unvolatilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvolatilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvolatilizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvolatilisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvolatilizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvolatiliseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unvulcanized ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unvulcanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvulcanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvulcanizedder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvulcanisedder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvulcanizeddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvulcaniseddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvulcanizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvulcaniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvulcanized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvulcanised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unvulgarize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unvulgarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvulgarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvulgarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvulgarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvulgarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvulgarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unvulgarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unvulgarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unwesternized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unwesternized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unwesternised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unwesternizeds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unwesterniseds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unwesternized's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unwesternised's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unwomanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unwomanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unwomanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "unwomanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unwomanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "unwomanized ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "unwomanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "unwomanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "updraft ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "updraft",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "updraught",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "updrafts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "updraughts",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "updraft's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "updraught's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "upward ",
  verified: false,
  level: 10,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "upward",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "upward",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "upwards",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adverb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uralitization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uralitization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uralitisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uralitizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uralitisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uralitize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uralitize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uralitise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uralitized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uralitised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uralitizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uralitising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uralitizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uralitises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uratemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uratemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urataemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "urbanization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "urbanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urbanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "urbanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urbanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "urbanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urbanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "urbanize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "urbanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urbanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "urbanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urbanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "urbanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urbanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "urbanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urbanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uredema ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uredema",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uroedema",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uremia ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uremias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uraemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "uremia's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uraemia's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uremic ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uremic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uraemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "urethrorrhea ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "urethrorrhea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urethrorrhoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uricemia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uricemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uricaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "uricemic ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "uricemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "uricaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "urinemia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "urinemia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urinaemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "urinemic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "urinemic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urinaemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "urohematin ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "urohematin",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urohaematin",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "urophein ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "urophein",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "urophaein",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "usability ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "usability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "useability",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "usability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "useability's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "usable ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "usable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "useable",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "utilitarianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "utilitarianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilitarianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilitarianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilitarianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilitarianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilitarianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilitarianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilitarianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "utilizability ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "utilizability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilisability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilizabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilisabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "utilizable ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "utilizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "utilization ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "utilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "utilize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "utilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "utilizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "utilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utiliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utilisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utilizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utiliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Utopianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Utopianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Utopianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Utopianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Utopianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Utopianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Utopianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utopianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utopianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utopianized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utopianised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utopianizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utopianising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utopianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utopianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "utopianizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "utopianizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utopianiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utopianizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utopianisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "utopianizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "utopianiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaagmer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaagmer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaagmaer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaccinization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaccinization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaccinisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaccinizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaccinisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaccinization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaccinisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vacuolization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vacuolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vacuolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vacuolizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vacuolisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vacuolization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vacuolisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vacuumize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vacuumize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vacuumise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vacuumized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vacuumised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vacuumizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vacuumising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vacuumizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vacuumises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vagabondize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vagabondize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vagabondise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vagabondized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vagabondised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vagabondizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vagabondising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vagabondizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vagabondises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vagabondizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vagabondizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vagabondiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vagabondizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vagabondisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaginule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaginule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaginulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vagrantize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vagrantize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vagrantise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vagrantizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vagrantises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "valor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "valor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "valour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "valors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "valours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "valor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "valour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "valorization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "valorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "valorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "valorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "valorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "valorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "valorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "valorize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "valorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "valorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "valorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "valorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "valorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "valorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "valorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "valorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "valorousness ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "valorousness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "valourousness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "valvule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "valvule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "valvulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vampirize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vampirize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vampirise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vampirized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vampirised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vampirizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vampirising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vampirizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vampirises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vandalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vandalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vandalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vandalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vandalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vandalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vandalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vandalize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vandalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vandalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vandalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vandalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vandalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vandalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vandalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vandalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vapor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vapor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "vapour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vapored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "vapoured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "vapouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vapor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "vapour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "vapourings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "vapouring's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vapors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "vapours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporability ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporability",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourability",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporabilities",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourabilities",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporability's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourability's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporescent ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporescent",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourescent",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporific ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporific",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourific",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporimeter ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporimeter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourimeter",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporimeter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourimeter's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporingly ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapouringly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporish ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporish",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourish",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporisher",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourisher",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporishest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourishest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporishness ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporishness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourishness",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporishnesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourishnesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporishness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourishness's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "see note vaporize",
  ],
  },
Cluster {
  header: "vaporization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "see note for \"vaporize\"",
  ],
  },
Cluster {
  header: "vaporize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "\"Vapor\" is spelled \"Vapour\" unless the word is \"vaporize\".  This",
    "seams odd to be, but I could not find any evidence that \"vapourize\"",
    "or \"vapourise\" is a legal word, although I could find some evidence",
    "it was used.  Nevertheless since \"vapourize/-ise\" doesn't seam to",
    "be a valid word, this spelling was not included.",
  ],
  },
Cluster {
  header: "vaporizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vaporiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "see note for \"vaporize\"",
  ],
  },
Cluster {
  header: "vaporless ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporlesses ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporlesses",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourlesses",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporlike ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporlike",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourlike",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporlikes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourlikes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporlike's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourlike's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporose ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporose",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourose",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporously ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporously",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourously",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vaporware ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vaporware",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "vapourware",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vapory ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vapory",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapoury",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapourier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vaporiest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vapouriest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "varicolored ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "varicolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "varicoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "varicoloredder",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "varicolouredder",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "varicoloreddest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "varicoloureddest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "varicoloreds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "varicoloureds",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "varicolored's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "varicoloured's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "varicolorous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "varicolorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "varicolourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vascularization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vascularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vascularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vascularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vascularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vascularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vascularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vascularize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vascularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vascularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vascularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vascularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vascularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vascularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vascularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vascularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vasectomize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vasectomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vasectomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vasectomized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vasectomised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vasectomizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vasectomising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vassalization ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vassalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vassalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vassalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vassalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vassalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vassalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vassalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vassalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vassalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vassalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vassalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Vaticanization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Vaticanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Vaticanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Vaticanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Vaticanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Vaticanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Vaticanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Vaticanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Vaticanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Vaticanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Vaticanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Vaticanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Vaticanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Vaticanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vavasor ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vavasor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vavasour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vavasors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vavasours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Ve ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Ve",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Voe",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Ve's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Voe's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vectorizable ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vectorizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vectorisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vectorization ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vectorization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vectorisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vectorizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vectorisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vectorization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vectorisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vectorize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vectorize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vectorise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vectorized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vectorised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vectorizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vectorising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vectorizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vectorises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vectorizer ",
  verified: false,
  level: 99,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vectorizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vectoriser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vectorizered",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vectorisered",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vectorizering",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vectorisering",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vectorizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vectorisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "veery ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "veery",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verey",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vegetablize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vegetablize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vegetablise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vegetablizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vegetablises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "velarization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "velarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "velarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "velarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "velarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "velarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "velarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "velarize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "velarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "velarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "velarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "velarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "velarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "velarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "velarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "velarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "veld ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "veld",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "veldt",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "velds",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "veldts",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "veld's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "veldt's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "venalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "venalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "venalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "venalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "venalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "venalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "venalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "venalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "venalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "venalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "venalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "venalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vendor ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vendor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vender",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vendors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "venders",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vendor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vender's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "venomization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "venomization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "venomisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "venomizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "venomisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "venomization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "venomisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "venomize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "venomize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "venomise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "venomizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "venomises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "venter ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "venter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ventre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "venters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ventres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "ventriloquize ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "ventriloquize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ventriloquise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ventriloquized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ventriloquised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ventriloquizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ventriloquising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "ventriloquizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "ventriloquises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "venule ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "venule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "venulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "veranda ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "veranda",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verandah",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "verandas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verandahs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "veranda's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verandah's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "verbalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "verbalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verbalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "verbalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verbalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "verbalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verbalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "verbalize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "verbalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verbalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "verbalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verbalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "verbalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verbalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "verbalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verbalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "verbalizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "verbalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verbaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "verbalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verbalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "verbalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "verbaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vermeil ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vermeiled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vermeilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vermeiling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vermeilling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vermeiles ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vermeiles",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vermeilles",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vermilion ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vermilion",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vermillion",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vermilion's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vermillion's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vermilionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vermilionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vermilionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vermilionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vermilionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vernacularization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vernacularization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernacularisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vernacularizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernacularisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vernacularization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernacularisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vernacularize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vernacularize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernacularise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vernacularized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernacularised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vernacularizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernacularising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vernacularizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernacularises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vernalization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vernalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vernalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vernalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vernalize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vernalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vernalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vernalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vernalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vernalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "versicolor ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "versicolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "versicolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "versicolorate ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "versicolorate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "versicolourate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "versicolored ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "versicolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "versicoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "versicolorous ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "versicolorous",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "versicolourous",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "versionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "versionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "versionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "versionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "versionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vesicule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vesicule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vesiculae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vestryize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vestryize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vestryise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vestryizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vestryises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "veter ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "veter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vetoer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "veter's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vetoer's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "veteranize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "veteranize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "veteranise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "veteranizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "veteranises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vial ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vialed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vialled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vialing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vialling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "victimizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "victimizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victimizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victimizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "victimization ",
  verified: true,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "victimization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victimizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victimization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "victimize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "victimize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victimized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victimizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victimizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "victimizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "victimizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victimizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victimizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victimiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Victorianize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Victorianize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Victorianise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Victorianizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Victorianises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Victorianize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Victorianise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "victual ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "victualed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victualled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victualing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victualling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "victualage ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "victualage",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victuallage",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victualages",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victuallages",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victualage's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victuallage's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "victualer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "victualer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victualler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victualers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victuallers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "victualer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victualler's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "victualess ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "victualess",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "victualless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "videodisk ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "videodisc",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "videodisk",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "videodisc's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "videodisk's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "videodiscs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "videodisks",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vier ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vire",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "viers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vires",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Vietnamization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Vietnamization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Vietnamisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Vietnamization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Vietnamisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Vietnamize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Vietnamize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Vietnamise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Vietnamized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Vietnamised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Vietnamizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Vietnamising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Vietnamizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Vietnamises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vigor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vigor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "vigour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vigors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "vigours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vigor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "vigour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vigorless ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vigorless",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vigourless",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "villagization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "villagization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "villagisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "villagizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "villagisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "viremia ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "viremia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "viraemia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "viremias",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "viraemias",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "viremic ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "viremic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "viraemic",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "virilization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "virilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "virilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "virilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "virilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "virilize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "virilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "virilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "virilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "virilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "virtualize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "virtualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "virtualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "virtualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "virtualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "virtuosi ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "virtuosi",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "virtuosos",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "virtuosoes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "visionize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "visionize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visionise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "visionizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visionises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "visor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "visor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vizor",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "visors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vizors",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "visor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vizor's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "visualizable ",
  verified: true,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "visualizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visualisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "visualization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "visualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "visualizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visualisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "visualization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visualisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "visualize ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "visualize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visualise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "visualized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visualised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "visualizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visualising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "visualizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visualises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "visualizer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "visualizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visualiser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "visualizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visualisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "visualizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visualiser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vitalization ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vitalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vitalize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vitalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitalizings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitalisings",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitalizing's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitalising's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vitalizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vitalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vitalizingly ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vitalizingly",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitalisingly",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vitaminization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vitaminization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitaminisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vitaminize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vitaminize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitaminise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitaminized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitaminised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitaminizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitaminising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitaminizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitaminises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vitriolizable ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vitriolizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitriolisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitriolizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitriolisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitriolizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitriolisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vitriolization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vitriolization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitriolisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitriolizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitriolisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitriolization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitriolisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vitriolize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vitriolize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitriolise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitriolized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitriolised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitriolizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitriolising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitriolizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitriolises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vitriolizer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vitriolizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitrioliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vitriolizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vitriolisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vivandier ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vivandier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vivandire",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "viver ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "viver",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vivre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vivers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vivres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "viverrine ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "viverrine",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "viverrinae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vizard ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vizard",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visard",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vizards",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visards",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vizied ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vizied",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visied",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vizies ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vizies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "visies",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vocalization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vocalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vocalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vocalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vocalize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vocalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vocalized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocalised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vocalizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocalising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vocalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vocalizer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vocalizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocaliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vocalizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocalisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vocalizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocaliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vocationalization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vocationalization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocationalisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vocationalizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocationalisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vocationalization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocationalisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vocationalize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vocationalize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocationalise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vocationalizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vocationalises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "volatilizable ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "volatilizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volatilizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volatilizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volatilizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volatilizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "volatilization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "volatilization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volatilizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volatilization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "volatilize ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "volatilize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volatilized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volatilizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volatilizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "volatilizer ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "volatilizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatiliser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volatilizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatilisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volatilizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volatiliser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "volcanization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "volcanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volcanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volcanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volcanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "volcanize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "volcanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volcanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volcanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volcanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volcanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volcanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volcanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "volcanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "voltize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "voltize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "voltise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "voltizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "voltises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vowelization ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vowelization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vowelisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vowelizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vowelisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vowelization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vowelisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vowelize ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vowelize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vowelise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vowelized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vowelised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vowelizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vowelising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vowelizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vowelises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vulcanizable ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vulcanizable",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisable",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanizabler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisabler",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanizables",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisables",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanizablest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisablest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanizable's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisable's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vulcanizate ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vulcanizate",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisate",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanizates",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisates",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanizate's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisate's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vulcanization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vulcanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vulcanize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vulcanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vulcanizer ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vulcanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulcanizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcaniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vulgarization ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vulgarization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulgarisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulgarizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulgarisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulgarization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulgarisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vulgarize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vulgarize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulgarise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulgarized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulgarised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulgarizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulgarising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulgarizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulgarises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "vulgarizer ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "vulgarizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulgariser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulgarizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulgarisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "vulgarizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulgariser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wackes ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wackes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wackoes",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wacky ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wacky",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "whacky",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wackier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "whackier",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wackiest",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "whackiest",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Wagnerize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Wagnerize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Wagnerise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Wagnerizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Wagnerises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Wagnerize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Wagnerise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wagon ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wagon",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waggon",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wagoned",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waggoned",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wagoning",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waggoning",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wagons",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waggons",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wagon's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waggon's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wagoneer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wagoneer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waggoneer",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wagoner ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wagoner",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waggoner",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wagoners",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waggoners",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wagoner's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waggoner's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wagonette ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wagonette",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waggonette",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wagonettes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waggonettes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wagonette's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waggonette's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wainscot ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wainscoted",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wainscotted",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wainscoting",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wainscotting",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wainscotings",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wainscottings",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wainscoting's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wainscotting's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wallah ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wallah",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "walla",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wallahs",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wallas",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wantonize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wantonize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wantonise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wantonized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wantonised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wantonizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wantonising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wantonizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wantonises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "warer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "warer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "warre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Warford ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Warford",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Warfourd",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Warford's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Warfourd's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "warm-up ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "warm-up",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "warmup",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "warm-ups",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "warmups",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "warm-up's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "warmup's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "warrantize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "warrantize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "warrantise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "waterbed ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "waterbed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "water bed",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "waterbeds",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "water beds",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "waterbed's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "water bed's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "watercolor ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "watercolor",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "watercolour",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "watercolored",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "watercoloured",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "watercoloring",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "watercolouring",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "watercolors",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "watercolours",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "watercolor's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "watercolour's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "watercolorist ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "watercolorist",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "watercolourist",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "watercolorists",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "watercolourists",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "watercolorist's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "watercolourist's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "weasel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "weaseled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "weaselled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "weaseling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "weaselling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "weatherize ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "weatherize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "weatherise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "weatherized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "weatherised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "weatherizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "weatherising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "weatherizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "weatherises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "weepy ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "weepy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "weepie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "weeviled ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "weeviled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "weevilled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "weftwize ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "weftwize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "weftwise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "weize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "weize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "weise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "weized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "weised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "weizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "weising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "weizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "weises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "welsh ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "welsh",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "welch",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "welshed",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "welched",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "welshing",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "welching",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "welshes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "welches",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wer ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "waer",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "werewolf ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "werewolf",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "werwolf",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "werewolves",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "werwolves",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "werewolf's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "werwolf's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "westernization ",
  verified: true,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "westernization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "westernisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "westernizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "westernisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "westernization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "westernisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "westernize ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "westernize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "westernise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "westernized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "westernised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "westernizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "westernising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "westernizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "westernises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "whimsy ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "whimsy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "whimsey",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whimsies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "whimseys",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whimsy's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "whimsey's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "whir ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "whir",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "whirr",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whirs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "whirrs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whir's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "whirr's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "whiskey ",
  verified: true,
  level: 20,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "whiskeys",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "whiskies",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("drink"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whiskey",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "whisky",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("drink"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whiskey's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "whisky's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("drink"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whiskey",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("code word"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whiskies",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("code word"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whiskey's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: true, note: None,
  description: Some("code word"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Whitmanize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Whitmanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Whitmanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Whitmanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Whitmanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Whitmanize's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Whitmanise's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "whiz ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "whiz",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "whizz",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whizzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whizzed",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whizzing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whiz's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "whizz's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whiz",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "whizz",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wiz",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whiz's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "whizz's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wiz's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whizzes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wizzes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wiz",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some("wizard"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wizz",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("Ox: informal, chiefly N. Amer."),
  description: Some("as in \"gee whiz\" "),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "whodunit ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "whodunit",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "whodunnit",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whodunits",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "whodunnits",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "whodunit's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "whodunnit's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wigeon ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wigeon",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "widgeon",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wigeon's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "widgeon's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wigeon",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "widgeons",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "wigeons",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "widgeon",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl"),
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "willful ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "willful",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "wilful",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "willfuler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "wilfuler",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "willfulest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "wilfulest",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "willfully ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "willfully",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wilfully",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "willfullier",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wilfullier",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "willfulliest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wilfulliest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "willfulness ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "willfulness",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wilfulness",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "willfulness's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wilfulness's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "willy ",
  verified: false,
  level: 55,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "willy",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "willie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "winterization ",
  verified: true,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "winterization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "winterisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "winterizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "winterisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "winterization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "winterisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "winterize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "winterize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "winterise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "winterized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "winterised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "winterizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "winterising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "winterizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "winterises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wintry ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wintry",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wintery",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wintrier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "winterier",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wintriest",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "winteriest",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wisteria ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wisteria",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wistaria",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wisterias",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wistarias",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wisteria's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wistaria's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wizen ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wizen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wisen",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wolf ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wolf",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wolves",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wolfs",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Possible), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wolf",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wolfs",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: Some(Pos::Verb), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "womanization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "womanization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "womanisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "womanizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "womanisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "womanization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "womanisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "womanize ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "womanize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "womanise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "womanized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "womanised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "womanizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "womanising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "womanizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "womanises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "womanizer ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "womanizer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "womaniser",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "womanizers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "womanisers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "womanizer's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "womaniser's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "woning ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "woning",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wonning",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "wonings",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wonnings",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woning's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wonning's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "woodcockize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "woodcockize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woodcockise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woodcockizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woodcockises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "woolen ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "woolen",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "woollen",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woolenner",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "woollenner",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woolennest",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "woollennest",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woolens",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "woollens",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woolen's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "woollen's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "woolenization ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "woolenization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woollenisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woollenization",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woolenizations",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woollenisations",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woollenizations",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woolenization's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woollenisation's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woollenization's",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "woolenize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "woolenize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woollenise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woollenize",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woolenizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woollenises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woollenizes",
        types: &[
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "woolly ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "woolly",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wooly",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woollier",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woolier",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woolliest",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wooliest",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Adjective), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woolly",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wooly",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "woolie",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woollies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "woolies",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "woolly's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wooly's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "woolie's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: Some(Pos::Noun), archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "wornil ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "wornil",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "wournil",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "worship ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "worshiped",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "worshipped",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "worshiping",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "worshipping",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "worshiper ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "worshiper",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "worshipper",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "worshipers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "worshippers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "worshiper's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "worshipper's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "yak ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "yak",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yaks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yak",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: Some("pl"),
  description: Some(":1 "),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yak's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":1"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yak",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yack",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yakked",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yacked",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yakking",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yacking",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yaks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yacks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":2"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yak",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yack",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yaks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yacks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yak's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yack's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":3"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yuck",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yuk",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":4"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yuk",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yuck",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yak",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "yock",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":5"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yuks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yucks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yocks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "yaks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":5"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yuk's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yuck's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yock's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
      Variant {
        word: "yak's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":5"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yuk",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yuck",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":6"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yukked",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yucked",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":6"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yukking",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yucking",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":6"),
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yuks",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yucks",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: Some(":6"),
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "yeshiva ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "yeshiva",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yeshivah",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yeshivas",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yeshivot",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yeshivahs",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yeshiva's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yeshivah's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "yodel ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "yodeled",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yodelled",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yodeling",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yodelling",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "yodeler ",
  verified: true,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "yodeler",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yodeller",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yodelers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yodellers",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yodeler's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yodeller's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "yogi ",
  verified: false,
  level: 50,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "yogi",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yogin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yogis",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yogins",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yogi's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yogin's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "yogurt ",
  verified: true,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "yogurt",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yoghurt",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yoghourt",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yogourt",
        types: &[
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yogurt's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yoghurt's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yoghourt's",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yogourt's",
        types: &[
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "yogurts",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
        ],
      },
      Variant {
        word: "yoghurts",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yoghourts",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Seldom), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "yogourts",
        types: &[
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zaffer ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zaffer",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zaffre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zaffers",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zaffres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zea ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zoea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zeas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zoeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zoea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zoaea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zoea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zooea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zoeas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zooeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zoea's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zoaea's",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zeroize ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zeroize",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zeroise",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zeroized",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zeroised",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zeroizing",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zeroising",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zeroizes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zeroises",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zoeae ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zoeae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zooeae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zoeal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zoeal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zooeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zoecium ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zoecium",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zooecium",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zoecia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zooecia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zombie ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zombie",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zombi",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zombies",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zombis",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zombie's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zombi's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zonesthesia ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zonesthesia",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zonaesthesia",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zonule ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zonule",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zonulae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zooglea ",
  verified: false,
  level: 70,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zooglea",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zoogloea",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zoogleas",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zoogloeas",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zoogleae ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zoogleae",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zoogloeae",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zoogleal ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zoogleal",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zoogloeal",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "zygenid ",
  verified: false,
  level: 80,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zygenid",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "zygaenid",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "Koran ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Koran",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "Quran",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Qur'an",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Korans",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "Qurans",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Qur'ans",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "Koran's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "Quran's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Qur'an's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "Quran is more correct bases on a modern transliterated from Arabic",
    "to English, but both M-W and Oxford has Koran marked as the primary",
    "form.  Technically \"Qur'an\" more correct but that doesn't seam to be",
    "used as much.",
  ],
  },
Cluster {
  header: "Koranic ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "Koranic",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Eq), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Eq), num: None,},
        ],
      },
      Variant {
        word: "Quranic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Qur'anic",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "See note for Quran",
  ],
  },
Cluster {
  header: "virtualization ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "virtualization",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
        ],
      },
      Variant {
        word: "virtualisation",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "volcanism ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "volcanism",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanism",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volcanisms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanisms",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "volcanism's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "vulcanism's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "eukaryote ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "eukaryote",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eucaryote",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eukaryotes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eucaryotes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eukaryote's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eucaryote's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "eukaryotic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "eucaryotic",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "prokaryote ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "prokaryote",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "procaryote",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prokaryotes",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "procaryotes",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prokaryote's",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "procaryote's",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "prokaryotic",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "procaryotic",
        types: &[
          Type {category: Category::American, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "adrenaline ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "adrenaline",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adrenalin",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adrenalines",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "adrenalin's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "adrenalin's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "Adrenalin is a trademark but some dictionaries allow the lowercase",
    "version as an alternative spelling for adrenaline.",
  ],
  },
Cluster {
  header: "dumpster ",
  verified: false,
  level: 40,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "dumpster",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Dumpster",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dumpsters",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Dumpsters",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "dumpster's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Dumpster's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "Dumpster was originally a trademark",
  ],
  },
Cluster {
  header: "zipper ",
  verified: false,
  level: 35,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "zipper",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Zipper",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zippers",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Zippers",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "zipper's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "Zipper's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
    "Zipper was originally a trademark",
  ],
  },
Cluster {
  header: "sponge_cake ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "sponge cake",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spongecake",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sponge cakes",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spongecakes",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "sponge cake's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "spongecake's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "name-drop ",
  verified: false,
  level: 95,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "name-drop",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "namedrop",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "name-drops",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "namedrops",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "name-drop's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "namedrop's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "name-dropped",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "namedropped",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "name-dropping",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "namedropping",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "name-droppings",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "namedroppings",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "name-dropping's",
        types: &[
          Type {category: Category::Other, tag: None, num: None,},
        ],
      },
      Variant {
        word: "namedropping's",
        types: &[
          Type {category: Category::Other, tag: Some(Tag::Seldom), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gigameter ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gigameter",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gigametre",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gigameters",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::Canadian, tag: Some(Tag::Variant), num: None,},
        ],
      },
      Variant {
        word: "gigametres",
        types: &[
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "megagram ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "megagram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "megagramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "megagrams",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "megagrammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "gigagram ",
  verified: true,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "gigagram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gigagramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "gigagrams",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "gigagrammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
Cluster {
  header: "microgram ",
  verified: false,
  level: 60,
  entries: &[
  Entry {
    variants: &[
      Variant {
        word: "microgram",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "microgramme",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  Entry {
    variants: &[
      Variant {
        word: "micrograms",
        types: &[
          Type {category: Category::American, tag: None, num: None,},
          Type {category: Category::BritishIse, tag: None, num: None,},
          Type {category: Category::Canadian, tag: None, num: None,},
          Type {category: Category::BritishIze, tag: None, num: None,},
          Type {category: Category::Australian, tag: None, num: None,},
        ],
      },
      Variant {
        word: "nanogrammes",
        types: &[
          Type {category: Category::BritishIse, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::BritishIze, tag: Some(Tag::Variant), num: None,},
          Type {category: Category::Australian, tag: Some(Tag::Variant), num: None,},
        ],
      },
  ],
  pos: None, archaic: false, note: None,
  description: None,
  comment: None,
  },
  ],
  notes: &[
  ],
  },
];
